#define NO_MESSAGE 1

#ifndef MMTRACE
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "re.h"

# ifdef NO_MESSAGE
#  define message(fmt, ...)
# else
#  define message(fmt, ...)	printf(fmt "\n", ##__VA_ARGS__)
# endif
#endif

#include <GL/gl.h>
#include "nv_hw.h"
#include "nv_objects.h"
#include "objects.h"

#define INSTANCE_MEM(i)		all_regs[(NV_PRAMIN + (instanceMem << 4)  + i * 4)/4]

const object_store nv_objects[];

const char* print_format(unsigned int value, void*format) {
	static char buf[128];
	snprintf(buf, sizeof(buf), format?format:"%x", value);
	return buf;
}

const char* print_float(unsigned int value, void*format) {
	static char buf[128];
	double dval = (double)*(float*)&value;
#ifdef MMTRACE
	int dint = (int)dval;
	int dpart = (dval - dint) * 1000000;
	if (dpart < 0) {
		dpart = - dpart;
	}
	snprintf(buf, sizeof(buf), "%d.%06d", dint, dpart);
#else
	snprintf(buf, sizeof(buf), format?format:"%f", dval);
#endif
	return buf;
}

const char* print_macro(unsigned int value, void*data) {
	data_store* names = (data_store*)data;
	static char buf[15];
	int i=0;

	while (names[i].name != NULL) {
		if (names[i].id == value) {
			return names[i].name;
		}
		i++;
	}

	snprintf(buf, sizeof(buf), "0x%x", value);

	return buf;
}

const char* print_macro16(unsigned int value, void*data) {
	data_store* names = (data_store*)data;
	static char buf[128];
	int i=0;
	int i1=-1,i2=-1;

	while (names[i].name != NULL) {
		if (names[i].id == (value&0xffff)) {
			i1=i;
		}
		if (names[i].id == (value>>16)) {
			i2=i;
		}
		i++;
	}

	snprintf(buf, sizeof(buf), "(%s,%s)", i1==-1?"UNKNOWN":names[i1].name,i2==-1?"UNKNOWN":names[i2].name);

	return buf;
}

const char* print_object(unsigned int value, void*data) {
	static char buf[BUFSIZ];
	int type = find_object_type(value);
	const char *typename = "UNKNOWN";
	int i;

	for(i=0;nv_objects[i].name != NULL;i++)
		if ((nv_objects[i].id==type)&&(nv_objects[i].card_type&card_family)) {
			typename = nv_objects[i].name;
			break;
		}
	snprintf(buf, BUFSIZ, "object %08x type %04x (%s)", value, type, typename);
	return buf;
}

const char* print_enable(unsigned int value, void*data) {
	if (value)
		return data;
	else
		return "";
}

const char* print_class(unsigned int value, void*data) {
	static char buf[BUFSIZ];
	const char *typename = "UNKNOWN";
	int i;

	for(i=0;nv_objects[i].name != NULL;i++)
		if ((nv_objects[i].id==value)&&(nv_objects[i].card_type&card_family)) {
			typename = nv_objects[i].name;
			break;
		}
	snprintf(buf, BUFSIZ, "%04x (%s)", value, typename);
	return buf;
}

/** \fn    const char* print_clip_planes(unsigned int value, void*data)
    \brief Print out, which clipping planes are active.

    See NV30_TCL_PRIMITIVE_3D_SET_CLIPPING_PLANES for some more infos
*/
const char* print_clip_planes(unsigned int value, void*data) {
        static char buf[BUFSIZ];
        int i = 0;

	if (card_family == NV30) {
           snprintf(buf, BUFSIZ, "NV30 unreliable, val=%d", value);
           return buf;
        }    
        if (value == 0) {
           return "all disabled";
        }
        strcpy(buf, "Enabled: ");
        while (i < 16) {
           char val[20];
           if (value & (1 << (1 + i*2 ))) {
              snprintf(val, 20, "%d, ", (i >> 1)+1);
              strcat(buf, val);
           }
           i++;
        }
        return buf;
        
}

const char *print_store(unsigned int value, void *data) {
   static char buf[BUFSIZ];
   int    what;
  
  what = (value & 0xff000000) >> 24;
  if (what == 1)
      {
        value = (value & 0x00ffffff) >> 4;
        snprintf(buf, BUFSIZ, "occl. query #%d result in reg. %08x", value, value *4+2);
      }
    else strcpy(buf, "Color buffer(?)");
  return buf; 
}


const char *print_query(unsigned int value, void *data) {
    
   if (value == 1)
      return "Enable occlusion query";
    else return "Enable color buffer(?)";    
}

static int format_data(char *buf, int len, const data_print_store *dp, unsigned int value);

#define N_(name)	name, # name
#define GL_(name)	GL_##name, # name
#define __(name)	name, 1, 1, # name
#define A_(name)	name(0), name ## __SIZE, (name (name ## __SIZE) - name(0))/name ## __SIZE , # name "(%d)"
#define M_(name)	name, 16, 4, # name "(%d)"

#define PRINT_BITFIELD(bf, func, data, name)	{0?bf, 1?bf, func, data, name}
#define PRINT_FP32	{0, 31, &print_float, NULL, NULL}
#define PRINT_X32	{0, 31, &print_format, "0x%08x", NULL}
#define PRINT_UINT	{0, 31, &print_format, "%u", NULL}
#define PRINT_MACRO(data)	{0, 31, &print_macro, data, NULL}
#define PRINT_MACRO16(data)	{0, 31, &print_macro16, data, NULL}
#define PRINT_FUNC(func)    {0, 31, &func, NULL, NULL}
#define PRINT_ENABLE_BIT(bit, name)	{bit, bit, &print_enable, name, NULL}
#define PRINT_OBJECT	{0, 31, &print_object, NULL, NULL}

static data_store begin_end_data[] = {
    { N_(STOP)				},
    { N_(POINTS)			},
    { N_(LINES)				},
    { N_(LINE_LOOP)			},
    { N_(LINE_STRIP)			},
    { N_(TRIANGLES)			},
    { N_(TRIANGLE_STRIP)		},
    { N_(TRIANGLE_FAN)			},
    { N_(QUADS)				},
    { N_(QUAD_STRIP)			},
    { N_(POLYGON)			},
    {}
};

static data_store bool_data[] = {
    { N_(TRUE)				},
    { N_(FALSE)				},
    {}
};

static data_store gl_comparison_op_data[] = {
    { GL_(NEVER) 			},
    { GL_(ALWAYS) 			},
    { GL_(LESS) 			},
    { GL_(LEQUAL) 			},
    { GL_(EQUAL) 			},
    { GL_(GEQUAL) 			},
    { GL_(GREATER) 			},
    { GL_(NOTEQUAL) 			},
    {}
};

static data_store gl_blend_func_data[] = {
    { GL_(ONE)				},
    { GL_(ZERO)				},
    { GL_(SRC_COLOR)			},
    { GL_(ONE_MINUS_SRC_COLOR)		},
    { GL_(DST_COLOR)			},
    { GL_(ONE_MINUS_DST_COLOR)		},
    { GL_(SRC_ALPHA)			},
    { GL_(ONE_MINUS_SRC_ALPHA)		},
    { GL_(DST_ALPHA)			},
    { GL_(ONE_MINUS_DST_ALPHA)		},
    { GL_(SRC_ALPHA_SATURATE)		},
    { GL_(CONSTANT_COLOR)		},
    { GL_(ONE_MINUS_CONSTANT_COLOR)	},
    { GL_(CONSTANT_ALPHA)		},
    { GL_(ONE_MINUS_CONSTANT_ALPHA)	},
    {}
};

static data_store gl_blend_eqn_data[] = {
    { GL_(FUNC_ADD)			},
    { GL_(FUNC_SUBTRACT)		},
    { GL_(FUNC_REVERSE_SUBTRACT)	},
    { GL_(MIN)				},
    { GL_(MAX)				},
    {}
};

static data_store gl_logic_op_data[] = {
    { GL_(CLEAR)			},
    { GL_(AND)				},
    { GL_(AND_REVERSE)			},
    { GL_(COPY)				},
    { GL_(NOOP)				},
    { GL_(XOR)				},
    { GL_(OR)				},
    { GL_(NOR)				},
    { GL_(EQUIV)			},
    { GL_(INVERT)			},
    { GL_(AND_INVERTED)			},
    { GL_(OR_REVERSE)			},
    { GL_(NAND)				},
    { GL_(COPY_INVERTED)		},
    { GL_(OR_INVERTED)			},
    { GL_(SET)				},
    {}
};

static data_store gl_face_data[] = {
    { GL_(FRONT)			},
    { GL_(BACK)				},
    { GL_(FRONT_AND_BACK)		},
    {}
};

static data_store gl_stencil_mod_data[] = {
    { GL_(KEEP)				},
    { GL_(ZERO)				},
    { GL_(REPLACE)			},
    { GL_(INCR)				},
    { GL_(DECR)				},
    { GL_(INVERT)			},
    { GL_(INCR_WRAP)			},
    { GL_(DECR_WRAP)			},
    {}
};

static data_store gl_polygon_mode_data[] = {
    { GL_(FILL)				},
    { GL_(POINT)			},
    { GL_(LINE)				},
    {}
};

static data_store gl_front_face_data[] = {
    { GL_(CW)				},
    { GL_(CCW)				},
    {}
};

static data_store gl_shade_model_data[] = {
    { N_(GL_FLAT)			},
    { N_(GL_SMOOTH)			},
    {}
};

static data_store gl_texgen_mode_data[] = {
    { N_(GL_OBJECT_LINEAR)		},
    { N_(GL_EYE_LINEAR)			},
    {}
};

static data_store gl_logic_op[] = {
    { 0x00, "CLEAR"		},
    { 0x01, "NOR"		},
    { 0x02, "AND_INVERTED"	},
    { 0x03, "COPY_INVERTED"	},
    { 0x04, "AND_REVERSE"	},
    { 0x05, "INVERT"		},
    { 0x06, "XOR"		},
    { 0x07, "NAND"		},
    { 0x08, "AND"		},
    { 0x09, "EQUIV"		},
    { 0x0a, "NOOP"		}, /* Note: useless operation, no need to draw */
    { 0x0b, "OR_INVERTED"	},
    { 0x0c, "COPY"		},
    { 0x0d, "OR_REVERSE"	},
    { 0x0e, "OR"		},
    { 0x0f, "SET"		},
    {}
};

static data_store nv10_rc_mapping[] = {
    { 0x00, "UNSIGNED_IDENTITY_NV"	},
    { 0x01, "UNSIGNED_INVERT_NV"	},
    { 0x02, "EXPAND_NORMAL_NV"		},
    { 0x03, "EXPAND_NEGATE_NV"		},
    { 0x04, "HALF_BIAS_NORMAL_NV"	},
    { 0x05, "HALF_BIAS_NEGATE_NV"	},
    { 0x06, "SIGNED_IDENTITY_NV"	},
    { 0x07, "SIGNED_NEGATE_NV"		},
    {}
};

static data_store nv10_rc_cu_alpha[] = {
    { 0x00, "BLUE"	},
    { 0x01, "ALPHA"	},
    {}
};

static data_store nv10_rc_cu_rgb[] = {
    { 0x00, "RGB"	},
    { 0x01, "ALPHA"	},
    {}
};

static data_store nv10_rc_input[] = {
    { 0x00, "ZERO"		},
    { 0x01, "CONSTANT_COLOR0_NV"},
    { 0x02, "CONSTANT_COLOR1_NV"},
    { 0x03, "FOG"		},
    { 0x04, "PRIMARY_COLOR_NV"	},
    { 0x05, "SECONDARY_COLOR_NV"},
    { 0x08, "TEXTURE1_ARB"	},
    { 0x09, "TEXTURE0_ARB"	},
    { 0x0c, "SPARE0_NV"		},
    { 0x0d, "SPARE1_NV"		},
    { 0x0e, "SPARE0_PLUS_SECONDARY_COLOR_NV"	},
    { 0x0f, "E_TIMES_F_NV"	},
    {}
};

static data_store nv10_rc_scale[] = {
    { 0x00, "NONE"		},
    { 0x01, "SCALE_BY_TWO_NV"	},
    { 0x02, "SCALE_BY_FOUR_NV"	},
    { 0x03, "SCALE_BY_ONE_HALF_NV"	},
    {}
};

static data_store nv10_rc_bias[] = {
    { 0x00, "NONE"		},
    { 0x01, "BIAS_BY_NEGATIVE_ONE_HALF_NV"	},
    {}
};

static data_store nv10_tx_format[] = {
	{ 0x00, "L8"		},
	{ 0x01, "A8"		},
	{ 0x02, "R5G5B5A1"	},
	{ 0x03, "A8_RECT"	},
	{ 0x04, "R4G4B4A4"	},
	{ 0x06, "R8G8B8A8"	},
	{ 0x0b, "INDEX8"	},
	{ 0x0c, "DXT1"		},
	{ 0x0e, "DXT3"		},
	{ 0x0f, "DXT5"		},
	{ 0x10, "R5G5B5A1_RECT"	},	/* NV20+ only for values >= 0x10 */
	{ 0x12, "R8G8B8A8_RECT"	},
	{ 0x13, "L8_RECT"	},
	{ 0x1a, "L8A8"		},
	{ 0x1b, "A8_RECT_2"	}, /* ??? */
	{ 0x1d, "R4G4B4A4_RECT"	},
	{ 0x20, "L8A8_RECT"	},
	{}
};

static data_store nv10_fog_mode[]={
    { 0x0800, "EXP"	},	/* is GL_EXP in GL/gl.h */
    { 0x0802, "EXP"	},	/* is not GL_EXP in GL/gl.h */
    { 0x0803, "EXP2"	},	/* is not GL_EXP2 in GL/gl.h */
    { 0x0804, "LINEAR"	},	/* is not GL_LINEAR in GL/gl.h */
    { 0x2601, "LINEAR"},	/* is GL_LINEAR in GL/gl.h */
    {}
};

static data_store nv10_fog_coord_dist[]={
    { 0x00, "coord_src = FOG_COORDINATE"	},
    { 0x01, "coord_src = FRAGMENT_DEPTH | distance_mode = EYE_RADIAL_NV"	},
    { 0x02, "coord_src = FRAGMENT_DEPTH | distance_mode = EYE_PLANE_ABSOLUTE_NV"		},
    { 0x03, "coord_src = FOG_COORDINATE | distance_mode = N/A"},
    {}
};

static data_store nv10_light_model_color_control[]={
    { 0x00, "SINGLE_COLOR"		},
    { 0x01, "SEPARATE_SPECULAR_COLOR"	},
    {}
};

static data_store vertex_data_type[]={
    { 0x01, "SHORT"	},
    { 0x02, "FLOAT"	},
    { 0x04, "BYTE"	},
    {}
};

static data_store nv10_tx_anisotropy[]={	/* nv10,nv20,nv30 */
    { 0x00, "1"	},
    { 0x01, "2"	},
    { 0x02, "4"	},	/* nv20+ */
    { 0x03, "8"	},
    {}
};

static data_store nv20_point_sprite_r_mode[]={
    { 0x00, "GL_ZERO"	},
    { 0x01, "GL_R"	},
    { 0x02, "GL_S"	},
    {}
};

static data_store nv40_tx_anisotropy[]={	/* nv40+ */
    { 0x00, "1"	},
    { 0x01, "2"	},
    { 0x02, "4"	},
    { 0x03, "NV40_6_NV30_8"	},
    { 0x04, "8"	},
    { 0x05, "10"	},
    { 0x06, "12"	},
    { 0x07, "16"	},
    {}
};

static data_store nv10_viewport_clip_mode[] = {
    { 0x00, "clip outside"	},
    { 0x01, "clip inside"	},
};

const char *
nv30_tx_swizzle_print(unsigned int value, void *data)
{
	static char buf[100]; snprintf(buf, sizeof(buf), "swizzle: ");
	char *bp = strlen(buf) + buf;
	int p;

	for (p=3;p>=0;p--) {
		int p0 = (p*2)+8;
		int p1 = (p*2)+0;
		switch ((value & (3 << p0)) >> p0) {
		case NV30_TCL_PRIMITIVE_3D_TX_SWIZZLE_S0_ZERO:
			bp[p] = '0';
			break;
		case NV30_TCL_PRIMITIVE_3D_TX_SWIZZLE_S0_ONE:
			bp[p] = '1';
			break;
		default:
			switch ((value & (3 << p1)) >> p1) {
			case NV30_TCL_PRIMITIVE_3D_TX_SWIZZLE_S1_X: bp[p] = 'x'; break;
			case NV30_TCL_PRIMITIVE_3D_TX_SWIZZLE_S1_Y: bp[p] = 'y'; break;
			case NV30_TCL_PRIMITIVE_3D_TX_SWIZZLE_S1_Z: bp[p] = 'z'; break;
			case NV30_TCL_PRIMITIVE_3D_TX_SWIZZLE_S1_W: bp[p] = 'w'; break;
			}
		}
	}
	return buf;
}

static data_store nv30_tx_type[] = {
	{ 0x0, "POT"	},
	{ 0x1, "NPOT"	},
	{ 0x3, "RECT"	},
	{}
};

static data_store nv30_tx_format[] = {
	{ 0x01, "L8"	},
	{ 0x02, "A1R5G5B5" },
	{ 0x03, "A4R4G4B4" },
	{ 0x04, "R5G6B5" },
	{ 0x05, "A8R8G8B8" },
	{ 0x06, "DXT1" },
	{ 0x07, "DXT3" },
	{ 0x08, "DXT5" },
	{ 0x14, "L16" },
	{ 0x15, "L16A16" },
	{ 0x18, "L8A8" },
	{ 0x19, "SL8A8" },
	{ 0x1a, "RGBA_F16" },
	{ 0x1b, "RGBA_F32" },
	{ 0x1c, "L_F32" },
	{ 0x1f, "LA_F16" },
	{}
};

static data_store nv30_tx_wrap[] = {
	{ 1,	"REPEAT"		},
	{ 2, 	"MIRRORED_REPEAT"	},
	{ 3,	"CLAMP_TO_EDGE"	},
	{ 4,	"CLAMP_TO_BORDER"	},
	{ 5,	"CLAMP"	},
	{}
};

static data_store nv30_tx_filter[] = {
	{ 1,	"NEAREST"		},
	{ 2,	"LINEAR"		},
	{ 3, 	"NEAREST_MIPMAP_NEAREST"},
	{ 4, 	"LINEAR_MIPMAP_NEAREST"	},
	{ 5, 	"NEAREST_MIPMAP_LINEAR"	},
	{ 6, 	"LINEAR_MIPMAP_LINEAR"	},
	{}
};

static data_store nv04_image_pattern_format[] = {
	{ 1,	"A16R5G6B5"		},
	{ 2, 	"X16A1R5G5B5"		},
	{ 3, 	"A8R8G8B8"		},
	{}
};

static data_store nv04_image_pattern_shape[] = {
	{ 0, "8x8"	},
	{ 1, "64x1"	},
	{ 2, "1x64"	},
	{}
};

static data_store nv04_image_pattern_select[] = {
	{ 0x01, "MONOCHROME"	},
	{ 0x02, "COLOR"		},
	{}
};


static data_store nv04_surface_format[] = {
	{ 0x01,	"Y8"			},
	{ 0x02,	"X1R5G5B5_Z1R5G5B5"	},
	{ 0x03,	"X1R5G5B5_O1R5G5B5"	},
	{ 0x04,	"R5G6B5"		},
	{ 0x05,	"Y16"			},
	{ 0x06,	"X8R8G8B8_Z8R8G8B8"	},
	{ 0x07,	"X8R8G8B8_O8R8G8B8"	},
	{ 0x08,	"X1A7R8G8B8_Z1A7R8G8B8"	},
	{ 0x09,	"X1A7R8G8B8_O1A7R8G8B8"	},
	{ 0x0a,	"A8R8G8B8"		},
	{ 0x0b,	"Y32"			},
	{}
};

static data_store nv04_context_surfaces_3d_cmft[] = {
	{ 0x01,	"X1R5G5B5_Z1R5G5B5"	},
	{ 0x02,	"X1R5G5B5_O1R5G5B5"	},
	{ 0x03,	"R5G6B5"		},
	{ 0x04,	"X8R8G8B8_Z8R8G8B8"	},
	{ 0x05,	"X8R8G8B8_O8R8G8B8"	},
	{ 0x06,	"X1A7R8G8B8_Z1A7R8G8B8"	},
	{ 0x07,	"X1A7R8G8B8_O1A7R8G8B8"	},
	{ 0x08,	"A8R8G8B8"		},
	{}
};

static data_store nv04_context_surfaces_3d_type[] = {
	{ 1, "pitch"			},
	{ 2, "swizzle"			},
	{}
};

static data_store nv04_dx5_func[] = {
	{ 0x01, "NEVER"			},
	{ 0x02, "LESS"			},
	{ 0x03, "EQUAL"			},
	{ 0x04, "LEQUAL"		},
	{ 0x05, "GREATER"		},
	{ 0x06, "NOTEQUAL"		},
	{ 0x07, "GEQUAL"		},
	{ 0x08, "ALWAYS"		},
	{}
};

static data_store nv04_dx5_bfunc[] = {
	{ 0x01,	"0"			},
	{ 0x02,	"1"			},
	{ 0x03, "src_col"		},
	{ 0x04, "inv_src_col"		},
	{ 0x05,	"src_a"			},
	{ 0x06,	"inv_src_a"		},
	{ 0x09,	"dst_col"		},
	{ 0x0a,	"inv_dst_col"		},
	{}
};

static data_store nv04_dx5_shademode[] = {
	{ 0x01, "FLAT"		},
	{ 0x02, "SMOOTH"	},
	{}
};

static data_store nv05_image_from_cpu_format[] = {
	{ 0x01, "R5G6B5"		},
	{ 0x02, "A1R5G5B5"	},
	{ 0x03, "X1R5G5B5"	},
	{ 0x04, "A8R8G8B8"	},
	{ 0x05, "X8R8G8B8"	},
	{}
};

static data_store nv04_dx5_wrap[] = {
	{ 0x01, "REPEAT"	},
	{ 0x02, "MIRRORED"	},
	{ 0x03, "CLAMP"		},
	{ 0x04, "CLAMP_BORDER"	},
	{ 0x08, "WRAP_ENABLE"	},
	{}
};

static data_store nv04_dx5_tx_color[] = {
	{ 0x01, "Y8"		},
	{ 0x02, "A1R5G5B5"	},
	{ 0x03, "X1R5G5B5"	},
	{ 0x04, "A4R4G4B4"	},
	{ 0x05, "R5G6B5"	},
	{ 0x06, "A8R8G8B8"	},
	{ 0x07, "X8R8G8B8"	},
	{}
};

static data_store nv04_dx5_textureblend[] = {
					   // RGBA	RGB
					   //
	{ 0x01, "DECAL"			}, // REPLACE
	{ 0x02, "MODULATE"		}, //
	{ 0x03, "DECAL_ALPHA"		}, // DECAL
	{ 0x04, "MODULATE_ALPHA"	}, // MODULATE	MODULATE
	{ 0x05, "DECAL_MASK"		}, //		DECAL,REPLACE
	{ 0x06, "MODULATE_MASK"		},
	{ 0x07, "COPY"			},
	{ 0x08, "ADD"			}, //		ADD
	{}
};

static data_store nv04_multitex_alpha[] = {
	{ 0x00, "a"	},
	{ 0x01, "1-a"	},
	// 0x02, 0x03 ???
	{}
};
static data_store nv04_multitex_color[] = {
	{ 0x00, "c"	},
	{ 0x01, "1-c"	},
	{ 0x02, "a"	},
	{ 0x03, "1-a"	},
	{}
};
static data_store nv04_multitex_arg[] = {
	{ 0x01, "ZERO"		},
	{ 0x02, "FACTOR"	},	// get from NV04_DX6_MULTITEX_TRIANGLE_COMBINE_FACTOR
	{ 0x03, "PRIMARY_COLOR"	},
	{ 0x04, "PREVIOUS"	},
	{ 0x05, "TEXTURE0"	},
	{ 0x06, "TEXTURE1"	},
	{ 0x07, "TEXTURELOD"	},	// ?
	{}
};
static data_store nv04_multitex_operation[] = {
	{ 0x01, "ADD"		},	// a0 * a1 + a2 * a3
	{ 0x02, "ADD2"		},
	{ 0x03, "ADD4"		},
	{ 0x04, "ADDSIGNED"	},
	{ 0x05, "MUX"		},	//?
	{ 0x06, "ADDCOMPLEMENT"	},	//?
	{ 0x07, "ADDSIGNED2"	},
	{}
};

static data_store nv04_scaled_image_from_memory_format[] = {
	{ 0x01, "A1R5G5B5"		},
	{ 0x02, "X1R5G5B5"		},
	{ 0x03, "A8R8G8B8"		},
	{ 0x04, "X8R8G8B8"		},
	{ 0x05, "V8YB8U8YA8"		},
	{ 0x06, "YB8V8YA8U8"		},
	{ 0x07, "R5G6B5"		},
	{}
};

// nv03 only has first three
static data_store nv04_operations[] = {
	{ 0x0, "SRCCOPY_AND"	},
	{ 0x1, "ROP_AND"	},
	{ 0x2, "BLEND_AND"	},
	{ 0x3, "SRCCOPY"	},
	{ 0x4, "SRCCOPY_PREMULT"},
	{ 0x5, "BLEND_PREMULT"	},
	{}
};

static data_store nv04_gdi_rectangle_text_format[] = {
	{ 0x01, "X16R5G6B5"	},
	{ 0x02, "X17R5G5B5"	},
	{ 0x03, "X8R8G8B8"	},
	{}
};

static data_store nv05_scaled_image_from_memory_conversion[] = {
	{ 0x00, "dither"	},
	{ 0x01, "truncate"	},
	{ 0x02, "subtr_truncate"},
	{}
};

static data_store nv_dac_format[] = {
	{ 0x00, "8 bpp" },
	{ 0x01, "16 bpp" },
	{ 0x02, "15 bpp" },
	{ 0x03, "24 bpp" },
	{}
};

static data_store nv04_dx5_cullmode[] = {
	{ 0x01, "NONE"	},
	{ 0x02, "CW"	},
	{ 0x03, "CCW"	},
	{}
};

static data_store nv04_stencilop[] = {
	{ 0x01, "KEEP"		},
	{ 0x02, "ZERO"		},
	{ 0x03, "REPLACE"	},
	{ 0x04, "INCR"		},
	{ 0x05, "DECR"		},
	{ 0x06, "INVERT"	},
	{ 0x07, "INCR_WRAP"	},
	{ 0x08, "DECR_WRAP"	},
	{}
};

static data_store nv04_scaled_image_from_memory_origin[] = {
	{ 0x01, "CENTER"},
	{ 0x02, "CORNER"},
	{}
};

static data_store nv04_scaled_image_from_memory_filter[] = {
	{ 0x01, "POINT_SAMPLE"},
	{ 0x02, "BILINEAR"},
	{}
};

static data_store nv_mono_format[] = {
	{ 0x01, "lowEndian"	},
	{ 0x02, "bigEndian"	},
	{}
};

static data_store nv_context_color_format[] = {
	{ 0x00, "INVALID"		},
	{ 0x01, "Y8"			},
	{ 0x02, "X16A8Y8"		},
	{ 0x03, "X24Y8"			},
	{ 0x06, "A1R5G5B5"		},
	{ 0x07, "X1R5G5B5"		},
	{ 0x08, "X16A1R5G5B5"		},
	{ 0x09, "X17R5G5B5"		},
	{ 0x0A, "R5G6B5"		},
	{ 0x0B, "A16R5G6B5"		},
	{ 0x0C, "X16R5G6B5"		},
	{ 0x0D, "A8R8G8B8"		},
	{ 0x0E, "X8R8G8B8"		},
	{ 0x0F, "Y16"			},
	{ 0x10, "A16Y16"		},
	{ 0x11, "X16Y16"		},
	{ 0x12, "V8YB8U8YA8"		},
	{ 0x13, "YB8V8YA8U8"		},
	{ 0x14, "Y32"			},
	{}
};

static data_store nv20_tx_shader_op[] = {
	{ 0x00, "NONE"		},
	{ 0x01, "TEXTURE_2D" 	},	// maybe any texture???

	{ 0x04, "PASS_THROUGH"	},
	{ 0x05, "CULL_FRAGMENT"	},

	{ 0x09, "DOT_PRODUCT_TEXTURE_2D" },
	{ 0x0f, "DEPENDANT_AR"	},
	{ 0x10, "DEPENDANT_GB"	},
	{ 0x11, "DOT_PRODUCT"	},
	{}
};

static data_store nv20_tx_shader_cull_mode[] = {
	{ 0, "gggg"	},
	{ 1, "lggg"	},
	{ 2, "glgg"	},
	{ 3, "llgg"	},
	{ 4, "gglg"	},
	{ 5, "lglg"	},
	{ 6, "gllg"	},
	{ 7, "lllg"	},
	{ 8, "gggl"	},
	{ 9, "lggl"	},
	{ 10, "glgl"	},
	{ 11, "llgl"	},
	{ 12, "ggll"	},
	{ 13, "lgll"	},
	{ 14, "glll"	},
	{ 15, "llll"	},
	{}
};

const object_store nv_objects[] = {
	{N_(NV01_CONTEXT_CLIP_RECTANGLE), NV03|NV04|NV10|NV15|NV20|NV40|G70, 0,0, {
		{__(NV01_CONTEXT_CLIP_RECTANGLE_NOP),	{PRINT_X32,	{}} },
		{__(NV01_CONTEXT_CLIP_RECTANGLE_NOTIFY),	{PRINT_X32,	{}} },
		{__(NV01_CONTEXT_CLIP_RECTANGLE_DMA_NOTIFY),{PRINT_OBJECT,	{}} },
		{__(NV01_CONTEXT_CLIP_RECTANGLE_SET_POINT),	{
			PRINT_BITFIELD(NV01_CONTEXT_CLIP_RECTANGLE_SET_POINT_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV01_CONTEXT_CLIP_RECTANGLE_SET_POINT_Y, print_format, "%d", "y"),
			{}} },
		{__(NV01_CONTEXT_CLIP_RECTANGLE_SET_SIZE),	{
			PRINT_BITFIELD(NV01_CONTEXT_CLIP_RECTANGLE_SET_SIZE_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV01_CONTEXT_CLIP_RECTANGLE_SET_SIZE_HEIGHT, print_format, "%d", "height"),
			{}} },
		{}},
	},

	{N_(NV_MEMORY_TO_MEMORY_FORMAT), NV04|NV10|NV15|NV20|NV30|NV40|G70|G80, 0,0, {
		{__(NV_MEMORY_TO_MEMORY_FORMAT_NOP),		{             {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_NOTIFY),		{             {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_DMA_NOTIFY),	{ PRINT_OBJECT, {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_OBJECT_IN),	{ PRINT_OBJECT, {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_OBJECT_OUT),	{ PRINT_OBJECT, {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_OFFSET_IN),	{ PRINT_FUNC(record_offset_in), {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_OFFSET_OUT),	{ PRINT_FUNC(record_offset_out), {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_PITCH_IN),	{ PRINT_UINT, {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_PITCH_OUT),	{ PRINT_UINT, {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_LINE_LENGTH_IN),	{ PRINT_FUNC(record_length_in), {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_LINE_COUNT),	{ PRINT_FUNC(record_dma_op), {}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_FORMAT),		{
			PRINT_BITFIELD( 2:0, print_format, "%d", "src_inc"),
			PRINT_BITFIELD(10:8, print_format, "%d", "dst_inc"),
			{}} },
		{__(NV_MEMORY_TO_MEMORY_FORMAT_BUF_NOTIFY),	{ PRINT_X32, {}} },
		{}},
	},

	{N_(NV03_PRIMITIVE_RASTER_OP), NV03|NV04|NV10|NV15|NV20|NV30|NV40|G70, 0,0, {
		{__(NV03_PRIMITIVE_RASTER_OP_NOP),	{PRINT_X32,	{}} },
		{__(NV03_PRIMITIVE_RASTER_OP_NOTIFY),	{PRINT_X32,	{}} },
		{__(NV03_PRIMITIVE_RASTER_OP_DMA_NOTIFY),{PRINT_OBJECT,	{}} },
		{__(NV03_PRIMITIVE_RASTER_OP_LOGIC_OP),	{
			PRINT_BITFIELD(NV03_PRIMITIVE_RASTER_OP_LOGIC_OP_CMD, print_macro, gl_logic_op, "logic_op"),
			{}} },
		{}},
	},

	{N_(NV04_GDI_RECTANGLE_TEXT), NV04|NV10|NV15|NV20|NV30|NV40|G70, 0,0, {
		{__(NV04_GDI_RECTANGLE_TEXT_NOP),         {PRINT_X32, {}} },
		{__(NV04_GDI_RECTANGLE_TEXT_NOTIFY),         {PRINT_X32, {}} },
		{__(NV04_GDI_RECTANGLE_TEXT_SET_DMA_NOTIFY), {PRINT_OBJECT, {}}},
		{__(NV04_GDI_RECTANGLE_TEXT_SET_DMA_FONTS), {PRINT_OBJECT, {}}},
		{__(NV04_GDI_RECTANGLE_TEXT_PATTERN), {PRINT_OBJECT, {}}},
		{__(NV04_GDI_RECTANGLE_TEXT_ROP5), {PRINT_OBJECT, {}}},
		{__(NV04_GDI_RECTANGLE_TEXT_SURFACE), {PRINT_OBJECT, {}}},
		{__(NV04_GDI_RECTANGLE_TEXT_OPERATION), {PRINT_MACRO(nv04_operations), {}}},
		{__(NV04_GDI_RECTANGLE_TEXT_FORMAT), {PRINT_MACRO(nv04_gdi_rectangle_text_format), {}}},
		{__(NV04_GDI_RECTANGLE_TEXT_MONO_FORMAT),	{PRINT_MACRO(nv_mono_format),	{}}},
		{__(NV04_GDI_RECTANGLE_TEXT_BLOCK_LEVEL1_TL),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "left"),
			PRINT_BITFIELD(31:16, print_format, "%d", "top"),
			{}} },
		{__(NV04_GDI_RECTANGLE_TEXT_BLOCK_LEVEL1_BR),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "right"),
			PRINT_BITFIELD(31:16, print_format, "%d", "bottom"),
			{}} },
		{__(NV04_GDI_RECTANGLE_TEXT_FILL_VALUE),		{PRINT_X32,	{}} },
		{__(NV04_GDI_RECTANGLE_TEXT_BLOCK_LEVEL2_TL),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "left"),
			PRINT_BITFIELD(31:16, print_format, "%d", "top"),
			{}} },
		{__(NV04_GDI_RECTANGLE_TEXT_BLOCK_LEVEL2_BR),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "right"),
			PRINT_BITFIELD(31:16, print_format, "%d", "bottom"),
			{}} },
		{}},
	},

	{N_(NV04_SWIZZLED_SURFACE), NV04|NV10|NV15, 0,0, {
		{__(NV04_SWIZZLED_SURFACE_DMA_NOTIFY),	{PRINT_OBJECT,	{}} },
		{__(NV04_SWIZZLED_SURFACE_DMA_IMAGE),	{PRINT_OBJECT,	{}} },
		{__(NV04_SWIZZLED_SURFACE_FORMAT),	{
			PRINT_BITFIELD(NV04_SWIZZLED_SURFACE_FORMAT_HEIGHT, print_format, "%d", "log2_height"),
			PRINT_BITFIELD(NV04_SWIZZLED_SURFACE_FORMAT_WIDTH, print_format, "%d", "log2_width"),
			PRINT_BITFIELD(NV04_SWIZZLED_SURFACE_FORMAT_COLOR, print_macro, nv04_surface_format, "color"),
			{}} },
		/* offset in the fb memory */
		{__(NV04_SWIZZLED_SURFACE_OFFSET),			{PRINT_X32,	{}} },
		{}},
	},

	{N_(NV04_CONTEXT_SURFACES_3D), NV04, 0,0, {
		{__(NV04_CONTEXT_SURFACES_3D_DMA_NOTIFY),	{PRINT_OBJECT,	{}} },
		{__(NV04_CONTEXT_SURFACES_3D_DMA_COLOR),	{PRINT_OBJECT,	{}} },
		{__(NV04_CONTEXT_SURFACES_3D_DMA_ZETA),		{PRINT_OBJECT,	{}} },
		{__(NV04_CONTEXT_SURFACES_3D_CLIP_HORIZONTAL),	{
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_CLIP_HORIZONTAL_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_CLIP_HORIZONTAL_WIDTH, print_format, "%d", "width"),
			{}} },
		{__(NV04_CONTEXT_SURFACES_3D_CLIP_VERTICAL),	{
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_CLIP_VERTICAL_Y, print_format, "%d", "y"),
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_CLIP_VERTICAL_HEIGHT, print_format, "%d", "height"),
			{}} },
		{__(NV04_CONTEXT_SURFACES_3D_FORMAT),		{
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_FORMAT_COLOR, print_macro, nv04_context_surfaces_3d_cmft, "color"),
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_FORMAT_TYPE, print_macro, nv04_context_surfaces_3d_type, "type"),
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_FORMAT_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_FORMAT_HEIGHT, print_format, "%d", "height"),
			{}} },
		{__(NV04_CONTEXT_SURFACES_3D_CLIP_SIZE),	{
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_CLIP_SIZE_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_CLIP_SIZE_HEIGHT, print_format, "%d", "height"),
			{}} },
		{__(NV04_CONTEXT_SURFACES_3D_PITCH),		{
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_PITCH_COLOR, print_format, "%d", "color"),
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_3D_PITCH_ZETA, print_format, "%d", "zeta"),
			{}} },
		{__(NV04_CONTEXT_SURFACES_3D_OFFSET_COLOR),	{PRINT_X32,		{}} },
		{__(NV04_CONTEXT_SURFACES_3D_OFFSET_ZETA),	{PRINT_X32,		{}} },
		{}},
	},

#define NV04_TEXTURE_FORMAT	\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FORMAT_COLOR, print_macro, nv04_dx5_tx_color, "color"),	\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FORMAT_MIPMAPS, print_format, "%d", "mipmaps"),		\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FORMAT_SIZE_U, print_format, "%d", "log_u"),			\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FORMAT_SIZE_V, print_format, "%d", "log_v"),			\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FORMAT_WRAP_S, print_macro, nv04_dx5_wrap, "wrap_s"),		\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FORMAT_WRAP_T, print_macro, nv04_dx5_wrap, "wrap_t"),		\
			{}

#define NV04_TEXTURE_FILTER	\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FILTER_MAG, print_macro, nv30_tx_filter, "magfilter"),	\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FILTER_MIN, print_macro, nv30_tx_filter, "minfilter"),	\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FILTER_LODBIAS, print_format, "%d", "lodbias"),		\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FILTER_DITHER, print_enable, "dither", NULL),		\
			{}
#define NV04_TRIANGLE_BLEND	\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_BLEND_ENABLE, print_macro, bool_data, "benable"),			\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_BLEND_DST, print_macro, nv04_dx5_bfunc, "dst"),			\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_BLEND_SRC, print_macro, nv04_dx5_bfunc, "src"),			\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_BLEND_FOG, print_enable, "fog", NULL),				\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_BLEND_SPECULAR, print_enable, "specular", NULL),				\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_BLEND_TEXTUREPERSPECTIVE, print_macro, bool_data, "tex_perspective"),	\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_BLEND_SHADEMODE, print_macro, nv04_dx5_shademode, "shademode"),	\
			{}
#define NV04_TRIANGLE_CONTROL	\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_CONTROL_ALPHAREF, print_format, "0x%02x", "alpharef"),		\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_CONTROL_ALPHAFUNC, print_macro, nv04_dx5_func, "alphafunc"),		\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_CONTROL_ALPHATEST, print_macro, bool_data, "alphaenable"),		\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_CONTROL_ZENABLE, print_macro, bool_data, "zenable"),			\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_CONTROL_ZWRITE, print_macro, bool_data, "zwrite"),			\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_CONTROL_ZFUNC, print_macro, nv04_dx5_func, "zfunc"),			\
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_CONTROL_CULLMODE, print_macro, nv04_dx5_cullmode, "cullmode"),	\
			{}
	{N_(NV04_DX5_TEXTURED_TRIANGLE), NV04, 0,0, {
		{__(NV04_DX5_TEXTURED_TRIANGLE_NOP),		{		{}} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_NOTIFY),		{		{}} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_DMA_NOTIFY),	{PRINT_OBJECT,	{}} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_DMA_1),		{PRINT_OBJECT,	{}} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_DMA_2),		{PRINT_OBJECT,	{}} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_SURFACE),	{PRINT_OBJECT,	{}} },

		{__(NV04_DX5_TEXTURED_TRIANGLE_COLOR_KEY),	{PRINT_X32,	{}} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_OFFSET),	{PRINT_X32,	{}} },

		{__(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FORMAT),	{
			// 0xa1
			NV04_TEXTURE_FORMAT
			} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_FILTER),	{
			// 0x1010
			NV04_TEXTURE_FILTER
			} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_BLEND),		{
			PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_BLEND_TEXTURE, print_macro, nv04_dx5_textureblend, "texture"),
			// 0x20
			NV04_TRIANGLE_BLEND
			} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_CONTROL),	{
			// 0x40002000
			NV04_TRIANGLE_CONTROL
			} },
		{__(NV04_DX5_TEXTURED_TRIANGLE_FOG_COLOR),	{PRINT_X32,	{}} },

		{A_(NV04_DX5_TEXTURED_TRIANGLE_TLVERTEX_SX),	{PRINT_FP32,	{}} },
		{A_(NV04_DX5_TEXTURED_TRIANGLE_TLVERTEX_SY),	{PRINT_FP32,	{}} },
		{A_(NV04_DX5_TEXTURED_TRIANGLE_TLVERTEX_SZ),	{PRINT_FP32,	{}} },
		{A_(NV04_DX5_TEXTURED_TRIANGLE_INV_W),		{PRINT_FP32,	{}} },
		{A_(NV04_DX5_TEXTURED_TRIANGLE_COLOR),		{PRINT_X32,	{}} },
		{A_(NV04_DX5_TEXTURED_TRIANGLE_SPECULAR),	{PRINT_X32,	{}} },
		{A_(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_S),	{PRINT_FP32,	{}} },
		{A_(NV04_DX5_TEXTURED_TRIANGLE_TEXTURE_T),	{PRINT_FP32,	{}} },

		{__(NV04_DX5_TEXTURED_TRIANGLE_DRAW), {
				PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_DRAW_V0, print_format, "%d", "v0"),
				PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_DRAW_V1, print_format, "%d", "v1"),
				PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_DRAW_V2, print_format, "%d", "v2"),
				PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_DRAW_V3, print_format, "%d", "v3"),
				PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_DRAW_V4, print_format, "%d", "v4"),
				PRINT_BITFIELD(NV04_DX5_TEXTURED_TRIANGLE_DRAW_V5, print_format, "%d", "v5"),
									      	{}} },
		{}},
	},

	{N_(NV04_DX6_MULTITEX_TRIANGLE), NV04|NV10|NV15, 0,0, {
		{__(NV04_DX6_MULTITEX_TRIANGLE_NOP),		{		{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_NOTIFY),		{		{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_DMA_NOTIFY),	{PRINT_OBJECT,	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_DMA_1),		{PRINT_OBJECT,	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_DMA_2),		{PRINT_OBJECT,	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_SURFACE),	{PRINT_OBJECT,	{}} },

		{__(NV04_DX6_MULTITEX_TRIANGLE_OFFSET0),	{PRINT_X32,	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_OFFSET1),	{PRINT_X32,	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_FORMAT0),	{
			NV04_TEXTURE_FORMAT
			} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_FORMAT1),	{
			// 0xa1
			NV04_TEXTURE_FORMAT
			} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_FILTER0),	{
			// 0x1010
			NV04_TEXTURE_FILTER
			} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_FILTER1),	{
			// 0x1010
			NV04_TEXTURE_FILTER
			} },
#define NV04_MULTITEX_COMBINE_UNIT(a)	\
		PRINT_BITFIELD(31:29, print_macro, nv04_multitex_operation, NULL),	\
		PRINT_BITFIELD(1:0, print_macro, nv04_multitex_ ## a, NULL),	\
		PRINT_BITFIELD(4:2, print_macro, nv04_multitex_arg, "arg0"),	\
		PRINT_BITFIELD(9:8, print_macro, nv04_multitex_ ## a, NULL),	\
		PRINT_BITFIELD(12:10, print_macro, nv04_multitex_arg, "arg1"),	\
		PRINT_BITFIELD(17:16, print_macro, nv04_multitex_ ## a, NULL),	\
		PRINT_BITFIELD(20:18, print_macro, nv04_multitex_arg, "arg2"),	\
		PRINT_BITFIELD(25:24, print_macro, nv04_multitex_ ## a, NULL),	\
		PRINT_BITFIELD(28:26, print_macro, nv04_multitex_arg, "arg3"),	\
		{}
		{__(NV04_DX6_MULTITEX_TRIANGLE_COMBINE_0_ALPHA),	{NV04_MULTITEX_COMBINE_UNIT(alpha),	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_COMBINE_0_COLOR),	{NV04_MULTITEX_COMBINE_UNIT(color),	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_COMBINE_1_ALPHA),	{NV04_MULTITEX_COMBINE_UNIT(alpha),	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_COMBINE_1_COLOR),	{NV04_MULTITEX_COMBINE_UNIT(color),	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_COMBINE_FACTOR),		{PRINT_X32,	{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_BLEND),		{
			// 0x20
			NV04_TRIANGLE_BLEND
			} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_CONTROL0),	{
			// 40002000
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL0_R_WRITE, print_macro, bool_data, "red_write"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL0_G_WRITE, print_macro, bool_data, "green_write"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL0_B_WRITE, print_macro, bool_data, "blue_write"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL0_A_WRITE, print_macro, bool_data, "alpha_write"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL0_A_WRITE, print_macro, bool_data, "alpha_write"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL0_S_WRITE, print_macro, bool_data, "stencil_write"),
			NV04_TRIANGLE_CONTROL
			} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_CONTROL1),	{
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL1_STENCILENABLE, print_macro, bool_data, "stencil_enable"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL1_STENCIL_MASK_WRITE, print_format, "0x%02x", "stencil_mask_write"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL1_STENCIL_MASK_READ, print_format, "0x%02x", "stencil_mask_read"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL1_STENCILREF, print_format, "0x%02x", "stencilref"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL1_STENCILFUNC, print_macro, nv04_dx5_func, "stencilfunc"),
			{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_CONTROL2),	{
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL2_STENCIL_FAIL, print_macro, nv04_stencilop, "stencil_fail"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL2_STENCIL_ZFAIL, print_macro, nv04_stencilop, "stencil_zfail"),
			PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_CONTROL2_STENCIL_ZPASS, print_macro, nv04_stencilop, "stencil_zpass"),
			{}} },
		{__(NV04_DX6_MULTITEX_TRIANGLE_FOG_COLOR),	{PRINT_X32,	{}} },

		{A_(NV04_DX6_MULTITEX_TRIANGLE_TLVERTEX_SX),	{PRINT_FP32,	{}} },
		{A_(NV04_DX6_MULTITEX_TRIANGLE_TLVERTEX_SY),	{PRINT_FP32,	{}} },
		{A_(NV04_DX6_MULTITEX_TRIANGLE_TLVERTEX_SZ),	{PRINT_FP32,	{}} },
		{A_(NV04_DX6_MULTITEX_TRIANGLE_INV_W),	{PRINT_FP32,	{}} },	\
		{A_(NV04_DX6_MULTITEX_TRIANGLE_COLOR),	{PRINT_X32,	{}} },	\
		{A_(NV04_DX6_MULTITEX_TRIANGLE_SPECULAR),	{PRINT_X32,	{}} },
		{A_(NV04_DX6_MULTITEX_TRIANGLE_TEXTURE0_S),	{PRINT_FP32,	{}} },
		{A_(NV04_DX6_MULTITEX_TRIANGLE_TEXTURE0_T),	{PRINT_FP32,	{}} },
		{A_(NV04_DX6_MULTITEX_TRIANGLE_TEXTURE1_S),	{PRINT_FP32,	{}} },
		{A_(NV04_DX6_MULTITEX_TRIANGLE_TEXTURE1_T),	{PRINT_FP32,	{}} },

		{__(NV04_DX6_MULTITEX_TRIANGLE_DRAW), {
				PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_DRAW_V0, print_format, "%d", "v0"),
				PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_DRAW_V1, print_format, "%d", "v1"),
				PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_DRAW_V2, print_format, "%d", "v2"),
				PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_DRAW_V3, print_format, "%d", "v3"),
				PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_DRAW_V4, print_format, "%d", "v4"),
				PRINT_BITFIELD(NV04_DX6_MULTITEX_TRIANGLE_DRAW_V5, print_format, "%d", "v5"),
									      	{}} },
		{}},
	},

	{N_(NV04_COLOR_KEY), NV04|NV10|NV15|NV20|NV40, 0,0, {
		{__(NV04_COLOR_KEY_SET_DMA_NOTIFY),	{PRINT_OBJECT,	{}} },
		{__(NV04_COLOR_KEY_FORMAT), {PRINT_MACRO(nv04_image_pattern_format),	{}} },
		{__(NV04_COLOR_KEY_VALUE),  {PRINT_X32,					{}} },
		{}},
	},

	{N_(NV04_SOLID_LINE), NV04, 0,0, {
		{__(NV04_SOLID_LINE_CLIP_RECTANGLE),			{PRINT_OBJECT,	{}} },
		{__(NV04_SOLID_LINE_PATTERN),				{PRINT_OBJECT,	{}} },
		{__(NV04_SOLID_LINE_ROP),				{PRINT_OBJECT,	{}} },
		{__(NV04_SOLID_LINE_SURFACE),				{PRINT_OBJECT,	{}} },
		{__(NV04_SOLID_LINE_OPERATION),				{PRINT_MACRO(nv04_operations),	{}} },
		{__(NV04_SOLID_LINE_COLOR_FORMAT),			{PRINT_MACRO(nv04_gdi_rectangle_text_format),	{}} },
		{__(NV04_SOLID_LINE_COLOR_VALUE),			{PRINT_X32,	{}} },
		{__(NV04_SOLID_LINE_START), {
			PRINT_BITFIELD(NV04_SOLID_LINE_START_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV04_SOLID_LINE_START_Y, print_format, "%d", "y"),
			{}} },
		{__(NV04_SOLID_LINE_END), {
			PRINT_BITFIELD(NV04_SOLID_LINE_END_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV04_SOLID_LINE_END_Y, print_format, "%d", "y"),
			{}} },
		{}},
	},

	{N_(NV04_UNK005E), NV04, 0,0, {
		{__(NV04_UNK005E_SET_SURFACE),				{PRINT_OBJECT,	{}} },
		{__(NV04_UNK005E_UNK02fc),				{PRINT_X32,	{}} },	/* 3 */
		{__(NV04_UNK005E_UNK0300),				{PRINT_X32,	{}} },	/* 3 */
		{__(NV04_UNK005E_COUNTER),				{		{}} },
		// 0x400 = 0x8
		// 0x404 = 0x00010001
		{}}
	},

	{N_(NV05_SCALED_IMAGE_FROM_MEMORY), NV04, 0,0, {
		{__(NV05_SCALED_IMAGE_FROM_MEMORY_SURFACE),		{PRINT_OBJECT,	{}} },
		{__(NV05_SCALED_IMAGE_FROM_MEMORY_COLOR_CONVERSION),	{PRINT_MACRO(nv05_scaled_image_from_memory_conversion), {}} },
		{__(NV05_SCALED_IMAGE_FROM_MEMORY_OPERATION),		{PRINT_MACRO(nv04_operations),	{}} },
		{}},
	},

	{N_(NV04_SCALED_IMAGE_FROM_MEMORY), NV04, 0,0, {
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_NOP),         {PRINT_X32, {}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_NOTIFY),         {PRINT_X32, {}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_DMA_NOTIFY),		{PRINT_OBJECT,	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_DMA_IMAGE),		{PRINT_OBJECT,	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_PATTERN),		{PRINT_OBJECT,	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_ROP),			{PRINT_OBJECT,	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_BETA1),		{PRINT_OBJECT,	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_BETA4),		{PRINT_OBJECT,	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_SURFACE),		{PRINT_OBJECT,	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_COLOR_FORMAT),		{PRINT_MACRO(nv04_scaled_image_from_memory_format),	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_OPERATION),		{PRINT_MACRO(nv04_operations),	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_CLIP_POS),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
											{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_CLIP_SIZE),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "width"),
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
											{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_OUT_POS),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
											{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_OUT_SIZE),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "width"),
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
											{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_DU_DX),		{
			PRINT_BITFIELD(31:20, print_format, "%d", "int"),
			PRINT_BITFIELD(19:0, print_format, "%d", "frac_mul_0x100000"),
											{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_DV_DY),		{
			PRINT_BITFIELD(31:20, print_format, "%d", "int"),
			PRINT_BITFIELD(19:0, print_format, "%d", "frac_mul_0x100000"),
											{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_SIZE),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "width"),
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
											{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_FORMAT),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "pitch"), 
			PRINT_BITFIELD(23:16, print_macro, nv04_scaled_image_from_memory_origin, "origin"), 
			PRINT_BITFIELD(31:24, print_macro, nv04_scaled_image_from_memory_filter, "filter"), 
											{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_OFFSET),		{ PRINT_X32,	{}} },
		{__(NV04_SCALED_IMAGE_FROM_MEMORY_POINT),		{
			PRINT_BITFIELD(31:20, print_format, "%d", "u_int"),
			PRINT_BITFIELD(19:16, print_format, "%d", "u_frac_mul_0x10"),
			PRINT_BITFIELD(15:4, print_format, "%d", "v_int"),
			PRINT_BITFIELD(3:0, print_format, "%d", "v_frac_mul_0x10"),
											{}} },
		{}},
	},

	{N_(NV_IMAGE_FROM_CPU), NV04, 0,0, {
		{__(NV_IMAGE_FROM_CPU_DMA_NOTIFY),		{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_FROM_CPU_CLIP_RECTANGLE),	{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_FROM_CPU_PATTERN),		{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_FROM_CPU_ROP),			{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_FROM_CPU_SURFACE),		{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_FROM_CPU_OPERATION),		{PRINT_MACRO(nv04_operations),	{}}},
		{__(NV_IMAGE_FROM_CPU_FORMAT),		{PRINT_MACRO(nv05_image_from_cpu_format), {}}},
		{}},
	},

	{N_(NV05_IMAGE_FROM_CPU), NV04, 0,0, {
		{__(NV05_IMAGE_FROM_CPU_DMA_NOTIFY),		{PRINT_OBJECT, {}}},
		{__(NV05_IMAGE_FROM_CPU_CLIP_RECTANGLE),	{PRINT_OBJECT, {}}},
		{__(NV05_IMAGE_FROM_CPU_PATTERN),		{PRINT_OBJECT, {}}},
		{__(NV05_IMAGE_FROM_CPU_ROP),			{PRINT_OBJECT, {}}},
		{__(NV05_IMAGE_FROM_CPU_BETA1),			{PRINT_OBJECT, {}}},
		{__(NV05_IMAGE_FROM_CPU_BETA4),			{PRINT_OBJECT, {}}},
		{__(NV05_IMAGE_FROM_CPU_SURFACE),		{PRINT_OBJECT, {}}},
		{__(NV05_IMAGE_FROM_CPU_OPERATION),		{PRINT_MACRO(nv04_operations),	{}}},
		{__(NV05_IMAGE_FROM_CPU_FORMAT),		{PRINT_MACRO(nv05_image_from_cpu_format), {}}},
		{__(NV05_IMAGE_FROM_CPU_POINT),			{
			PRINT_BITFIELD(NV05_IMAGE_FROM_CPU_POINT_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV05_IMAGE_FROM_CPU_POINT_Y, print_format, "%d", "y"),
			{}} },
		{__(NV05_IMAGE_FROM_CPU_SIZE_OUT),			{
			PRINT_BITFIELD(NV05_IMAGE_FROM_CPU_SIZE_OUT_WIDTH, print_format, "%d", "x"),
			PRINT_BITFIELD(NV05_IMAGE_FROM_CPU_SIZE_OUT_HEIGHT, print_format, "%d", "y"),
			{}} },
		{__(NV05_IMAGE_FROM_CPU_SIZE_IN),			{
			PRINT_BITFIELD(NV05_IMAGE_FROM_CPU_SIZE_IN_WIDTH, print_format, "%d", "x"),
			PRINT_BITFIELD(NV05_IMAGE_FROM_CPU_SIZE_IN_HEIGHT, print_format, "%d", "y"),
			{}} },
		{A_(NV05_IMAGE_FROM_CPU_COLOR),	{PRINT_X32, {}}},
		{}},
	},

	{N_(NV_IMAGE_BLIT), NV04|NV10|NV15|NV20, 0,0, {
		{__(NV_IMAGE_BLIT_SET_DMA_NOTIFY),		{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_BLIT_SET_COLOR_KEY),		{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_BLIT_SET_CLIP_RECTANGLE),	{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_BLIT_SET_PATTERN),		{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_BLIT_SET_ROP5),		{PRINT_OBJECT, {}}},
		{__(NV_IMAGE_BLIT_SET_SURFACES_2D),		{PRINT_OBJECT, {}}},

		{__(NV_IMAGE_BLIT_OPERATION),		{PRINT_MACRO(nv04_operations), {}}},

		{__(NV_IMAGE_BLIT_POINT_IN),		{
			PRINT_BITFIELD(NV_IMAGE_BLIT_POINT_IN_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV_IMAGE_BLIT_POINT_IN_Y, print_format, "%d", "y"),
			{}} },
		{__(NV_IMAGE_BLIT_POINT_OUT),		{
			PRINT_BITFIELD(NV_IMAGE_BLIT_POINT_OUT_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV_IMAGE_BLIT_POINT_OUT_Y, print_format, "%d", "y"),
			{}} },
		{__(NV_IMAGE_BLIT_SIZE),		{
			PRINT_BITFIELD(NV_IMAGE_BLIT_SIZE_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV_IMAGE_BLIT_SIZE_HEIGHT, print_format, "%d", "height"),
			{}} },
		{}},
	},

	{N_(NV11_IMAGE_BLIT), NV15|NV20, NV_IMAGE_BLIT,NV10, {
		/* 0x120 */
		/* 0x124 */
		/* 0x128 */
		{}},
	},

	{N_(NV30_IMAGE_BLIT), NV30|NV40|G70, NV11_IMAGE_BLIT,NV15, {
		/* 0x194 */
		/* 0x198 */
		{}},
	},

	{N_(NV10_TCL_PRIMITIVE_3D), NV10, 0,0, {
		{__(NV10_TCL_PRIMITIVE_3D_NOP),			{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_NOTIFY),		{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_SET_DMA_NOTIFY),	{PRINT_OBJECT,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_SET_DMA_IN_MEMORY0),	{PRINT_OBJECT,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_SET_DMA_IN_MEMORY1),	{PRINT_OBJECT,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_SET_DISPLAY_LIST),	{PRINT_OBJECT,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_SET_DMA_IN_MEMORY2),	{PRINT_OBJECT,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_SET_DMA_IN_MEMORY3),	{PRINT_OBJECT,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_HORIZ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "width"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_VERT),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
			PRINT_BITFIELD(15:0, print_format, "%d", "y"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_BUFFER_FORMAT),		{
			PRINT_BITFIELD(15:8, print_macro, nv04_context_surfaces_3d_type, "type"),
			PRINT_BITFIELD(7:0, print_macro, nv04_context_surfaces_3d_cmft, "color"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_BUFFER_PITCH),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "depth/stencil buffer pitch"),
			PRINT_BITFIELD(15:0, print_format, "%d", "color buffer pitch"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_OFFSET),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_DEPTH_OFFSET),	{PRINT_X32,	{}} },

#define NV10_TCL_PRIMITIVE_3D_TX_UNIT(n)	\
		{__(NV10_TCL_PRIMITIVE_3D_TX_OFFSET(n)),	{PRINT_X32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_TX_FORMAT(n)),	{	\
			PRINT_BITFIELD(31:28, print_macro, nv30_tx_wrap, "wrap_t"),	\
			PRINT_BITFIELD(27:24, print_macro, nv30_tx_wrap, "wrap_s"),	\
			PRINT_BITFIELD(23:20, print_format, "%d", "log2_height"),	\
			PRINT_BITFIELD(19:16, print_format, "%d", "log2_width"),	\
			PRINT_BITFIELD(15:12, print_format, "%d", "lod"),	\
			PRINT_BITFIELD(11:11, print_macro, bool_data, "npot"),	\
			PRINT_BITFIELD(10:7, print_macro, nv10_tx_format, "format"),	\
			PRINT_BITFIELD(2:2, print_macro, bool_data, "cube_map"),	\
			{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_TX_ENABLE(n)),	{	\
			PRINT_BITFIELD(30:30, print_macro, bool_data, "enable"),	\
			PRINT_BITFIELD(5:4, print_macro, nv10_tx_anisotropy, "anisotropy"),	\
			{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_TX_NPOT_PITCH(n)),	{	\
			PRINT_BITFIELD(31:16, print_format, "%d", "pitch"),	\
			{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_TX_NPOT_SIZE(n)),	{	\
			PRINT_BITFIELD(31:16, print_format, "%d", "width"),	\
			PRINT_BITFIELD(15:0, print_format, "%d", "height"),	\
			{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_TX_FILTER(n)),	{	\
			PRINT_BITFIELD(31:28, print_macro, nv30_tx_filter, "mag_filter"),	\
			PRINT_BITFIELD(27:24, print_macro, nv30_tx_filter, "min_filter"),	\
			{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_TX_PALETTE_OFFSET(n)),	{PRINT_X32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_TX_MATRIX_ENABLE(n)),	{PRINT_MACRO(bool_data),{}} },	\
		{M_(NV10_TCL_PRIMITIVE_3D_TX_MATRIX(n)),	{PRINT_FP32,	{}} },

		NV10_TCL_PRIMITIVE_3D_TX_UNIT(0)
		NV10_TCL_PRIMITIVE_3D_TX_UNIT(1)

#define NV10_TCL_PRIMITIVE_3D_RC_UNIT(n)	\
		{__(NV10_TCL_PRIMITIVE_3D_RC_IN_ALPHA(n)),	{	\
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vara_mapping"),	\
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_alpha, "vara_component_usage"),	\
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vara_input"),	\
			\
			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varb_mapping"),	\
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_alpha, "varb_component_usage"),	\
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varb_input"),	\
			\
			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varc_mapping"),	\
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_alpha, "varc_component_usage"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varc_input"),	\
			\
			PRINT_BITFIELD(7:5, print_macro, nv10_rc_mapping, "vard_mapping"),	\
			PRINT_BITFIELD(4:4, print_macro, nv10_rc_cu_alpha, "vard_component_usage"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "vard_input"),	\
			{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_RC_IN_RGB(n)),	{	\
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vara_mapping"),	\
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_rgb, "vara_component_usage"),	\
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vara_input"),	\
			\
			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varb_mapping"),	\
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_rgb, "varb_component_usage"),	\
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varb_input"),	\
			\
			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varc_mapping"),	\
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_rgb, "varc_component_usage"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varc_input"),	\
			\
			PRINT_BITFIELD(7:5, print_macro, nv10_rc_mapping, "vard_mapping"),	\
			PRINT_BITFIELD(4:4, print_macro, nv10_rc_cu_rgb, "vard_component_usage"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "vard_input"),	\
			{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_RC_OUT_ALPHA(n)),	{	\
			PRINT_BITFIELD(17:16, print_macro, nv10_rc_scale, "scale"),	\
			PRINT_BITFIELD(15:15, print_macro, nv10_rc_bias, "bias"),	\
			PRINT_BITFIELD(14:14, print_macro, bool_data, "mux_sum"),	\
			PRINT_BITFIELD(13:13, print_macro, bool_data, "ab_dot_product"),	\
			PRINT_BITFIELD(12:12, print_macro, bool_data, "cd_dot_product"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "sum_output"),	\
			PRINT_BITFIELD(7:4, print_macro, nv10_rc_input, "ab_output"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "cd_output"),	\
			{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_RC_OUT_RGB(n)),	{	\
			PRINT_BITFIELD(29:28, print_format, "%d", "rc1_tx_units_enabled"),	\
			PRINT_BITFIELD(27:27, print_macro, bool_data, "rc1_rc_enabled"),	\
			PRINT_BITFIELD(17:16, print_macro, nv10_rc_scale, "scale"),	\
			PRINT_BITFIELD(15:15, print_macro, nv10_rc_bias, "bias"),	\
			PRINT_BITFIELD(14:14, print_macro, bool_data, "mux_sum"),	\
			PRINT_BITFIELD(13:13, print_macro, bool_data, "ab_dot_product"),	\
			PRINT_BITFIELD(12:12, print_macro, bool_data, "cd_dot_product"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "sum_output"),	\
			PRINT_BITFIELD(7:4, print_macro, nv10_rc_input, "ab_output"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "cd_output"),	\
			{}} },

		NV10_TCL_PRIMITIVE_3D_RC_UNIT(0)
		NV10_TCL_PRIMITIVE_3D_RC_UNIT(1)

		{__(NV10_TCL_PRIMITIVE_3D_RC_COLOR0),		{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_RC_COLOR1),		{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_RC_FINAL0),		{
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vara_mapping"),
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_rgb, "vara_component_usage"),
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vara_input"),

			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varb_mapping"),
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_rgb, "varb_component_usage"),
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varb_input"),

			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varc_mapping"),
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_rgb, "varc_component_usage"),
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varc_input"),

			PRINT_BITFIELD(7:5, print_macro, nv10_rc_mapping, "vard_mapping"),
			PRINT_BITFIELD(4:4, print_macro, nv10_rc_cu_rgb, "vard_component_usage"),
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "vard_input"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_RC_FINAL1),		{
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vare_mapping"),
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_rgb, "vare_component_usage"),
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vare_input"),

			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varf_mapping"),
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_rgb, "varf_component_usage"),
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varf_input"),

			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varg_mapping"),
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_rgb, "varg_component_usage"),
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varg_input"),

			PRINT_BITFIELD(7:7, print_macro, bool_data, "color_sum_clamp"),
			{}} },
		/* 0x290 */
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_MODEL),		{
			PRINT_BITFIELD(16:16, print_macro, bool_data, "local_viewer"),
			PRINT_BITFIELD(1:1, print_macro, nv10_light_model_color_control, "color_control"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_MATERIAL_ENABLE),	{
			PRINT_BITFIELD(3:3, print_macro, bool_data, "specular"),
			PRINT_BITFIELD(2:2, print_macro, bool_data, "diffuse"),
			PRINT_BITFIELD(1:1, print_macro, bool_data, "ambient"),
			PRINT_BITFIELD(0:0, print_macro, bool_data, "emission"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FOG_MODE),		{PRINT_MACRO(nv10_fog_mode),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FOG_COORD_DIST),	{PRINT_MACRO(nv10_fog_coord_dist),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FOG_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FOG_COLOR),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_MODE),	{PRINT_MACRO(nv10_viewport_clip_mode),	{}} },

#define NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(n)	\
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_HORIZ(n)),	{	\
			PRINT_BITFIELD(27:27, print_macro, bool_data, "enable clip at x2"),	\
			PRINT_BITFIELD(26:16, print_format, "%d", "x2"),	\
			PRINT_BITFIELD(11:11, print_macro, bool_data, "enable clip at x1"),	\
			PRINT_BITFIELD(10:0, print_format, "%d", "x1"),	\
			{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_VERT(n)),	{	\
			PRINT_BITFIELD(27:27, print_macro, bool_data, "enable clip at y2"),	\
			PRINT_BITFIELD(26:16, print_format, "%d", "y2"),	\
			PRINT_BITFIELD(11:11, print_macro, bool_data, "enable clip at y1"),	\
			PRINT_BITFIELD(10:0, print_format, "%d", "y1"),	\
			{}} },

		NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(0)
		NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(1)
		NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(2)
		NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(3)
		NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(4)
		NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(5)
		NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(6)
		NV10_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(7)

		{__(NV10_TCL_PRIMITIVE_3D_ALPHA_FUNC_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_BLEND_FUNC_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_CULL_FACE_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_DEPTH_TEST_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_DITHER_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_LIGHTING_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_PARAMETERS_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_SMOOTH_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_LINE_SMOOTH_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POLYGON_SMOOTH_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_WEIGHT_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_STENCIL_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POLYGON_OFFSET_POINT_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POLYGON_OFFSET_LINE_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POLYGON_OFFSET_FILL_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_ALPHA_FUNC_FUNC),		{PRINT_MACRO(gl_comparison_op_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_ALPHA_FUNC_REF),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_BLEND_FUNC_SRC),	{PRINT_MACRO(gl_blend_func_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_BLEND_FUNC_DST),	{PRINT_MACRO(gl_blend_func_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_BLEND_COLOR),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_BLEND_EQUATION),	{PRINT_MACRO(gl_blend_eqn_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_DEPTH_FUNC),		{PRINT_MACRO(gl_comparison_op_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_MASK),		{
			PRINT_BITFIELD(31:24, print_macro, bool_data, "a"),
			PRINT_BITFIELD(23:16, print_macro, bool_data, "r"),
			PRINT_BITFIELD(15:8, print_macro, bool_data, "g"),
			PRINT_BITFIELD(7:0, print_macro, bool_data, "b"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_DEPTH_WRITE_ENABLE),	{PRINT_MACRO(bool_data),{}} },
		{__(NV10_TCL_PRIMITIVE_3D_STENCIL_MASK),	{PRINT_X32,		{}} },
		{__(NV10_TCL_PRIMITIVE_3D_STENCIL_FUNC_FUNC),	{PRINT_MACRO(gl_comparison_op_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_STENCIL_FUNC_REF),	{PRINT_X32,		{}} },
		{__(NV10_TCL_PRIMITIVE_3D_STENCIL_FUNC_MASK),	{PRINT_X32,		{}} },
		{__(NV10_TCL_PRIMITIVE_3D_STENCIL_OP_FAIL),	{PRINT_MACRO(gl_stencil_mod_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_STENCIL_OP_ZFAIL),	{PRINT_MACRO(gl_stencil_mod_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_STENCIL_OP_ZPASS),	{PRINT_MACRO(gl_stencil_mod_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_SHADE_MODEL),		{PRINT_MACRO(gl_shade_model_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_LINE_WIDTH),		{PRINT_UINT,		{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POLYGON_OFFSET_FACTOR),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POLYGON_OFFSET_UNITS),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POLYGON_MODE_FRONT),	{PRINT_MACRO(gl_polygon_mode_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POLYGON_MODE_BACK),	{PRINT_MACRO(gl_polygon_mode_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_DEPTH_RANGE_NEAR),	{PRINT_FP32,		{}} },
		{__(NV10_TCL_PRIMITIVE_3D_DEPTH_RANGE_FAR),	{PRINT_FP32,		{}} },
		{__(NV10_TCL_PRIMITIVE_3D_CULL_FACE),		{PRINT_MACRO(gl_face_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FRONT_FACE),		{PRINT_MACRO(gl_front_face_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_NORMALIZE_ENABLE),	{PRINT_MACRO(bool_data),{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_MATERIAL_R),	{PRINT_FP32,{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_MATERIAL_G),	{PRINT_FP32,{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_MATERIAL_B),	{PRINT_FP32,{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_MATERIAL_A),	{PRINT_FP32,{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_CONTROL),	{
			PRINT_BITFIELD(0:0, print_macro, nv10_light_model_color_control, "color_control"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_ENABLED_LIGHTS),	{
			PRINT_BITFIELD(14:14, print_macro, bool_data, "light 7"),
			PRINT_BITFIELD(12:12, print_macro, bool_data, "light 6"),
			PRINT_BITFIELD(10:10, print_macro, bool_data, "light 5"),
			PRINT_BITFIELD(8:8, print_macro, bool_data, "light 4"),
			PRINT_BITFIELD(6:6, print_macro, bool_data, "light 3"),
			PRINT_BITFIELD(4:4, print_macro, bool_data, "light 2"),
			PRINT_BITFIELD(2:2, print_macro, bool_data, "light 1"),
			PRINT_BITFIELD(0:0, print_macro, bool_data, "light 0"),
			{}} },
		{A_(NV10_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE),	{PRINT_MACRO(gl_texgen_mode_data),	{}} }, /*wtf? */
		{__(NV10_TCL_PRIMITIVE_3D_VIEW_MATRIX_ENABLE),	{
			PRINT_BITFIELD(2:2, print_macro, bool_data, "projection"),
			PRINT_BITFIELD(1:1, print_macro, bool_data, "modelview0"),
			PRINT_BITFIELD(0:0, print_macro, bool_data, "modelview1"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_SIZE),		{PRINT_X32,		{}} },
		/* 0x3f4 */
		{M_(NV10_TCL_PRIMITIVE_3D_MODELVIEW0_MATRIX),	{PRINT_FP32,	{}} },
		{M_(NV10_TCL_PRIMITIVE_3D_MODELVIEW1_MATRIX),	{PRINT_FP32,	{}} },
		{M_(NV10_TCL_PRIMITIVE_3D_INVERSE_MODELVIEW0_MATRIX),		{PRINT_FP32,	{}} },
		{M_(NV10_TCL_PRIMITIVE_3D_INVERSE_MODELVIEW1_MATRIX),		{PRINT_FP32,	{}} },
		{M_(NV10_TCL_PRIMITIVE_3D_PROJECTION_MATRIX),	{PRINT_FP32,	{}} },

#define NV10_TCL_PRIMITIVE_3D_CLIP_PLANE(n)	\
		{__(NV10_TCL_PRIMITIVE_3D_CLIP_PLANE_A(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_CLIP_PLANE_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_CLIP_PLANE_C(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_CLIP_PLANE_D(n)),	{PRINT_FP32,	{}} },

		NV10_TCL_PRIMITIVE_3D_CLIP_PLANE(0)
		NV10_TCL_PRIMITIVE_3D_CLIP_PLANE(1)
		NV10_TCL_PRIMITIVE_3D_CLIP_PLANE(2)
		NV10_TCL_PRIMITIVE_3D_CLIP_PLANE(3)
		NV10_TCL_PRIMITIVE_3D_CLIP_PLANE(4)
		NV10_TCL_PRIMITIVE_3D_CLIP_PLANE(5)
		NV10_TCL_PRIMITIVE_3D_CLIP_PLANE(6)
		NV10_TCL_PRIMITIVE_3D_CLIP_PLANE(7)

		{__(NV10_TCL_PRIMITIVE_3D_FOG_EQUATION_CONSTANT),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FOG_EQUATION_LINEAR),		{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FOG_EQUATION_QUADRATIC),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_A),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_B),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_C),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_D),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_E),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_F),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_MODEL_FRONT_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_R),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_MODEL_FRONT_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_G),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_MODEL_FRONT_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_B),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_SCALE_X),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_SCALE_Y),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_SCALE_Z),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_SCALE_W),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_PARAMETER_A),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_PARAMETER_B),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_PARAMETER_C),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_PARAMETER_D),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_PARAMETER_E),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_PARAMETER_F),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_PARAMETER_G),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_POINT_PARAMETER_H),	{PRINT_FP32,	{}} },

#define NV10_TCL_PRIMITIVE_3D_LIGHT(n)	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_AMBIENT_R(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_AMBIENT_G(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_AMBIENT_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_DIFFUSE_R(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_DIFFUSE_G(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_DIFFUSE_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_SPECULAR_R(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_SPECULAR_G(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_SPECULAR_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_HALF_VECTOR_X(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_HALF_VECTOR_Y(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_HALF_VECTOR_Z(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_DIRECTION_X(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_DIRECTION_Y(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_DIRECTION_Z(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_SPOT_CUTOFF_A(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_SPOT_EXPONENT(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_SPOT_CUTOFF_B(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_SPOT_DIR_X(n)),				{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_SPOT_DIR_Y(n)),				{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_SPOT_DIR_Z(n)),				{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_SPOT_CUTOFF_C(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_POSITION_X(n)),				{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_POSITION_Y(n)),				{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_POSITION_Z(n)),				{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_CONSTANT_ATTENUATION(n)),		{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_LINEAR_ATTENUATION(n)),			{PRINT_FP32,	{}} },	\
		{__(NV10_TCL_PRIMITIVE_3D_LIGHT_QUADRATIC_ATTENUATION(n)),		{PRINT_FP32,	{}} },

		NV10_TCL_PRIMITIVE_3D_LIGHT(0)
		NV10_TCL_PRIMITIVE_3D_LIGHT(1)
		NV10_TCL_PRIMITIVE_3D_LIGHT(2)
		NV10_TCL_PRIMITIVE_3D_LIGHT(3)
		NV10_TCL_PRIMITIVE_3D_LIGHT(4)
		NV10_TCL_PRIMITIVE_3D_LIGHT(5)
		NV10_TCL_PRIMITIVE_3D_LIGHT(6)
		NV10_TCL_PRIMITIVE_3D_LIGHT(7)

		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_POS_3F_X),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_POS_3F_Y),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_POS_3F_Z),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_POS_4F_X),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_POS_4F_Y),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_POS_4F_Z),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_POS_4F_W),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_NOR_3F_X),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_NOR_3F_Y),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_NOR_3F_Z),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_NOR_3I_XY),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_NOR_3I_Z),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "z"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL_4F_R),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL_4F_G),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL_4F_B),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL_4F_A),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL_3F_R),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL_3F_G),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL_3F_B),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL_4I),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL2_3F_R),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL2_3F_G),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL2_3F_B),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_COL2_3I),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX0_2F_S),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX0_2F_T),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX0_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_S),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_T),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_R),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_Q),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX0_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX0_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX1_2F_S),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX1_2F_T),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX1_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_S),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_T),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_R),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_Q),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX1_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_TX1_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_FOG_1F),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_WGH_1F),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_EDGEFLAG_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{A_(NV10_TCL_PRIMITIVE_3D_VERTEX_ATTR),		{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_VALIDATE),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_OFFSET_POS),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_FORMAT_POS),	{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_OFFSET_COL),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_FORMAT_COL),	{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_OFFSET_COL2),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_FORMAT_COL2),	{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_OFFSET_TX0),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_FORMAT_TX0),	{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_OFFSET_TX1),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_FORMAT_TX1),	{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_OFFSET_NOR),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_FORMAT_NOR),	{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_OFFSET_WGH),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_FORMAT_WGH),	{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_OFFSET_FOG),	{PRINT_X32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_FORMAT_FOG),	{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_LOGIC_OP_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_COLOR_LOGIC_OP_OP),		{PRINT_MACRO(gl_logic_op_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_BEGIN_END),	{PRINT_MACRO(begin_end_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_INDEX_DATA),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "index1"),
			PRINT_BITFIELD(15:0, print_format, "%d", "index0"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_BUFFER_BEGIN_END),	{PRINT_MACRO(begin_end_data),	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_BUFFER_DRAW_ARRAYS),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "count_minus_1"),
			PRINT_BITFIELD(15:0, print_format, "%d", "first"),
			{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_ORIGIN_X),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_ORIGIN_Y),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_ORIGIN_Z),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VIEWPORT_ORIGIN_W),	{PRINT_FP32,	{}} },
		{__(NV10_TCL_PRIMITIVE_3D_VERTEX_ARRAY_DATA),	{PRINT_FP32,	{}} },
		{}},
	},
	{N_(NV11_TCL_PRIMITIVE_3D), NV15, NV10_TCL_PRIMITIVE_3D,NV10, {
		/* 0x120 */
		/* 0x124 */
		/* 0x128 */
		/* 0x3f0 */
		{}},
	},
	{N_(NV17_TCL_PRIMITIVE_3D), NV15, NV11_TCL_PRIMITIVE_3D,NV15, {
		{__(NV17_TCL_PRIMITIVE_3D_SET_DMA_IN_MEMORY4),		{PRINT_OBJECT,	{}} },
		{__(NV17_TCL_PRIMITIVE_3D_SET_DMA_IN_MEMORY5),		{PRINT_OBJECT,	{}} },
		/* 0x258 */
		/* 0x25c */
		{__(NV17_TCL_PRIMITIVE_3D_COLOR_MASK_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		/* 0x3f8 */
		/* 0x3fc */
		{__(NV17_TCL_PRIMITIVE_3D_LMA_DEPTH_BUFFER_PITCH),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "pitch"),
			{}} },
		{__(NV17_TCL_PRIMITIVE_3D_LMA_DEPTH_BUFFER_OFFSET),	{PRINT_X32,	{}} },
		{__(NV17_TCL_PRIMITIVE_3D_LMA_DEPTH_FILL_VALUE),	{PRINT_X32,	{}} },
		{__(NV17_TCL_PRIMITIVE_3D_LMA_DEPTH_CLEAR_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		/* 0xd74 */
		/* 0xd84 */
		{__(NV17_TCL_PRIMITIVE_3D_LMA_DEPTH_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{}},
	},

	{N_(NV10_IMAGE_FROM_CPU), NV10|NV15|NV20|NV30|NV40|G70, 0,0, {
		{__(NV10_IMAGE_FROM_CPU_SET_DMA_NOTIFY),		{PRINT_OBJECT,	{}}},
		{__(NV10_IMAGE_FROM_CPU_SET_CONTEXT_CLIP_RECTANGLE),	{PRINT_OBJECT,	{}}},
		{__(NV10_IMAGE_FROM_CPU_SET_IMAGE_PATTERN),		{PRINT_OBJECT,	{}}},
		{__(NV10_IMAGE_FROM_CPU_SET_RASTER_OP),			{PRINT_OBJECT,	{}}},
		{__(NV10_IMAGE_FROM_CPU_SET_CONTEXT_SURFACES_2D),	{PRINT_OBJECT,	{}}},
		{__(NV10_IMAGE_FROM_CPU_OPERATION),	{PRINT_MACRO(nv04_operations),	{}}},
		{__(NV10_IMAGE_FROM_CPU_FORMAT),	{PRINT_MACRO(nv05_image_from_cpu_format), {}}},
		{__(NV10_IMAGE_FROM_CPU_POINT),	{
			PRINT_BITFIELD(NV10_IMAGE_FROM_CPU_POINT_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV10_IMAGE_FROM_CPU_POINT_Y, print_format, "%d", "y"),
			{}} },
		{__(NV10_IMAGE_FROM_CPU_SIZE_OUT),			{
			PRINT_BITFIELD(NV10_IMAGE_FROM_CPU_SIZE_OUT_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV10_IMAGE_FROM_CPU_SIZE_OUT_HEIGHT, print_format, "%d", "height"),
			{}} },
		{__(NV10_IMAGE_FROM_CPU_SIZE_IN),			{
			PRINT_BITFIELD(NV10_IMAGE_FROM_CPU_SIZE_IN_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV10_IMAGE_FROM_CPU_SIZE_IN_HEIGHT, print_format, "%d", "height"),
			{}} },
		{__(NV10_IMAGE_FROM_CPU_HLINE),	{PRINT_X32,	{}} },
		{}},
	},

	{N_(NV10_PRIMITIVE_2D), NV10|NV15|NV20|NV30|NV40|G70, 0,0, {
		{__(NV10_PRIMITIVE_2D_SET_DMA_NOTIFY),	{PRINT_OBJECT,	{}} },
		{__(NV10_PRIMITIVE_2D_SET_SURFACE),		{PRINT_OBJECT,	{}} },
		{__(NV10_PRIMITIVE_2D_SET_FORMAT),	{PRINT_MACRO(nv05_image_from_cpu_format),	{}} },
		{__(NV10_PRIMITIVE_2D_SET_POINT),	{
			PRINT_BITFIELD(NV10_PRIMITIVE_2D_SET_POINT_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV10_PRIMITIVE_2D_SET_POINT_Y, print_format, "%d", "y"),
			{}} },
		{__(NV10_PRIMITIVE_2D_SET_SIZE),	{
			PRINT_BITFIELD(NV10_PRIMITIVE_2D_SET_SIZE_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV10_PRIMITIVE_2D_SET_SIZE_HEIGHT, print_format, "%d", "height"),
			{}} },
		{__(NV10_PRIMITIVE_2D_SET_CLIP_HORIZ),	{
			PRINT_BITFIELD(NV10_PRIMITIVE_2D_SET_CLIP_HORIZ_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV10_PRIMITIVE_2D_SET_CLIP_HORIZ_X, print_format, "%d", "x"),
			{}} },
		{__(NV10_PRIMITIVE_2D_SET_CLIP_VERT),	{
			PRINT_BITFIELD(NV10_PRIMITIVE_2D_SET_CLIP_VERT_HEIGHT, print_format, "%d", "height"),
			PRINT_BITFIELD(NV10_PRIMITIVE_2D_SET_CLIP_VERT_Y, print_format, "%d", "y"),
			{}} },
		{A_(NV10_PRIMITIVE_2D_SET_DATA),	{PRINT_X32,	{}} },
		{}},
	},

	{N_(NV10_VIDEO_DISPLAY), NV10|NV15|NV20|NV30|NV40|G70, 0,0, {
		{__(NV10_VIDEO_DISPLAY_COUNTER),	{PRINT_X32,	{}} },
		{__(NV10_VIDEO_DISPLAY_SET_DMA_FROM_MEMORY),	{PRINT_OBJECT,	{}} },
		{__(NV10_VIDEO_DISPLAY_SET_DMA_IN_MEMORY0),	{PRINT_OBJECT,	{}} },
		{__(NV10_VIDEO_DISPLAY_SET_DMA_IN_MEMORY1),	{PRINT_OBJECT,	{}} },
		{__(NV10_VIDEO_DISPLAY_SET_OBJECT3),		{PRINT_OBJECT,	{}} },
		{__(NV10_VIDEO_DISPLAY_SIZE),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
			PRINT_BITFIELD(15:0, print_format, "%d", "width"),
			{}} },
		{__(NV10_VIDEO_DISPLAY_OFFSET),		{PRINT_X32,	{}} },
		{}},
	},

	{N_(NV10_UNK0072), NV10|NV15|NV20|NV40|G70|G80, 0,0, {
		{__(NV10_UNK0072_COUNTER),		{PRINT_X32,	{}} },
		{__(NV40_UNK0072_SET_OBJECT),		{PRINT_OBJECT,	{}} },
		{__(NV10_UNK0072_SET_DMA_NOTIFY),	{PRINT_OBJECT,	{}} },
		{}},
	},

	{N_(NV10_SCALED_IMAGE_FROM_MEMORY), NV10|NV15|NV20|NV30|NV40|G70, 0,0, {
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_COUNTER),		{PRINT_X32,	{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_SET_DMA_IN_MEMORY),	{PRINT_OBJECT,	{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_SET_RASTER_OP),	{PRINT_OBJECT,	{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_SET_IMAGE_PATTERN),	{PRINT_OBJECT,	{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_SET_SURFACE),		{PRINT_OBJECT,	{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_OPERATION),		{PRINT_MACRO(nv04_operations),	{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_CLIP_POS),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
			{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_CLIP_SIZE),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "width"),
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
			{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_OUT_POS),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
			{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_OUT_SIZE),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "width"),
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
			{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_SIZE),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "width"),
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
			{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_FORMAT),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "pitch"), 
			{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_OFFSET),		{ PRINT_X32,	{}} },
		{__(NV10_SCALED_IMAGE_FROM_MEMORY_POINT),		{
			PRINT_BITFIELD(31:20, print_format, "%d", "u_int"),
			PRINT_BITFIELD(19:16, print_format, "%d", "u_frac_mul_0x10"),
			PRINT_BITFIELD(15:4, print_format, "%d", "v_int"),
			PRINT_BITFIELD(3:0, print_format, "%d", "v_frac_mul_0x10"),
			{}} },
		{}},
	},

	{N_(NV10_CONTEXT_SURFACES_2D), NV10|NV15|NV20|NV30|NV40|G70, 0,0, {
		{__(NV10_CONTEXT_SURFACES_2D_SET_DMA_NOTIFY),	{PRINT_OBJECT,	{}} },
		{__(NV10_CONTEXT_SURFACES_2D_SET_DMA_IN_MEMORY0),	{PRINT_OBJECT,	{}} },
		{__(NV10_CONTEXT_SURFACES_2D_SET_DMA_IN_MEMORY1),	{PRINT_OBJECT,	{}} },
		{__(NV10_CONTEXT_SURFACES_2D_FORMAT),		{
			PRINT_BITFIELD(NV10_CONTEXT_SURFACES_2D_FORMAT_COLOR, print_macro, nv04_surface_format, "color"),
			PRINT_BITFIELD(NV10_CONTEXT_SURFACES_2D_FORMAT_TYPE, print_macro, nv04_context_surfaces_3d_type, "type"),
			PRINT_BITFIELD(NV10_CONTEXT_SURFACES_2D_FORMAT_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV10_CONTEXT_SURFACES_2D_FORMAT_HEIGHT, print_format, "%d", "height"),
			{}} },
		{__(NV10_CONTEXT_SURFACES_2D_PITCH),		{
			PRINT_BITFIELD(NV10_CONTEXT_SURFACES_2D_PITCH_SRC, print_format, "%d", "src"),
			PRINT_BITFIELD(NV10_CONTEXT_SURFACES_2D_PITCH_DST, print_format, "%d", "dst"),
			{}} },
		{__(NV10_CONTEXT_SURFACES_2D_OFFSET_SRC),	{PRINT_X32,		{}} },
		{__(NV10_CONTEXT_SURFACES_2D_OFFSET_DST),	{PRINT_X32,		{}} },
		{}},
	},

	{N_(NV04_CONTEXT_SURFACES_2D), NV04|NV10|NV15, 0,0, {
		{__(NV04_CONTEXT_SURFACES_2D_NOTIFY),		{PRINT_X32,		{}} },
		{__(NV04_CONTEXT_SURFACES_2D_SET_DMA_NOTIFY),		{PRINT_OBJECT,		{}} },
		{__(NV04_CONTEXT_SURFACES_2D_SET_DMA_IMAGE_SRC),	{PRINT_OBJECT,		{}} },
		{__(NV04_CONTEXT_SURFACES_2D_SET_DMA_IMAGE_DST),	{PRINT_OBJECT,		{}} },
		{__(NV04_CONTEXT_SURFACES_2D_FORMAT),		{PRINT_MACRO(nv04_surface_format), {}} },
		{__(NV04_CONTEXT_SURFACES_2D_PITCH),		{
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_2D_PITCH_SRC, print_format, "%d", "src"),
			PRINT_BITFIELD(NV04_CONTEXT_SURFACES_2D_PITCH_DST, print_format, "%d", "dst"),
			{}} },
		{__(NV04_CONTEXT_SURFACES_2D_OFFSET_SRC),	{PRINT_X32,		{}} },
		{__(NV04_CONTEXT_SURFACES_2D_OFFSET_DST),	{PRINT_X32,		{}} },
		{}},
	},

	{N_(NV04_IMAGE_PATTERN), NV04|NV10|NV15|NV20|NV30|NV40|G70, 0,0, {
		{__(NV04_IMAGE_PATTERN_COLOR_FORMAT),	{PRINT_MACRO(nv04_image_pattern_format), {}}},
		{__(NV04_IMAGE_PATTERN_MONO_FORMAT),	{PRINT_MACRO(nv_mono_format),	{}}},
		{__(NV04_IMAGE_PATTERN_SELECT),		{PRINT_MACRO(nv04_image_pattern_select), {}}},
		{__(NV04_IMAGE_PATTERN_MONOCHROME_SHAPE),	{PRINT_MACRO(nv04_image_pattern_shape), 	{}}},
		{__(NV04_IMAGE_PATTERN_MONOCHROME_COLOR0),	{PRINT_X32, 	{}}},
		{__(NV04_IMAGE_PATTERN_MONOCHROME_COLOR1),	{PRINT_X32, 	{}}},
		{__(NV04_IMAGE_PATTERN_MONOCHROME_PATTERN0),	{PRINT_X32, 	{}}},
		{__(NV04_IMAGE_PATTERN_MONOCHROME_PATTERN1),	{PRINT_X32, 	{}}},
		{}},
	},


	{N_(NV20_SWIZZLED_SURFACE), NV20|NV30|NV40|G70, 0,0, {
		{__(NV20_SWIZZLED_SURFACE_SET_OBJECT0),	{PRINT_OBJECT,	{}} },
		{__(NV20_SWIZZLED_SURFACE_SET_OBJECT1),	{PRINT_OBJECT,	{}} },
		{__(NV20_SWIZZLED_SURFACE_FORMAT),	{
			PRINT_BITFIELD(NV20_SWIZZLED_SURFACE_FORMAT_HEIGHT, print_format, "%d", "log2_height"),
			PRINT_BITFIELD(NV20_SWIZZLED_SURFACE_FORMAT_WIDTH, print_format, "%d", "log2_width"),
			PRINT_BITFIELD(NV20_SWIZZLED_SURFACE_FORMAT_COLOR, print_macro, nv04_surface_format, "color"),
			{}} },
		/* offset in the fb memory */
		{__(NV20_SWIZZLED_SURFACE_OFFSET),			{PRINT_X32,	{}} },
		{}},
	},


	{N_(NV20_TCL_PRIMITIVE_3D), NV20, 0,0, {
		{__(NV20_TCL_PRIMITIVE_3D_NOP),			{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_NOTIFY),		{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT0),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT1),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT2),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT3),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT4),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT5),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT6),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT7),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT8),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT9),		{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SET_OBJECT10),	{PRINT_OBJECT,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_HORIZ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "width"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_VERT),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
			PRINT_BITFIELD(15:0, print_format, "%d", "y"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_BUFFER_FORMAT),			{
			PRINT_BITFIELD(15:8, print_macro, nv04_context_surfaces_3d_type, "type"),
			PRINT_BITFIELD(7:0, print_macro, nv04_context_surfaces_3d_cmft, "color"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_BUFFER_PITCH),			{
			PRINT_BITFIELD(31:16, print_format, "%d", "zs_pitch"),
			PRINT_BITFIELD(15:0, print_format, "%d", "color_pitch"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_OFFSET),			{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_DEPTH_OFFSET),			{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LMA_DEPTH_BUFFER_PITCH),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "pitch"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LMA_DEPTH_BUFFER_OFFSET),		{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_CONTROL),			{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MATERIAL_CONTROL),		{
			PRINT_BITFIELD(14:14, print_macro, bool_data, "back_specular"),
			PRINT_BITFIELD(12:12, print_macro, bool_data, "back_ambient"),
			PRINT_BITFIELD(10:10, print_macro, bool_data, "back_diffuse"),
			PRINT_BITFIELD(8:8, print_macro, bool_data, "back_emission"),
			PRINT_BITFIELD(6:6, print_macro, bool_data, "front_specular"),
			PRINT_BITFIELD(4:4, print_macro, bool_data, "front_ambient"),
			PRINT_BITFIELD(2:2, print_macro, bool_data, "front_diffuse"),
			PRINT_BITFIELD(0:0, print_macro, bool_data, "front_emission"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FOG_MODE),				{PRINT_MACRO(nv10_fog_mode),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FOG_COORD_DIST),			{PRINT_MACRO(nv10_fog_coord_dist),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FOG_ENABLE),				{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FOG_COLOR),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_MODE),			{PRINT_MACRO(nv10_viewport_clip_mode),	{}} },

#define NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(n)	\
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_HORIZ(n)),	{	\
			PRINT_BITFIELD(31:16, print_format, "%d", "x2"),	\
			PRINT_BITFIELD(15:0, print_format, "%d", "x1"),	\
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_VERT(n)),	{	\
			PRINT_BITFIELD(31:16, print_format, "%d", "y2"),	\
			PRINT_BITFIELD(15:0, print_format, "%d", "y1"),	\
			{}} },

		NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(0)
		NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(1)
		NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(2)
		NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(3)
		NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(4)
		NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(5)
		NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(6)
		NV20_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(7)

		{__(NV20_TCL_PRIMITIVE_3D_ALPHA_FUNC_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_BLEND_FUNC_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CULL_FACE_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_DEPTH_TEST_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_DITHER_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LIGHTING_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_SMOOTH_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_PARAMETERS_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LINE_SMOOTH_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_SMOOTH_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_STENCIL_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_OFFSET_POINT_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_OFFSET_LINE_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_OFFSET_FILL_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_ALPHA_FUNC_FUNC),			{PRINT_MACRO(gl_comparison_op_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_ALPHA_FUNC_REF),			{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_BLEND_FUNC_SRC),			{PRINT_MACRO(gl_blend_func_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_BLEND_FUNC_DST),			{PRINT_MACRO(gl_blend_func_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_BLEND_COLOR),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_BLEND_EQUATION),			{PRINT_MACRO(gl_blend_eqn_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_DEPTH_FUNC),				{PRINT_MACRO(gl_comparison_op_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MASK),		{
			PRINT_BITFIELD(31:24, print_macro, bool_data, "a"),
			PRINT_BITFIELD(23:16, print_macro, bool_data, "r"),
			PRINT_BITFIELD(15:8, print_macro, bool_data, "g"),
			PRINT_BITFIELD(7:0, print_macro, bool_data, "b"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_DEPTH_WRITE_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_STENCIL_MASK),			{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_STENCIL_FUNC_FUNC),			{PRINT_MACRO(gl_comparison_op_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_STENCIL_FUNC_REF),			{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_STENCIL_FUNC_MASK),			{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_STENCIL_OP_FAIL),			{PRINT_MACRO(gl_stencil_mod_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_STENCIL_OP_ZFAIL),			{PRINT_MACRO(gl_stencil_mod_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_STENCIL_OP_ZPASS),			{PRINT_MACRO(gl_stencil_mod_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SHADE_MODEL),				{PRINT_MACRO(gl_shade_model_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LINE_WIDTH),				{PRINT_X32,			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_OFFSET_FACTOR),		{PRINT_FP32,			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_OFFSET_UNITS),		{PRINT_FP32,			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_MODE_FRONT),			{PRINT_MACRO(gl_polygon_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_MODE_BACK),			{PRINT_MACRO(gl_polygon_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_DEPTH_RANGE_NEAR),			{PRINT_FP32,			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_DEPTH_RANGE_FAR),			{PRINT_FP32,			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CULL_FACE),				{PRINT_MACRO(gl_face_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FRONT_FACE),				{PRINT_MACRO(gl_front_face_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_NORMALIZE_ENABLE),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MATERIAL_FRONT_R),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MATERIAL_FRONT_G),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MATERIAL_FRONT_B),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MATERIAL_FRONT_A),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SEPARATE_SPECULAR_ENABLE),		{PRINT_MACRO(bool_data),	{}} },

/* NV20_TCL_PRIMITIVE_3D_ENABLED_LIGHTS looks strange but seems to be correct at least
 *  on NV28
 */
		{__(NV20_TCL_PRIMITIVE_3D_ENABLED_LIGHTS),			{
			PRINT_BITFIELD(13:13, print_macro, bool_data, "light 7"),
			PRINT_BITFIELD(15:15, print_macro, bool_data, "light 6"),
			PRINT_BITFIELD(9:9, print_macro, bool_data, "light 5"),
			PRINT_BITFIELD(11:11, print_macro, bool_data, "light 4"),
			PRINT_BITFIELD(5:5, print_macro, bool_data, "light 3"),
			PRINT_BITFIELD(7:7, print_macro, bool_data, "light 2"),
			PRINT_BITFIELD(1:1, print_macro, bool_data, "light 1"),
			PRINT_BITFIELD(3:3, print_macro, bool_data, "light 0"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(0)),		{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(1)),		{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(2)),		{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(3)),		{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(4)),		{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(5)),		{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(6)),		{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(7)),		{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV20_TCL_PRIMITIVE_3D_TX_MATRIX_ENABLE(0)),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_TX_MATRIX_ENABLE(1)),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_TX_MATRIX_ENABLE(2)),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_TX_MATRIX_ENABLE(3)),			{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_SIZE),				{PRINT_FP32,	{}} },
		{M_(NV20_TCL_PRIMITIVE_3D_MODELVIEW_MATRIX),			{PRINT_FP32,	{}} },
		{M_(NV20_TCL_PRIMITIVE_3D_INVERSE_MODELVIEW_MATRIX),		{PRINT_FP32,	{}} },
		{M_(NV20_TCL_PRIMITIVE_3D_PROJECTION_MATRIX),			{PRINT_FP32,	{}} },
		{M_(NV20_TCL_PRIMITIVE_3D_TX_MATRIX(0)),			{PRINT_FP32,	{}} },
		{M_(NV20_TCL_PRIMITIVE_3D_TX_MATRIX(1)),			{PRINT_FP32,	{}} },
		{M_(NV20_TCL_PRIMITIVE_3D_TX_MATRIX(2)),			{PRINT_FP32,	{}} },
		{M_(NV20_TCL_PRIMITIVE_3D_TX_MATRIX(3)),			{PRINT_FP32,	{}} },

#define NV20_TCL_PRIMITIVE_3D_CLIP_PLANE(n)	\
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_A(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_C(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_CLIP_PLANE_D(n)),	{PRINT_FP32,	{}} },

		NV20_TCL_PRIMITIVE_3D_CLIP_PLANE(0)
		NV20_TCL_PRIMITIVE_3D_CLIP_PLANE(1)
		NV20_TCL_PRIMITIVE_3D_CLIP_PLANE(2)
		NV20_TCL_PRIMITIVE_3D_CLIP_PLANE(3)
		NV20_TCL_PRIMITIVE_3D_CLIP_PLANE(4)
		NV20_TCL_PRIMITIVE_3D_CLIP_PLANE(5)
/* NOTE: I suspect the following values are true but my hardware has only 6 planes */
		NV20_TCL_PRIMITIVE_3D_CLIP_PLANE(6)
		NV20_TCL_PRIMITIVE_3D_CLIP_PLANE(7)

		{__(NV20_TCL_PRIMITIVE_3D_FOG_EQUATION_CONSTANT),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FOG_EQUATION_LINEAR),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FOG_EQUATION_QUADRATIC),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_A),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_B),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_C),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_D),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_E),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_F),		{PRINT_FP32,	{}} },

		{__(NV20_TCL_PRIMITIVE_3D_POINT_SPRITE),		{
			PRINT_BITFIELD(11:11, print_macro, bool_data, "coord_replace"),
			PRINT_BITFIELD(2:1, print_macro, nv20_point_sprite_r_mode, "r_mode"),
			PRINT_BITFIELD(0:0, print_macro, bool_data, "enable"),
			{}} },

		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_OX),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_OY),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_DEPTH_AVG_S),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_UNKNOWN_A),	{PRINT_FP32,	{}} },

		{__(NV20_TCL_PRIMITIVE_3D_POINT_PARAMETER_A),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_PARAMETER_B),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_PARAMETER_C),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_PARAMETER_D),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_PARAMETER_E),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_PARAMETER_F),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_PARAMETER_G),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_POINT_PARAMETER_H),	{PRINT_FP32,	{}} },

		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_PX_DIV2),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_PY_DIV2),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_DEPTH_HALF_S),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_UNKNOWN_B),	{PRINT_FP32,	{}} },

		{__(NV20_TCL_PRIMITIVE_3D_VP_UPLOAD_INST0),	{PRINT_FUNC(vshader_set_inst0), {}} },
		{__(NV20_TCL_PRIMITIVE_3D_VP_UPLOAD_INST1),	{PRINT_FUNC(vshader_set_inst1), {}} },
		{__(NV20_TCL_PRIMITIVE_3D_VP_UPLOAD_INST2),	{PRINT_FUNC(vshader_set_inst2), {}} },
		{__(NV20_TCL_PRIMITIVE_3D_VP_UPLOAD_INST3),	{PRINT_FUNC(vshader_set_inst3_print), {}} },
		{__(NV20_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_X),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_Y),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_Z),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_W),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_ID),	{PRINT_X32, {}} },

		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_MODEL_FRONT_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_R),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_MODEL_FRONT_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_G),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_MODEL_FRONT_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_B),	{PRINT_FP32,	{}} },

#define NV20_TCL_PRIMITIVE_3D_TX_UNIT(n)	\
		{__(NV20_TCL_PRIMITIVE_3D_TX_OFFSET(n)),	{PRINT_X32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_TX_FORMAT(n)),	{	\
			PRINT_BITFIELD(27:24, print_format, "%d", "log2_height"),	\
			PRINT_BITFIELD(23:20, print_format, "%d", "log2_width"),	\
			PRINT_BITFIELD(19:16, print_format, "%d", "lod"),		\
			PRINT_BITFIELD(15:8, print_macro, nv10_tx_format, "format"),	\
			PRINT_BITFIELD(2:2, print_macro, bool_data, "cube_map"),	\
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_TX_WRAP(n)),	{ \
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_WRAP_S, print_macro, nv30_tx_wrap, "wrap_s"),	\
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_WRAP_T, print_macro, nv30_tx_wrap, "wrap_t"),	\
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_WRAP_R, print_macro, nv30_tx_wrap, "wrap_r"),	\
			{}} }, \
		{__(NV20_TCL_PRIMITIVE_3D_TX_ENABLE(n)),	{	\
			PRINT_BITFIELD(30:30, print_macro, bool_data, "enable"),	\
			PRINT_BITFIELD(5:4, print_macro, nv10_tx_anisotropy, "anisotropy"),	\
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_TX_NPOT_PITCH(n)),	{	\
			PRINT_BITFIELD(31:16, print_format, "%d", "pitch"),		\
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_TX_FILTER(n)),	{	\
			PRINT_BITFIELD(31:24, print_macro, nv30_tx_filter, "mag_filter"),	\
			PRINT_BITFIELD(23:16, print_macro, nv30_tx_filter, "min_filter"),	\
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_TX_NPOT_SIZE(n)),	{	\
			PRINT_BITFIELD(31:16, print_format, "%d", "width"),	\
			PRINT_BITFIELD(15:0, print_format, "%d", "height"),	\
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_TX_PALETTE_OFFSET(n)), {PRINT_X32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_TX_BORDER_COLOR(n)), {	\
			PRINT_BITFIELD(31:24, print_macro, bool_data, "a"),	\
			PRINT_BITFIELD(23:16, print_macro, bool_data, "r"),	\
			PRINT_BITFIELD(15:8, print_macro, bool_data, "g"),	\
			PRINT_BITFIELD(7:0, print_macro, bool_data, "b"),	\
			{}} },

		NV20_TCL_PRIMITIVE_3D_TX_UNIT(0)
		NV20_TCL_PRIMITIVE_3D_TX_UNIT(1)
		NV20_TCL_PRIMITIVE_3D_TX_UNIT(2)
		NV20_TCL_PRIMITIVE_3D_TX_UNIT(3)

		{__(NV20_TCL_PRIMITIVE_3D_RC_ENABLE),	{
			PRINT_BITFIELD(3:0, print_format, "%d", "number of rc enabled"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_TX_SHADER_OP),	{
			PRINT_BITFIELD(4:0, print_macro, nv20_tx_shader_op, "op0"),
			PRINT_BITFIELD(9:5, print_macro, nv20_tx_shader_op, "op1"),
			PRINT_BITFIELD(14:10, print_macro, nv20_tx_shader_op, "op2"),
			PRINT_BITFIELD(19:15, print_macro, nv20_tx_shader_op, "op3"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_TX_SHADER_CULL_MODE),		{
			PRINT_BITFIELD(3:0, print_macro, nv20_tx_shader_cull_mode, "cull0"),
			PRINT_BITFIELD(7:4, print_macro, nv20_tx_shader_cull_mode, "cull1"),
			PRINT_BITFIELD(11:8, print_macro, nv20_tx_shader_cull_mode, "cull2"),
			PRINT_BITFIELD(15:12, print_macro, nv20_tx_shader_cull_mode, "cull3"),
			{}} },

		{__(NV20_TCL_PRIMITIVE_3D_TX_SHADER_PREVIOUS),		{
			PRINT_BITFIELD(17:16, print_format, "%d", "prev2"),
			PRINT_BITFIELD(21:20, print_format, "%d", "prev3"),
			{}} },

		{__(NV20_TCL_PRIMITIVE_3D_RC_COLOR0),		{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_RC_COLOR1),		{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_RC_FINAL0),		{
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vara_mapping"),
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_rgb, "vara_component_usage"),
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vara_input"),

			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varb_mapping"),
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_rgb, "varb_component_usage"),
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varb_input"),

			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varc_mapping"),
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_rgb, "varc_component_usage"),
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varc_input"),

			PRINT_BITFIELD(7:5, print_macro, nv10_rc_mapping, "vard_mapping"),
			PRINT_BITFIELD(4:4, print_macro, nv10_rc_cu_rgb, "vard_component_usage"),
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "vard_input"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_RC_FINAL1),		{
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vare_mapping"),
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_rgb, "vare_component_usage"),
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vare_input"),

			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varf_mapping"),
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_rgb, "varf_component_usage"),
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varf_input"),

			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varg_mapping"),
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_rgb, "varg_component_usage"),
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varg_input"),

			PRINT_BITFIELD(7:7, print_macro, bool_data, "color_sum_clamp"),
			{}} },

#define NV20_TCL_PRIMITIVE_3D_RC_UNIT(n)	\
		{__(NV20_TCL_PRIMITIVE_3D_RC_IN_ALPHA(n)),	{	\
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vara_mapping"),	\
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_alpha, "vara_component_usage"),	\
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vara_input"),	\
			\
			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varb_mapping"),	\
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_alpha, "varb_component_usage"),	\
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varb_input"),	\
			\
			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varc_mapping"),	\
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_alpha, "varc_component_usage"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varc_input"),	\
			\
			PRINT_BITFIELD(7:5, print_macro, nv10_rc_mapping, "vard_mapping"),	\
			PRINT_BITFIELD(4:4, print_macro, nv10_rc_cu_alpha, "vard_component_usage"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "vard_input"),	\
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_RC_IN_RGB(n)),	{	\
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vara_mapping"),	\
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_rgb, "vara_component_usage"),	\
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vara_input"),	\
			\
			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varb_mapping"),	\
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_rgb, "varb_component_usage"),	\
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varb_input"),	\
			\
			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varc_mapping"),	\
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_rgb, "varc_component_usage"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varc_input"),	\
			\
			PRINT_BITFIELD(7:5, print_macro, nv10_rc_mapping, "vard_mapping"),	\
			PRINT_BITFIELD(4:4, print_macro, nv10_rc_cu_rgb, "vard_component_usage"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "vard_input"),	\
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_RC_CONSTANT_COLOR0(n)),		{ \
			PRINT_BITFIELD(31:24, print_format, "%d", "a"), \
			PRINT_BITFIELD(23:16, print_format, "%d", "r"), \
			PRINT_BITFIELD(15:8, print_format, "%d", "g"), \
			PRINT_BITFIELD(7:0, print_format, "%d", "b"), \
			{}} }, \
		{__(NV20_TCL_PRIMITIVE_3D_RC_CONSTANT_COLOR1(n)),		{ \
			PRINT_BITFIELD(31:24, print_format, "%d", "a"), \
			PRINT_BITFIELD(23:16, print_format, "%d", "r"), \
			PRINT_BITFIELD(15:8, print_format, "%d", "g"), \
			PRINT_BITFIELD(7:0, print_format, "%d", "b"), \
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_RC_OUT_ALPHA(n)),	{	\
			PRINT_BITFIELD(17:16, print_macro, nv10_rc_scale, "scale"),	\
			PRINT_BITFIELD(15:15, print_macro, nv10_rc_bias, "bias"),	\
			PRINT_BITFIELD(14:14, print_macro, bool_data, "mux_sum"),	\
			PRINT_BITFIELD(13:13, print_macro, bool_data, "ab_dot_product"),	\
			PRINT_BITFIELD(12:12, print_macro, bool_data, "cd_dot_product"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "sum_output"),	\
			PRINT_BITFIELD(7:4, print_macro, nv10_rc_input, "ab_output"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "cd_output"),	\
			{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_RC_OUT_RGB(n)),	{	\
			PRINT_BITFIELD(17:16, print_macro, nv10_rc_scale, "scale"),	\
			PRINT_BITFIELD(15:15, print_macro, nv10_rc_bias, "bias"),	\
			PRINT_BITFIELD(14:14, print_macro, bool_data, "mux_sum"),	\
			PRINT_BITFIELD(13:13, print_macro, bool_data, "ab_dot_product"),	\
			PRINT_BITFIELD(12:12, print_macro, bool_data, "cd_dot_product"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "sum_output"),	\
			PRINT_BITFIELD(7:4, print_macro, nv10_rc_input, "ab_output"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "cd_output"),	\
			{}} },

		NV20_TCL_PRIMITIVE_3D_RC_UNIT(0)
		NV20_TCL_PRIMITIVE_3D_RC_UNIT(1)
		NV20_TCL_PRIMITIVE_3D_RC_UNIT(2)
		NV20_TCL_PRIMITIVE_3D_RC_UNIT(3)
		NV20_TCL_PRIMITIVE_3D_RC_UNIT(4)
		NV20_TCL_PRIMITIVE_3D_RC_UNIT(5)
		NV20_TCL_PRIMITIVE_3D_RC_UNIT(6)
		NV20_TCL_PRIMITIVE_3D_RC_UNIT(7)

#define NV20_TCL_PRIMITIVE_3D_LIGHT(n)	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_POSITION_X(n)),				{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_POSITION_Y(n)),				{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_POSITION_Z(n)),				{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_HALF_VECTOR_X(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_HALF_VECTOR_Y(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_HALF_VECTOR_Z(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_DIRECTION_X(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_DIRECTION_Y(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_DIRECTION_Z(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_AMBIENT_R(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_AMBIENT_G(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_AMBIENT_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_DIFFUSE_R(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_DIFFUSE_G(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_DIFFUSE_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_SPECULAR_R(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_SPECULAR_G(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_SPECULAR_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_BACK_SIDE_PRODUCT_AMBIENT(n)),		{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_BACK_SIDE_PRODUCT_DIFFUSE(n)),		{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_BACK_SIDE_PRODUCT_SPECULAR(n)),		{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_CONSTANT_ATTENUATION(n)),		{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_LINEAR_ATTENUATION(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_QUADRATIC_ATTENUATION(n)),		{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_SPOT_CUTOFF_A(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_SPOT_EXPONENT(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_SPOT_CUTOFF_B(n)),			{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_SPOT_DIR_X(n)),				{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_SPOT_DIR_Y(n)),				{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_SPOT_DIR_Z(n)),				{PRINT_FP32,	{}} },	\
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_SPOT_CUTOFF_C(n)),			{PRINT_FP32,	{}} },

		NV20_TCL_PRIMITIVE_3D_LIGHT(0)
		NV20_TCL_PRIMITIVE_3D_LIGHT(1)
		NV20_TCL_PRIMITIVE_3D_LIGHT(2)
		NV20_TCL_PRIMITIVE_3D_LIGHT(3)
		NV20_TCL_PRIMITIVE_3D_LIGHT(4)
		NV20_TCL_PRIMITIVE_3D_LIGHT(5)
		NV20_TCL_PRIMITIVE_3D_LIGHT(6)
		NV20_TCL_PRIMITIVE_3D_LIGHT(7)

		{__(NV20_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_A),		{PRINT_FP32,	{}} },	
		{__(NV20_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_B),		{PRINT_FP32,	{}} },	
		{__(NV20_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_C),		{PRINT_FP32,	{}} },	
		{__(NV20_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_D),		{PRINT_FP32,	{}} },	
		{__(NV20_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_E),		{PRINT_FP32,	{}} },	
		{__(NV20_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_F),		{PRINT_FP32,	{}} },	


		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_ENABLE),		{PRINT_MACRO(bool_data),	{}} },

#define NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(n)		{__(NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_PATTERN(n)),		{PRINT_X32,	{}} },
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(0)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(1)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(2)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(3)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(4)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(5)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(6)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(7)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(8)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(9)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(10)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(11)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(12)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(13)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(14)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(15)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(16)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(17)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(18)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(19)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(20)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(21)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(22)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(23)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(24)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(25)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(26)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(27)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(28)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(29)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(30)
		NV20_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_P(31)

		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_POS_3F_X),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_POS_3F_Y),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_POS_3F_Z),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_POS_4F_X),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_POS_4F_Y),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_POS_4F_Z),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_POS_4F_W),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_POS_4I_XY),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_POS_4I_ZW),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "w"),
			PRINT_BITFIELD(15:0, print_format, "%d", "z"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_NOR_3F_X),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_NOR_3F_Y),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_NOR_3F_Z),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_NOR_3I_XY),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_NOR_3I_Z),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "z"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL_4F_R),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL_4F_G),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL_4F_B),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL_4F_A),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL_3F_R),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL_3F_G),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL_3F_B),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL_4I),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL2_3F_R),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL2_3F_G),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL2_3F_B),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_COL2_3I),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX0_2F_S),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX0_2F_T),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX0_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_S),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_T),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_R),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_Q),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX0_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX0_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX1_2F_S),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX1_2F_T),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX1_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_S),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_T),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_R),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_Q),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX1_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX1_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX2_2F_S),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX2_2F_T),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX2_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX2_4F_S),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX2_4F_T),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX2_4F_R),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX2_4F_Q),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX2_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX2_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX3_2F_S),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX3_2F_T),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX3_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX3_4F_S),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX3_4F_T),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX3_4F_R),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX3_4F_Q),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX3_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_TX3_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_FOG_1F),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_EDGE_FLAG),		{PRINT_MACRO(bool_data),	{}} },

		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR0_POS), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR1_WGH), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR2_NOR), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR3_COL), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR4_COL2), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR5_FOG), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR6), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR7), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR8_TX0), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR9_TX1), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR10_TX2), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR11_TX3), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR12_TX4), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR13_TX5), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR14_TX6), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{__(NV20_TCL_PRIMITIVE_3D_VB_POINTER_ATTR15_TX7), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "enabled"), /* ? */
			PRINT_BITFIELD(28:0, print_format, "0x%08x", "offset"),
			{}}},
		{A_(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR),		{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR0_POS),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR1_WGH),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR2_NOR),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR3_COL),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR4_COL2),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR5_FOG),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR6),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR7),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR8_TX0),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR9_TX1),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR10_TX2),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR11_TX3),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR12_TX4),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR13_TX5),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR14_TX6),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_ATTR15_TX7),			{
			PRINT_BITFIELD(15:8, print_format, "%d", "stride"),
			PRINT_BITFIELD(7:4, print_format, "%d", "fields"),
			PRINT_BITFIELD(3:0, print_macro, vertex_data_type, "type"),
			{}} },

		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_MODEL_BACK_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_R),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_MODEL_BACK_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_G),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_MODEL_BACK_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_B),	{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MATERIAL_BACK_A),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MATERIAL_BACK_R),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MATERIAL_BACK_G),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_MATERIAL_BACK_B),		{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_LOGIC_OP_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_COLOR_LOGIC_OP_OP),			{PRINT_MACRO(gl_logic_op_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_LIGHT_MODEL_TWO_SIDE_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_BEGIN_END),				{PRINT_MACRO(begin_end_data),	{}} },

		{__(NV20_TCL_PRIMITIVE_3D_SCISSOR_X2_X1),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "x2"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x1"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_SCISSOR_Y2_Y1),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "y2"),
			PRINT_BITFIELD(15:0, print_format, "%d", "y1"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLEAR_VALUE_DEPTH),				{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLEAR_VALUE_ARGB),				{PRINT_X32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_CLEAR_WHICH_BUFFERS),	{
			PRINT_BITFIELD(7:7, print_format, "%d", "clear color a"),
			PRINT_BITFIELD(6:6, print_format, "%d", "clear color b"),
			PRINT_BITFIELD(5:5, print_format, "%d", "clear color g"),
			PRINT_BITFIELD(4:4, print_format, "%d", "clear color r"),
			PRINT_BITFIELD(1:1, print_format, "%d", "clear depth"),
			PRINT_BITFIELD(0:0, print_format, "%d", "clear stencil"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_INDEX_DATA),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "index1"),
			PRINT_BITFIELD(15:0, print_format, "%d", "index0"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VB_VERTEX_BATCH), {
			PRINT_BITFIELD(31:24, print_format, "%d", "count_vertices"),
			PRINT_BITFIELD(23:0, print_format, "%d", "offset_vertices"),
			{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VERTEX_DATA),				{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_ORIGIN_X),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_ORIGIN_Y),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_ORIGIN_Z),			{PRINT_FP32,	{}} },
		{__(NV20_TCL_PRIMITIVE_3D_VIEWPORT_ORIGIN_W),			{PRINT_FP32,	{}} },
		{}},
	},

	{N_(NV30_TCL_PRIMITIVE_3D), NV30, 0,0, {
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT0),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT1),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT2),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT3),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT4),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT5),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_VB_SRC0_OBJECT),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_VB_SRC1_OBJECT),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT6),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT7),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT8),	{PRINT_OBJECT,	{}} },
                {__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT8B), {PRINT_OBJECT, {}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT9),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_OBJECT10),	{PRINT_OBJECT,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_COLOR_BUFFER_DIM0),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_XOFFS, print_format, "%d", "x_offset"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_COLOR_BUFFER_DIM1),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_HEIGHT, print_format, "%d", "height"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_YOFFS, print_format, "%d", "y_offset"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BUFFER0_PITCH),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "zs_pitch"),
			PRINT_BITFIELD(15:0, print_format, "%d", "color0_pitch"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR0_OFFSET),	{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_DEPTH_OFFSET),	{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR1_OFFSET),	{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BUFFER1_PITCH),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "color1 buffer pitch"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_ENABLED_BUFFERS), {
			PRINT_BITFIELD(0:0, print_macro, bool_data, "BUF0"),
			PRINT_BITFIELD(1:1, print_macro, bool_data, "BUF1"),
			PRINT_BITFIELD(2:2, print_macro, bool_data, "BUF2"),
			PRINT_BITFIELD(3:3, print_macro, bool_data, "BUF3"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LMA_DEPTH_BUFFER_PITCH),		{
			PRINT_BITFIELD(15:0, print_format, "%d", "pitch"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LMA_DEPTH_BUFFER_OFFSET),		{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_TX_MATRIX_ENABLE(0)),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_TX_MATRIX_ENABLE(1)),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_TX_MATRIX_ENABLE(2)),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_TX_MATRIX_ENABLE(3)),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BUFFER2_PITCH),	{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BUFFER3_PITCH),	{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BUFFER2_OFFSET),	{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BUFFER3_OFFSET),	{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_COLOR_BUFFER_OFS0),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_XOFFS, print_format, "%d", "x_offset"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_COLOR_BUFFER_OFS1),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_HEIGHT, print_format, "%d", "height"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_YOFFS, print_format, "%d", "y_offset"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_DITHER_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_ALPHA_FUNC_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_ALPHA_FUNC_FUNC),	{PRINT_MACRO(gl_comparison_op_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_ALPHA_FUNC_REF),	{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BLEND_FUNC_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BLEND_FUNC_SRC),	{PRINT_MACRO16(gl_blend_func_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BLEND_FUNC_DST),	{PRINT_MACRO16(gl_blend_func_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BLEND_COLOR),		{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BLEND_EQUATION),	{PRINT_MACRO16(gl_blend_eqn_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_MASK),		{
			PRINT_BITFIELD(31:24, print_macro, bool_data, "a"),
			PRINT_BITFIELD(23:16, print_macro, bool_data, "r"),
			PRINT_BITFIELD(15:8, print_macro, bool_data, "g"),
			PRINT_BITFIELD(7:0, print_macro, bool_data, "b"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_BACK_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_BACK_MASK),		{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_BACK_FUNC_FUNC),	{PRINT_MACRO(gl_comparison_op_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_BACK_FUNC_REF),	{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_BACK_FUNC_MASK),	{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_BACK_OP_FAIL),	{PRINT_MACRO(gl_stencil_mod_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_BACK_OP_ZFAIL),	{PRINT_MACRO(gl_stencil_mod_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_BACK_OP_ZPASS),	{PRINT_MACRO(gl_stencil_mod_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_FRONT_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_FRONT_MASK),		{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_FRONT_FUNC_FUNC),	{PRINT_MACRO(gl_comparison_op_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_FRONT_FUNC_REF),	{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_FRONT_FUNC_MASK),	{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_FRONT_OP_FAIL),	{PRINT_MACRO(gl_stencil_mod_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_FRONT_OP_ZFAIL),	{PRINT_MACRO(gl_stencil_mod_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_STENCIL_FRONT_OP_ZPASS),	{PRINT_MACRO(gl_stencil_mod_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SHADE_MODEL),		{PRINT_MACRO(gl_shade_model_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FOG_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FOG_COLOR),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_LOGIC_OP_ENABLE),	{PRINT_MACRO(bool_data),{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_LOGIC_OP_OP),	 	{PRINT_MACRO(gl_logic_op_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_NORMALIZE_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_DEPTH_RANGE_NEAR),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_DEPTH_RANGE_FAR),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_MATERIAL_FRONT_R),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_MATERIAL_FRONT_G),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_MATERIAL_FRONT_B),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_MATERIAL_FRONT_A),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LINE_WIDTH_SMOOTH),	{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LINE_SMOOTH_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(0)),	{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(1)),	{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(2)),	{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(3)),	{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(4)),	{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(5)),	{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(6)),	{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_ENABLE(7)),	{PRINT_MACRO(gl_texgen_mode_data),		{}} },
		{M_(NV30_TCL_PRIMITIVE_3D_MODELVIEW_MATRIX),		{PRINT_FP32,	{}} },
		{M_(NV30_TCL_PRIMITIVE_3D_INVERSE_MODELVIEW_MATRIX),	{PRINT_FP32,	{}} },
		{M_(NV30_TCL_PRIMITIVE_3D_PROJECTION_MATRIX),		{PRINT_FP32,	{}} },
		{M_(NV30_TCL_PRIMITIVE_3D_TX_MATRIX(0)),	{PRINT_FP32,	{}} },
		{M_(NV30_TCL_PRIMITIVE_3D_TX_MATRIX(1)),	{PRINT_FP32,	{}} },
		{M_(NV30_TCL_PRIMITIVE_3D_TX_MATRIX(2)),	{PRINT_FP32,	{}} },
		{M_(NV30_TCL_PRIMITIVE_3D_TX_MATRIX(3)),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SCISSOR_WIDTH_XPOS),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_XOFFS, print_format, "%d", "x_offset"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SCISSOR_HEIGHT_YPOS),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_HEIGHT, print_format, "%d", "height"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_YOFFS, print_format, "%d", "y_offset"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FOG_COORD_DIST),	{PRINT_MACRO(nv10_fog_coord_dist),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FOG_MODE),		{PRINT_MACRO(nv10_fog_mode),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FOG_EQUATION_CONSTANT),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FOG_EQUATION_LINEAR),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FOG_EQUATION_QUADRATIC),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FP_ACTIVE_PROGRAM),	{PRINT_FUNC(fshader_from_offset),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_RC_COLOR0),		{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_RC_COLOR1),		{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_RC_FINAL0),		{
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vara_mapping"),
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_rgb, "vara_component_usage"),
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vara_input"),

			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varb_mapping"),
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_rgb, "varb_component_usage"),
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varb_input"),

			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varc_mapping"),
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_rgb, "varc_component_usage"),
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varc_input"),

			PRINT_BITFIELD(7:5, print_macro, nv10_rc_mapping, "vard_mapping"),
			PRINT_BITFIELD(4:4, print_macro, nv10_rc_cu_rgb, "vard_component_usage"),
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "vard_input"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_RC_FINAL1),		{
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vare_mapping"),
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_rgb, "vare_component_usage"),
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vare_input"),

			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varf_mapping"),
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_rgb, "varf_component_usage"),
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varf_input"),

			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varg_mapping"),
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_rgb, "varg_component_usage"),
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varg_input"),

			PRINT_BITFIELD(7:7, print_macro, bool_data, "color_sum_clamp"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_RC_ENABLE),	{
			PRINT_BITFIELD(3:0, print_format, "%d", "number of rc enabled"),
			{}} },

#define NV30_TCL_PRIMITIVE_3D_RC_UNIT(n)	\
		{__(NV30_TCL_PRIMITIVE_3D_RC_IN_ALPHA(n)),	{	\
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vara_mapping"),	\
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_alpha, "vara_component_usage"),	\
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vara_input"),	\
			\
			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varb_mapping"),	\
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_alpha, "varb_component_usage"),	\
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varb_input"),	\
			\
			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varc_mapping"),	\
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_alpha, "varc_component_usage"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varc_input"),	\
			\
			PRINT_BITFIELD(7:5, print_macro, nv10_rc_mapping, "vard_mapping"),	\
			PRINT_BITFIELD(4:4, print_macro, nv10_rc_cu_alpha, "vard_component_usage"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "vard_input"),	\
			{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_RC_IN_RGB(n)),	{	\
			PRINT_BITFIELD(31:29, print_macro, nv10_rc_mapping, "vara_mapping"),	\
			PRINT_BITFIELD(28:28, print_macro, nv10_rc_cu_rgb, "vara_component_usage"),	\
			PRINT_BITFIELD(27:24, print_macro, nv10_rc_input, "vara_input"),	\
			\
			PRINT_BITFIELD(23:21, print_macro, nv10_rc_mapping, "varb_mapping"),	\
			PRINT_BITFIELD(20:20, print_macro, nv10_rc_cu_rgb, "varb_component_usage"),	\
			PRINT_BITFIELD(19:16, print_macro, nv10_rc_input, "varb_input"),	\
			\
			PRINT_BITFIELD(15:13, print_macro, nv10_rc_mapping, "varc_mapping"),	\
			PRINT_BITFIELD(12:12, print_macro, nv10_rc_cu_rgb, "varc_component_usage"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "varc_input"),	\
			\
			PRINT_BITFIELD(7:5, print_macro, nv10_rc_mapping, "vard_mapping"),	\
			PRINT_BITFIELD(4:4, print_macro, nv10_rc_cu_rgb, "vard_component_usage"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "vard_input"),	\
			{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_RC_CONSTANT_COLOR0(n)),	{ \
			PRINT_BITFIELD(31:24, print_format, "%d", "a"), \
			PRINT_BITFIELD(23:16, print_format, "%d", "r"), \
			PRINT_BITFIELD(15:8, print_format, "%d", "g"), \
			PRINT_BITFIELD(7:0, print_format, "%d", "b"), \
			{}} }, \
		{__(NV30_TCL_PRIMITIVE_3D_RC_CONSTANT_COLOR1(n)),	{ \
			PRINT_BITFIELD(31:24, print_format, "%d", "a"), \
			PRINT_BITFIELD(23:16, print_format, "%d", "r"), \
			PRINT_BITFIELD(15:8, print_format, "%d", "g"), \
			PRINT_BITFIELD(7:0, print_format, "%d", "b"), \
			{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_RC_OUT_ALPHA(n)),	{	\
			PRINT_BITFIELD(17:16, print_macro, nv10_rc_scale, "scale"),	\
			PRINT_BITFIELD(15:15, print_macro, nv10_rc_bias, "bias"),	\
			PRINT_BITFIELD(14:14, print_macro, bool_data, "mux_sum"),	\
			PRINT_BITFIELD(13:13, print_macro, bool_data, "ab_dot_product"),	\
			PRINT_BITFIELD(12:12, print_macro, bool_data, "cd_dot_product"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "sum_output"),	\
			PRINT_BITFIELD(7:4, print_macro, nv10_rc_input, "ab_output"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "cd_output"),	\
			{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_RC_OUT_RGB(n)),	{	\
			PRINT_BITFIELD(17:16, print_macro, nv10_rc_scale, "scale"),	\
			PRINT_BITFIELD(15:15, print_macro, nv10_rc_bias, "bias"),	\
			PRINT_BITFIELD(14:14, print_macro, bool_data, "mux_sum"),	\
			PRINT_BITFIELD(13:13, print_macro, bool_data, "ab_dot_product"),	\
			PRINT_BITFIELD(12:12, print_macro, bool_data, "cd_dot_product"),	\
			PRINT_BITFIELD(11:8, print_macro, nv10_rc_input, "sum_output"),	\
			PRINT_BITFIELD(7:4, print_macro, nv10_rc_input, "ab_output"),	\
			PRINT_BITFIELD(3:0, print_macro, nv10_rc_input, "cd_output"),	\
			{}} },

		NV30_TCL_PRIMITIVE_3D_RC_UNIT(0)
		NV30_TCL_PRIMITIVE_3D_RC_UNIT(1)
		NV30_TCL_PRIMITIVE_3D_RC_UNIT(2)
		NV30_TCL_PRIMITIVE_3D_RC_UNIT(3)
		NV30_TCL_PRIMITIVE_3D_RC_UNIT(4)
		NV30_TCL_PRIMITIVE_3D_RC_UNIT(5)
		NV30_TCL_PRIMITIVE_3D_RC_UNIT(6)
		NV30_TCL_PRIMITIVE_3D_RC_UNIT(7)

		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_XOFFS, print_format, "%d", "x_offset"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_1),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_HEIGHT, print_format, "%d", "height"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_YOFFS, print_format, "%d", "y_offset"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_MODEL_FRONT_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_R),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_MODEL_FRONT_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_G),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_MODEL_FRONT_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_B),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_XFRM_OX),		{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_XFRM_OY),		{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_XFRM_NPF_DIV2),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_XFRM_UNK0_0x0),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_XFRM_PX_DIV2),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_XFRM_PY_DIV2),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_XFRM_FMN_DIV2),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VIEWPORT_XFRM_UNK1_0x0),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POLYGON_OFFSET_FILL_ENABLE),	{PRINT_MACRO(bool_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POLYGON_OFFSET_LINE_ENABLE),	{PRINT_MACRO(bool_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POLYGON_OFFSET_POINT_ENABLE),	{PRINT_MACRO(bool_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_DEPTH_FUNC),		{PRINT_MACRO(gl_comparison_op_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_DEPTH_WRITE_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_DEPTH_TEST_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POLYGON_OFFSET_FACTOR),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POLYGON_OFFSET_UNITS),	{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_NOR_3I_XY),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_NOR_3I_Z),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "z"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_INST0),	{PRINT_FUNC(vshader_set_inst0), {}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_INST1),	{PRINT_FUNC(vshader_set_inst1), {}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_INST2),	{PRINT_FUNC(vshader_set_inst2), {}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_INST3),	{PRINT_FUNC(vshader_set_inst3_print), {}} },
#define NV30_TCL_PRIMITIVE_3D_CLIP_PLANE(n)	\
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_A(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_C(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_CLIP_PLANE_D(n)),	{PRINT_FP32,	{}} },

		NV30_TCL_PRIMITIVE_3D_CLIP_PLANE(0)
		NV30_TCL_PRIMITIVE_3D_CLIP_PLANE(1)
		NV30_TCL_PRIMITIVE_3D_CLIP_PLANE(2)
		NV30_TCL_PRIMITIVE_3D_CLIP_PLANE(3)
		NV30_TCL_PRIMITIVE_3D_CLIP_PLANE(4)
		NV30_TCL_PRIMITIVE_3D_CLIP_PLANE(5)
/* NOTE: I suspect the following values are true but my hardware has only 6 planes */
		NV30_TCL_PRIMITIVE_3D_CLIP_PLANE(6)
		NV30_TCL_PRIMITIVE_3D_CLIP_PLANE(7)
#define NV30_TCL_PRIMITIVE_3D_LIGHT(n)	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_AMBIENT_R(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_AMBIENT_G(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_AMBIENT_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_DIFFUSE_R(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_DIFFUSE_G(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_DIFFUSE_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_SPECULAR_R(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_SPECULAR_G(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_FRONT_SIDE_PRODUCT_SPECULAR_B(n)),	{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_HALF_VECTOR_X(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_HALF_VECTOR_Y(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_HALF_VECTOR_Z(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_DIRECTION_X(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_DIRECTION_Y(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_DIRECTION_Z(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_SPOT_CUTOFF_A(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_SPOT_EXPONENT(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_SPOT_CUTOFF_B(n)),			{PRINT_FP32,	{}} },  \
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_SPOT_DIR_X(n)),				{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_SPOT_DIR_Y(n)),				{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_SPOT_DIR_Z(n)),				{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_SPOT_CUTOFF_C(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_POSITION_X(n)),				{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_POSITION_Y(n)),				{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_POSITION_Z(n)),				{PRINT_FP32,	{}} },  \
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_CONSTANT_ATTENUATION(n)),		{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_LINEAR_ATTENUATION(n)),			{PRINT_FP32,	{}} },	\
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_QUADRATIC_ATTENUATION(n)),		{PRINT_FP32,	{}} },

		NV30_TCL_PRIMITIVE_3D_LIGHT(0)
		NV30_TCL_PRIMITIVE_3D_LIGHT(1)
		NV30_TCL_PRIMITIVE_3D_LIGHT(2)
		NV30_TCL_PRIMITIVE_3D_LIGHT(3)
		NV30_TCL_PRIMITIVE_3D_LIGHT(4)
		NV30_TCL_PRIMITIVE_3D_LIGHT(5)
		NV30_TCL_PRIMITIVE_3D_LIGHT(6)
		NV30_TCL_PRIMITIVE_3D_LIGHT(7)
		
		{__(NV30_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_A),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_B),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_C),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_D),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_E),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FRONT_MATERIAL_SHININESS_F),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_ENABLED_LIGHTS),			{
			PRINT_BITFIELD(15:15, print_macro, bool_data, "light 7"),
			PRINT_BITFIELD(13:13, print_macro, bool_data, "light 6"),
			PRINT_BITFIELD(11:11, print_macro, bool_data, "light 5"),
			PRINT_BITFIELD(9:9, print_macro, bool_data, "light 4"),
			PRINT_BITFIELD(7:7, print_macro, bool_data, "light 3"),
			PRINT_BITFIELD(5:5, print_macro, bool_data, "light 2"),
			PRINT_BITFIELD(3:3, print_macro, bool_data, "light 1"),
			PRINT_BITFIELD(0:0, print_macro, bool_data, "light 0"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_CLIPPING_PLANES),	{{0,31, &print_clip_planes, NULL, NULL},   {}} },
		{__(NV30_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_ENABLE),	{PRINT_MACRO(bool_data),   {}} },
		{A_(NV30_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_PATTERN),	{PRINT_X32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_3F_X),		{PRINT_FP32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_3F_Y),		{PRINT_FP32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_3F_Z),		{PRINT_FP32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_3F_W),		{PRINT_FP32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VB_POINTER), {
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VB_POINTER_SOURCE, print_format, "%d", "source"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VB_POINTER_OFFSET, print_format, "0x%08x", "offset"),
			{}}},

		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_UNK_0),	{PRINT_X32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTXFMT), {
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VTXFMT_STRIDE, print_format, "%d", "stride"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VTXFMT_NCOMP, print_format, "%d", "ncomp"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VTXFMT_TYPE, print_macro, vertex_data_type, "type"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_MODEL_BACK_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_R),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_MODEL_BACK_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_G),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LIGHT_MODEL_BACK_SIDE_PRODUCT_AMBIENT_PLUS_EMISSION_B),	{PRINT_FP32,	{}} },

		{__(NV30_TCL_PRIMITIVE_3D_COLOR_MATERIAL_BACK_R),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_MATERIAL_BACK_G),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_MATERIAL_BACK_B),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_COLOR_MATERIAL_BACK_A),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_OCC_QUERY_OR_COLOR_BUFF_ENABLE),	{{0, 31, &print_query, NULL, NULL}, {} } },
		{__(NV30_TCL_PRIMITIVE_3D_STORE_RESULT),			{{0, 31, &print_store, NULL, NULL}, {} } },
		{__(NV30_TCL_PRIMITIVE_3D_BEGIN_END),		{PRINT_MACRO(begin_end_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VB_ELEMENT_U16), {
			PRINT_BITFIELD(31:24, print_format, "0x%04x", "1"),
			PRINT_BITFIELD(15:0, print_format, "0x%04x", "0"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VB_ELEMENT_U32),		{PRINT_X32,		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VB_VERTEX_BATCH), {
			PRINT_BITFIELD(31:24, print_format, "%d", "count_vertices"),
			PRINT_BITFIELD(23:0, print_format, "%d", "offset_vertices"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_DATA),			{PRINT_FP32,		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_SET_DISPLAY_LIST_MEM_OFFSET),	{ PRINT_X32, {}} },
		{__(NV30_TCL_PRIMITIVE_3D_EXECUTE_DISPLAY_LIST),	{
			PRINT_BITFIELD(31:24, print_format, "0x%04x", "length"),
			PRINT_BITFIELD(23:0, print_format, "0x%04x", "start offset"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POLYGON_MODE_FRONT),	{PRINT_MACRO(gl_polygon_mode_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POLYGON_MODE_BACK),	{PRINT_MACRO(gl_polygon_mode_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CULL_FACE),		{PRINT_MACRO(gl_face_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_FRONT_FACE),		{PRINT_MACRO(gl_front_face_data),		{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POLYGON_SMOOTH_ENABLE),{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CULL_FACE_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_TX_DEPTH_UNIT), {
			PRINT_BITFIELD(31:20, print_format, "%d", "depth"),
			PRINT_BITFIELD(19:0, print_format, "%d", "NPOT pitch"),
											{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_2F_X),		{PRINT_FP32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_2F_Y),		{PRINT_FP32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_2I),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_2I_X, print_format, "%d", "x"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_2I_Y, print_format, "%d", "y"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_COL_4I),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_COL2_3I),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_TX0_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_TX0_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_TX1_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_TX1_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_TX2_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_TX2_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_TX3_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_TX3_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },

		{A_(NV30_TCL_PRIMITIVE_3D_TX_ADDRESS_UNIT), {PRINT_X32, {}} },
		{A_(NV30_TCL_PRIMITIVE_3D_TX_FORMAT_UNIT), {
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_FORMAT_MIPMAP_COUNT, print_format, "%d", "mipmap"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_FORMAT_TYPE, print_macro, nv30_tx_type, "type"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_FORMAT_FORMAT, print_macro, nv30_tx_format, "format"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_FORMAT_NCOMP, print_format, "%d", "ncomp"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_FORMAT_CUBIC, print_macro, bool_data, "cubic"),
											{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_TX_WRAP_UNIT), {
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_WRAP_S, print_macro, nv30_tx_wrap, "wrap_s"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_WRAP_T, print_macro, nv30_tx_wrap, "wrap_t"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_WRAP_R, print_macro, nv30_tx_wrap, "wrap_r"),
											{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_TX_ENABLE_UNIT), {
			PRINT_BITFIELD(31:31, print_macro, bool_data, "nv40_enable"),
			PRINT_BITFIELD(30:30, print_macro, bool_data, "nv30_enable"),
			PRINT_BITFIELD(6:4, print_macro, nv40_tx_anisotropy, "anisotropy"),
			{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_TX_SWIZZLE_UNIT), {PRINT_FUNC(nv30_tx_swizzle_print), {}} },
		{A_(NV30_TCL_PRIMITIVE_3D_TX_FILTER_UNIT), {
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_FILTER_MIN, print_macro, nv30_tx_filter, "filter_min"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_TX_FILTER_MAG, print_macro, nv30_tx_filter, "filter_mag"),
											{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_TX_XY_DIM_UNIT), {
			PRINT_BITFIELD(31:16, print_format, "%d", "width"),
			PRINT_BITFIELD(15:0, print_format, "%d", "height"),
											{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_TX_BORDER_COLOR_UNIT), {
			PRINT_BITFIELD(31:24, print_macro, bool_data, "a"),
			PRINT_BITFIELD(23:16, print_macro, bool_data, "r"),
			PRINT_BITFIELD(15:8, print_macro, bool_data, "g"),
			PRINT_BITFIELD(7:0, print_macro, bool_data, "b"),
			{}} },

		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_4F_X),		{PRINT_FP32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_4F_Y),		{PRINT_FP32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_4F_Z),		{PRINT_FP32,	{}} },
		{A_(NV30_TCL_PRIMITIVE_3D_VTX_ATTR_4F_W),		{PRINT_FP32,	{}} },

		{__(NV30_TCL_PRIMITIVE_3D_FP_CONTROL),      {
			PRINT_BITFIELD(7:7, print_macro, bool_data, "uses_kil"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_UNK1D6C_OFFSET),	{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_UNK1D70_VALUE),	{PRINT_X32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLEAR_VALUE_DEPTH),	{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLEAR_VALUE_ARGB), {
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "r"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "b"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_CLEAR_WHICH_BUFFERS),	{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_DO_VERTICES),		{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LINE_STIPPLE_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_LINE_STIPPLE_PATTERN),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "factor"),
			PRINT_BITFIELD(31:16, print_format, "%04x", "pattern"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_A),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_B),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_C),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_D),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_E),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_BACK_MATERIAL_SHININESS_F),		{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VERTEX_FOG_1F),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_FROM_ID),	{PRINT_FUNC(vshader_set_next_id),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_PROGRAM_START_ID),	{PRINT_UINT,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_PARAMETER_A),		{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_PARAMETER_B),		{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_PARAMETER_C),		{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_PARAMETER_D),		{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_PARAMETER_E),		{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_PARAMETER_F),		{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_PARAMETER_G),		{PRINT_FP32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_PARAMETER_H),		{PRINT_FP32,			{}} },
                {__(NV30_TCL_PRIMITIVE_3D_POINT_SIZE),                  {PRINT_FP32,                    {}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_PARAMETERS_ENABLE),	{PRINT_MACRO(bool_data),	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_POINT_SPRITE),		{
			PRINT_BITFIELD(11:11, print_macro, bool_data, "coord_replace"),
			PRINT_BITFIELD(2:1, print_macro, nv20_point_sprite_r_mode, "r_mode"),
			PRINT_BITFIELD(0:0, print_macro, bool_data, "enable"),
			{}} },

		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_ID),	{PRINT_X32,			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P0_X),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P0_Y),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P0_Z),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P0_W),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P1_X),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P1_Y),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P1_Z),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P1_W),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P2_X),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P2_Y),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P2_Z),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P2_W),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P3_X),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P3_Y),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P3_Z),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_P3_W),	{PRINT_FP32,	{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_IN_REG), {
			PRINT_BITFIELD(0:0,  print_macro, bool_data, "POS"),
			PRINT_BITFIELD(1:1,  print_macro, bool_data, "WEIGHT"),
			PRINT_BITFIELD(2:2,  print_macro, bool_data, "NORMAL"),
			PRINT_BITFIELD(3:3,  print_macro, bool_data, "COL0"),
			PRINT_BITFIELD(4:4,  print_macro, bool_data, "COL1"),
			PRINT_BITFIELD(5:5,  print_macro, bool_data, "FOGC"),
			PRINT_BITFIELD(8:8,  print_macro, bool_data, "TEX0"),
			PRINT_BITFIELD(9:9,  print_macro, bool_data, "TEX1"),
			PRINT_BITFIELD(10:10,  print_macro, bool_data, "TEX2"),
			PRINT_BITFIELD(11:11,  print_macro, bool_data, "TEX3"),
			PRINT_BITFIELD(12:12,  print_macro, bool_data, "TEX4"),
			PRINT_BITFIELD(13:13,  print_macro, bool_data, "TEX5"),
			PRINT_BITFIELD(14:14,  print_macro, bool_data, "TEX6"),
			PRINT_BITFIELD(15:15,  print_macro, bool_data, "TEX7"),
			{}} },
		{__(NV30_TCL_PRIMITIVE_3D_VP_OUT_REG), {
			PRINT_BITFIELD(0:0,  print_macro, bool_data, "COL0"),
			PRINT_BITFIELD(1:1,  print_macro, bool_data, "COL1"),
			PRINT_BITFIELD(2:2,  print_macro, bool_data, "BFC0"),
			PRINT_BITFIELD(3:3,  print_macro, bool_data, "BFC1"),
			PRINT_BITFIELD(4:4,  print_macro, bool_data, "FOGC"),
			PRINT_BITFIELD(5:5,  print_macro, bool_data, "PSZ"),
			PRINT_BITFIELD(6:6,  print_macro, bool_data, "CLP0"),
			PRINT_BITFIELD(7:7,  print_macro, bool_data, "CLP1"),
			PRINT_BITFIELD(8:8,  print_macro, bool_data, "CLP2"),
			PRINT_BITFIELD(9:9,  print_macro, bool_data, "CLP3"),
			PRINT_BITFIELD(10:10,  print_macro, bool_data, "CLP4"),
			PRINT_BITFIELD(11:11,  print_macro, bool_data, "CLP5"),
			PRINT_BITFIELD(14:14,  print_macro, bool_data, "TEX0"),
			PRINT_BITFIELD(15:15,  print_macro, bool_data, "TEX1"),
			PRINT_BITFIELD(16:16,  print_macro, bool_data, "TEX2"),
			PRINT_BITFIELD(17:17,  print_macro, bool_data, "TEX3"),
			PRINT_BITFIELD(18:18,  print_macro, bool_data, "TEX4"),
			PRINT_BITFIELD(19:19,  print_macro, bool_data, "TEX5"),
			PRINT_BITFIELD(20:20,  print_macro, bool_data, "TEX6"),
			PRINT_BITFIELD(21:21,  print_macro, bool_data, "TEX7"),
			{}} },
		{}},
	},

	{N_(NV40_TCL_PRIMITIVE_3D), NV40|G70, NV30_TCL_PRIMITIVE_3D,NV30, {
		{__(NV40_TCL_PRIMITIVE_3D_COLOR_MASK_BUFFER123),	{
			PRINT_BITFIELD(15:15, print_macro, bool_data, "buffer3 b"),
			PRINT_BITFIELD(14:14, print_macro, bool_data, "buffer3 g"),
			PRINT_BITFIELD(13:13, print_macro, bool_data, "buffer3 r"),
			PRINT_BITFIELD(12:12, print_macro, bool_data, "buffer3 a"),
			PRINT_BITFIELD(11:11, print_macro, bool_data, "buffer2 b"),
			PRINT_BITFIELD(10:10, print_macro, bool_data, "buffer2 g"),
			PRINT_BITFIELD(9:9, print_macro, bool_data, "buffer2 r"),
			PRINT_BITFIELD(8:8, print_macro, bool_data, "buffer2 a"),
			PRINT_BITFIELD(7:7, print_macro, bool_data, "buffer1 b"),
			PRINT_BITFIELD(6:6, print_macro, bool_data, "buffer1 g"),
			PRINT_BITFIELD(5:5, print_macro, bool_data, "buffer1 r"),
			PRINT_BITFIELD(4:4, print_macro, bool_data, "buffer1 a"),
			{}} },
		{}},
	},

	{N_(NV30_CLEAR_BUFFER), NV30|NV40|G70, 0,0, {
		{__(NV30_CLEAR_BUFFER_SET_DMA_NOTIFY),		{PRINT_OBJECT,	{}} },
		{__(NV30_CLEAR_BUFFER_SET_IMAGE_PATTERN),	{PRINT_OBJECT,	{}} },
		{__(NV30_CLEAR_BUFFER_SET_RASTER_OP),		{PRINT_OBJECT,	{}} },
		{__(NV30_CLEAR_BUFFER_SET_CONTEXT_SURFACE_2D),	{PRINT_OBJECT,	{}} },
		{__(NV30_CLEAR_BUFFER_UNK002fc),		{PRINT_OBJECT,	{}} },
		{}},
	},

	{N_(NV50_TCL_PRIMITIVE_3D), G80, 0,0, {
		{A_(NV50_TCL_PRIMITIVE_3D_SET_OBJECT_0),		{PRINT_OBJECT,	{}} },
		{A_(NV50_TCL_PRIMITIVE_3D_SET_OBJECT_1),		{PRINT_OBJECT,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_FOG_1F),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_2F_X),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_2F_Y),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX0_2F_S),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX0_2F_T),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX1_2F_S),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX1_2F_T),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX2_2F_S),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX2_2F_T),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX3_2F_S),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX3_2F_T),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_3F_X),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_3F_Y),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_3F_Z),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_NOR_3F_X),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_NOR_3F_Y),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_NOR_3F_Z),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL_3F_R),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL_3F_G),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL_3F_B),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL2_3F_R),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL2_3F_G),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL2_3F_B),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_4F_X),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_4F_Y),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_4F_Z),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_4F_W),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL_4F_R),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL_4F_G),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL_4F_B),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL_4F_A),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_S),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_T),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_R),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX0_4F_Q),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_S),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_T),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_R),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX1_4F_Q),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX2_4F_S),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX2_4F_T),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX2_4F_R),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX2_4F_Q),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX3_4F_S),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX3_4F_T),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX3_4F_R),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX3_4F_Q),		{PRINT_FP32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX0_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX1_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX2_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX3_2I),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_4I_XY),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_POS_4I_ZW),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "w"),
			PRINT_BITFIELD(15:0, print_format, "%d", "z"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX0_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX0_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX1_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX1_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX2_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX2_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX3_4I_ST),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "t"),
			PRINT_BITFIELD(15:0, print_format, "%d", "s"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_TX3_4I_RQ),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "q"),
			PRINT_BITFIELD(15:0, print_format, "%d", "r"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_NOR_3I_XY),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "y"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_NOR_3I_Z),	{
			PRINT_BITFIELD(15:0, print_format, "%d", "z"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL_4I),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_COL2_3I),	{
			PRINT_BITFIELD(31:24, print_format, "%d", "a"),
			PRINT_BITFIELD(23:16, print_format, "%d", "b"),
			PRINT_BITFIELD(15:8, print_format, "%d", "g"),
			PRINT_BITFIELD(7:0, print_format, "%d", "r"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_UNK0_X),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_UNK0_Y),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_UNK0_Z),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_UNK1_X),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_UNK1_Y),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_UNK1_Z),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_DEPTH_RANGE_NEAR),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_DEPTH_RANGE_FAR),		{PRINT_FP32,		{}} },

#define NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(n)	\
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_HORIZ(n)),	{	\
			PRINT_BITFIELD(31:16, print_format, "%d", "x2"),	\
			PRINT_BITFIELD(15:0, print_format, "%d", "x1"),	\
			{}} },	\
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_VERT(n)),	{	\
			PRINT_BITFIELD(31:16, print_format, "%d", "y2"),	\
			PRINT_BITFIELD(15:0, print_format, "%d", "y1"),	\
			{}} },

		NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(0)
		NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(1)
		NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(2)
		NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(3)
		NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(4)
		NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(5)
		NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(6)
		NV50_TCL_PRIMITIVE_3D_VIEWPORT_CLIP_REGION(7)

		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_BUFFER_FIRST),		{PRINT_X32, 		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_BUFFER_COUNT),		{PRINT_X32, 		{}} },
 		{__(NV50_TCL_PRIMITIVE_3D_CLEAR_COLOR_R),		{PRINT_FP32,		{}} },
 		{__(NV50_TCL_PRIMITIVE_3D_CLEAR_COLOR_G),		{PRINT_FP32,		{}} },
 		{__(NV50_TCL_PRIMITIVE_3D_CLEAR_COLOR_B),		{PRINT_FP32,		{}} },
 		{__(NV50_TCL_PRIMITIVE_3D_CLEAR_COLOR_A),		{PRINT_FP32,		{}} },
 		{__(NV50_TCL_PRIMITIVE_3D_CLEAR_DEPTH),			{PRINT_FP32,		{}} },
 		{__(NV50_TCL_PRIMITIVE_3D_CLEAR_STENCIL),		{PRINT_X32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POLYGON_MODE_FRONT),		{PRINT_MACRO(gl_polygon_mode_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POLYGON_MODE_BACK),		{PRINT_MACRO(gl_polygon_mode_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POLYGON_SMOOTH_ENABLE),	{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POLYGON_OFFSET_POINT_ENABLE),	{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POLYGON_OFFSET_LINE_ENABLE),	{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POLYGON_OFFSET_FILL_ENABLE),	{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_SCISSOR_WIDTH_XPOS),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "w"),
			PRINT_BITFIELD(15:0, print_format, "%d", "x"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_SCISSOR_HEIGHT_YPOS),	{
			PRINT_BITFIELD(31:16, print_format, "%d", "h"),
			PRINT_BITFIELD(15:0, print_format, "%d", "y"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_ID),		{PRINT_X32, 		{}} },
 		{__(NV50_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_X),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_Y),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_Z),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VP_UPLOAD_CONST_W),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_FRONT_FUNC_REF),	{PRINT_X32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_FRONT_MASK),		{PRINT_X32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_FRONT_FUNC_MASK),	{PRINT_X32,	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_DEPTH_TEST_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_SHADE_MODEL),			{PRINT_MACRO(gl_shade_model_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_DEPTH_WRITE_ENABLE),		{PRINT_MACRO(bool_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_ALPHA_FUNC_ENABLE),		{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_DEPTH_FUNC),			{PRINT_MACRO(gl_comparison_op_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_ALPHA_FUNC_REF),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_ALPHA_FUNC_FUNC),		{PRINT_MACRO(gl_comparison_op_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_COLOR_R),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_COLOR_G),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_COLOR_B),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_COLOR_A),		{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_EQUATION_RGB),		{PRINT_MACRO(gl_blend_eqn_data),	{}} },
		/* FIXME: bit 14 for blend_func_src,dst should be masked out to get gl_data -> enable flag ? */
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_FUNC_SRC_RGB),		{PRINT_MACRO(gl_blend_func_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_FUNC_DST_RGB),		{PRINT_MACRO(gl_blend_func_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_EQUATION_ALPHA),	{PRINT_MACRO(gl_blend_eqn_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_FUNC_SRC_ALPHA),	{PRINT_MACRO(gl_blend_func_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_BLEND_FUNC_DST_ALPHA),	{PRINT_MACRO(gl_blend_func_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_BACK_ENABLE),		{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_BACK_OP_FAIL),	{PRINT_MACRO(gl_stencil_mod_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_BACK_OP_ZFAIL),	{PRINT_MACRO(gl_stencil_mod_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_BACK_OP_ZPASS),	{PRINT_MACRO(gl_stencil_mod_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_BACK_FUNC_FUNC),	{PRINT_MACRO(gl_comparison_op_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_BACK_FUNC_REF),	{PRINT_X32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_BACK_MASK),		{PRINT_X32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_BACK_FUNC_MASK),	{PRINT_X32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_LINE_WIDTH),			{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POINT_SIZE),			{PRINT_FP32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POLYGON_OFFSET_FACTOR),	{PRINT_FP32, 		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_LINE_SMOOTH_ENABLE),		{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_FRONT_ENABLE),	{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_FRONT_OP_FAIL),	{PRINT_MACRO(gl_stencil_mod_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_FRONT_OP_ZFAIL),	{PRINT_MACRO(gl_stencil_mod_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_FRONT_OP_ZPASS),	{PRINT_MACRO(gl_stencil_mod_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_STENCIL_FRONT_FUNC_FUNC),	{PRINT_MACRO(gl_comparison_op_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POLYGON_OFFSET_UNITS),	{PRINT_FP32, 		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_BEGIN),		{PRINT_X32, 		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_END),			{PRINT_X32, 		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VERTEX_DATA),			{PRINT_FP32, 		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_LINE_STIPPLE_ENABLE),		{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_LINE_STIPPLE_PATTERN),	{
			PRINT_BITFIELD(23:8, print_format, "%04x", "pattern"),
			PRINT_BITFIELD(7:0, print_format, "%d", "factor"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_ENABLE),	{PRINT_MACRO(bool_data),{}} },
		{A_(NV50_TCL_PRIMITIVE_3D_POLYGON_STIPPLE_PATTERN),	{PRINT_X32,		{}} },
		{__(NV50_TCL_PRIMITIVE_3D_CULL_FACE_ENABLE),		{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_FRONT_FACE),			{PRINT_MACRO(gl_front_face_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_CULL_FACE),			{PRINT_MACRO(gl_face_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_LOGIC_OP_ENABLE),		{PRINT_MACRO(bool_data),{}} },
		{__(NV50_TCL_PRIMITIVE_3D_LOGIC_OP_OP),			{PRINT_MACRO(gl_logic_op_data),	{}} },
		{__(NV50_TCL_PRIMITIVE_3D_CLEAR_BUFFERS),	{
			PRINT_BITFIELD(5:2, print_format, "%d", "color"),
			PRINT_BITFIELD(1:1, print_format, "%d", "stencil"),
			PRINT_BITFIELD(0:0, print_format, "%d", "depth"),
			{}} },
		{A_(NV50_TCL_PRIMITIVE_3D_COLOR_MASK),	{
			PRINT_BITFIELD(15:12, print_macro, bool_data, "a"),
			PRINT_BITFIELD(11:8, print_macro, bool_data, "b"),
			PRINT_BITFIELD(7:4, print_macro, bool_data, "g"),
			PRINT_BITFIELD(3:0, print_macro, bool_data, "r"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_WIDTH, print_format, "%d", "width"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_XOFFS, print_format, "%d", "x_offset"),
			{}} },
		{__(NV50_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_1),	{
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_HEIGHT, print_format, "%d", "height"),
			PRINT_BITFIELD(NV30_TCL_PRIMITIVE_3D_VIEWPORT_DIMS_0_YOFFS, print_format, "%d", "y_offset"),
			{}} },
		{}},
	},

	// Software (DMA engine) objects
	{0x02, "NV_DMA_FROM_MEMORY", NVALL, 0,0, { {}}},
	{0x03, "NV_DMA_TO_MEMORY", NVALL, 0,0, { {}}},
	{0x3D, "NV_DMA_IN_MEMORY", NVALL, 0,0, { {}}},
	// Unknown objects
	{0x46, "NvType0046", NV04, 0,0, {
		{0x180, 1, 1, "NvType0046" "_DMA_NOTIFY", {PRINT_OBJECT, {}}},
		{0x184, 1, 1, "NvType0046" "_DMA_MEM_1", {PRINT_OBJECT, {}}},
		{0x188, 1, 1, "NvType0046" "_DMA_MEM_2", {PRINT_OBJECT, {}}},
		{0x18c, 1, 1, "NvType0046" "_DMA_3", {PRINT_OBJECT, {}}},
		{0x190, 1, 1, "NvType0046" "_DMA_4", {PRINT_OBJECT, {}}},
		{0x194, 1, 1, "NvType0046" "_OBJ_5", {PRINT_OBJECT, {}}},
		{0x198, 1, 1, "NvType0046" "_OBJ_6", {PRINT_OBJECT, {}}},
/*
		{0x11C, 1, 1, "NvType0046" "_UNK11C", {		{}}},	// = 0
		{0x118, 1, 1, "NvType0046" "_UNK118", {		{}}},	// = 0
		{0x224, 1, 1, "NvType0046" "_UNK224", {		{}}},	// = 0
		{0x300, 1, 1, "NvType0046" "_UNK300", {		{}}},	// = 0
		{0x308, 1, 1, "NvType0046" "_UNK308", {		{}}},	// = 0
		{0x310, 1, 1, "NvType0046" "_UNK310", {		{}}},	// = 0x400 or 0
		{0x314, 1, 1, "NvType0046" "_UNK314", {		{}}},	// = 0
		{0x318, 1, 1, "NvType0046" "_UNK318", {		{}}},	// = 0x400 or 0
		{0x320, 1, 1, "NvType0046" "_UNK320", {		{}}},	// = 0
		{0x324, 1, 1, "NvType0046" "_UNK324", {		{}}},	// = 0
		{0x328, 1, 1, "NvType0046" "_UNK328", {		{}}},	// = 0
		{0x32C, 1, 1, "NvType0046" "_UNK32C", {		{}}},	// = 0
*/
		// both pitches are equal. what's the difference?
		{0x304, 1, 1, "NvType0046" "_PITCH1",	{PRINT_X32,	{}}},
		{0x30C, 1, 1, "NvType0046" "_PITCH2",	{PRINT_X32,	{}}},
		{0x340, 1, 1, "NvType0046" "_SIZE",	{
			PRINT_BITFIELD(15:0, print_format, "%d", "width"),
			PRINT_BITFIELD(31:16, print_format, "%d", "height"),
			{}}},
		{0x344, 1, 1, "NvType0046" "_WIDTH",	{
			PRINT_BITFIELD(15:0, print_format, "%d", "visible_width"),
			PRINT_BITFIELD(31:16, print_format, "%d", "blank_width"),
			{}}},
		{0x348, 1, 1, "NvType0046" "_HSYNC",	{
			PRINT_BITFIELD(15:0, print_format, "%d", "hsync_start"),
			PRINT_BITFIELD(31:16, print_format, "%d", "hsync_len"),
			{}}},
		{0x34c, 1, 1, "NvType0046" "_HEIGHT",	{
			PRINT_BITFIELD(15:0, print_format, "%d", "visible_height"),
			PRINT_BITFIELD(31:16, print_format, "%d", "blank_height"),
			{}}},
		{0x350, 1, 1, "NvType0046" "_VSYNC",	{
			PRINT_BITFIELD(15:0, print_format, "%d", "vsync_start"),
			PRINT_BITFIELD(31:16, print_format, "%d", "vsync_len"),
			{}}},
		{0x354, 1, 1, "NvType0046" "_FULL_SIZE",	{
			PRINT_BITFIELD(15:0, print_format, "%d", "full_width"),
			PRINT_BITFIELD(31:16, print_format, "%d", "full_height"),
			{}}},
		{0x358, 1, 1, "NvType0046" "_PIXEL_CLK",	{ // in KHz
			PRINT_X32,
			{}}},
		{0x35C, 1, 1, "NvType0046" "_FLAGS", {
			PRINT_BITFIELD(1:1, print_macro, bool_data, "doublescan"),
			PRINT_BITFIELD(3:3, print_macro, bool_data, "neg_hsync"),
			PRINT_BITFIELD(4:4, print_macro, bool_data, "neg_vsync"),
			PRINT_BITFIELD(17:16, print_macro, nv_dac_format, "depth"),
//			PRINT_BITFIELD(28:28, .......),					 = 1
			{}}},
		{}},
	},
	{0x47, "NvType0047", NV04, 0,0, {
//		{0x120, 1, 1, "NvType0047" "_UNK120", {		{}}}, // 1
//		{0x124, 1, 1, "NvType0047" "_UNK124", {		{}}}, // 1
		{0x180, 1, 1, "NvType0047" "_DMA_NOTIFY", {PRINT_OBJECT, {}}},
		{0x19C, 1, 1, "NvType0047" "_UNK19C", {		{}}}, // maybe PRINT_OBJECT: memory
		{0x1A0, 1, 1, "NvType0047" "_UNK1A0", {		{}}}, // maybe PRINT_OBJECT: memory
		{}},
	},
	{0x49, "NvType0049", NV04, 0,0, {
		{0x180, 1, 1, "NvType0049" "_DMA_NOTIFY", {PRINT_OBJECT, {}}},
		{0x184, 1, 1, "NvType0049" "_DMA_MEM_1", {PRINT_OBJECT, {}}},
		{0x188, 1, 1, "NvType0049" "_DMA_MEM_2", {PRINT_OBJECT, {}}},
		{}},
	},
	{0x4D, "NvType004D", NV04, 0,0, {
		{0x180, 1, 1, "NvType004D" "_DMA_NOTIFY", {PRINT_OBJECT, {}}},
		{0x184, 1, 1, "NvType004D" "_DMA_MEM_1", {PRINT_OBJECT, {}}},
		{0x188, 1, 1, "NvType004D" "_DMA_MEM_2", {PRINT_OBJECT, {}}},
		{0x18c, 1, 1, "NvType004D" "_DMA_MEM_3", {PRINT_OBJECT, {}}},
		{0x190, 1, 1, "NvType004D" "_DMA_MEM_4", {PRINT_OBJECT, {}}},
		{}},
	},
	{0,NULL, NVALL, 0,0, {
		{}},
	},
};

typedef struct obj_s obj_t;

struct obj_s {
	unsigned int parent;
	unsigned int name;
	unsigned int type;
	int dumped;

	obj_t *next;
};

static obj_t *mt_object_list = NULL;

static obj_t *find_object(unsigned int name)
{
	obj_t *temp;

	for (temp = mt_object_list; temp != NULL; temp = temp->next) {
		if (temp->name == name) {
			return temp;
		}
	}

	return NULL;
}

void object_create(unsigned int parent, unsigned int name, unsigned int type)
{
	/* Don't add existing object */
	if (find_object(name)) {
		return;
	}

	/* Clear file if list is empty */
	if (!mt_object_list) {
		FILE *f = fopen("card_objects.txt", "w");
		if (f) {
			fclose(f);
		}
	}

	obj_t *temp = malloc(sizeof(obj_t));

	temp->parent = parent;
	temp->name = name;
	temp->type = type;
	temp->dumped = 0;

	temp->next = mt_object_list;
	mt_object_list = temp;
}

unsigned int get_object_type(unsigned int name) {
	obj_t *temp;

	for (temp = mt_object_list; temp != NULL; temp = temp->next) {
		if (temp->name == name) {
			return temp->type & 0xff;
		}
	}

	return 0xffffffff;
}

static unsigned int find_object_instance(unsigned int name) {
	unsigned int htval = all_regs[NV_PFIFO_RAMHT/4];

	unsigned int ht_base;
	unsigned int ht_size;
	unsigned int ht_bits;
//	unsigned int ht_search;

	/* FIXME: fix it for G80 */

	if (card_family == NV03) {
		ht_base = 0x1000 * ((htval >> 12) & 0x0f);
		ht_size = 0x1000 << ((htval >> 16) & 0x03);
		ht_bits = ((htval >> 16) & 0x03) + 9;
	} else {
		ht_base = 0x1000 * ((htval >> 4) & 0x1f);
		ht_size = 0x1000 << ((htval >> 16) & 0x03);
		ht_bits = ((htval >> 16) & 0x03) + 9;
//		ht_search = 0x10 << ((htval >> 24) & 0x03);
	}

	unsigned int hash = 0;
	unsigned int context;

	int i;
	unsigned int temp;

	temp = name;
	for (i = 32; i > 0; i -= ht_bits) {
		hash ^= (temp & ((1 << ht_bits) - 1));
		temp >>= ht_bits;
	}
	if (fifo_channel != -1)
		hash ^= fifo_channel << (ht_bits - 4);
	hash <<= 3;

//	message("# Object hash is %08x", hash);

	int found = 0;

	if (all_regs[(NV_PRAMIN + ht_base + hash)/4] == name)
		found = 1;

	if (found == 0) {
//		message("# Object not found at expected place. bruteforcing the table. PLEASE REPORT!");
		// brute force hash!
		for(i=NV_PRAMIN + ht_base ;i<NV_PRAMIN+ht_base + ht_size; i+=4) {
			if (all_regs[i/4] == name) {
				if  (fifo_channel != -1) {
					if (card_family >= NV40) {
						if (((all_regs[(i+4)/4] >> 23) & 0x1F) != fifo_channel) {
							if (!is_name_dumped(name)) {
								printf("# HT search: possible match discarded (channel mismatch) handle=0x%08x,context=0x%08x\n",
									all_regs[i/4], all_regs[(i+4)/4]);
							}
							continue;
						}
					} else if (((all_regs[(i+4)/4] >> 24) & 0x1F) != fifo_channel) {
						if (!is_name_dumped(name)) {
							printf("# HT search: possible match discarded (channel mismatch) handle=0x%08x,context=0x%08x\n",
								all_regs[i/4], all_regs[(i+4)/4]);
						}
						continue;
					}
				}
				hash = i - NV_PRAMIN - ht_base;
				found = 1;
				break;
			}
		}
	}

	if (found == 0) {
		return 0xffffffff;
	}

//	message("# Object found at %08x, %08x", hash, hash ^ ohash);

	context = all_regs[(NV_PRAMIN + ht_base + hash + 4) / 4];
	return context;
}

static data_store dither_str[] = {
	{ 0, "compat"	},
	{ 1, "dither"	},
	{ 2, "truncate"	},
	{ 3, "subtract_truncate" },
	{}
};

static data_store dma_access[] = {
	{ 0, "rw"		},
	{ 1, "ro"		},
	{ 2, "wo"		},
	{}
};

static data_store dma_target[] = {
	{ 0, "NV mem"		},
	{ 1, "NV mem tiled"	},
	{ 2, "PCI"		},
	{ 3, "AGP"		},
	{}
};

static const object_store nv3_imem[] = {
	[ENGINE_SW] = {N_(ENGINE_SW), NV03 | NV04 | NV10 | NV15 | NV20 | NV30 , 0,0, {
		{__(0), {
			{0, 11, print_class, NULL, "class"},
			PRINT_ENABLE_BIT(12, "page table"),
			PRINT_ENABLE_BIT(13, "page entry linear"),
			{14, 15, print_macro, dma_access, "dma_access"},
			{16, 17, print_macro, dma_target, "dma_target"},
			{20, 31, print_format, "%03x", "dma adjust"},
			{}} },
		{__(1), {
			{0, 31, print_format, "%08x", "dma limit"},
			{}} },
		{__(2), {
			// 0x01 -- ??
			{12, 31, print_format, "%05x", "dma page address"},
			{1, 1, print_macro, bool_data, "r/w"},
			{}} },
		{__(3), {
			// 0x01 -- ??
			{12, 31, print_format, "%05x", "dma page address"},
			{1, 1, print_macro, bool_data, "r/w"},
			{}} },
		{}},
	},
	[ENGINE_GRAPHICS] = {N_(ENGINE_GRAPHICS), NV03 | NV04 | NV10 | NV15 | NV20 | NV30 , 0,0, {
		{__(0),	{
			{0, 11, print_class, NULL, "class"},
			PRINT_ENABLE_BIT(12, "chroma key"),
			PRINT_ENABLE_BIT(13, "user clip"),
			PRINT_ENABLE_BIT(14, "swizzle"),
			{15, 17, print_macro, nv04_operations, "patch"},
			PRINT_ENABLE_BIT(18, "synchronized"),
			PRINT_ENABLE_BIT(19, "big endian"),
			{20, 21, print_macro, dither_str, "dither"},
			// 22 ???
			PRINT_ENABLE_BIT(23, "single step"),
			{24, 24, print_macro, bool_data, "patch valid"},
			PRINT_ENABLE_BIT(25, "surface0"),
			PRINT_ENABLE_BIT(26, "surface1"),
			PRINT_ENABLE_BIT(27, "pattern"),
			PRINT_ENABLE_BIT(28, "rop"),
			PRINT_ENABLE_BIT(29, "beta1"),
			PRINT_ENABLE_BIT(30, "beta4"),
			// 31
			// 32
			{}} },
		{__(1),	{
			// FIXME: formats!
			{0, 7, print_macro, nv_mono_format, "mono format"},
			{8, 15, print_macro, nv_context_color_format, "color format"},
			{16, 31, print_format, "%04x", "notify address"},
			{}} },
		{__(2),	{
			{0, 15, print_format, "%04x", "dma0 address"},
			{16, 31, print_format, "%04x", "dma1 address"},
			{}} },
		{__(3),	{
			{}} },
		{}},
	},
	{}
};

unsigned int read_fb_ramin(unsigned instanceMem, int offset) {
	
	unsigned int ramin_block = (instanceMem >> 15) + 1;
	unsigned int ramin_inst  = (instanceMem & 0x7FFF) << 4;

	return fb[((fb_size - (ramin_block*512*1024) + ramin_inst)/4) + offset];	
}

unsigned int nv50_instmem_rv32(unsigned int offset)
{
	unsigned int old_bar0_pramin = all_regs[0x1700/4];
	unsigned int val;

	all_regs[0x1700/4] = (offset & 0xffff0000) >> 16;
	val = all_regs[(NV_PRAMIN + (offset & 0x0000ffff))/4];

	all_regs[0x1700/4] = old_bar0_pramin;
	return val;
}

int boot0_chipset(void)
{
	int pmc_boot_0;

	if (card_family < NV10)
		return 0x04;

	pmc_boot_0   = all_regs[0];
	pmc_boot_0  &= 0x0ff00000;
	pmc_boot_0 >>= 20;
	return pmc_boot_0;
}

unsigned int nv50_ramfc_offset(void)
{
	unsigned int ramfc;

	if (fifo_channel == -1) {
		message("# AIII, can't do this unless we know the channel id!\n");
		return ~0;
	}

	ramfc = all_regs[(0x2600 + (fifo_channel * 4))/4];
	if (!(ramfc & (1<<31))) {
		message("# AIII, channel marked disabled?\n");
		return ~0;
	}

	ramfc &= ~(1<<31);
	if (boot0_chipset() != 0x50)
		ramfc <<= 8;
	else
		ramfc <<= 12; /*XXX: untested, but "nv" does this..*/

	return ramfc;
}

unsigned int nv50_chan_pramin_base(void)
{
	unsigned int pramin;

	if (boot0_chipset() != 0x50) {
		unsigned int v = nv50_ramfc_offset() + 0x98;
		pramin = nv50_instmem_rv32(v) << 12;
	} else {
		pramin = nv50_ramfc_offset();
	}

	return pramin;
}

static unsigned int nv50_find_object_instance(unsigned int name) {
	unsigned int ht_base;
	int i;

	ht_base  = nv50_chan_pramin_base();
	ht_base += (nv50_instmem_rv32(nv50_ramfc_offset() + 0x80) & 0xffff) << 4;

	/* brute force.. */
	for (i=0; i<0x1000; i+=8) {
		if (nv50_instmem_rv32(ht_base + i) == name)
			return nv50_instmem_rv32(ht_base + i + 4);
	}

	return ~0;
}

unsigned int nv50_find_object_type_in_ramin(unsigned int name)
{
	unsigned int instance;
	unsigned int context;
	int i, dump_obj = 1;
	obj_t *obj;
	FILE *output = NULL;

	obj = find_object(name);

	context = nv50_find_object_instance(name);
	if (context == 0xffffffff) {
		if (!is_name_dumped(name))
			message("# Object not found in hash table! REPORT ASAP!!!!");
		return 0;
	}

	/* Add obj if we have needed context */
	if (!obj) {
		object_create(0, name, 0);
		obj = find_object(name);
	}

	instance  = nv50_chan_pramin_base();
	instance += (context & 0xffff) << 4;

	if (obj) {
		if (obj->dumped) {
			dump_obj = 0;
		}
	}
	if (dump_obj) {
		output = fopen("card_objects.txt", "a+");
		if (output) {
			fprintf(output, "%08x %08x", name, context);
			for (i=0; i<4; i++) {
				fprintf(output, " %08x", nv50_instmem_rv32(instance + (i*4)));
			}
			fprintf(output, "\n");
			fclose(output);
		}
		if (obj) {
			obj->dumped = 1;
		}
	}

	/* dump a bit of info about the object, plus first few dwords */
	if (!is_name_dumped(name)) {
		message("#--OBJ-- Handle : 0x%08x\n", name);
		message("#--OBJ-- Context: 0x%08x\n", context);
		message("#--OBJ-- Engine: %d\n", (context & (3<<20)) >> 20);
		message("#--OBJ-- Object:\n");
		for (i=0; i<4; i++)
			message("#--OBJ--\t+0x%04x = 0x%08x\n", i, nv50_instmem_rv32(instance + (i*4)));
	}

	return nv50_instmem_rv32(instance) & 0xffff;
}

unsigned int find_object_type_in_ramin(unsigned int name) {
	unsigned int instanceMem;
	unsigned int context;
	int i, dump_obj = 1;
	char buf[BUFSIZ];
	int read_ramin_from_fb_only = 0;
	obj_t *obj;
	FILE *output = NULL;

	if (!is_name_dumped(name)) {
		message("# Searching for object %08x", name);
	}

	obj = find_object(name);
	if (obj) {
		if (obj->dumped) {
			dump_obj = 0;
		}
	}

	context = find_object_instance(name);
	if (context == 0xffffffff) {
		if (!is_name_dumped(name)) {
			message("# Object not found in hash table! REPORT ASAP!!!!");
		}
		return 0;
	}

	/* Add obj if we have needed context */
	if (!obj) {
		object_create(0, name, 0);
		obj = find_object(name);
	}

	if (!is_name_dumped(name)) {
		message("# Context is %08x", context);
	}

	if (card_family >= NV40)
	instanceMem = context & 0xfffff;
	else
	instanceMem = context & 0xffff;

	int engine;
	int channel = -1;
	unsigned int data;

	switch (card_family) {
		case NV03:
			engine = (context >> 23) & 1;
			channel = (context >> 24) & 0x1f;
			break;
		case NV04:
		case NV10:
		case NV15:
		case NV20:
		case NV30:
			engine = (context >> 16) & 0x3;
			channel = (context >> 24) & 0x1f;
			break;
		case NV40:
		case G70:
		case G80:	/* FIXME: check it's right for G80 */
			engine = (context >> 20) & 0x3;
			channel = (context >> 23) & 0x1f; /* guess about mask */
			int unk = context & ~((0x1f<<23)|(0x3<<20)|(0xfffff<<0));
			if (unk != 0) {
				if (!is_name_dumped(name)) {
					message("# Unknown bits in context: %08x PLEASE REPORT!", unk);
				}
			}
			
			if (card_family == G70) {
				// For G70 cards (at least, not sure about G80)
				// do not read the ramin from the pci resource 1 (all_regs). 
				// Otherwise, you will get Xids, Xlockups, and other bad
				// things.
				// However, we cannot just blacklist G70 and make them 
				// only read ramin from the fb. This is because cards that 
				// use shared memory (Nv40+) do not store the ramin in the fb.
				//
				// Use this fb ramin read check to see if the card
				// stores the ramin in the fb or not.  
				// It seems to be reliable for most if not all cards. 
				if (read_fb_ramin(0x4, 2) == 0x1000) 
					read_ramin_from_fb_only = 1;
			}
			break;
	}

	switch (engine) {
		case ENGINE_SW:
			if (!is_name_dumped(name)) {
				message("# Software object");
			}
			break;
		case ENGINE_GRAPHICS:
			if (!is_name_dumped(name)) {
				message("# Graphics object");
			}
			break;
		default:
			if (!is_name_dumped(name)) {
				message("# Unknown object engine, please REPORT!");
			}
			break;
	}

	if (dump_obj) {
		output = fopen("card_objects.txt", "a+");
		if (output) {
			fprintf(output, "%08x %08x", name, context);
		}
	}

	if (card_family >= NV40) {
		int dumped = 0;
		/* Always dump from fb, except for some nv40/g70 */
		int dump_from_fb = 1;
		/*if (card_family==NV40)*/ {
			if (pci_device_id == 0x00f8) {
				dump_from_fb = 0;
			}
			switch(pci_device_id & 0xfff0) {
				case 0x01d0:
				case 0x0240:
				case 0x03d0:
					dump_from_fb = 0;
					break;
				default:
					break;
			}
		}
		if (!read_ramin_from_fb_only) {
			for (i = 0; i < 4; i++) {
				data = INSTANCE_MEM(i);
				// FIXME: format_data!
				buf[0] = '\0';
				if (output && dump_obj && !dump_from_fb) {
					fprintf(output, " %08x", data);
					dumped = 1;
				}
				if (!is_name_dumped(name)) {
					message("# instanceMem[%d] = %08x", i, data);
				}
			}
		}
		if (fb) {
			for (i = 0; i < 4; i++) {
				data = read_fb_ramin(instanceMem, i);
				if (output && dump_obj && dump_from_fb) {
					fprintf(output, " %08x", data);
					dumped = 1;
				}
				if (!is_name_dumped(name)) {
					message("# instanceMemFB[%d] = 0x%08x", i, data);
				}
			}
		}
		if (!dumped) {
			/* Output dummy data */
			for (i = 0; i < 4; i++) {
				if (output && dump_obj) {
					fprintf(output, " %08x", 0);
				}
			}
		}
	} else {
		for (i = 0; i < 4; i++) {
			data = INSTANCE_MEM(i);
			format_data(buf, BUFSIZ, nv3_imem[engine].fields[i].data, data);
			if (output && dump_obj) {
				fprintf(output, " %08x", data);
			}
			if (!is_name_dumped(name)) {
				message("# %s[%d] = %08x = %s", nv3_imem[engine].name, i, data, buf);
			}
		}
	}

	if (output) {
		if (dump_obj) {
			fprintf(output, "\n");
		}
		fclose(output);
	}
	if (obj) {
		obj->dumped = 1;
	}

	if (channel != -1) {
		if (!is_name_dumped(name)) {
			message("# Channel %d", channel);
		}
	}
	
	if (read_ramin_from_fb_only)
		return read_fb_ramin(instanceMem, 0) & 0xfff;
	else 
		return INSTANCE_MEM(0) & 0xfff;
}

unsigned int find_object_type(unsigned int name) {
	unsigned int type;
	unsigned int ramin_type;
	
	switch (card_family) {
	case G80:
		ramin_type = nv50_find_object_type_in_ramin(name);
		break;
	default:
		ramin_type = find_object_type_in_ramin(name);
		break;
	}

	if (ramin_type && card_family < NV40)
		return (ramin_type & 0xff);

	type = get_object_type(name);

	if (type == 0xffffffff) {
		type = ramin_type;
		if (!is_name_dumped(name)) {
			message("# Object not found in the table, guessed type %04x!!!", type);
		}
		object_create(0xffffffff, name, type);
	}

	return type;
}

char *format_load(int object_type, int offset, unsigned int value, int gpu_family) {
	static char buf[BUFSIZ];
	int pos = 0;
	const object_store *obj;
	const object_field_store *field;
	int i;

	for(i=0;nv_objects[i].name != NULL;i++)
		if ((nv_objects[i].id==object_type) && (nv_objects[i].card_type & gpu_family))
			break;

	obj = &nv_objects[i];

	field = &obj->fields[0];

	while (field->name != NULL) {
		if (offset >= field->offset && 
		    offset <= field->offset + (field->size - 1) * field->stride &&
		    (offset - field->offset) % field->stride == 0) {
			break;
		}

		field ++;
	}

	if (field->name == NULL) {
		if (obj->name == NULL) {
			memset(buf+pos, ' ', 20); pos += 20;
			pos += snprintf(buf+pos, BUFSIZ-pos, "NvType%04x", object_type);

			memset(buf+pos, ' ', 6); pos += 6;
			pos += snprintf(buf+pos, BUFSIZ-pos, "[0x%04x/4] = ", offset);
		} else if (offset != 0) {
			/* Try with parent object */
			if (nv_objects[i].parent_object && nv_objects[i].parent_gpu) {
				return format_load(nv_objects[i].parent_object, offset, value, nv_objects[i].parent_gpu);
			} else {
				pos += snprintf(buf+pos, BUFSIZ-pos, "%30s", obj->name);

				memset(buf+pos, ' ', 6); pos += 6;
				pos += snprintf(buf+pos, BUFSIZ-pos, "[0x%04x/4] = ", offset);
			}
		} else {
			memset(buf+pos, ' ', 40-strlen(obj->name)); pos += 40-strlen(obj->name);
			pos += snprintf(buf+pos, BUFSIZ-pos, "%s_NAME = ", obj->name);
		}

	} else {
		char buf2[BUFSIZ];
		int b2len = snprintf(buf2, BUFSIZ, field->name, (offset - field->offset) / field->stride);
		if (b2len < 45) {
			memset(buf+pos, ' ', 45-b2len); pos += 45-b2len;
		}
		pos += snprintf(buf+pos, BUFSIZ-pos, "%s = ", buf2);
	}

	if (offset == 0) {
		const char *str = print_object(value, NULL);

		strncpy(buf+pos, str, BUFSIZ-pos);
		pos += strlen(buf+pos);

		return buf;
	}

	const data_print_store* dp = &field->data[0];

	pos += format_data(buf+pos, BUFSIZ-pos, dp, value);

	return buf;
}



static int format_data(char *buf, int len, const data_print_store *dp, unsigned int value) {
	int pos = 0;
	unsigned int bitmask = 0;
	int i;

	buf[0] = '\0';

	int unk = 0;

	if (dp[0].func == NULL) {
		dp = user_data_print;
		unk = 1;
	}

	for (i = 0; dp[i].func != NULL; i++) {
		const char *str = dp[i].func(
				(value >> dp[i].lowbit) & (0xFFFFFFFF >> (31-dp[i].highbit + dp[i].lowbit)),
				dp[i].data);

		bitmask |= (0xFFFFFFFF >> (31-dp[i].highbit + dp[i].lowbit)) << dp[i].lowbit; 

		if (i != 0 && (dp[i].name != NULL || *str != '\0')) {
			strncpy(buf+pos, " | ", 3);
			pos += 3;
		}
		if (dp[i].name != NULL) {
			strncpy(buf+pos, dp[i].name, len-pos);
			pos += strlen(buf+pos);
			strncpy(buf+pos, " = ", len-pos);
			pos += 3;
		}
		strncpy(buf+pos, str, len-pos);
		pos += strlen(buf+pos);
	}

	if (unk) {
		bitmask = 0;
	}

	if ((value & ~bitmask) != 0) {
		snprintf(buf+pos, len-pos, " | UNKNOWN = %08x", value & ~bitmask);
	}

	return pos;
}

typedef struct _dmarec
{
	unsigned int from;
	unsigned int from_end;
	unsigned int to;
}
dmarec;

#define MAX_DMARECS 100
static dmarec  dmarecs[MAX_DMARECS];
static int next_rec = 0;

static unsigned int cur_offset_in;
static unsigned int cur_offset_out;
static unsigned int cur_length_in;
const char *
record_offset_in(unsigned int value, void *data)
{
	static char buf[11];
	cur_offset_in = value;
	snprintf(buf, sizeof(buf), "0x%08x", value);
	return buf;
}

const char *
record_offset_out(unsigned int value, void *data)
{
	static char buf[11];
	cur_offset_out = value;
	snprintf(buf, sizeof(buf), "0x%08x", value);
	return buf;
}

const char *
record_length_in(unsigned int value, void *data)
{
	static char buf[11];
	cur_length_in = value;
	snprintf(buf, sizeof(buf), "0x%08x", value);
	return buf;
}

const char *
record_dma_op(unsigned int value, void *data) {
	static char buf[10];
	int rec = -1;
	unsigned int ofs_in_len = (cur_length_in * value);
	unsigned int ofs_in_end = ofs_in_len + cur_offset_in;
	int i;

	/* If this overlaps a previous DMA, overwrite the previous record */
	for (i=0;i<next_rec;i++) {
		// starts before and overlaps
		if ((cur_offset_in <= dmarecs[i].from && ofs_in_end >= dmarecs[i].from) // start before, overlaps
				|| (cur_offset_in >= dmarecs[i].from && cur_offset_in <= dmarecs[i].from_end)) {// start inside
			rec = i;	
			break;
		}
	}
	if (rec == -1) {
		rec = next_rec;
		next_rec++;
	}

	if ((rec+1) < MAX_DMARECS) {
		dmarecs[rec].from     = cur_offset_in;
		dmarecs[rec].from_end = ofs_in_end;
		dmarecs[rec].to       = cur_offset_out;
	}

	snprintf(buf, 10, "%d", value); // line count
	return buf;
}

unsigned int
find_dma_src_from_dest(unsigned int to)
{
	int i;
	for (i=0;i<next_rec;i++)
		if (dmarecs[i].to == to) return dmarecs[i].from;
	return -1;
}

void clean_dmarecs() {
//	dmarecs[0].to = dmarecs[0].from;
	next_rec = 0;
}

static unsigned int name_list[256];

int is_name_dumped(int name)
{
	static int firsttime=1;
	int i;

	if (firsttime) {
		memset(name_list, 0, sizeof(name_list));
		firsttime=0;
	}

	for (i=0;i<256;i++) {
		if (name_list[i]==name) {
			return 1;
		}
		if (name_list[i]==0) {
			break;
		}
	}
	return 0;
}

void set_name_dumped(int name)
{
	/* Add at first available position */

	int i;
	for (i=0;i<256;i++) {
		if (name_list[i]==name) {
			/* Already present */
			break;
		}
		if (name_list[i]==0) {
			name_list[i]=name;
			break;
		}
	}
}
