#include <fcntl.h>
#include <math.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include <SDL.h>
#include <GL/gl.h>
#include <GL/glext.h>
#include <GL/glu.h>

#include "re.h"
#include "nv_hw.h"
#include "regl.h"
#include "objects.h"

extern unsigned int pci_device_id;

#define TEST_PROLOGUE redirect_output(__FUNCTION__)
#define TEST_EPILOGUE close_output()
#define __(a)	a, #a
#define N_(name)	name, # name

#define TEXTURE_WIDTH 16
#define TEXTURE_HEIGHT 16
#define TEXTURE_DEPTH 16	/* for 3d texture */
#define NPOT_TEXTURE_WIDTH 15	/* for nv_texture_rectangle or arb_npot */
#define NPOT_TEXTURE_HEIGHT 13

#define TEST_SPOTLIGHT_PARAMS 0
#define TEST_POINTPARAMETERS_PARAMS 0

#define CHECK_ERROR check_opengl_error(__LINE__)

#define _gl_error_switch(x) case x: erstr = #x; break;
static void check_opengl_error(int lineno)
{
	const char *erstr;
	int err = regl.GetError();
	switch(err)
	{
		case GL_NO_ERROR:
			return;
		_gl_error_switch(GL_INVALID_ENUM)
		_gl_error_switch(GL_INVALID_VALUE)
		_gl_error_switch(GL_INVALID_OPERATION)
		_gl_error_switch(GL_STACK_OVERFLOW)
		_gl_error_switch(GL_STACK_UNDERFLOW)
		_gl_error_switch(GL_OUT_OF_MEMORY)
		default:
			erstr = "unknown";
	}
	
	fprintf(stderr, "GL error %s (0x%04x) before line %d.\n",
	        erstr, err, lineno);
}
#undef _gl_error_switch

static void print_commented(const char *text)
{
	unsigned char c;
	printf("# ");
	while ((c = *text++)) {
		printf("%c", c);
		if (c=='\n') {
			printf("# ");
		}
	}
	printf("\n");
}

// draws a dumb triangle. usefull to force flushing of state changes to the card
void tri()
{
	regl.Begin(GL_TRIANGLES);
	regl.Vertex3i(1,0,0);
	regl.Vertex3i(0,1,0);
	regl.Vertex3i(0,0,1);
	regl.End();
}

void tex1_tri()
{
	regl.Begin(GL_TRIANGLES);
	regl.TexCoord2f(0.0,0.0);
	regl.Vertex3f(1.0,0.0,0.0);
	regl.TexCoord2f(0.0,1.0);
	regl.Vertex3f(0.0,1.0,0.0);
	regl.TexCoord2f(1.0,1.0);
	regl.Vertex3f(0.0,0.0,1.0);
	regl.End();
}

void tex_tri()
{
	regl.Begin(GL_TRIANGLES);
	regl.MultiTexCoord2f(GL_TEXTURE0, 1.0, 0.0);
	regl.MultiTexCoord2f(GL_TEXTURE1, 1.0, 0.0);
	regl.MultiTexCoord2f(GL_TEXTURE2, 1.0, 0.0);
	regl.MultiTexCoord2f(GL_TEXTURE3, 1.0, 0.0);
	regl.Vertex3i(1,0,0);
	regl.MultiTexCoord2f(GL_TEXTURE0, 0.0, 1.0);
	regl.MultiTexCoord2f(GL_TEXTURE1, 0.0, 1.0);
	regl.MultiTexCoord2f(GL_TEXTURE2, 0.0, 1.0);
	regl.MultiTexCoord2f(GL_TEXTURE3, 0.0, 1.0);
	regl.Vertex3i(0,1,0);
	regl.MultiTexCoord2f(GL_TEXTURE0, 0.0, 0.0);
	regl.MultiTexCoord2f(GL_TEXTURE1, 0.0, 0.0);
	regl.MultiTexCoord2f(GL_TEXTURE2, 0.0, 0.0);
	regl.MultiTexCoord2f(GL_TEXTURE3, 0.0, 0.0);
	regl.Vertex3i(0,0,1);
	regl.End();
}

void test_default(void)
{
	TEST_PROLOGUE;
	printf("# Testing default state of OpenGL context\n");
	dump_before();
	tri();
	dump_after(0);
	TEST_EPILOGUE;
}

void test_primitives(void)
{
	typedef struct {
		int num_points;
		data_store prim;
	} prim_t;

	const prim_t primitives[]={
		{ 1, { N_(GL_POINTS) } },
		{ 2, { N_(GL_LINES) } },
		{ 3, { N_(GL_LINE_LOOP) } },
		{ 3, { N_(GL_LINE_STRIP) } },
		{ 3, { N_(GL_TRIANGLES) } },
		{ 4, { N_(GL_TRIANGLE_STRIP) } },
		{ 4, { N_(GL_TRIANGLE_FAN) } },
		{ 4, { N_(GL_QUADS) } },
		{ 6, { N_(GL_QUAD_STRIP) } },
		{ 5, { N_(GL_POLYGON) } }
	};
	int i, j;

	TEST_PROLOGUE;

	for(i=0; i<sizeof(primitives)/sizeof(prim_t); i++) {
		printf("# Testing %s\n", primitives[i].prim.name);
		dump_before();
		regl.Begin(primitives[i].prim.id);
		for (j=0; j<primitives[i].num_points; j++) {
			regl.Vertex3i(1+j*3, 2+j*3, 3+j*3);
		}
		regl.End();
		dump_after(0);
	}
	TEST_EPILOGUE;
}

void *create_random_texture(int w, int h)
{
	uint32_t *texture;
	int i;

	texture = malloc(w*h*sizeof(uint32_t));
	if (!texture) {
		printf("# Can not allocate mem for texture\n");
		return NULL;
	}

	for (i=0; i<w*h; i++) {
		texture[i] = random();
	}

	return texture;
}

void test_arb_vertex_buffer_object(void)
{
#ifndef GL_ARB_vertex_buffer_object
	printf("# Sorry, no GL_ARB_vertex_buffer_object\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_vertex_buffer_object")) {
		fprintf(stderr,"No GL_ARB_vertex_buffer_object extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing ARB_vertex_buffer_object\n");

#define NUM_PRIMITIVES  8
#define NUM_VERTICES    (4*NUM_PRIMITIVES)
#define NUM_CLIENTSTATE 6
	GLfloat vtxdata[NUM_VERTICES*NUM_CLIENTSTATE];
	GLuint  bufid;

	dump_before();

	regl.GenBuffersARB(1, &bufid);
	regl.BindBufferARB(GL_ARRAY_BUFFER_ARB, bufid);
	regl.BufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(vtxdata), vtxdata, GL_STATIC_DRAW);

	regl.EnableClientState(GL_VERTEX_ARRAY);
	regl.EnableClientState(GL_COLOR_ARRAY);
	//regl.EnableClientState(GL_SECONDARY_COLOR_ARRAY);
	//regl.EnableClientState(GL_TEXTURE_COORD_ARRAY);
	//regl.EnableClientState(GL_NORMAL_ARRAY);
	//regl.EnableClientState(GL_FOG_COORD_ARRAY);

	regl.VertexPointer(3, GL_FLOAT, 0, NULL);
	regl.ColorPointer (3, GL_FLOAT, 0, NULL);
	//regl.SecondaryColorPointer (3, GL_FLOAT, 0, NULL);
	//regl.TexCoordPointer (4, GL_FLOAT, 0, NULL);
	//regl.NormalPointer (GL_FLOAT, 0, NULL);
	//regl.FogCoordPointer (GL_FLOAT, 0, NULL);

	regl.DrawArrays(GL_TRIANGLES, 0, NUM_VERTICES);

	dump_after(0);

	regl.DisableClientState(GL_VERTEX_ARRAY);
	regl.DisableClientState(GL_COLOR_ARRAY);
	//regl.DisableClientState(GL_SECONDARY_COLOR_ARRAY);
	//regl.DisableClientState(GL_TEXTURE_COORD_ARRAY);
	//regl.DisableClientState(GL_NORMAL_ARRAY);
	//regl.DisableClientState(GL_FOG_COORD_ARRAY);

	regl.DeleteBuffers(1, &bufid);
	TEST_EPILOGUE;
#endif
}


void test_arb_vertex_buffer_object_indexed(void)
{
#ifndef GL_ARB_vertex_buffer_object
	printf("# Sorry, no GL_ARB_vertex_buffer_object\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_vertex_buffer_object")) {
		fprintf(stderr,"No GL_ARB_vertex_buffer_object extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing ARB_vertex_buffer_object\n");

#define NUM_TRIANGLES   4000
#define NUM_VERTICES    (4*NUM_PRIMITIVES)
	
	uint16_t vtxdata[NUM_VERTICES*4];
	uint16_t inddata[NUM_TRIANGLES*3] = { 0 };
	GLuint  bufid;
	GLuint  indid;
	

	dump_before();

	regl.GenBuffersARB(1, &bufid);
	regl.GenBuffersARB(1, &indid);
	
	regl.BindBufferARB(GL_ARRAY_BUFFER_ARB, bufid);
	regl.BufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(vtxdata), vtxdata, GL_STATIC_DRAW);
	
	regl.BindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, indid);
	regl.BufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, sizeof(inddata), inddata, GL_STATIC_DRAW);
	
	

	regl.EnableClientState(GL_VERTEX_ARRAY);
	//regl.EnableClientState(GL_COLOR_ARRAY);
	//regl.EnableClientState(GL_SECONDARY_COLOR_ARRAY);
	//regl.EnableClientState(GL_TEXTURE_COORD_ARRAY);
	//regl.EnableClientState(GL_NORMAL_ARRAY);
	//regl.EnableClientState(GL_FOG_COORD_ARRAY);

	regl.VertexPointer(4, GL_SHORT, 0, NULL);
	//regl.ColorPointer (3, GL_FLOAT, 0, NULL);
	//regl.SecondaryColorPointer (3, GL_FLOAT, 0, NULL);
	//regl.TexCoordPointer (4, GL_FLOAT, 0, NULL);
	//regl.NormalPointer (GL_FLOAT, 0, NULL);
	//regl.FogCoordPointer (GL_FLOAT, 0, NULL);

	regl.DrawElements(GL_TRIANGLES, NUM_TRIANGLES * 3, GL_UNSIGNED_SHORT, 0);

	dump_after(0);

	regl.DisableClientState(GL_VERTEX_ARRAY);
	regl.DisableClientState(GL_COLOR_ARRAY);
	//regl.DisableClientState(GL_SECONDARY_COLOR_ARRAY);
	//regl.DisableClientState(GL_TEXTURE_COORD_ARRAY);
	//regl.DisableClientState(GL_NORMAL_ARRAY);
	//regl.DisableClientState(GL_FOG_COORD_ARRAY);

	regl.DeleteBuffers(1, &bufid);
	regl.DeleteBuffers(1, &indid);
	TEST_EPILOGUE;
#endif
}

void test_blend()
{
	TEST_PROLOGUE;
	printf("# Enable blending\n");
	dump_before();
	regl.Enable(GL_BLEND);
	tri();
	dump_after(0);

	printf("# 1 1\n");
	dump_before();
	regl.BlendFunc(GL_ONE,GL_ONE);
	tri();
	dump_after(0);

	printf("# Change color and func with color\n");
	dump_before();
	regl.BlendColor(0.2,0.3,0.4,0.5);
	regl.BlendFunc(GL_CONSTANT_COLOR,GL_ONE_MINUS_CONSTANT_COLOR);
	tri();
	dump_after(0);

	printf("# 0 0\n");
	dump_before();
	regl.BlendFunc(GL_ZERO,GL_ZERO);
	tri();
	dump_after(0);

	printf("# 1 0\n");
	dump_before();
	regl.BlendFunc(GL_ONE,GL_ZERO);
	tri();
	dump_after(0);

	printf("# 0 1\n");
	dump_before();
	regl.BlendFunc(GL_ZERO,GL_ONE);
	tri();
	dump_after(0);

	printf("# a (1-a)\n");
	dump_before();
	regl.BlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	tri();
	dump_after(0);

	printf("# 1 (1-a)\n");
	dump_before();
	regl.BlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
	tri();
	dump_after(0);

	printf("# Disable blending\n");
	dump_before();
	regl.BlendFunc(GL_ONE,GL_ZERO);
	regl.BlendColor(0.0,0.0,0.0,0.0);
	regl.Disable(GL_BLEND);
	tri();
	dump_after(0);
	TEST_EPILOGUE;
}

/**
  \fn    void test_scissor()
  \brief Test scissor functionality

  Please see doc/nv40_scissor.txt for further information
*/
void test_scissor()
{
        TEST_PROLOGUE;
	dump_before();

        // Now set scissor to smaller red sub region
        regl.Scissor(64, 80, 96, 112);
        regl.Enable(GL_SCISSOR_TEST);
        printf("# Enabled scissor %d\n", regl.IsEnabled(GL_SCISSOR_TEST));
        tri();
        // Turn scissor back off for next render
        regl.Disable(GL_SCISSOR_TEST);
        printf("# Disabled scissor %d\n", regl.IsEnabled(GL_SCISSOR_TEST));
        dump_after(0);
        TEST_EPILOGUE;
}

void test_ext_blend_func_separate()
{
#ifndef GL_EXT_blend_func_separate
	printf("# Sorry, no GL_EXT_blend_func_separate\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_blend_func_separate")) {
		fprintf(stderr,"No GL_EXT_blend_func_separate extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_ext_blend_func_separate()\n");

	dump_before();
	regl.Enable(GL_BLEND);
	regl.BlendFuncSeparateEXT(
			/* Sr */ GL_CONSTANT_COLOR,
			/* Dr */ GL_ONE_MINUS_DST_COLOR,
			/* Sa */ GL_SRC_ALPHA_SATURATE,
			/* Da */ GL_SRC_COLOR
			);
	tri();

	regl.BlendFuncSeparateEXT(GL_ONE,GL_ZERO,GL_ONE,GL_ZERO);
	tri();
	regl.Disable(GL_BLEND);
	tri();
	dump_after(0);

	TEST_EPILOGUE;
#endif
}

void test_ext_blend_equation_separate()
{
#ifndef GL_EXT_blend_equation_separate
	printf("# Sorry, no GL_EXT_blend_equation_separate\n");
#else
	int has_blend_equation_separate = 0;
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	const char *version = (const char *)regl.GetString(GL_VERSION);

	if (strstr(extensions, "GL_EXT_blend_equation_separate")) {
		has_blend_equation_separate = 1;
	} else if (version[0]>='2') {
		has_blend_equation_separate = 1;
	}

	if (!has_blend_equation_separate) {
		fprintf(stderr,"No GL_EXT_blend_equation_separate extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_ext_blend_equation_separate()\n");

	dump_before();
	regl.Enable(GL_BLEND);
	regl.BlendEquationSeparate(GL_FUNC_REVERSE_SUBTRACT, GL_FUNC_SUBTRACT);
	tri();

	regl.BlendEquationSeparate(GL_FUNC_ADD,GL_FUNC_ADD);
	tri();
	regl.Disable(GL_BLEND);
	tri();
	dump_after(0);

	TEST_EPILOGUE;
#endif
}

void test_ext_blend_logic_op()
{
#ifndef GL_EXT_blend_logic_op
	printf("# Sorry, no GL_EXT_blend_logic_op\n");
#else
	typedef struct {
		GLenum	value;
		char*	name;
	} logic_op;
	
	logic_op operations[16] = {
		{__(GL_CLEAR)},
		{ __(GL_AND) },
		{ __(GL_AND_REVERSE) },
		{ __(GL_COPY) },
		{ __(GL_AND_INVERTED) },
		{ __(GL_NOOP) },
		{ __(GL_XOR) },
		{ __(GL_OR) },
		{ __(GL_NOR) },
		{ __(GL_EQUIV) },
		{ __(GL_INVERT) },
		{ __(GL_OR_REVERSE) },
		{ __(GL_COPY_INVERTED) },
		{ __(GL_OR_INVERTED) },
		{ __(GL_NAND) },
		{ __(GL_SET) }
	};

	int i, use_ext_blend_logic_op = 0, use_color_logic_op = 0;
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	const char *version = (const char *)regl.GetString(GL_VERSION);

	if (strstr(extensions, "GL_EXT_blend_logic_op")) {
		use_ext_blend_logic_op = 1;
	} else if (strncmp(version,"1.1",3)>=0) {
		use_color_logic_op = 1;
	}

	if (!use_ext_blend_logic_op && !use_color_logic_op) {
		fprintf(stderr,"No GL_EXT_blend_logic_op extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_ext_blend_logic_op()\n");

	regl.Enable(GL_BLEND);
	if (use_ext_blend_logic_op) {
		regl.BlendEquationEXT(GL_LOGIC_OP);
	} else {
		regl.Enable(GL_COLOR_LOGIC_OP);
	}

	for (i=0; i<16; i++) {
		printf("\n# Testing %s\n\n", operations[i].name);
		dump_before();
 		regl.LogicOp(operations[i].value);
 		tri();
		dump_after(0);
	}

	regl.LogicOp(GL_COPY);
	regl.Disable(GL_COLOR_LOGIC_OP);
	regl.BlendEquation(GL_FUNC_ADD);
	regl.Disable(GL_BLEND);
	tri();

	TEST_EPILOGUE;
#endif
}

void test_depth()
{
	TEST_PROLOGUE;
	printf("# depth test enable\n");
	dump_before();
	regl.Enable(GL_DEPTH_TEST);
	tri();
	dump_after(0);

	printf("# depth test GREATER\n");
	dump_before();
	regl.DepthFunc(GL_GREATER);
	tri();
	dump_after(0);

	printf("# depth test LESS\n");
	dump_before();
	regl.DepthFunc(GL_LESS);
	tri();
	dump_after(0);

	printf("# depth mask FALSE\n");
	dump_before();
	regl.DepthMask(GL_FALSE);
	tri();
	dump_after(0);

	printf("# depth mask TRUE\n");
	dump_before();
	regl.DepthMask(GL_TRUE);
	tri();
	dump_after(0);

	printf("# depth range 0.12-0.56\n");
	dump_before();
	regl.DepthRange(0.12,0.56);
	tri();
	dump_after(0);

	printf("# depth range 0.0-1.0\n");
	dump_before();
	regl.DepthRange(0.0,1.0);
	tri();
	dump_after(0);

	printf("# depth test disabled\n");
	dump_before();
	regl.Disable(GL_DEPTH_TEST);
	tri();
	dump_after(0);
	TEST_EPILOGUE;
}

void test_color_mask(void)
{
	TEST_PROLOGUE;
	int i;

	dump_before();
	for (i=0; i<16; i++) {
		regl.ColorMask(
			(i & 1) ? GL_TRUE : GL_FALSE,
			((i>>1) & 1) ? GL_TRUE : GL_FALSE,
			((i>>2) & 1) ? GL_TRUE : GL_FALSE,
			((i>>3) & 1) ? GL_TRUE : GL_FALSE
		);
		tri();
	}
	dump_after(0);
	TEST_EPILOGUE;
}

void test_alpha()
{
	TEST_PROLOGUE;
	printf("# enable alpha test EQUAL 0.5f\n");
	dump_before();
	regl.Enable(GL_ALPHA_TEST);
	regl.AlphaFunc(GL_EQUAL,0.5f);
	tri();
	dump_after(0);

	printf("# enable alpha test LESS 0.5f\n");
	dump_before();
	regl.Enable(GL_ALPHA_TEST);
	regl.AlphaFunc(GL_LESS,0.5f);
	tri();
	dump_after(0);

	printf("# enable alpha test GREATER 0.5f\n");
	dump_before();
	regl.Enable(GL_ALPHA_TEST);
	regl.AlphaFunc(GL_GREATER,0.5f);
	tri();
	dump_after(0);

	printf("# disable alpha test\n");
	dump_before();
	regl.AlphaFunc(GL_ALWAYS,0.0);
	regl.Disable(GL_ALPHA_TEST);
	tri();
	dump_after(0);
	TEST_EPILOGUE;
}

void test_viewport()
{
	TEST_PROLOGUE;
	printf("# 30,60, 120x240\n");
	dump_before();
	regl.Viewport(30,60,120,240);
	tri();
	dump_after(0);

	printf("# 60,30, 240x120\n");
	dump_before();
	regl.Viewport(60,30,240,120);
	tri();
	dump_after(0);

	/* Restore default full size */
	printf("# 0 0 512 512\n");
	dump_before();
	regl.Viewport(0,0,512,512);
	tri();
	dump_after(0);

	TEST_EPILOGUE;
}

void test_lights()
{
	TEST_PROLOGUE;
	int i;
	regl.Enable(GL_LIGHTING);
	for(i=0;i<8;i++)
	{
		printf("# Testing LIGHT%i\n", i);
		
		dump_before();
		GLfloat amb[4] = {0.1f * i,0.0f * i,0.27f * i,1.0f};
		GLfloat dif[4] = {0.0f * i,0.2f * i,0.2f * i,1.0f};
		GLfloat spe[4] = {0.3f * i,0.0f * i,0.3f * i,1.0f};
		GLfloat pos[4] = {0.0f * i,0.4f * i,0.27f * i,1.0f};
		GLfloat spotdir[4] = {0.1f * i,0.3f * i,0.15f * i,1.0f * i};
		regl.Lightfv(GL_LIGHT0+i,GL_AMBIENT, amb);
		regl.Lightfv(GL_LIGHT0+i,GL_DIFFUSE, dif);
		regl.Lightfv(GL_LIGHT0+i,GL_SPECULAR, spe);
		regl.Lightfv(GL_LIGHT0+i,GL_POSITION, pos);
		regl.Lightfv(GL_LIGHT0+i,GL_SPOT_DIRECTION, spotdir);
		regl.Lightf(GL_LIGHT0+i,GL_SPOT_CUTOFF, 0.2f * i);
		regl.Lightf(GL_LIGHT0+i,GL_SPOT_EXPONENT, 2.0f * i);
		regl.Lightf(GL_LIGHT0+i,GL_CONSTANT_ATTENUATION, 0.5f * i);
		regl.Lightf(GL_LIGHT0+i,GL_LINEAR_ATTENUATION, 0.75f * i);
		regl.Lightf(GL_LIGHT0+i,GL_QUADRATIC_ATTENUATION, 0.95f * i);
		regl.Enable(GL_LIGHT0+i);		
		tri();
		dump_after(0);
	}

	printf("# disabling all lights\n");
	dump_before();
	for (i=0;i<8;i++)
		regl.Disable(GL_LIGHT0+i);
	regl.Disable(GL_LIGHTING);
	tri();
	dump_after(0);
	TEST_EPILOGUE;
}

void test_light_single(void)
{
	int num_light = GL_LIGHT0;
	GLfloat mat_ambient[]={0.01,0.02,0.03,0.04};
	GLfloat mat_diffuse[]={0.05,0.06,0.07,0.08};
	GLfloat mat_specular[]={0.09,0.10,0.11,0.12};
	GLfloat mat_emission[]={0.13,0.14,0.15,0.16};
	GLfloat mat_shininess[]={17.0};

	GLfloat lig_ambient[]={0.18,0.19,0.20,0.21};
	GLfloat lig_diffuse[]={0.22,0.23,0.24,0.25};
	GLfloat lig_specular[]={0.26,0.27,0.28,0.29};
	GLfloat lig_position[]={0.30,0.31,0.32,0.33};

	GLfloat lig_position_w0[]={0.30,0.31,0.32,0.0};
	GLfloat lig_position_w1[]={0.30,0.31,0.32,1.0};

	GLfloat lig_attconst = 0.40;
	GLfloat lig_attline = 0.41;
	GLfloat lig_attquad = 0.42;

	GLfloat lig_spotdir[]={0.50,0.51,0.52};
	GLfloat lig_spotexp = 1.53;
	GLfloat lig_spotcutoff = 60.0;

	TEST_PROLOGUE;
	printf("# test_light_single\n");
	dump_before();
	regl.Enable(GL_LIGHTING);
	regl.Enable(num_light);

	regl.Materialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
	regl.Materialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
	regl.Materialfv(GL_FRONT, GL_SPECULAR, mat_specular);
	regl.Materialfv(GL_FRONT, GL_EMISSION, mat_emission);
	regl.Materialfv(GL_FRONT, GL_SHININESS, mat_shininess);
	tri();

	regl.Lightfv(num_light, GL_AMBIENT, lig_ambient);
	regl.Lightfv(num_light, GL_DIFFUSE, lig_diffuse);
	regl.Lightfv(num_light, GL_SPECULAR, lig_specular);
	regl.Lightfv(num_light, GL_POSITION, lig_position);
	tri();

	regl.Lightfv(num_light, GL_POSITION, lig_position_w0);
	tri();

	regl.Lightfv(num_light, GL_POSITION, lig_position_w1);
	tri();

	regl.Lightf(num_light, GL_CONSTANT_ATTENUATION, lig_attconst);
	regl.Lightf(num_light, GL_LINEAR_ATTENUATION, lig_attline);
	regl.Lightf(num_light, GL_QUADRATIC_ATTENUATION, lig_attquad);
	tri();

	regl.Lightfv(num_light, GL_SPOT_DIRECTION, lig_spotdir);
	regl.Lightf(num_light, GL_SPOT_EXPONENT, lig_spotexp);
	regl.Lightf(num_light, GL_SPOT_CUTOFF, lig_spotcutoff);
	tri();

	dump_after(0);

#if TEST_SPOTLIGHT_PARAMS
	int i;
	GLfloat lig_spotdir_test[]={0.0,0.0,1.0};
	float spot_length = lig_spotdir_test[0]*lig_spotdir_test[0];
	spot_length += lig_spotdir_test[1]*lig_spotdir_test[1];
	spot_length += lig_spotdir_test[2]*lig_spotdir_test[2];
	spot_length = sqrt(spot_length);

	regl.Lightfv(num_light, GL_SPOT_DIRECTION, lig_spotdir_test);

	/* Cutoff angle test */
	regl.Lightf(num_light, GL_SPOT_EXPONENT, 1.0);
	for (i=10;i<=80;i+=10) {
		printf("# spot light cut off angle: %d\n", i);
		float cos_cutoff = cos(((float) i) * M_PI / 180);
		printf("#  cos(cutoff) = %f\n", cos_cutoff);
		printf("#  spot a = f1(cutoff,exponent)\n");
		printf("#  spot b = f2(exponent)\n");
		printf("#  spot c = f3(cutoff,exponent)\n");
		printf("#  spot dir x = %f\n", (1.0 / (cos_cutoff - 1.0)) * (lig_spotdir_test[0]/spot_length));
		printf("#  spot dir y = %f\n", (1.0 / (cos_cutoff - 1.0)) * (lig_spotdir_test[1]/spot_length));
		printf("#  spot dir z = %f\n", (1.0 / (cos_cutoff - 1.0)) * (lig_spotdir_test[2]/spot_length));
		printf("#  spot d = f4(cutoff,exponent) = %f\n", 1.0 + 1.0 / (cos_cutoff - 1.0) );
		dump_before();
		regl.Lighti(num_light, GL_SPOT_CUTOFF, i);
		tri();
		dump_after(0);
	}

	/* Exponent test */
	regl.Lighti(num_light, GL_SPOT_CUTOFF, 60); /* cos(cutoff) = 0.5 */
	for (i=1;i<=10;i++) {
		printf("# spot light exponent: %d\n", i);
		printf("#  spot a = f1(cutoff,exponent)\n");
		printf("#  spot b = f2(exponent)\n");
		printf("#  spot c = f3(cutoff,exponent)\n");
		printf("#  spot d = f4(cutoff,exponent)\n");
		dump_before();
		regl.Lighti(num_light, GL_SPOT_EXPONENT, i);
		tri();
		dump_after(0);
	}
#endif

	regl.Lighti(num_light, GL_SPOT_CUTOFF, 180);
	regl.Lightf(num_light, GL_SPOT_EXPONENT, 1.0);
	regl.Disable(num_light);
	regl.Disable(GL_LIGHTING);
	TEST_EPILOGUE;
}

void test_light_two_sides()
{
	TEST_PROLOGUE;
	dump_before();
	regl.Enable(GL_LIGHTING);
	regl.LightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	regl.Materialf(GL_FRONT, GL_SHININESS, 1.1);
	regl.Materialf(GL_BACK, GL_SHININESS, 1.3);
	tri();
	dump_after(0);

	regl.LightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
	regl.Disable(GL_LIGHTING);
	TEST_EPILOGUE;
}

void test_shade_model()
{
	TEST_PROLOGUE;
	printf("# test shade model FLAT\n");
	dump_before();
	regl.ShadeModel(GL_FLAT);
	tri();
	dump_after(0);

	printf("# test shade model SMOOTH\n");
	dump_before();
	regl.ShadeModel(GL_SMOOTH);
	tri();
	dump_after(0);
	TEST_EPILOGUE;
}

void test_enables()
{
	int i;
	data_store enable_tests[]=
	{
		{ N_(GL_DITHER)		},
		{ N_(GL_NORMALIZE)	},
		{ N_(GL_FOG)		},
		{ N_(GL_LINE_SMOOTH)	},
		{ N_(GL_LINE_STIPPLE)	},
		{ N_(GL_LIGHTING)	},
		{ N_(GL_MINMAX)		},
		{ N_(GL_POINT_SMOOTH)	},
		{ N_(GL_POLYGON_SMOOTH)	},
		{ N_(GL_POLYGON_STIPPLE)},
		{ N_(GL_COLOR_LOGIC_OP)	},
		{ N_(GL_SCISSOR_TEST)	},
		{ N_(GL_STENCIL_TEST)	},
		{ N_(GL_TEXTURE_GEN_Q)	},
		{ N_(GL_TEXTURE_GEN_R)	},
		{ N_(GL_TEXTURE_GEN_S)	},
		{ N_(GL_TEXTURE_GEN_T)	},
		{ N_(GL_TEXTURE_1D)	},
		{ N_(GL_TEXTURE_2D)	},
		{ N_(GL_TEXTURE_3D)	},
	};

	TEST_PROLOGUE;
	regl.Enable(GL_BLEND);

	for(i=0;i<sizeof(enable_tests)/sizeof(enable_tests[0]);i++)
	{
		printf("# test disable/enable %s\n",enable_tests[i].name);
		dump_before();
		regl.Enable(enable_tests[i].id);
		tri();
		regl.Disable(enable_tests[i].id);
		tri();
		dump_after(0);
	}

	regl.Disable(GL_BLEND);
	TEST_EPILOGUE;
}

void test_line_hint()
{
	TEST_PROLOGUE;
        printf("# test line Hint (GL_FASTEST)\n");
        dump_before();
        regl.Enable (GL_LINE_SMOOTH);
        regl.Hint(GL_LINE_SMOOTH_HINT, GL_FASTEST);
	regl.LineWidth((float)2.0);
	tri();
	dump_after(0);
	printf("# test line Hint (NICEST)\n");
	dump_before();
        regl.Hint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	regl.LineWidth((float)1.0);
	tri();
	dump_after(0);
	printf("# test line Hint (FASTEST)\n");
	dump_before();
        regl.Hint(GL_LINE_SMOOTH_HINT, GL_FASTEST);
	regl.LineWidth((float)3.0);
	tri();
	dump_after(0);
	
	TEST_EPILOGUE;
}

void test_line_width()
{
	TEST_PROLOGUE;
	printf("# test line width\n");
	int i;
	for(i=1;i<10;i++)
	{
		dump_before();
		regl.LineWidth((float)i);
		tri();
		dump_after(0);
	}
	
	TEST_EPILOGUE;
}

void test_line_stipple()
{
	TEST_PROLOGUE;
	int i;

	printf("# test line stipple\n");
	dump_before();
	//regl.Enable(GL_LINE_SMOOTH);
	//regl.Hint(GL_LINE_SMOOTH, GL_NICEST);
	regl.LineWidth(2.6);
	regl.Enable(GL_LINE_STIPPLE);
	for (i=0; i<4; i++) {
		regl.LineStipple(i+1, 0x5555 << i);
		regl.Begin(GL_LINES);
			regl.Vertex2f(0.1,1.9);
			regl.Vertex2f(1.8,0.2);
		regl.End();
	}	
	regl.Disable(GL_LINE_SMOOTH);
	regl.Disable(GL_LINE_STIPPLE);
	dump_after(0);
	TEST_EPILOGUE;
}

void test_polygon_stipple()
{
	// Stipple
	GLubyte map[] = 
	{ 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00 };

	typedef struct {
			GLenum	value;
			char*	name;
			int vals[2];
	} pixel_format;

	/* the values are ordered so that the default value is
	 * the last one used
	 */
	pixel_format formats[] =
	{
		{__(GL_PACK_SWAP_BYTES), {GL_TRUE, GL_FALSE} },
		{__(GL_PACK_LSB_FIRST), {GL_TRUE, GL_FALSE} },
		{__(GL_PACK_ROW_LENGTH), {5, 0} },
		{__(GL_PACK_SKIP_PIXELS), {1, 0} },
		{__(GL_PACK_SKIP_ROWS), {1, 0} },
		{__(GL_PACK_ALIGNMENT), {1, 2} },
		{__(GL_PACK_ALIGNMENT), {8, 4} },
		{__(GL_UNPACK_ALIGNMENT), {1, 2} },
		{__(GL_UNPACK_ALIGNMENT), {8, 4} },
		{__(GL_UNPACK_SWAP_BYTES), {GL_TRUE, GL_FALSE} },
		{__(GL_UNPACK_LSB_FIRST), {GL_TRUE, GL_FALSE} },
		{__(GL_UNPACK_ROW_LENGTH), {5, 0} },
		{__(GL_UNPACK_SKIP_PIXELS), {5, 0} },
		{__(GL_UNPACK_SKIP_ROWS), {1, 0} }
	};
	
	TEST_PROLOGUE;
	int i;

	printf("# test polygon stipple\n");

	printf("\n# Enable GL_POLYGON_STIPPLE\n\n");
	dump_before();
	regl.Enable(GL_POLYGON_STIPPLE);
	dump_after(0);

	for(i=0; i < 2*(sizeof(formats)/sizeof(formats[0])); i++)
	{
		dump_before();
		printf("\n# Testing %s, value %d\n",
		       formats[i/2].name, formats[i/2].vals[i%2]);
		regl.PixelStorei(formats[i/2].value, formats[i/2].vals[i%2]);
		regl.PolygonStipple(map);
		regl.Begin(GL_POLYGON);
			regl.Vertex2f(-0.5f, 1.0f);
			regl.Vertex2f(0.5f, 1.0f);
			regl.Vertex2f(1.0f, 0.5f);
			regl.Vertex2f(1.0f, -0.5f);
			regl.Vertex2f(0.5f, -1.0f);
			regl.Vertex2f(-0.5f, -1.0f);
			regl.Vertex2f(-1.0f, -0.5f);
			regl.Vertex2f(-1.0f, 0.5f);
		regl.End();
		dump_after(0);
	}

	printf("\n# Disable GL_POLYGON_STIPPLE\n\n");
	dump_before();
	regl.Disable(GL_POLYGON_STIPPLE);
	dump_after(0);
	
	TEST_EPILOGUE;
}

void test_polygon_offset()
{
	TEST_PROLOGUE;
	printf("# test glPolygonOffset\n");
	dump_before();
	regl.PolygonOffset(0.3f,0.4f);
	tri();
	dump_after(0);

	printf("# test disable/enable GL_POLYGON_OFFSET_FILL\n");
	dump_before();
	regl.Enable(GL_POLYGON_OFFSET_FILL);
	tri();
	dump_after(0);

	dump_before();
	regl.Disable(GL_POLYGON_OFFSET_FILL);
	tri();
	dump_after(0);

	printf("# test disable/enable GL_POLYGON_OFFSET_LINE\n");
	dump_before();
	regl.Enable(GL_POLYGON_OFFSET_LINE);
	tri();
	dump_after(0);

	dump_before();
	regl.Disable(GL_POLYGON_OFFSET_LINE);
	tri();
	dump_after(0);

	printf("# test disable/enable GL_POLYGON_OFFSET_POINT\n");
	dump_before();
	regl.Enable(GL_POLYGON_OFFSET_POINT);
	tri();
	dump_after(0);

	dump_before();
	regl.Disable(GL_POLYGON_OFFSET_POINT);
	tri();
	dump_after(0);

	TEST_EPILOGUE;
}

void test_cull_face()
{
	TEST_PROLOGUE;

	printf("# enable GL_CULL_FACE\n");
	dump_before();
	regl.Enable(GL_CULL_FACE);
	tri();
	dump_after(0);

	printf("# test cull GL_FRONT\n");
	dump_before();
	regl.CullFace(GL_FRONT);
	tri();
	dump_after(0);

	printf("# test cull GL_FRONT_AND_BACK\n");
	dump_before();
	regl.CullFace(GL_FRONT_AND_BACK);
	tri();
	dump_after(0);

	printf("# test cull GL_BACK\n");
	dump_before();
	regl.CullFace(GL_BACK);
	tri();
	dump_after(0);

	printf("# disable GL_CULL_FACE\n");
	dump_before();
	regl.Disable(GL_CULL_FACE);
	tri();
	dump_after(0);
	TEST_EPILOGUE;
}

void test_front_face()
{
	TEST_PROLOGUE;
	dump_before();

	regl.FrontFace(GL_CW);
	tri();
	regl.FrontFace(GL_CCW);	/* default */
	tri();

	dump_after(0);
	TEST_EPILOGUE;
}

void test_polygon_mode()
{
	TEST_PROLOGUE;
	dump_before();

	regl.PolygonMode(GL_FRONT_AND_BACK, GL_POINT);
	tri();

	regl.PolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	tri();

	regl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	tri();

	regl.PolygonMode(GL_FRONT, GL_POINT);
	tri();

	regl.PolygonMode(GL_FRONT, GL_LINE);
	tri();

	regl.PolygonMode(GL_FRONT, GL_FILL);
	tri();

	regl.PolygonMode(GL_BACK, GL_POINT);
	tri();

	regl.PolygonMode(GL_BACK, GL_LINE);
	tri();

	regl.PolygonMode(GL_BACK, GL_FILL);
	tri();

	dump_after(0);
	regl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);	/* default */
	TEST_EPILOGUE;
}

void test_matrix()
{
	TEST_PROLOGUE;
	dump_before();
	printf("# loading identity to the modelview matrix\n");
	regl.MatrixMode(GL_MODELVIEW);
	regl.LoadIdentity();
	tri();
	dump_after(0);

	dump_before();
	printf("# loading identity to the projection matrix\n");
	regl.MatrixMode(GL_PROJECTION);
	regl.LoadIdentity();
	tri();
	dump_after(0);
	TEST_EPILOGUE;
}

void test_clear()
{
	TEST_PROLOGUE;
	dump_before();
	printf("# clearing depth buffer\n");
	regl.ClearDepth(0.1);
	regl.Clear(GL_DEPTH_BUFFER_BIT);
	tri();
	dump_after(0);

	dump_before();
	printf("# clearing depth buffer with stencil\n");
	regl.Enable(GL_STENCIL_TEST);
	regl.ClearDepth(0.1);
	regl.Clear(GL_DEPTH_BUFFER_BIT);
	tri();
	regl.Disable(GL_STENCIL_TEST);
	dump_after(0);

	dump_before();
	printf("# clearing color buffer\n");
	regl.ClearColor(0.2,0.3,0.4,0.5);
	regl.Clear(GL_COLOR_BUFFER_BIT);
	tri();
	dump_after(0);

	dump_before();
	printf("# clearing color+depth buffers\n");
	regl.Clear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	tri();
	dump_after(0);

	dump_before();
	printf("# clearing stencil buffer\n");
	regl.ClearStencil(0x5a);
	regl.Clear(GL_STENCIL_BUFFER_BIT);
	tri();
	dump_after(0);

	dump_before();
	printf("# clearing accum buffer\n");
	regl.ClearAccum(0.7,0.8,0.9,1.0);
	regl.Clear(GL_ACCUM_BUFFER_BIT);
	tri();
	dump_after(0);

	TEST_EPILOGUE;
}

void test_tex()
{
	TEST_PROLOGUE;
	GLuint tid[10];
	GLuint tdata_uint[128*128*4];

	printf("# Testing multitexture, unit 0&1 enabled, default texenv state\n");

	dump_before();

	regl.GenTextures(10, tid);

	regl.ActiveTexture(GL_TEXTURE0);
	regl.Enable(GL_TEXTURE_2D);
	regl.MatrixMode(GL_TEXTURE);
	regl.Scalef(0.1,0.2,0.3);
	regl.BindTexture(GL_TEXTURE_2D, tid[0]);
	regl.TexImage2D(GL_TEXTURE_2D, 0,
			GL_RGBA,
			64, 64,
			0,
			GL_RGBA, GL_UNSIGNED_INT,
			tdata_uint
			);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	regl.ActiveTexture(GL_TEXTURE1);
	regl.Enable(GL_TEXTURE_2D);
	regl.MatrixMode(GL_TEXTURE);
	regl.Scalef(0.4,0.5,0.6);
	regl.BindTexture(GL_TEXTURE_2D, tid[1]);
	regl.TexImage2D(GL_TEXTURE_2D, 0,
			GL_RGB,
			32, 64,
			0,
			GL_RGBA, GL_UNSIGNED_INT,
			tdata_uint
			);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.MatrixMode(GL_MODELVIEW);
	regl.Translatef(-0.5, 0.5, 0.0);
	regl.Begin(GL_QUADS);
		regl.MultiTexCoord2f(GL_TEXTURE0, 0.0, 0.0);
		regl.MultiTexCoord2f(GL_TEXTURE1, 0.0, 0.0);
		regl.Vertex3f(0.0, 0.0, 0.0);
		regl.MultiTexCoord2f(GL_TEXTURE0, 0.0, 1.0);
		regl.MultiTexCoord2f(GL_TEXTURE1, 0.0, 1.0);
		regl.Vertex3f(0.0, 1.0, 0.0);
		regl.MultiTexCoord2f(GL_TEXTURE0, 1.0, 1.0);
		regl.MultiTexCoord2f(GL_TEXTURE1, 1.0, 1.0);
		regl.Vertex3f(1.0, 1.0, 0.0);
		regl.MultiTexCoord2f(GL_TEXTURE0, 1.0, 0.0);
		regl.MultiTexCoord2f(GL_TEXTURE1, 1.0, 0.0);
		regl.Vertex3f(1.0, 0.0, 0.0);
	regl.End();

	dump_after(0);

	regl.MatrixMode(GL_TEXTURE);
	regl.ActiveTexture(GL_TEXTURE1);
	regl.LoadIdentity();
	regl.Disable(GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE0);
	regl.LoadIdentity();
	regl.Disable(GL_TEXTURE_2D);
	regl.MatrixMode(GL_MODELVIEW);
	regl.LoadIdentity();

	TEST_EPILOGUE;
}

void test_arb_multitexture()
{
#ifndef GL_ARB_multitexture
	printf("# Sorry, no GL_ARB_multitexture\n");
#else
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_multitexture")) {
		fprintf(stderr,"No GL_ARB_multitexture extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_arb_multitexture()\n");

	int max_textures;
	regl.GetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &max_textures);
	printf("# max texture units: %d\n", max_textures);

	int i;
	GLuint tid;

	dump_before();

	regl.GenTextures(1, &tid);

	for (i=0;i<max_textures;i++) {
		regl.ActiveTexture(GL_TEXTURE0_ARB+i);
		regl.Enable(GL_TEXTURE_2D);

		regl.BindTexture(GL_TEXTURE_2D, tid);
		regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH, TEXTURE_HEIGHT,
			0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
		regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

		regl.MatrixMode(GL_TEXTURE);
		regl.Scalef(1.0*i,2.0*i,3.0*i);

		regl.Begin(GL_TRIANGLES);
			regl.MultiTexCoord2f(GL_TEXTURE0_ARB+i, 0.0, 1.0*i);
			regl.Vertex3f(0.0, 0.0, 0.0);
			regl.MultiTexCoord2f(GL_TEXTURE0_ARB+i, 2.0*i, 3.0*i);
			regl.Vertex3f(0.0, 1.0, 0.0);
			regl.MultiTexCoord2f(GL_TEXTURE0_ARB+i, 4.0*i, 5.0*i);
			regl.Vertex3f(1.0, 1.0, 0.0);
		regl.End();
	}	

	/* Test for multiple paletted texture */
	if (strstr(extensions, "GL_EXT_paletted_texture")) {
		regl.ColorTableEXT(GL_TEXTURE_2D, GL_RGBA, 256,
			GL_RGBA, GL_UNSIGNED_BYTE, texture);

		for (i=0;i<max_textures;i++) {
			regl.ActiveTexture(GL_TEXTURE0_ARB+i);
			regl.Enable(GL_TEXTURE_2D);

			regl.BindTexture(GL_TEXTURE_2D, tid);
			regl.TexImage2D(GL_TEXTURE_2D, 0, GL_COLOR_INDEX8_EXT, TEXTURE_WIDTH, TEXTURE_HEIGHT,
				0, GL_COLOR_INDEX, GL_UNSIGNED_BYTE, texture);
			regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

			regl.MatrixMode(GL_TEXTURE);
			regl.Scalef(1.0*i,2.0*i,3.0*i);

			regl.Begin(GL_TRIANGLES);
				regl.MultiTexCoord2f(GL_TEXTURE0_ARB+i, 0.0, 1.0*i);
				regl.Vertex3f(0.0, 0.0, 0.0);
				regl.MultiTexCoord2f(GL_TEXTURE0_ARB+i, 2.0*i, 3.0*i);
				regl.Vertex3f(0.0, 1.0, 0.0);
				regl.MultiTexCoord2f(GL_TEXTURE0_ARB+i, 4.0*i, 5.0*i);
				regl.Vertex3f(1.0, 1.0, 0.0);
			regl.End();
		}
	} else {
		printf("Can not test multitexture with textured palettes\n");
	}	

	dump_after(0);

	/* Disable all texture units */
	for (i=0;i<max_textures;i++) {
		regl.ActiveTexture(GL_TEXTURE0_ARB+i);
		regl.MatrixMode(GL_TEXTURE);
		regl.LoadIdentity();
		regl.Disable(GL_TEXTURE_2D);
	}
	regl.MatrixMode(GL_MODELVIEW);
	free(texture);
	TEST_EPILOGUE;

#endif
}

void test_display_list()
{
	TEST_PROLOGUE;
	printf("# creating a display list.\n");
	dump_before();
	GLuint i;
	int j;
	i=regl.GenLists(1);
  
	printf("# i=%d\n", i);
        srand(time(NULL) & 0x00ffffff);
	regl.NewList(i,GL_COMPILE);
	regl.Begin(GL_TRIANGLES);
	for(j=1; j <= 1000; j++)
	{
		regl.Vertex3i(rand()%20,rand()%20,rand()%20);
// 		regl.Vertex3i(j /10.0f,j /10.0f,j /10.0f);
	}
	regl.End();
	regl.EndList();
	dump_after(0);
	printf("# calling the display list.\n");
	dump_before();
	regl.CallList(i);
	dump_after(0);
	printf("# calling the display list again.\n");
	dump_before();
	regl.CallList(i);
	dump_after(0);
	TEST_EPILOGUE;
}

void test_clip_plane()
{
	int i, max_planes;
	GLdouble planes[4];
	
	TEST_PROLOGUE;
	printf("# test_clip_plane()\n");

	dump_before();

	regl.GetIntegerv(GL_MAX_CLIP_PLANES, &max_planes);
	printf("# %d clip planes available\n", max_planes);

	for (i=0; i<max_planes; i++) {
		regl.Enable(GL_CLIP_PLANE0+i);
		planes[0] = (1.0 * (float) i);
		planes[1] = planes[0] + 10.0;
		planes[2] = planes[0] + 100.0;
		planes[3] = planes[0] + 1000.0;
		regl.ClipPlane(GL_CLIP_PLANE0+i,planes);
		tri();
	}

	/* Disable all clip planes */
	for (i=0; i<max_planes; i++) {
		regl.Disable(GL_CLIP_PLANE0+i);
	}
	dump_after(0);
	TEST_EPILOGUE;
}

void test_clip_plane_after_vertex()
{
	TEST_PROLOGUE;
	int i, j, m;
	GLdouble planes[4*6]={
		0.0, 0.01, 0.02, 0.03,
		0.04, 0.05, 0.06, 0.07,
		0.08, 0.09, 0.10, 0.11,
		0.12, 0.13, 0.14, 0.15,
		0.16, 0.17, 0.18, 0.19,
		0.20, 0.21, 0.22, 0.23
	};
	printf("# Clipping planes after short vertex program test\n");
#ifdef GL_ARB_vertex_program
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_vertex_program")) {
		fprintf(stderr,"No GL_ARB_vertex_program extension.\n");
		return;
	} else printf("# Testing GL_ARB_vertex_program extension.\n");

	TEST_PROLOGUE;

	const char *vp =
		"!!ARBvp1.0"
		"TEMP t0;\n"
		""
		"ADD t0, vertex.color, vertex.position;\n"
		"MAD result.position, vertex.position, t0, t0.x;\n"
		"END";

	dump_before();
	GLuint s;
	regl.Enable(GL_VERTEX_PROGRAM_ARB);
	regl.GenProgramsARB(1, &s);
	regl.BindProgramARB(GL_VERTEX_PROGRAM_ARB, s);
	regl.GetError();
	regl.ProgramStringARB(GL_VERTEX_PROGRAM_ARB,
			GL_PROGRAM_FORMAT_ASCII_ARB, strlen(vp), vp);
	if (regl.GetError()) 
	{ 
		regl.DeleteProgramsARB(1, &s);
		regl.Disable(GL_VERTEX_PROGRAM_ARB);
		printf("# err - s\n"); 
		return; 
	}
	tri();
#else
        #warning  No GL_ARB_vertex_program defined or available
#endif
	regl.GetIntegerv(GL_MAX_CLIP_PLANES, &m);
	printf("# %d clip planes available\n", m);
	for (i=1; i<=m; i++) {
		printf("# %d clip planes.\n", i);

		dump_before();
		for (j=0;j<i;j++) {
			regl.Enable(GL_CLIP_PLANE0+j);
			regl.ClipPlane(GL_CLIP_PLANE0+j,&planes[j*4]);
		}
		tri();
		for (j=0;j<i;j++) {
			regl.Disable(GL_CLIP_PLANE0+j);
		}
		dump_after(0);
	}
	regl.DeleteProgramsARB(1, &s);
	regl.Disable(GL_VERTEX_PROGRAM_ARB);
	TEST_EPILOGUE;
}

#define N 12
void test_arb_occlusion_query() 
{
#ifdef GL_ARB_occlusion_query
        GLuint queries[N];  
        GLuint sampleCount;  
        GLint available;  
        GLint bitsSupported;  
        int    i;
  
        // check to make sure functionality is supported  
        regl.GetQueryiv(GL_SAMPLES_PASSED, GL_QUERY_COUNTER_BITS_ARB, &bitsSupported); 
        if (bitsSupported == 0) {  
            // render scene without using occlusion queries  
            printf("# No Queries supported\n");
            return;
        }  

	TEST_PROLOGUE;
        printf("# test_arb_occlusion_query\n");
        dump_before();

        regl.GenQueriesARB(N, queries);  
        regl.ColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);  
        regl.DepthMask(GL_FALSE);  
	dump_after(0);
        printf("# Created query\n");
        // also disable texturing and any fancy shaders  
        for (i = 0; i < N; i++) {  
            printf("\n# Query start %d\n", i);
            dump_before();
            regl.BeginQueryARB(GL_SAMPLES_PASSED_ARB, queries[i]);  
            // render bounding box for object i  
            tri();
            regl.EndQueryARB(GL_SAMPLES_PASSED_ARB);  
            dump_after(0);
        }  
        printf("# Queries all done\n");
        dump_before();
        regl.Flush();  

	int delay = 0;
        do {  
		regl.GetQueryObjectivARB(queries[N-1],  
			GL_QUERY_RESULT_AVAILABLE_ARB,  
			&available);  
		printf("# Available=%d\n", available);
		SDL_Delay(500);	/* wait 500ms */
		delay += 500;	/* we waited 500ms more */
        } while (!available && (delay<10000));	/* till result avail or 10 second elapsed */
  
        regl.ColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);  
        regl.DepthMask(GL_TRUE);  
	if (available) {
	        for (i = 0; i < N; i++) {  
        	    regl.GetQueryObjectuivARB(queries[i], GL_QUERY_RESULT_ARB,  
                                      &sampleCount);  
	            printf("# Sample Count %d\n", sampleCount);
	        }
	} else {
		printf("# Query result not available\n");
	}
        regl.DeleteQueriesARB(N, queries);
        dump_after(0);
        TEST_EPILOGUE;
#endif
}

void test_arb_vertex_program()
{
#ifndef GL_ARB_vertex_program
	printf("# Sorry, no GL_ARB_vertex_program\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_vertex_program")) {
		fprintf(stderr,"No GL_ARB_vertex_program extension.\n");
		return;
	}

	char vp[4096];
	int i;
	const char *vps[]  = {
		"MOV result.position, vertex.position;",
		"MOV result.position, vertex.color;",
		"MOV result.position, vertex.color.secondary;",
		"MOV result.position, vertex.texcoord[0];",
		"MOV result.position, vertex.texcoord[1];",
		"MOV result.position, vertex.texcoord[2];",
		"MOV result.position, vertex.texcoord[3];",
		"MOV result.position, vertex.weight;",
		"MOV result.position, vertex.normal;",
		"MOV result.position, vertex.fogcoord;",
		"MOV result.color, vertex.position;",
		"MOV result.color.secondary, vertex.position;",
		"MOV result.texcoord[0], vertex.position;",
		"MOV result.texcoord[1], vertex.position;",
		"MOV result.texcoord[2], vertex.position;",
		"MOV result.texcoord[3], vertex.position;",
		"MOV result.color.back, vertex.position;",
		"MOV result.color.back.secondary, vertex.position;",
		"MOV result.fogcoord, vertex.position;",
		"MOV result.pointsize, vertex.position;",
		"MOV t0, vertex.position; ADD result.position, vertex.color, t0;",
		"MOV t0, vertex.position; SUB result.position, vertex.color, t0;",
		"MOV t0, vertex.position; SUB result.position, t0, vertex.color;",
		"SUB t0, vertex.position, vertex.color; ABS result.position, t0;",
		"MOV t0, vertex.position; ADD result.position.x, vertex.color, t0;",
		"MOV t0, vertex.position; MUL result.position.y, vertex.color, t0;",	
		"MOV t0, vertex.position; DP3 result.position.z, vertex.color, t0;",
		"MOV t0, vertex.position; DP4 result.position.w, vertex.color, t0;",
		"MOV t0, vertex.position; DPH result.position, vertex.color, t0;",
		"MOV t0, vertex.position; DST result.position, vertex.color, t0;",
		"MOV t0, vertex.position; MIN result.position, vertex.color, t0;",
		"MOV t0, vertex.position; MAX result.position, vertex.color, t0;",
		"MOV t0, vertex.position; SLT result.position, vertex.color, t0;",
		"MOV t0, vertex.position; SGE result.position, vertex.color, t0;",

		"MOV t0, vertex.position;"
		"MAD result.position, vertex.color, t0, {1.0, 0.50, 0.25, 0.00};",

		"MOV t0, vertex.position;"
		"MAD result.position, vertex.color, t0, {1.0, 0.50, 0.50, 0.00};",

		"LIT result.position, vertex.position;",
		"RCP result.position.x, vertex.position.x;",
		"RSQ result.position.y, vertex.position.x;",
		"EXP result.position.z, vertex.position.x;",
		"LOG result.position.w, vertex.position.x;",
		"LIT result.position, vertex.position.x;",
		"EX2 result.position, vertex.position.x;",
		"LG2 result.position, vertex.position.x;",

		"ADD t0, vertex.color, vertex.position;"
		"RCP t1, vertex.position.x;"
		"MAD result.position, t0, t1.x, vertex.position;",

		"ADD t0.x, vertex.color, vertex.position;"
		"RCP t1, vertex.position.x;"
		"MAD result.position, t0, t1.x, vertex.position;",

		"ADD t0.y, vertex.color, vertex.position;"
		"RCP t1, vertex.position.x;"
		"MAD result.position, t0, t1.x, vertex.position;",

		"ADD t0.z, vertex.color, vertex.position;"
		"RCP t1, vertex.position.x;"
		"MAD result.position, t0, t1.x, vertex.position;",

		"ADD t0.w, vertex.color, vertex.position;"
		"RCP t1, vertex.position.x;"
		"MAD result.position, t0, t1.x, vertex.position;",

		"ADD t0, vertex.color, vertex.position;"
		"RCP t1.x, vertex.position.x;"
		"MAD result.position, t0, t1.x, vertex.position;",

		"ADD t0, vertex.color, vertex.position;"
		"RCP t1.y, vertex.position.x;"
		"MAD result.position, t0, t1.y, vertex.position;",

		"ADD t0, vertex.color, vertex.position;"
		"RCP t1.z, vertex.position.x;"
		"MAD result.position, t0, t1.z, vertex.position;",

		"ADD t0, vertex.color, vertex.position;"
		"RCP t1.w, vertex.position.x;"
		"MAD result.position, t0, t1.w, vertex.position;",

		"RCP t0, vertex.position.x;"
		"ADD t1, vertex.color, vertex.position;"
		"MAD result.position, t1, t0.x, vertex.position;",

		"ADD t0, vertex.position, {0.0, 0.1, 0.2, 0.3};"
		"MAD result.position, t0, vertex.color, {0.12, 0.13, 0.14, 0.15};",

		"ARL A0.x, vertex.position.x; MOV result.position, mat[A0.x + 1];",
		"ARL A0.x, vertex.position.x; MOV result.position, mat[A0.x + 2];",
		"ARL A0.x, vertex.position.x; MOV result.position, mat[A0.x + 3];",
		"FLR result.position, vertex.position;",
		"FRC result.position, vertex.position;",
		"POW result.position, vertex.position.x, vertex.position.y;",
		"XPD result.position, vertex.position, vertex.color;",

		"RCP t0.x, vertex.position.x;\n"
		"RCP t0.y, vertex.position.y;\n"
		"RCP t0.z, vertex.position.z;\n"
		"RCP t0.w, vertex.position.w;\n"
		"RCP t1.x, vertex.color.x;\n"
		"RCP t1.y, vertex.color.y;\n"
		"RCP t1.z, vertex.color.z;\n"
		"RCP t1.w, vertex.color.w;\n"
		"MAD result.position, t0, vertex.position, t1;\n",

		NULL
	};

	TEST_PROLOGUE;
	printf("# test_arb_vertex_program\n");

	regl.Enable(GL_VERTEX_PROGRAM_ARB);

	i = 0;
	while (vps[i]) {
		snprintf(vp, sizeof(vp),
				"!!ARBvp1.0\n"
				"TEMP t0, t1, t2;\n"
				"PARAM mat[4] = { state.matrix.mvp };\n"
				"ADDRESS A0;\n"
				"%s\n"
				"END\n",
				vps[i]);
		printf("# Testing:\n");
		print_commented(vp);
		dump_before();
		GLuint s;
		regl.GenProgramsARB(1, &s);
		regl.BindProgramARB(GL_VERTEX_PROGRAM_ARB, s);
		regl.GetError();
		regl.ProgramStringARB(GL_VERTEX_PROGRAM_ARB,
				GL_PROGRAM_FORMAT_ASCII_ARB, strlen(vp), vp);
		if (regl.GetError()) 
		{ 
			regl.DeleteProgramsARB(1, &s);
			printf("# err - s\n"); 
			break; 
		}
		tri();
		dump_after(0);
		regl.DeleteProgramsARB(1, &s);

		i++;
	}

	regl.Disable(GL_VERTEX_PROGRAM_ARB);
	TEST_EPILOGUE;
#endif
}

void test_nv_vertex_program1_1()
{
#ifndef GL_NV_vertex_program1_1
	printf("# Sorry, no GL_NV_vertex_program1_1\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_vertex_program1_1")) {
		fprintf(stderr,"No GL_NV_vertex_program1_1 extension.\n");
		return;
	}

	GLuint vpobj;
	const char *vp =
		"!!VP1.1\n"
		"RCC o[HPOS], v[OPOS].x;\n"
		"END\n";

	TEST_PROLOGUE;
	printf("# test_nv_vertex_program1_1\n");
	dump_before();
	regl.Enable(GL_VERTEX_PROGRAM_NV);
	regl.GenProgramsNV(1, &vpobj);
	regl.LoadProgramNV(GL_VERTEX_PROGRAM_NV, vpobj, strlen(vp), (const GLubyte*)vp);
	regl.BindProgramNV(GL_VERTEX_PROGRAM_NV, vpobj);
	tri();
	dump_after(0);
	regl.DeleteProgramsNV(1, &vpobj);
	regl.Disable(GL_VERTEX_PROGRAM_NV);
	
	TEST_EPILOGUE;
#endif
}

void test_nv_vertex_program2()
{
#ifndef GL_NV_vertex_program2
	printf("# Sorry, no GL_NV_vertex_program2\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_vertex_program2")) {
		fprintf(stderr,"No GL_NV_vertex_program2 extension.\n");
		return;
	}

	char vp[4096];
	GLuint vpobj;
	int i;

	const char *vps[] = {
		"MOV R0, v[COL0]; SEQ o[HPOS], v[OPOS], R0;",
		"MOV R0, v[COL0]; SFL o[HPOS], v[OPOS], R0;",
		"MOV R0, v[COL0]; SGT o[HPOS], v[OPOS], R0;",
		"MOV R0, v[COL0]; SLE o[HPOS], v[OPOS], R0;",
		"MOV R0, v[COL0]; SNE o[HPOS], v[OPOS], R0;",
		"MOV R0, v[COL0]; STR o[HPOS], v[OPOS], R0;",
		"SSG o[HPOS], v[OPOS];",
		"COS o[HPOS], v[OPOS].x;",
		"SIN o[HPOS], v[OPOS].x;",

		"ARL A0.x, v[OPOS].x;"
		"ARR A0.y, v[OPOS].y;"
		"MOV R0, c[A0.x + 1];"
		"ADD o[HPOS], c[A0.y + 1], R0;",

		"ARL A0, c[0]; ARAC A0.xy, A0; ADD o[HPOS], c[A0.y + 1], c[A0.x + 2];",
		"ADDC o[HPOS], v[OPOS], c[0]; MOV o[COL0] (EQ.x), c[1];",
		"ADDC o[HPOS], v[OPOS], c[0]; MOV o[COL0] (GE.y), c[1];",
		"ADDC o[HPOS], v[OPOS], c[0]; MOV o[COL0] (GT.z), c[1];",
		"ADDC o[HPOS], v[OPOS], c[0]; MOV o[COL0] (LE.w), c[1];",
		"ADDC o[HPOS], v[OPOS], c[0]; MOV o[COL0] (LT)  , c[1];",
		"ADDC o[HPOS], v[OPOS], c[0]; MOV o[COL0] (NE.xxxy), c[1];",
		"ADDC o[HPOS], v[OPOS], c[0]; MOV o[COL0] (TR.xxxz), c[1];",
		"ADDC o[HPOS], v[OPOS], c[0]; MOV o[COL0] (FL.xxxw), c[1];",

		"ADDC o[HPOS], v[OPOS], c[0];"
		"BRA cond_lt (LT);"
		"MOV o[COL0], v[COL0];"
		"cond_lt:",

		"ADDC o[HPOS], v[OPOS], c[0];"
		"BRA cond_lt (LT);"
		"ADD R0, v[COL0], c[0];"
		"MUL o[COL0], R0, v[COL1];"
		"cond_lt:",

		"main:"
		"  ADDC o[HPOS], v[OPOS], c[0];"
		"  CAL cond_lt (LT);"
		"  BRA end;"
		"cond_lt:"
		"  ADD R0, v[COL0], c[0];"
		"  MUL o[COL0], R0, v[COL1];"
		"  RET;"
		"end:",

		"cond_lt:"
		"  ADD R0, v[COL0], c[0];"
		"  MUL o[COL0], R0, v[COL1];"
		"  RET;"
		"main:"
		"  ADDC o[HPOS], v[OPOS], c[0];"
		"  CAL cond_lt (LT);"
		"  BRA end;"
		"end:",
		
		"ARL A0, v[OPOS];\n"
		"ARL A1, v[COL0];\n"
		"SFL R0, R0, R0;\n"
		"bleh:\n"
		"   ADD R0, R0, c[A0.x + 1];\n"
		"   ADD R0, R0, c[A1.x + 1];\n"
		"   ARAC A1.xy, A1;\n"
		"   BRA bleh (GT.x);\n"
		"MOV o[HPOS], R0;\n",
		
		NULL
	};

	TEST_PROLOGUE;
	printf("# test_nv_vertex_program2\n");
	GLfloat current_color[4];
	regl.Enable(GL_VERTEX_PROGRAM_NV);
	/* You must set color to 0 to avoid a possible infinite loop above.
	 * This loop only terminates if color.x <= 0.
	 * The odd thing is that the lockup only occurs on my NVAA but not
	 * on my NV49.  Perhaps this condition is detected?*/ 
	regl.GetFloatv(GL_CURRENT_COLOR, current_color);
	regl.Color4f(0.0f,0.0f,0.0f,0.0f);

	i = 0;
	while (vps[i]) {
		snprintf(vp, sizeof(vp),
				"!!VP2.0\n"
				"%s\n"
				"END\n",
				vps[i]);
		printf("# Testing:\n");
		print_commented(vp);
		dump_before();
		regl.GenProgramsNV(1, &vpobj);
		regl.GetError();
		regl.LoadProgramNV(GL_VERTEX_PROGRAM_NV, vpobj, strlen(vp), (const GLubyte*)vp);
		if (regl.GetError()) 
		{ 
			regl.DeleteProgramsNV(1, &vpobj); 
			printf("# err - s\n"); 
			break; 
		}
		regl.BindProgramNV(GL_VERTEX_PROGRAM_NV, vpobj);
		regl.TrackMatrixNV(GL_VERTEX_PROGRAM_NV, 0, GL_MODELVIEW_PROJECTION_NV, GL_IDENTITY_NV);
		CHECK_ERROR;
		tri();
		dump_after(0);
		regl.DeleteProgramsNV(1, &vpobj);

		i++;
	}

	regl.Color4f(current_color[0],current_color[1],current_color[2],current_color[3]);
	regl.Disable(GL_VERTEX_PROGRAM_NV);
	TEST_EPILOGUE;
#endif
}

void test_glsl_vp()
{
#ifndef GL_ARB_shading_language_100
	printf("# Sorry, no GLSL\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_shading_language_100") ||
	    !strstr(extensions, "GL_ARB_shader_objects") ||
	    !strstr(extensions, "GL_ARB_fragment_shader") ||
	    !strstr(extensions, "GL_ARB_vertex_shader")) {
		fprintf(stderr,"No GLSL vertex program extension\n");
		return;
	}

#if 0
	const char *vp =
		"void main()\n"
		"{\n"
		"    vec4 pos = gl_Vertex;\n"
		"    float i;\n"
		""
		"    do {\n"
		"       pos.y += pos.x;\n"
		"       pos.z += pos.y;\n"
		"    } while (pos.y < 30.0);\n"
		"    gl_Position = pos;\n"
		"}\n";
#else
	const char *vp =
		"void main()\n"
		"{\n"
		"    if (gl_Vertex.x != 0.0) {\n"
		"        gl_Position   = gl_Vertex + gl_Color;\n"
		"        gl_FrontColor = gl_Color;\n"
		"    } else {\n"
		"        gl_Position   = gl_Color;\n"
		"        gl_FrontColor = gl_Vertex;\n"
		"    }\n"
		"}\n";
#endif

	const int vps = strlen(vp);
	GLuint po, vo;

	TEST_PROLOGUE;
	printf("# test_glsl_vp\n");
	dump_before();

	vo = regl.CreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	regl.ShaderSourceARB(vo, 1,  &vp, &vps);
	regl.CompileShaderARB(vo);

	po = regl.CreateProgramObjectARB();
	regl.AttachObjectARB(po, vo);
	regl.LinkProgramARB(po);
	regl.UseProgramObjectARB(po);

	tri();
	dump_after(0);
	
	regl.DeleteObjectARB(po);
	regl.DeleteObjectARB(vo);
	regl.UseProgramObjectARB(0);

	TEST_EPILOGUE;
#endif
}

void test_nv_vertex_program3()
{
#ifndef GL_NV_vertex_program3
	printf("# Sorry, no GL_NV_vertex_program3\n");
#else
/* This check shouldn't be needed nymore.
 * For me (NVAA, 8200) the crash seems to happens because of the PUSHA 
 * instruction.  So, a different vp is used for G80.
	if (card_family >= G80)
	{
		fprintf(stderr,"GL_NV_vertex_program3 crashes up until driver 100.14.19 on G80 - skipping.\n");
		return;
	}
*/
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_vertex_program3")) {
		fprintf(stderr,"No GL_NV_vertex_program3 extension.\n");
		return;
	}
	
	const char *vp =
		"!!ARBvp1.0\n"
		"OPTION NV_vertex_program3;\n"
		"PARAM count = { 10.0, 0.0, -1.0, 1.0 };\n"
		"PARAM count2 = {  2.0, 0.0, -1.0, 1.0 };\n"
		"ADDRESS A0;\n"
		"TEMP t0, r0;\n"
		""
		"sub1:"
		"  MAD r0, r0, vertex.color, vertex.position;\n"
		"  RET;\n"
		""
		"main:"
		"  ARL A0, count;\n"
		"  loopS:\n"
		"    PUSHA A0;\n"
		"    ARL A0, count2;\n"
		"    loop2S:\n"
		"      CAL sub1;\n"
		"      ARAC A0.xy, A0;\n"
		"      BRA loop2S (GT.x);\n"
		"    POPA A0;\n"
		"    ARAC A0.xy, A0;\n"
		"    BRA loopS (GT.x);\n"
		"  MOV result.position, r0;\n"
		"  BRA end;\n"
		""
		"end:\n"
		"END";
		
	const char *g80_vp =
		"!!ARBvp1.0\n"
		"OPTION NV_vertex_program3;\n"
		"PARAM count = { 10.0, 0.0, -1.0, 1.0 };\n"
		"PARAM count2 = {  2.0, 0.0, -1.0, 1.0 };\n"
		"ADDRESS A0, A1;\n"
		"TEMP t0, r0;\n"
		""
		"sub1:"
		"  MAD r0, r0, vertex.color, vertex.position;\n"
		"  RET;\n"
		""
		"main:"
		"  ARL A0, count;\n"
		"  loopS:\n"
		"  ARL A1, count2;\n"
		"    loop2S:\n"
		"      CAL sub1;\n"
		"      ARAC A1.xy, A1;\n"
		"      BRA loop2S (GT.x);\n"
		"    ARAC A0.xy, A0;\n"
		"    BRA loopS (GT.x);\n"
		"  MOV result.position, r0;\n"
		/* PUSHA causes lockups on G80 cards.
		"  ARL A0, count;\n"
		" PUSHA A0;\n"
		"  ARL A0, count2;\n"
		" POPA A0;\n"
		* */
		"  BRA end;\n"
		""
		"end:\n"
		"END";

	GLuint s;

	TEST_PROLOGUE;
	printf("# test_nv_vertex_program3\n");
 	dump_before();
	
	regl.Enable(GL_VERTEX_PROGRAM_ARB);
	regl.GenProgramsARB(1, &s);
	regl.BindProgramARB(GL_VERTEX_PROGRAM_ARB, s);
	regl.GetError();
	if (card_family < G80)
	{
		regl.ProgramStringARB(GL_VERTEX_PROGRAM_ARB,
				GL_PROGRAM_FORMAT_ASCII_ARB, strlen(vp), vp);
	}
	else
	{
		regl.ProgramStringARB(GL_VERTEX_PROGRAM_ARB,
				GL_PROGRAM_FORMAT_ASCII_ARB, strlen(g80_vp), g80_vp);
	}
	if (regl.GetError()) 
	{ 
		regl.DeleteProgramsARB(1, &s);
		regl.Disable(GL_VERTEX_PROGRAM_ARB);
		printf("# err - s\n"); 
		return; 
	}
	tri();
	dump_after(0);

	regl.DeleteProgramsARB(1, &s);
	regl.Disable(GL_VERTEX_PROGRAM_ARB);
	TEST_EPILOGUE;
#endif
}

void test_arb_fragment_program()
{
#ifndef GL_ARB_fragment_program
	printf("# Sorry, no GL_ARB_fragment_program\n");
#else
	char *extensions = (char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_fragment_program")) {
		fprintf(stderr, "No GL_ARB_fragment_program extension\n");
		return;
	}

	char fp[4096];
	GLuint fp_obj;
	int i;

	const char *fps[] = {
/*      ABS            v       v        absolute value */
		"ADD t0, fragment.color, {0.21};\n"
		"ABS result.color, t0;\n",
/*      ADD            v,v     v        add */
		"ADD result.color.x, fragment.color, fragment.texcoord[0];\n",
		"ADD result.color.y, fragment.color, fragment.texcoord[0];\n",
		"ADD result.color.z, fragment.color, fragment.texcoord[0];\n",
		"ADD result.color.w, fragment.color, fragment.texcoord[0];\n",
		"ADD result.color, fragment.color, fragment.texcoord[0];\n",
/*      CMP            v,v,v   v        compare */ /////////////////////////////////////////////////////////////////////////////////////
		"ADD t0, fragment.color, {0.21};\n"
		"ADD t1, fragment.texcoord[0], {0.43};\n"
		"CMP result.color, t0, t1, fragment.texcoord[0];\n",
/*      COS            s       ssss     cosine with reduction to [-PI,PI] */
		"COS result.color, fragment.color.x;\n",
		"COS result.color, fragment.color.y;\n",
		"COS result.color, fragment.color.z;\n",
		"COS result.color, fragment.color.w;\n",
/*      DP3            v,v     ssss     3-component dot product */
		"DP3 result.color, fragment.color, fragment.color.secondary;\n",
		"DP3 result.color, fragment.color, fragment.texcoord[0];\n",
		"DP3 result.color, fragment.color, fragment.texcoord[1];\n",
		"DP3 result.color, fragment.color, fragment.texcoord[2];\n",
		"DP3 result.color, fragment.color, fragment.texcoord[3];\n",
		"DP3 result.color, fragment.color, fragment.fogcoord;\n",
		"DP3 result.color, fragment.color, fragment.position;\n",
/*      DP4            v,v     ssss     4-component dot product */
		"DP4 result.color, fragment.color, fragment.color.secondary;\n",
/*      DPH            v,v     ssss     homogeneous dot product */
		"DPH result.color, fragment.color, fragment.color.secondary;\n",
/*      DST            v,v     v        distance vector */
		"DST result.color, fragment.color, fragment.color.secondary;\n",
/*      EX2            s       ssss     exponential base 2 */
		"EX2 result.color, fragment.color.x;\n",
/*      FLR            v       v        floor */
		"FLR result.color, fragment.color;\n",
/*      FRC            v       v        fraction */
		"FRC result.color, fragment.color;\n",
/*      KIL            v       v        kill fragment */
		"KIL fragment.color;\n"
		"MOV result.color, fragment.color;\n",
/*      LG2            s       ssss     logarithm base 2 */
		"LG2 result.color, fragment.color.x;\n",
/*      LIT            v       v        compute light coefficients */
		"LIT result.color, fragment.color;\n",
/*      LRP            v,v,v   v        linear interpolation */
		"LRP result.color, fragment.color, fragment.color.secondary, fragment.texcoord[0];\n",
/*      MAD            v,v,v   v        multiply and add */
		"MAD result.color, fragment.color, fragment.color.secondary, fragment.texcoord[0];\n",
/*      MAX            v,v     v        maximum */
		"MAX result.color, fragment.color, fragment.color.secondary;\n",
/*      MIN            v,v     v        minimum */
		"MIN result.color, fragment.color, fragment.color.secondary;\n",
/*      MOV            v       v        move */
		"MOV result.color, fragment.color;\n",
/*      MUL            v,v     v        multiply */
		"MUL result.color, fragment.color, fragment.color.secondary;\n",
/*      POW            s,s     ssss     exponentiate */
		"POW result.color, fragment.color.x, fragment.color.secondary.x;\n",
		"POW result.color, fragment.color.x, fragment.color.secondary.y;\n",
		"POW result.color, fragment.color.x, fragment.color.secondary.z;\n",
		"POW result.color, fragment.color.x, fragment.color.secondary.w;\n",
/*      RCP            s       ssss     reciprocal */
		"RCP result.color, fragment.color.x;\n",
/*      RSQ            s       ssss     reciprocal square root */
		"RSQ result.color, fragment.color.x;\n",
/*      SCS            s       ss--     sine/cosine without reduction */
		"SCS result.color, fragment.color.x;\n",
/*      SGE            v,v     v        set on greater than or equal */
		"SGE result.color, fragment.color, fragment.color.secondary;\n",
/*      SIN            s       ssss     sine with reduction to [-PI,PI] */
		"SIN result.color, fragment.color.x;\n",
/*      SLT            v,v     v        set on less than */
		"SLT result.color, fragment.color, fragment.color.secondary;\n",
/*      SUB            v,v     v        subtract */
		"SUB result.color, fragment.color, fragment.color.secondary;\n",
/*      SWZ            v       v        extended swizzle */
		"SWZ result.color, fragment.color, x, y, z, w;\n",
		"SWZ result.color, fragment.color, y, y, z, w;\n",
		"SWZ result.color, fragment.color, z, y, z, w;\n",
		"SWZ result.color, fragment.color, w, y, z, w;\n",
		"SWZ result.color, fragment.color, 0, y, z, w;\n",
		"SWZ result.color, fragment.color, 1, y, z, w;\n",
/*      TEX            v,u   v        texture sample */
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"ADD result.color, t0, t1;\n",
/*      TEX            v,u,t   v        texture sample *//////////////////////////////////////////////////////////////////////////////////////
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"TEX t2, fragment.texcoord[2], texture[2], 2D;\n"
		"MAD result.color, t0, t1, t2;\n",
/*      TEX            v,u,t   v        texture sample *//////////////////////////////////////////////////////////////////////////////////////
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"TEX t2, fragment.texcoord[2], texture[2], 2D;\n"
		"TEX t3, fragment.texcoord[3], texture[3], 2D;\n"
		"ADD t0,t1,t0;\n"
		"MAD result.color, t0, t1, t2;\n",
/*      TEX            v,u,t   v        texture sample *//////////////////////////////////////////////////////////////////////////////////////
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[0], texture[1], 2D;\n"
		"TEX t2, fragment.texcoord[1], texture[2], 2D;\n"
		"TEX t3, fragment.texcoord[1], texture[3], 2D;\n"
		"ADD t0,t1,t0;\n"
		"MAD result.color, t0, t1, t2;\n",
/*      TEX            v,u,t   v        texture sample *//////////////////////////////////////////////////////////////////////////////////////
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"TEX t2, fragment.texcoord[2], texture[2], 2D;\n"
		"TEX t3, fragment.texcoord[3], texture[3], 2D;\n"
		"TEX t4, fragment.texcoord[4], texture[4], 2D;\n"
		"TEX t5, fragment.texcoord[5], texture[5], 2D;\n"
		"TEX t6, fragment.texcoord[6], texture[6], 2D;\n"
		"TEX t7, fragment.texcoord[7], texture[7], 2D;\n"
		"MAD t8,t0,t1,t2;\n"
		"MAD t9,t3,t4,t5;\n"
		"MAD t10,t6,t7,t0;\n"
		"MAD t11,t6,t7,t1;\n"
		"MAD t12,t8,t9,t10;\n"
		"ADD result.color, t12, t11;\n",
/*      TEX            v,u,t   v        texture sample *//////////////////////////////////////////////////////////////////////////////////////
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"TEX t2, fragment.texcoord[2], texture[2], 2D;\n"
		"ADD t0,t1,t0;\n"
		"MAD result.color, t0, t1, t2;\n",
/*      TEX            v,u,t   v        texture sample *//////////////////////////////////////////////////////////////////////////////////////
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"TEX t2, fragment.texcoord[2], texture[2], 2D;\n"
		"ADD t0,t1,t0;\n"
		"ADD result.color, t0, t2;\n",
/*      TEX            v,u,t   v        texture sample */
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"ADD t0,fragment.color,t0;\n"
		"ADD result.color, t0, t1;\n",
/*      TEX            v,u,t   v        texture sample */
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"ADD t0,fragment.color,t0;\n"
		"MAD result.color, t0, t1, t0;\n",
/*      TEX            v,u,t   v        texture sample */
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"TEX t2, fragment.texcoord[2], texture[2], 2D;\n"
		"TEX t3, fragment.texcoord[3], texture[3], 2D;\n"
		"ADD t0, t0, t1;\n"
		"ADD t0, t0, t2;\n"
		"ADD t0, t0, t3;\n"
		"ADD result.color, t0, t1;\n",
/*      TEX            v,u,t   v        texture sample */
		"TEX t0, fragment.texcoord[0], texture[0], 2D;\n"
		"TEX t1, fragment.texcoord[1], texture[1], 2D;\n"
		"TEX t2, fragment.texcoord[2], texture[2], 2D;\n"
		"TEX t3, fragment.texcoord[3], texture[3], 2D;\n"
		"MAD t4, t0, t1, t2;\n"
		"ADD result.color, t3, t4;\n",
/*      TXB            v,u,t   v        texture sample with bias */
		"TXB result.color, fragment.texcoord[0], texture[0], 2D;\n",
/*      TXP            v,u,t   v        texture sample with projection */
		"TXP result.color, fragment.texcoord[0], texture[0], 2D;\n",
/*      XPD            v,v     v        cross product */
		"XPD result.color, fragment.color, fragment.color.secondary;\n",
		NULL
	} ;

	TEST_PROLOGUE;
	printf("# test_arb_fragment_program\n");

	regl.Enable(GL_FRAGMENT_PROGRAM_ARB);
	i = 0;
	while (fps[i]) {
		snprintf(fp, sizeof(fp),
				"!!ARBfp1.0\n"
				"TEMP t0, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12;\n"
				"%s\n"
				"END\n",
				fps[i]);
		printf("# Testing:\n");
		print_commented(fp);
		dump_before();

		regl.GenProgramsARB(1, &fp_obj);
		regl.BindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fp_obj);
		regl.GetError();
		regl.ProgramStringARB(GL_FRAGMENT_PROGRAM_ARB,
				GL_PROGRAM_FORMAT_ASCII_ARB, strlen(fp), fp);
		if (regl.GetError()) 
		{ 
			regl.DeleteProgramsARB(1, &fp_obj);
			printf("# err - fp_obj\n"); 
			break; 
		}
		tri();
		dump_after(0);
		regl.DeleteProgramsARB(1, &fp_obj);

		i++;
	}
	regl.Disable(GL_FRAGMENT_PROGRAM_ARB);

	TEST_EPILOGUE;
#endif
}

void test_nv_fragment_program()
{
#ifndef GL_NV_fragment_program
	printf("# Sorry, no GL_NV_fragment_program\n");
#else
	char *extensions = (char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_fragment_program")) {
		fprintf(stderr, "No GL_NV_fragment_program extension\n");
		return;
	}

	char fp[4096];
	GLuint fp_obj;
	int i;

	const char *fps[] = {
/*      DDX[RH ][C][_SAT]    v       v        derivative relative to x */
		"DDX o[COLR], f[COL0];\n",
/*      DDY[RH ][C][_SAT]    v       v        derivative relative to y */
		"DDY o[COLR], f[COL0];\n",
/*      KIL                  none    none     conditionally discard fragment */
		"MOV R0, f[COL0];\n"
		"ADDC R0, f[COL1], R0;\n"
		"KIL LT;\n"
		"MOV o[COLR], R0;\n",
/*      PK2H                 v       ssss     pack two 16-bit floats */
		"PK2H o[COLR], f[COL0];\n",
/*      PK2US                v       ssss     pack two unsigned 16-bit scalars */
		"PK2US o[COLR], f[COL0];\n",
/*      PK4B                 v       ssss     pack four signed 8-bit scalars */
		"PK4B o[COLR], f[COL0];\n",
/*      PK4UB                v       ssss     pack four unsigned 8-bit scalars */
		"PK4UB o[COLR], f[COL0];\n",
/*      RFL[RH ][C][_SAT]    v,v     v        reflection vector */
		"MOV R0, f[COL1];\n RFL o[COLR], f[COL0], R0;\n",
/*      SEQ[RHX][C][_SAT]    v,v     v        set on equal */
		"MOV R0, f[COL1];\n SEQ o[COLR], f[COL0], R0;\n",
/*      SFL[RHX][C][_SAT]    v,v     v        set on false */
		"MOV R0, f[COL1];\n SFL o[COLR], f[COL0], R0;\n",
/*      SGT[RHX][C][_SAT]    v,v     v        set on greater than */
		"MOV R0, f[COL1];\n SGT o[COLR], f[COL0], R0;\n",
/*      SLE[RHX][C][_SAT]    v,v     v        set on less than or equal */
		"MOV R0, f[COL1];\n SLE o[COLR], f[COL0], R0;\n",
/*      SNE[RHX][C][_SAT]    v,v     v        set on not equal */
		"MOV R0, f[COL1];\n SNE o[COLR], f[COL0], R0;\n",
/*      STR[RHX][C][_SAT]    v,v     v        set on true */
		"MOV R0, f[COL1];\n STR o[COLR], f[COL0], R0;\n",
/*      TXD[C][_SAT]         v,v,v   v        texture lookup w/partials */
		"MOV R0, f[COL0];\n"
		"MOV R1, f[COL1];\n"
		"TXD o[COLR], f[TEX0], R0, R1, TEX0, 2D;\n",
/*      UP2H[C][_SAT]        s       v        unpack two 16-bit floats */
		"UP2H o[COLR], f[COL0].x;\n",
/*      UP2US[C][_SAT]       s       v        unpack two unsigned 16-bit scalars */
		"UP2US o[COLR], f[COL0].x;\n",
/*      UP4B[C][_SAT]        s       v        unpack four signed 8-bit scalars */
		"UP4B o[COLR], f[COL0].x;\n",
/*      UP4UB[C][_SAT]       s       v        unpack four unsigned 8-bit scalars */
		"UP4UB o[COLR], f[COL0].x;\n",
/*      X2D[RH ][C][_SAT]    v,v,v   v        2D coordinate transformation */
		"MOV R0, f[COL1];\n"
		"MOV R1, f[TEX0];\n"
		"X2D o[COLR], f[COL0], R0, R1;\n",
/* Conditional execution */
		"MOV R0, f[COL0];\nADDC o[COLR], R0, f[COL1];\nSGT o[COLR].x (EQ.xyzw), R0, f[TEX0];\n",
		"MOV R0, f[COL0];\nADDC o[COLR], R0, f[COL1];\nSGT o[COLR].x (GE.xyzw), R0, f[TEX0];\n",
		"MOV R0, f[COL0];\nADDC o[COLR], R0, f[COL1];\nSGT o[COLR].x (GT.xyzw), R0, f[TEX0];\n",
		"MOV R0, f[COL0];\nADDC o[COLR], R0, f[COL1];\nSGT o[COLR].x (LE.xyzw), R0, f[TEX0];\n",
		"MOV R0, f[COL0];\nADDC o[COLR], R0, f[COL1];\nSGT o[COLR].x (LT.xyzw), R0, f[TEX0];\n",
		"MOV R0, f[COL0];\nADDC o[COLR], R0, f[COL1];\nSGT o[COLR].x (NE.xyzw), R0, f[TEX0];\n",
		"MOV R0, f[COL0];\nADDC o[COLR], R0, f[COL1];\nSGT o[COLR].x (TR.xyzw), R0, f[TEX0];\n",
		"MOV R0, f[COL0];\nADDC o[COLR], R0, f[COL1];\nSGT o[COLR].x (FL.xyzw), R0, f[TEX0];\n",
		NULL
	};

	TEST_PROLOGUE;
	printf("# test_nv_fragment_program\n");

	regl.Enable(GL_FRAGMENT_PROGRAM_NV);
	i = 0;
	while (fps[i]) {
		snprintf(fp, sizeof(fp),
				"!!FP1.0\n"
				"%s\n"
				"END\n",
				fps[i]);
		printf("# Testing:\n");
		print_commented(fp);
		dump_before();
		regl.GenProgramsNV(1, &fp_obj);
		regl.GetError();
		regl.LoadProgramNV(GL_FRAGMENT_PROGRAM_NV, fp_obj, strlen(fp), (const GLubyte*)fp);
		if (regl.GetError()) 
		{ 
			regl.DeleteProgramsNV(1, &fp_obj);
			printf("# err - s\n"); 
			break;
		}
		regl.BindProgramNV(GL_FRAGMENT_PROGRAM_NV, fp_obj);
		// glTrackMatrixNV can only be called with a GL_VERTEX_PROGRAM_NV target.
		// Well, according to the mesa sources (src/mesa/shader/nvprogram.c).
		// regl.TrackMatrixNV(GL_FRAGMENT_PROGRAM_NV, 0, GL_MODELVIEW_PROJECTION_NV, GL_IDENTITY_NV);
		tri();
		dump_after(0);
		regl.DeleteProgramsNV(1, &fp_obj);

		i++;
	}
	regl.Disable(GL_FRAGMENT_PROGRAM_NV);

	TEST_EPILOGUE;
#endif
}

void test_nv_fragment_program_half()
{
#ifndef GL_NV_fragment_program
	printf("# Sorry, no GL_NV_fragment_program\n");
#else
	char *extensions = (char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_fragment_program")) {
		fprintf(stderr, "No GL_NV_fragment_program extension\n");
		return;
	}

	char fp[4096];
	GLuint fp_obj;
	int i;

	const char *fps[] = {
		"MOV H0, f[COL0];ADD o[COLH], R0, R1;\n",
		"MOV H1, f[COL0]; MOV H0, f[COL1];ADD o[COLH], H0, H1;\n",	
		NULL
	};

	TEST_PROLOGUE;
	printf("# test_nv_fragment_program\n");

	regl.Enable(GL_FRAGMENT_PROGRAM_NV);
	i = 0;
	while (fps[i]) {
		snprintf(fp, sizeof(fp),
				"!!FP1.0\n"
				"%s\n"
				"END\n",
				fps[i]);
		printf("# Testing:\n");
		print_commented(fp);
		dump_before();
		regl.GenProgramsNV(1, &fp_obj);
		regl.GetError();
		regl.LoadProgramNV(GL_FRAGMENT_PROGRAM_NV, fp_obj, strlen(fp), (const GLubyte*)fp);
		if (regl.GetError()) 
		{ 
			regl.DeleteProgramsNV(1, &fp_obj);
			printf("# err - s\n"); 
			break;
		}
		regl.BindProgramNV(GL_FRAGMENT_PROGRAM_NV, fp_obj);
		// glTrackMatrixNV can only be called with a GL_VERTEX_PROGRAM_NV target.
		// Well, according to the mesa sources (src/mesa/shader/nvprogram.c).
		// regl.TrackMatrixNV(GL_FRAGMENT_PROGRAM_NV, 0, GL_MODELVIEW_PROJECTION_NV, GL_IDENTITY_NV);
		tri();
		dump_after(0);
		regl.DeleteProgramsNV(1, &fp_obj);

		i++;
	}
	regl.Disable(GL_FRAGMENT_PROGRAM_NV);

	TEST_EPILOGUE;
#endif
}

void test_nv_fragment_program2()
{
#ifndef GL_NV_fragment_program2
	printf("# Sorry, no GL_NV_fragment_program2\n");
#else
	char *extensions = (char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_fragment_program2")) {
		fprintf(stderr, "No GL_NV_fragment_program2 extension\n");
		return;
	}

	char fp[4096];
	GLuint fp_obj;
	int i;

	const char *fps[] = {
/*      DIV      X X - X X  v,s     v        divide vector components by scalar */
		"DIV result.color, fragment.color, fragment.texcoord.x;\n",
/*      DP2      X X X X X  v,v     ssss     2-component dot product */
		"DP2 result.color, fragment.color, fragment.color.secondary;\n",
/*      DP2A     X X X X X  v,v,v   ssss     2-comp. dot product w/scalar add */
		"DP2A result.color, fragment.color, fragment.color.secondary, fragment.texcoord;\n",
/*      NRM      X X - X X  v       v        normalize 3-component vector */
		"ADD t0, fragment.color, {0.45};\nNRM result.color, t0;\n",
/*      TXL      - - - X X  v       v        texture same w/explicit LOD */
		"TXL result.color, fragment.texcoord[0], texture[0], 2D;\n",
		
/* LOOP/ENDLOOP + BRK + loop address reg */
#define LOOP(cn,ini,inc) \
		"MOV t0, fragment.color;\n" \
		"LOOP { "#cn", "#ini", "#inc" };\n" \
		"  ADD t0, t0, fragment.color;\n" \
		"  RCP t1.x, t0.w;\n" \
		"  MUL t0, t0, t1.x;\n" \
		"ENDLOOP;\n" \
		"MOV result.color, t0;\n"
		LOOP(250, 0, 1), LOOP(251, 0, 1), LOOP(252, 0, 1),
		LOOP(255, 0, 1), LOOP(255, 1, 1), LOOP(255, 2, 1),
		LOOP(255, 0, 1), LOOP(255, 0, 2), LOOP(255, 0, 3),

		"MOV t0, fragment.color;\n"
		"LOOP { 255, 0, 1 };\n"
		"  ADD t0, t0, fragment.color;\n"
		"  RCP t1.x, t0.w;\n"
		"  MAD t0, t0, t1.x, fragment.texcoord[1];\n"
		"ENDLOOP;\n"
		"MOV result.color, t0;\n",

		"MOV t0, fragment.color;\n"
		"LOOP { 2, 0, 1 };\n"
		"  ADD t0, t0, fragment.color;\n"
		"  RCP t1.x, t0.w;\n"
		"  MAD t0, t0, t1.x, fragment.texcoord[A0.x + 0];\n"
		"ENDLOOP;\n"
		"MOV result.color, t0;\n",

		"MOV t0, fragment.color;\n"
		"LOOP { 2, 0, 1 };\n"
		"  ADD t0, t0, fragment.color;\n"
		"  RCP t1.x, t0.w;\n"
		"  MAD t0, t0, t1.x, fragment.texcoord[A0.x + 1];\n"
		"ENDLOOP;\n"
		"MOV result.color, t0;\n",

		"MOV t0, fragment.color;\n"
		"LOOP { 2, 0, 1 };\n"
		"  ADD t0, t0, fragment.color;\n"
		"  RCP t1.x, t0.w;\n"
		"  MAD t0, t0, t1.x, fragment.texcoord[A0.x + 2];\n"
		"ENDLOOP;\n"
		"MOV result.color, t0;\n",

/* REP/ENDREP + BRK */
		#define REP(c) \
			"MOV t0, fragment.color;\n" \
			"REP "#c";\n" \
			"  ADD t0, t0, fragment.color;\n" \
			"  RCP t1.x, t0.w;\n" \
			"  BRK (LT.x);\n" \
			"  MAD t0, t0, t1.x, fragment.texcoord[A0.x + 2];\n" \
			"ENDREP;\n" \
			"MOV result.color, t0;\n"
		REP(2), REP(2), REP(2), REP(2),

/* CAL/RET */
		"main:\n"
		"  MOV t0, fragment.texcoord[0];\n"
		"  CAL adder;\n"
		"  MOV result.color, t0;\n"
		"  RET;\n"
		"adder:\n"
		"  ADD t0, t0, fragment.color;\n"
		"  RET;\n",

		"main:\n"
		"  MOV t0, fragment.texcoord[0];\n"
		"  CAL adder;\n"
		"  ADD t0, t0, fragment.texcoord[1];\n"
		"  ADD result.color, t0, fragment.texcoord[3];\n"
		"  RET;\n"
		"adder:\n"
		"  ADD t0, t0, fragment.color;\n"
		"  RET;\n",

		"adder:\n"
		"  ADD t0, t0, fragment.color;\n"
		"  RET;\n"
		"main:\n"
		"  MOV t0, fragment.texcoord[0];\n"
		"  CAL adder;\n"
		"  ADD t0, t0, fragment.texcoord[1];\n"
		"  ADD result.color, t0, fragment.texcoord[3];\n"
		"  RET;\n",

/* IF/ELSE/ENDIF */
		"SUBC t0, fragment.color, fragment.texcoord[0];\n"
		"IF GT;\n"
		"  MOV result.color, t0;\n"
		"ELSE;\n"
		"  ADD result.color, fragment.texcoord[1], t0;\n"
		"ENDIF;\n",

/* fragment.facing */
		"MOV result.color, fragment.facing;\n",
		NULL
	};

	TEST_PROLOGUE;
	printf("# test_nv_fragment_program2\n");

	regl.Enable(GL_FRAGMENT_PROGRAM_ARB);
	i = 0;
	while (fps[i]) {
		snprintf(fp, sizeof(fp),
				"!!ARBfp1.0\n"
				"OPTION NV_fragment_program2;\n"
				"TEMP t0, t1, t2;\n"
				"%s\n"
				"END\n",
				fps[i]);
		printf("# Testing:\n");
		print_commented(fp);
		dump_before();

		regl.GenProgramsARB(1, &fp_obj);
		regl.BindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fp_obj);
		regl.GetError();
		regl.ProgramStringARB(GL_FRAGMENT_PROGRAM_ARB,
				GL_PROGRAM_FORMAT_ASCII_ARB, strlen(fp), fp);
		if (regl.GetError()) 
		{ 
			regl.DeleteProgramsARB(1, &fp_obj);
			printf("# err - fp_obj\n"); 
			break;
		}
		tri();
		dump_after(0);
		regl.DeleteProgramsARB(1, &fp_obj);

		i++;
	}
	regl.Disable(GL_FRAGMENT_PROGRAM_ARB);

	TEST_EPILOGUE;
#endif
}

void test_color_material()
{
	TEST_PROLOGUE;
	printf("# test_color_material\n");

	dump_before();

	regl.Enable(GL_COLOR_MATERIAL);
	regl.Enable(GL_LIGHTING);

	int i,j;
	GLenum faces[3]={GL_FRONT_AND_BACK, GL_FRONT, GL_BACK};
	GLenum material[5]={GL_AMBIENT,GL_DIFFUSE,GL_SPECULAR,GL_EMISSION,GL_AMBIENT_AND_DIFFUSE};

	for (i=0;i<3;i++) {
		for (j=0;j<5;j++) {
			regl.ColorMaterial(faces[i], material[j]);
			regl.Color3f(0.1*i*j, 0.2*i*j, 0.3*i*j);
			tri();
		}
	}

	regl.Disable(GL_LIGHTING);
	regl.Disable(GL_COLOR_MATERIAL);
	dump_after(0);
	TEST_EPILOGUE;
}

void test_draw_pixels()
{
	TEST_PROLOGUE;
	int texture[8*8];
	printf("# test_draw_pixel\n");
	memset(texture, 0xaa, sizeof(texture));
	dump_before();
	regl.DrawPixels(7+4,5+4,GL_RGBA, GL_UNSIGNED_BYTE, texture);
	dump_after(0);
	TEST_EPILOGUE;
}

void test_logic_op()
{
	TEST_PROLOGUE;
/*	int i; */
	int texture[8*8];

	typedef struct {
			GLenum	value;
			char*	name;
	} logic_op;
	
	logic_op operations[16] = {
		{__(GL_CLEAR)},
		{ __(GL_AND) },
		{ __(GL_AND_REVERSE) },
		{ __(GL_COPY) },
		{ __(GL_AND_INVERTED) },
		{ __(GL_NOOP) },
		{ __(GL_XOR) },
		{ __(GL_OR) },
		{ __(GL_NOR) },
		{ __(GL_EQUIV) },
		{ __(GL_INVERT) },
		{ __(GL_OR_REVERSE) },
		{ __(GL_COPY_INVERTED) },
		{ __(GL_OR_INVERTED) },
		{ __(GL_NAND) },
		{ __(GL_SET) }
	};

	memset(texture, 0x5a, sizeof(texture));

#if 1
	int i;
	regl.Enable(GL_COLOR_LOGIC_OP);
	for (i=0; i<16; i++) {
		printf("\n# Testing %s\n\n", operations[i].name);
		dump_before();
 		regl.LogicOp(operations[i].value);
		regl.DrawPixels(7,5,GL_RGBA, GL_UNSIGNED_BYTE, texture);
 		tri();
		dump_after(0);
	}
	regl.Disable(GL_COLOR_LOGIC_OP);
#else
	// Skip initialization
	printf("\n\n#------------------------- Init --------------------\n");
	dump_before();
	dump_after(0);

	printf("\n\n#------------------------- RGB --------------------\n");
	dump_before();
// 	regl.Enable(GL_COLOR_LOGIC_OP);
// 	regl.Enable(GL_DITHER);
//         regl.LogicOp(operations[0].value);
	regl.DrawPixels(7,5,GL_RGBA, GL_UNSIGNED_BYTE, texture);
	dump_after(0);

	printf("\n\n#------------------------- RGBA --------------------\n");
	dump_before();
        regl.LogicOp(operations[1].value);
// 	regl.DrawPixels(7,5,GL_RGB, GL_UNSIGNED_BYTE, texture);
	dump_after(0);

/*
	regl.Enable(GL_COLOR_LOGIC_OP);
	for (i=0; i<16; i++) {
		printf("\n# Testing %s\n\n", operations[i].name);
		dump_before();
// 		regl.LogicOp(operations[i].value);
// 		regl.DrawPixels(7,5,GL_RGB, GL_UNSIGNED_BYTE, texture);
		regl.DrawPixels(7,5,GL_RGBA, GL_UNSIGNED_BYTE, texture);
// 		tri();
		dump_after(0);
	}
*/
	regl.Disable(GL_COLOR_LOGIC_OP);
#endif
	TEST_EPILOGUE;
}

void test_overlay(SDL_Surface *screen)
{
	TEST_PROLOGUE;
	SDL_Overlay *overlay;
	SDL_Rect rect = {10,20,30,40};

	dump_before();

	overlay = SDL_CreateYUVOverlay(100,100,SDL_YV12_OVERLAY, screen);
	if (overlay) {
		SDL_DisplayYUVOverlay(overlay, &rect);
		SDL_Delay(1000);
		SDL_FreeYUVOverlay(overlay);
	}

	dump_after(0);
	TEST_EPILOGUE;
}

void test_sgis_generate_mipmap(void)
{
#ifndef GL_GENERATE_MIPMAP_SGIS
	printf("# Sorry, no GL_GENERATE_MIPMAP_SGIS\n");
#else
	GLuint tid;
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_SGIS_generate_mipmap")) {
		fprintf(stderr,"No GL_SGIS_generate_mipmap extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_sgis_generate_mipmap()\n");

	regl.GenTextures(1, &tid);

	dump_before();
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
	                   GL_LINEAR_MIPMAP_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS,GL_TRUE);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,
	                TEXTURE_HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture);

	tri();

	dump_after(0);

	regl.DeleteTextures(1, &tid);
	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_nv_texture_rectangle(void)
{
#ifndef GL_NV_texture_rectangle
	printf("# Sorry, no GL_NV_texture_rectangle\n");
#else
	GLuint tid;
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_texture_rectangle")) {
		fprintf(stderr,"No GL_NV_texture_rectangle extension.\n");
		return;
	}

	texture = create_random_texture(NPOT_TEXTURE_WIDTH,NPOT_TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_nv_texture_rectangle()\n");

	regl.GenTextures(1, &tid);

	dump_before();

	regl.Enable(GL_TEXTURE_RECTANGLE_NV);
	regl.BindTexture(GL_TEXTURE_RECTANGLE_NV, tid);
	regl.TexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	regl.TexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_RGBA, NPOT_TEXTURE_WIDTH,NPOT_TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	regl.Begin(GL_TRIANGLES);
		regl.TexCoord2i(0,0);
		regl.Vertex3i(1,0,0);
		regl.TexCoord2i(0,NPOT_TEXTURE_HEIGHT);
		regl.Vertex3i(0,1,0);
		regl.TexCoord2i(NPOT_TEXTURE_WIDTH,NPOT_TEXTURE_HEIGHT);
		regl.Vertex3i(0,0,1);
	regl.End();

	dump_after(0);

	regl.DeleteTextures(1, &tid);
	free(texture);
	regl.Disable(GL_TEXTURE_RECTANGLE_NV);
	TEST_EPILOGUE;
#endif
}

void test_nv_register_combiners(void)
{
#ifndef GL_NV_register_combiners
	printf("# Sorry, no GL_NV_register_combiners\n");
#else
	const GLfloat color0[]={0.1,0.2,0.3,0.4};
	const GLfloat color1[]={0.5,0.6,0.7,0.8};
	const GLfloat black_color[]={0.0,0.0,0.0,0.0};
	const GLenum input_mapping[8]={
		GL_UNSIGNED_IDENTITY_NV,
		GL_UNSIGNED_INVERT_NV,
	        GL_EXPAND_NORMAL_NV,
	        GL_EXPAND_NEGATE_NV,
	        GL_HALF_BIAS_NORMAL_NV,
	        GL_HALF_BIAS_NEGATE_NV,
	        GL_SIGNED_IDENTITY_NV,
	        GL_SIGNED_NEGATE_NV
	};
	const GLenum input_param[10]={
		GL_ZERO,
		GL_CONSTANT_COLOR0_NV,
		GL_CONSTANT_COLOR1_NV,
		GL_FOG,
		GL_PRIMARY_COLOR_NV,
		GL_SECONDARY_COLOR_NV,
		GL_SPARE0_NV,
		GL_SPARE1_NV,
		GL_TEXTURE0_ARB,
		GL_TEXTURE1_ARB
	};
	int i, max_combiners;

	/* NOTE: Testing everything is a lengthy process, so only check one variable and one component at once */
	int cur_rc_variable = GL_VARIABLE_A_NV;	/* or B,C,D ; or also E,F,G for final */
	int cur_rc_component = GL_RGB;	/* or ALPHA */
	int cur_rc = GL_COMBINER0_NV;	/* general combiner to test */

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_register_combiners")) {
		fprintf(stderr,"No GL_NV_register_combiners extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_nv_register_combiners\n");

	regl.GetIntegerv(GL_MAX_GENERAL_COMBINERS_NV, &max_combiners);
	printf("# %d max general combiners\n", max_combiners);

	regl.Enable(GL_REGISTER_COMBINERS_NV);

	/*--- Global settings ---*/
	dump_before();

	printf("# Constant colors, color sum clamp\n");

	regl.CombinerParameterfvNV(GL_CONSTANT_COLOR0_NV,color0);
	regl.CombinerParameterfvNV(GL_CONSTANT_COLOR1_NV,color1);	
	regl.CombinerParameteriNV(GL_COLOR_SUM_CLAMP_NV, 0);
	tri();

	/* restore default */
	regl.CombinerParameterfvNV(GL_CONSTANT_COLOR0_NV, black_color);
	regl.CombinerParameterfvNV(GL_CONSTANT_COLOR1_NV, black_color);
	regl.CombinerParameteriNV(GL_COLOR_SUM_CLAMP_NV, 1);
	tri();

	dump_after(0);

	/*--- General combiners ---*/
	for(i=1; i<=max_combiners; i++)
	{
		dump_before();
		printf("# Enable %d register combiners\n", i);
		regl.CombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV, i);
		tri();
		dump_after(0);
	}

	/* Restore default */
	regl.CombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV, 1);

	/*--- For a general combiner, check input and outputs ---*/
	printf("# General combiner stage\n");

	/* Input */
	dump_before();
	printf("# General combiner, input\n");
	for (i=0; i<10; i++) {
		regl.CombinerInputNV(
			cur_rc,
			GL_RGB,		/* or ALPHA */
			cur_rc_variable,
			input_param[i],
			GL_UNSIGNED_IDENTITY_NV,
			cur_rc_component
		);
		tri();
	}
	dump_after(0);

	/* Input mapping */
	dump_before();
	printf("# General combiner, input mapping\n");
	for (i=0; i<8; i++) {
		regl.CombinerInputNV(
			cur_rc,
			GL_RGB,		/* or ALPHA */
			cur_rc_variable,
			GL_ZERO,
			input_mapping[i],
			cur_rc_component
		);
		tri();
	}
	dump_after(0);

	/* Restore default */
	regl.CombinerInputNV(cur_rc, GL_RGB, GL_VARIABLE_A_NV,
		GL_PRIMARY_COLOR_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	regl.CombinerInputNV(cur_rc, GL_RGB, GL_VARIABLE_B_NV,
		GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
	regl.CombinerInputNV(cur_rc, GL_RGB, GL_VARIABLE_C_NV,
		GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	regl.CombinerInputNV(cur_rc, GL_RGB, GL_VARIABLE_D_NV,
		GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	regl.CombinerInputNV(cur_rc, GL_ALPHA, GL_VARIABLE_A_NV,
		GL_PRIMARY_COLOR_NV, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
	regl.CombinerInputNV(cur_rc, GL_ALPHA, GL_VARIABLE_B_NV,
		GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_ALPHA);
	regl.CombinerInputNV(cur_rc, GL_ALPHA, GL_VARIABLE_C_NV,
		GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
	regl.CombinerInputNV(cur_rc, GL_ALPHA, GL_VARIABLE_D_NV,
		GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);

	/* Output, even more params to check */
	dump_before();
	printf("# General combiner, output\n");
	/* Only inputs from 4 to 8 are valid for glCombinerOutputNV.*/
	for (i=4; i<8; i++) {
		regl.CombinerOutputNV(
			cur_rc,
			GL_RGB,		/* or ALPHA */
			GL_DISCARD_NV,	/* abOutput */
			GL_DISCARD_NV,	/* cdOutput */
			input_param[i],	/* sumOutput */
			GL_NONE,	/* scale */
			GL_NONE,	/* bias */
			GL_FALSE,	/* abDotProduct */
			GL_FALSE,	/* cdDotProduct */
			GL_FALSE	/* muxSum */
		);
		tri();
	}
	dump_after(0);

	/* Restore default */
	regl.CombinerOutputNV(cur_rc, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV,
		GL_SPARE0_NV, GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);

	/*--- Final combiner ---*/
	printf("# Final combiner stage\n");

	/* Input */
	dump_before();
	printf("# Final combiner, input\n");
	for (i=0; i<10; i++) {
		regl.FinalCombinerInputNV(
			cur_rc_variable,
			input_param[i],
			GL_UNSIGNED_IDENTITY_NV,
			cur_rc_component);
		tri();
	}
	dump_after(0);

	/* Input mapping */
	dump_before();
	printf("# Final combiner, input mapping\n");
	/* Only first 2 mappings valid for glFinalCombinerInputNV.*/
	for (i=0; i<2; i++) {
		regl.FinalCombinerInputNV(
			cur_rc_variable,
			GL_ZERO,
			input_mapping[i],
			cur_rc_component);
		tri();
	}
	dump_after(0);

	/* Restore default */
	regl.FinalCombinerInputNV(GL_VARIABLE_A_NV, GL_FOG,
		GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
	regl.FinalCombinerInputNV(GL_VARIABLE_B_NV, GL_SPARE0_PLUS_SECONDARY_COLOR_NV,
		GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	regl.FinalCombinerInputNV(GL_VARIABLE_C_NV, GL_FOG,
		GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	regl.FinalCombinerInputNV(GL_VARIABLE_D_NV, GL_ZERO,
		GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	regl.FinalCombinerInputNV(GL_VARIABLE_E_NV, GL_ZERO,
		GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	regl.FinalCombinerInputNV(GL_VARIABLE_F_NV, GL_ZERO,
		GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	regl.FinalCombinerInputNV(GL_VARIABLE_G_NV, GL_SPARE0_NV,
		GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);

	regl.Disable(GL_REGISTER_COMBINERS_NV);
	TEST_EPILOGUE;
#endif
}

void test_nv_register_combiners2(void)
{
#ifndef GL_NV_register_combiners2
	printf("# Sorry, no GL_NV_register_combiners2\n");
#else
	const GLfloat color0[4]={0.1,0.2,0.3,0.4};
	const GLfloat color1[4]={0.5,0.6,0.7,0.8};
	const GLfloat black_color[4]={0.0,0.0,0.0,0.0};
	int i, max_combiners;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_register_combiners2")) {
		fprintf(stderr,"No GL_NV_register_combiners2 extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_nv_register_combiners2\n");

	regl.GetIntegerv(GL_MAX_GENERAL_COMBINERS_NV, &max_combiners);
	printf("# %d general combiners\n", max_combiners);

	regl.Enable(GL_REGISTER_COMBINERS_NV);
	regl.Enable(GL_PER_STAGE_CONSTANTS_NV);
	for(i=1; i<=max_combiners; i++)
	{
		dump_before();
		printf("# Enable %d register combiners\n", i);
		regl.CombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV, i);
		regl.CombinerStageParameterfvNV(GL_COMBINER0_NV+i-1, GL_CONSTANT_COLOR0_NV, color0);
		regl.CombinerStageParameterfvNV(GL_COMBINER0_NV+i-1, GL_CONSTANT_COLOR1_NV, color1);
		tri();
		dump_after(0);

		/* Restore default */
		regl.CombinerStageParameterfvNV(GL_COMBINER0_NV+i-1, GL_CONSTANT_COLOR0_NV, black_color);
		regl.CombinerStageParameterfvNV(GL_COMBINER0_NV+i-1, GL_CONSTANT_COLOR1_NV, black_color);
	}

	regl.CombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV, 1);
	regl.Disable(GL_PER_STAGE_CONSTANTS_NV);
	regl.Disable(GL_REGISTER_COMBINERS_NV);
	dump_after(0);
	TEST_EPILOGUE;
#endif
}

void test_nv_texture_env_combine4(void)
{
#ifndef GL_NV_texture_env_combine4
	printf("# Sorry, no GL_NV_texture_env_combine4\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_texture_env_combine4")) {
		fprintf(stderr,"No GL_NV_texture_env_combine4 extension.\n");
		return;
	}

	GLuint tid[10];
	GLuint tdata_uint[128*128*4];

	TEST_PROLOGUE;
	printf("# test_nv_texture_env_combine4\n");
	dump_before();

	regl.GenTextures(10, tid);

	regl.ActiveTexture(GL_TEXTURE0);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[0]);
	regl.TexImage2D(GL_TEXTURE_2D, 0,
			GL_RGBA,
			64, 64,
			0,
			GL_RGBA, GL_UNSIGNED_INT,
			tdata_uint
			);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	regl.ActiveTexture(GL_TEXTURE1);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[0]);
	regl.TexImage2D(GL_TEXTURE_2D, 0,
			GL_RGBA,
			32, 32,
			0,
			GL_RGBA, GL_UNSIGNED_INT,
			tdata_uint
			);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);

	GLfloat color[4] = {1.0, 0.5, 0.25, 0.125};
	regl.TexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, color);

//	regl.TexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE, 2);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_ADD);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA, GL_ADD);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA, GL_PREVIOUS);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA, GL_SRC_ALPHA);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_ALPHA, GL_TEXTURE);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA, GL_SRC_ALPHA);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB, GL_PRIMARY_COLOR);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_ALPHA, GL_PRIMARY_COLOR);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB, GL_SRC_COLOR);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_ALPHA, GL_SRC_ALPHA);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_SOURCE3_RGB_NV, GL_TEXTURE);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_SOURCE3_ALPHA_NV, GL_TEXTURE);

	regl.TexEnvi(GL_TEXTURE_ENV, GL_OPERAND3_RGB_NV, GL_SRC_COLOR);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_OPERAND3_ALPHA_NV, GL_SRC_ALPHA);

	tri();

	regl.ActiveTexture(GL_TEXTURE0);
	regl.Disable(GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE1);
	regl.Disable(GL_TEXTURE_2D);

	dump_after(0);
	TEST_EPILOGUE;
#endif
}

void test_ext_stencil_two_side(void)
{
#ifndef GL_EXT_stencil_two_side
	printf("# Sorry, no GL_EXT_stencil_two_side\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_stencil_two_side")) {
		fprintf(stderr,"No GL_EXT_stencil_two_side extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_ext_stencil_two_side()\n");

	dump_before();

        regl.Enable(GL_STENCIL_TEST);
        regl.Enable(GL_STENCIL_TEST_TWO_SIDE_EXT);
	tri();

        regl.ActiveStencilFaceEXT(GL_BACK);
	regl.Clear(GL_STENCIL_BUFFER_BIT);
        regl.StencilOp(GL_KEEP,            // stencil test fail
                    GL_KEEP,            // depth test fail
                    GL_DECR_WRAP_EXT);  // depth test pass
        regl.StencilMask(~0);
        regl.StencilFunc(GL_ALWAYS, 0, ~0);
	tri();

        regl.ActiveStencilFaceEXT(GL_FRONT);
	regl.Clear(GL_STENCIL_BUFFER_BIT);
        regl.StencilOp(GL_KEEP,            // stencil test fail
                    GL_KEEP,            // depth test fail
                    GL_INCR_WRAP_EXT);  // depth test pass
        regl.StencilMask(~0);
        regl.StencilFunc(GL_ALWAYS, 0, ~0);
	tri();

        regl.Disable(GL_STENCIL_TEST_TWO_SIDE_EXT);
        regl.Disable(GL_STENCIL_TEST);
	dump_after(0);
	TEST_EPILOGUE;
#endif
}

void test_swap_buffers(void)
{
	TEST_PROLOGUE;
	int i;
	printf("# test_swap_buffers\n");
	dump_before();
	//regl.Enable(GL_DEPTH_TEST);
	for (i=0;i<10;i++) {
		regl.Clear(GL_COLOR_BUFFER_BIT);
		tri();
		SDL_GL_SwapBuffers();
	}
	//regl.Disable(GL_DEPTH_TEST);
	dump_after(0);
	TEST_EPILOGUE;
}

void test_ext_vertex_weighting(void)
{
#ifndef GL_EXT_vertex_weighting
	printf("# Sorry, no GL_EXT_vertex_weighting\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_vertex_weighting")) {
		fprintf(stderr,"No GL_EXT_vertex_weighting extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_ext_vertex_weighting()\n");

	dump_before();
	tri();
	regl.Enable(GL_VERTEX_WEIGHTING_EXT);

	regl.Begin(GL_TRIANGLES);
		regl.VertexWeightfEXT(0.2);	regl.Vertex3i(1,2,3);
		regl.VertexWeightfEXT(0.3);	regl.Vertex3i(4,5,6);
		regl.VertexWeightfEXT(0.4);	regl.Vertex3i(7,8,9);
	regl.End();

	regl.MatrixMode(GL_MODELVIEW0_EXT);
	regl.Scalef(0.8,0.9,1.0);
	regl.MatrixMode(GL_MODELVIEW1_EXT);
	regl.Scalef(0.5,0.6,0.7);

	regl.Begin(GL_TRIANGLES);
		regl.VertexWeightfEXT(0.2);	regl.Vertex3i(1,2,3);
		regl.VertexWeightfEXT(0.3);	regl.Vertex3i(4,5,6);
		regl.VertexWeightfEXT(0.4);	regl.Vertex3i(7,8,9);
	regl.End();

	regl.Disable(GL_VERTEX_WEIGHTING_EXT);
	dump_after(0);

	regl.MatrixMode(GL_MODELVIEW0_EXT);
	regl.LoadIdentity();
	regl.MatrixMode(GL_MODELVIEW1_EXT);
	regl.LoadIdentity();

	TEST_EPILOGUE;
#endif
}

void test_stencil(void)
{
	TEST_PROLOGUE;
	printf("# test_stencil()\n");

	dump_before();

	regl.Enable(GL_STENCIL_TEST);
	regl.ClearStencil(0x21);

	regl.StencilMask(0x5a);
	regl.StencilFunc(GL_LESS, 0x44, 0x33);
	regl.StencilOp(GL_INCR, GL_DECR, GL_INVERT);

	tri();

	regl.Disable(GL_STENCIL_TEST);

	dump_after(0);
	TEST_EPILOGUE;
}

void test_nv_texgen_emboss(void)
{
#ifndef GL_NV_texgen_emboss
	printf("# Sorry, no GL_NV_texgen_emboss\n");
#else
	GLuint tid[1];
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_texgen_emboss")) {
		fprintf(stderr,"No GL_NV_texgen_emboss extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_nv_texgen_emboss()\n");

	regl.GenTextures(1, tid);

	dump_before();
	tri();

	regl.Enable(GL_LIGHTING);
	regl.Enable(GL_LIGHT3);

	regl.ClientActiveTextureARB(GL_TEXTURE0_ARB);
	regl.Enable(GL_TEXTURE_2D);

	regl.BindTexture(GL_TEXTURE_2D, tid[0]);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	regl.ClientActiveTextureARB(GL_TEXTURE1_ARB);
	regl.Enable(GL_TEXTURE_2D);

	regl.BindTexture(GL_TEXTURE_2D, tid[0]);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.TexGeni(GL_TEXTURE_2D, GL_TEXTURE_GEN_MODE, GL_EMBOSS_MAP_NV);
	regl.TexGenf(GL_TEXTURE_2D, GL_EMBOSS_CONSTANT_NV, 0.7);
	regl.TexGeni(GL_TEXTURE_2D, GL_EMBOSS_LIGHT_NV, GL_LIGHT3);

	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	tri();

	regl.ClientActiveTextureARB(GL_TEXTURE1_ARB);
	regl.Disable(GL_TEXTURE_2D);
	regl.ClientActiveTextureARB(GL_TEXTURE0_ARB);
	regl.Disable(GL_TEXTURE_2D);
	regl.Disable(GL_LIGHT3);
	regl.Disable(GL_LIGHTING);
	dump_after(0);

	regl.DeleteTextures(1, tid);
	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_texture_3d(void)
{
	GLuint tid;
	int i;
	int *texture = NULL;
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);

	texture = malloc(TEXTURE_WIDTH*TEXTURE_HEIGHT*TEXTURE_DEPTH*sizeof(int));
	if (!texture) {
		printf("# can not allocate mem for texture\n");
		return;
	}

	for (i=0; i<TEXTURE_WIDTH*TEXTURE_HEIGHT*TEXTURE_DEPTH; i++) {
		texture[i] = random();
	}

	TEST_PROLOGUE;
	printf("# test_texture_3d()\n");

	regl.GenTextures(1, &tid);

	dump_before();

	regl.Enable(GL_TEXTURE_3D);
	regl.BindTexture(GL_TEXTURE_3D, tid);
	regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.TexImage3D(GL_TEXTURE_3D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT,TEXTURE_DEPTH, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
	tex1_tri();
	regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	tex1_tri();
	regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_BORDER);
	tex1_tri();
	regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_REPEAT);
	tex1_tri();

#ifdef GL_ARB_texture_mirrored_repeat
	if (strstr(extensions, "GL_ARB_texture_mirrored_repeat")) {
		regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_MIRRORED_REPEAT_ARB);
		tex1_tri();
		regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_REPEAT);
		tex1_tri();
	} else
#endif
	{
		printf("# No ARB_texture_mirrored_repeat\n");
	}

#ifdef GL_EXT_texture_mirror_clamp
	if (strstr(extensions, "GL_EXT_texture_mirror_clamp")) {
		regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_MIRROR_CLAMP_EXT);
		tex1_tri();
		regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_MIRROR_CLAMP_TO_EDGE_EXT);
		tex1_tri();
		regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_MIRROR_CLAMP_TO_BORDER_EXT);
		tex1_tri();
		regl.TexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_REPEAT);
		tex1_tri();
	} else
#endif
	{
		printf("# No EXT_texture_mirror_clamp\n");
	}

	regl.Disable(GL_TEXTURE_3D);
	regl.DeleteTextures(1, &tid);
	dump_after(0);

	free(texture);
	TEST_EPILOGUE;
}

void test_fog()
{
	GLfloat fogColor[4] = {0.4, 0.5, 0.6, 0.7};

	TEST_PROLOGUE;
	printf("# test_fog()\n");
	dump_before();

	regl.Enable(GL_FOG);
	tri();
	regl.Fogi(GL_FOG_MODE, GL_LINEAR);
	tri();
	regl.Fogf(GL_FOG_DENSITY, 0.9);
	tri();
	regl.Fogf(GL_FOG_START, 0.3);
	tri();
	regl.Fogf(GL_FOG_END, 0.8);
	tri();	
	regl.Fogfv(GL_FOG_COLOR, fogColor);
	tri();

	regl.Fogf(GL_FOG_DENSITY, 1.0);
	regl.Fogf(GL_FOG_START, 0.0);
	regl.Fogf(GL_FOG_END, 1.0);
	regl.Fogi(GL_FOG_MODE, GL_EXP);
	regl.Disable(GL_FOG);
	dump_after(0);
	TEST_EPILOGUE;
}

void test_fog_hint()
{
	GLfloat fogColor[4] = {0.4, 0.5, 0.6, 0.7};

	TEST_PROLOGUE;
	printf("# test_fog_hint DONT_CARE()\n");
	dump_before();
        regl.Hint(GL_FOG_HINT, GL_DONT_CARE);
	regl.Enable(GL_FOG);
	regl.Fogi(GL_FOG_MODE, GL_LINEAR);
	regl.Fogf(GL_FOG_DENSITY, 0.1);
	regl.Fogf(GL_FOG_START, 0.2);
	regl.Fogf(GL_FOG_END, 0.3);
	regl.Fogfv(GL_FOG_COLOR, fogColor);
	tri();
	regl.Disable(GL_FOG);
	dump_after(0);
	printf("# test_fog_hint DONT_FAsTEST()\n");
	dump_before();
        regl.Hint(GL_FOG_HINT, GL_FASTEST);
	regl.Enable(GL_FOG);
	regl.Fogi(GL_FOG_MODE, GL_LINEAR);
	regl.Fogf(GL_FOG_DENSITY, 0.4);
	regl.Fogf(GL_FOG_START, 0.5);
	regl.Fogf(GL_FOG_END, 0.6);
	regl.Fogfv(GL_FOG_COLOR, fogColor);
	tri();
	regl.Disable(GL_FOG);
	dump_after(0);
	printf("# test_fog_hint DONT_NICEST()\n");
	dump_before();
        regl.Hint(GL_FOG_HINT, GL_NICEST);
	regl.Enable(GL_FOG);
	regl.Fogi(GL_FOG_MODE, GL_LINEAR);
	regl.Fogf(GL_FOG_DENSITY, 0.7);
	regl.Fogf(GL_FOG_START, 0.8);
	regl.Fogf(GL_FOG_END, 0.9);
	regl.Fogfv(GL_FOG_COLOR, fogColor);
	tri();
	regl.Disable(GL_FOG);
	dump_after(0);
	TEST_EPILOGUE;
}

void test_nv_fog_distance()
{
#ifndef GL_NV_fog_distance
	printf("# Sorry, no GL_NV_fog_distance\n");
#else
	GLint curFogDistMode;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_fog_distance")) {
		fprintf(stderr,"No GL_NV_fog_distance extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_nv_fog_distance()\n");
	dump_before();

	regl.Enable(GL_FOG);
	regl.GetIntegerv(GL_FOG_DISTANCE_MODE_NV, &curFogDistMode);

	regl.Fogi(GL_FOG_DISTANCE_MODE_NV, GL_EYE_RADIAL_NV);
	tri();
	regl.Fogi(GL_FOG_DISTANCE_MODE_NV, GL_EYE_PLANE);	/* signed but the same as _ABSOLUTE_NV */
	tri();
	regl.Fogi(GL_FOG_DISTANCE_MODE_NV, GL_EYE_PLANE_ABSOLUTE_NV);
	tri();

	dump_after(0);
	regl.Fogi(GL_FOG_DISTANCE_MODE_NV, curFogDistMode);
	regl.Disable(GL_FOG);
	TEST_EPILOGUE;
#endif
}

void test_ext_fog_coord()
{
#ifndef GL_EXT_fog_coord
	printf("# Sorry, no GL_EXT_fog_coord\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_fog_coord")) {
		fprintf(stderr,"No GL_EXT_fog_coord extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_ext_fog_coord()\n");
	dump_before();

	regl.Enable(GL_FOG);
	regl.FogCoordf(2.0);
	tri();
	regl.Fogi(GL_FOG_COORDINATE_SOURCE_EXT, GL_FOG_COORDINATE_EXT);
	tri();

	dump_after(0);
	regl.Fogi(GL_FOG_COORDINATE_SOURCE_EXT, GL_FRAGMENT_DEPTH_EXT);
	regl.FogCoordf(0.0);
	regl.Disable(GL_FOG);
	TEST_EPILOGUE;
#endif
}

void test_arb_texture_cube_map(void)
{
#ifndef GL_ARB_texture_cube_map
	printf("# Sorry, no GL_ARB_texture_cube_map\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_texture_cube_map")) {
		fprintf(stderr,"No GL_ARB_texture_cube_map.\n");
		return;
	}

	GLuint tid;
	int i;
	int *texture = NULL;

	texture = malloc(TEXTURE_WIDTH*TEXTURE_HEIGHT*6*sizeof(int));
	if (!texture) {
		printf("# can not allocate mem for texture\n");
		return;
	}

	for (i=0; i<TEXTURE_WIDTH*TEXTURE_HEIGHT*6; i++) {
		texture[i] = random();
	}

	TEST_PROLOGUE;
	printf("# test_arb_texture_cube_map()\n");

	regl.GenTextures(1, &tid);
	int maxtex;
	regl.GetIntegerv(GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB, &maxtex);
	printf("# max texture size for cube map: %d\n", maxtex);

	dump_before();

	regl.Enable(GL_TEXTURE_CUBE_MAP_ARB);
	regl.BindTexture(GL_TEXTURE_CUBE_MAP_ARB, tid);
	regl.TexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	for (i=0; i<6; i++) {
		regl.TexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB + i, 0, GL_RGBA,
			TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
			GL_RGBA, GL_UNSIGNED_BYTE, &texture[i*TEXTURE_WIDTH*TEXTURE_HEIGHT]);
	}

	regl.TexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
	regl.TexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
	regl.Enable(GL_TEXTURE_GEN_S);
	regl.Enable(GL_TEXTURE_GEN_T);
	tri();

	dump_after(0);

	regl.Disable(GL_TEXTURE_GEN_S);
	regl.Disable(GL_TEXTURE_GEN_T);
	regl.Disable(GL_TEXTURE_CUBE_MAP_ARB);
	regl.DeleteTextures(1, &tid);

	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_startup() {
	TEST_PROLOGUE;
	dump_before();
	regl.Clear(GL_COLOR_BUFFER_BIT);
	dump_after(1);
	TEST_EPILOGUE;
}

void test_texture_filter(void)
{
	GLuint tid;
	int *texture = NULL;

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_texture_filter()\n");

	regl.GenTextures(1, &tid);

	dump_before();

	regl.Enable(GL_TEXTURE_2D);

	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	tex1_tri();
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	tex1_tri();

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	tex1_tri();
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	tex1_tri();

	regl.Disable(GL_TEXTURE_2D);
	regl.DeleteTextures(1, &tid);
	dump_after(0);

	free(texture);
	TEST_EPILOGUE;
}

void test_texture_format(void)
{
#define NUM_TX_FORMATS 40

#define ADD_TX_FORMAT(name) \
	tx_formats_names[max_tx_formats] = # name;	\
	tx_formats[max_tx_formats++] = GL_##name;	

	GLuint tid[NUM_TX_FORMATS];
	int i, max_tx_formats = 0;
	int *texture = NULL;
	GLenum tx_formats[NUM_TX_FORMATS];
	const char *tx_formats_names[NUM_TX_FORMATS];
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	int has_nv_texture_rectangle = 0;

	texture = malloc(TEXTURE_WIDTH*TEXTURE_HEIGHT*sizeof(int));
	if (!texture) {
		printf("# can not allocate mem for texture\n");
		return;
	}

	for (i=0; i<TEXTURE_WIDTH*TEXTURE_HEIGHT; i++) {
		texture[i] = random();
	}

	TEST_PROLOGUE;
	printf("# test_texture_format()\n");

	regl.GenTextures(NUM_TX_FORMATS, tid);
	
	/* The card type comments tell on which gpus the format is hw accelerated. */

	ADD_TX_FORMAT(ALPHA8);
	ADD_TX_FORMAT(LUMINANCE8);
	ADD_TX_FORMAT(INTENSITY8);
	ADD_TX_FORMAT(RGBA4);
	ADD_TX_FORMAT(RGB4);
	ADD_TX_FORMAT(RGB5);
	ADD_TX_FORMAT(RGB5_A1);
	ADD_TX_FORMAT(RGBA8);
	ADD_TX_FORMAT(RGB8);

	if (strstr(extensions, "GL_EXT_paletted_texture")) {
		ADD_TX_FORMAT(COLOR_INDEX2_EXT);
		/* ADD_TX_FORMAT(COLOR_INDEX4_EXT); */
		ADD_TX_FORMAT(COLOR_INDEX8_EXT);	/* NV10-NV30 */
	} else {
		printf("#  GL_EXT_paletted_texture not present\n");
	}

	/* DEPTH_COMPONENT is not a valid texture internalformat
	ADD_TX_FORMAT(DEPTH_COMPONENT24);*/		/* NV10+ */

	if (strstr(extensions, "GL_EXT_texture_compression_s3tc")) {
		ADD_TX_FORMAT(COMPRESSED_RGBA_S3TC_DXT1_EXT);	/* NV10+ */
		ADD_TX_FORMAT(COMPRESSED_RGBA_S3TC_DXT3_EXT);	/* NV10+ */
		ADD_TX_FORMAT(COMPRESSED_RGBA_S3TC_DXT5_EXT);	/* NV10+ */
	} else {
		printf("#  GL_EXT_texture_compression_s3tc not present\n");
	}

	ADD_TX_FORMAT(LUMINANCE8_ALPHA8);			/* NV20+ */

	if (strstr(extensions, "GL_NV_texture_shader")) {
		ADD_TX_FORMAT(HILO16_NV);			/* NV20+ */
		ADD_TX_FORMAT(SIGNED_HILO16_NV);		/* NV20+ */
		ADD_TX_FORMAT(DSDT8_NV);			/* NV20+ */
		ADD_TX_FORMAT(DSDT8_MAG8_INTENSITY8_NV);	/* NV20+ */
	} else {
		printf("#  GL_NV_texture_shader not present\n");
	}

	if (strstr(extensions, "GL_NV_texture_shader2")) {
		ADD_TX_FORMAT(SIGNED_RGBA8_NV);			/* NV20+ */
		ADD_TX_FORMAT(SIGNED_LUMINANCE8_NV);		/* NV20+ */
		ADD_TX_FORMAT(SIGNED_LUMINANCE8_ALPHA8_NV);	/* NV20+ */
		ADD_TX_FORMAT(SIGNED_ALPHA8_NV);		/* NV20+ */
		ADD_TX_FORMAT(SIGNED_INTENSITY8_NV);		/* NV20+ */
		ADD_TX_FORMAT(SIGNED_RGB8_UNSIGNED_ALPHA8_NV);	/* NV20+ */
	} else {
		printf("#  GL_NV_texture_shader2 not present\n");
	}

	if (strstr(extensions, "GL_NV_texture_shader3")) {
		ADD_TX_FORMAT(HILO8_NV);			/* NV25+ */
		ADD_TX_FORMAT(SIGNED_HILO8_NV);			/* NV25+ */
	} else {
		printf("#  GL_NV_texture_shader3 not present\n");
	}

	ADD_TX_FORMAT(ALPHA16);					/* NV30+ */
	ADD_TX_FORMAT(LUMINANCE16);				/* NV30+ */
	ADD_TX_FORMAT(INTENSITY16);				/* NV30+ */

	if (strstr(extensions, "GL_NV_float_buffer")) {
	    // Try to use the GL_NV_texture_rectangle extension if we can.
	    // According to this page, 
	    // http://www.jcode.org/content/floating_point_textures.html,
	    // the GL_NV_float_buffer extension is not supported for 
	    // GL_TEXTURE_2D.  That is why the open gl errors occur.
		if (strstr(extensions, "GL_NV_texture_rectangle")) {
			has_nv_texture_rectangle = 1;
		}
		else {
			printf("#  GL_NV_texture_rectangle is not present for GL_NV_float_buffer.");
			printf(" Will fallback to using GL_TEXTURE_2D.\n");
		}
		
		ADD_TX_FORMAT(FLOAT_R32_NV);			/* NV30+ */
		ADD_TX_FORMAT(FLOAT_RGBA16_NV);			/* NV30+ */
		ADD_TX_FORMAT(FLOAT_RG32_NV);			/* NV30+ */
		ADD_TX_FORMAT(FLOAT_RGBA32_NV);			/* NV30+ */
		ADD_TX_FORMAT(FLOAT_RG16_NV);			/* NV40+ */
	} else {
		printf("#  GL_NV_float_buffer not present\n");
	}

	ADD_TX_FORMAT(LUMINANCE16_ALPHA16);			/* NV40+ */

	if (strstr(extensions, "GL_ATI_texture_float")) {
		ADD_TX_FORMAT(RGBA_FLOAT32_ATI);		/* NV40+ */
		ADD_TX_FORMAT(ALPHA_FLOAT32_ATI);		/* NV40+ */
		ADD_TX_FORMAT(INTENSITY_FLOAT32_ATI);		/* NV40+ */
		ADD_TX_FORMAT(LUMINANCE_FLOAT32_ATI);		/* NV40+ */
		ADD_TX_FORMAT(RGBA_FLOAT16_ATI);		/* NV40+ */
		ADD_TX_FORMAT(LUMINANCE_ALPHA_FLOAT16_ATI);	/* NV40+ */
	} else {
		printf("#  GL_ATI_texture_float not present\n");
	}

	uint32_t palette[4*256];
	for (i=0; i<256;i++) {
		palette[i] = (i<<24)|(i<<16)|(i<<8)|i;
	}

	for (i=0;i<max_tx_formats;i++) {
		GLenum tx_target = GL_TEXTURE_2D;
		int format = GL_RGBA;
		int has_palette = 0;

		switch(tx_formats[i])
		{
			case GL_HILO16_NV:
			case GL_SIGNED_HILO16_NV:
			case GL_HILO8_NV:
			case GL_SIGNED_HILO8_NV:
				format = GL_HILO_NV;
				break;
			case GL_DSDT8_NV:
				format = GL_DSDT_NV;
				break;
			case GL_DSDT8_MAG8_INTENSITY8_NV:
				format = GL_DSDT_MAG_VIB_NV;
				break;
			case GL_FLOAT_R32_NV:		
			case GL_FLOAT_RGBA16_NV:		
			case GL_FLOAT_RG32_NV:
			case GL_FLOAT_RGBA32_NV:
			case GL_FLOAT_RG16_NV:
				if (has_nv_texture_rectangle) {
					tx_target = GL_TEXTURE_RECTANGLE_NV;
				}
				break;
			case GL_COLOR_INDEX1_EXT:
			case GL_COLOR_INDEX2_EXT:
			case GL_COLOR_INDEX4_EXT:
			case GL_COLOR_INDEX8_EXT:
				format = GL_COLOR_INDEX;
				has_palette = 1;
				break;
			default:
				break;
		}

		regl.Enable(tx_target);

		regl.BindTexture(tx_target, tid[i]);
		regl.TexParameteri(tx_target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		regl.TexParameteri(tx_target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		if (has_palette) {
			regl.ColorTableEXT(tx_target, GL_RGBA,
				1<<(1<<(tx_formats[i]-GL_COLOR_INDEX1_EXT)),
				GL_RGBA, GL_UNSIGNED_BYTE, palette);
		}

		printf("# TEXformat %d: %s\n", i, tx_formats_names[i]);

		dump_before();
		
		regl.TexImage2D(tx_target, 0, tx_formats[i], TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
			format, GL_UNSIGNED_BYTE, texture);
		tex1_tri();
		dump_after(0);

		regl.Disable(tx_target);
	}

	regl.DeleteTextures(NUM_TX_FORMATS, tid);

	free(texture);
	TEST_EPILOGUE;
}

void test_texture_wrap(void)
{
	GLuint tid;
	GLenum wrap_coord[2]={GL_TEXTURE_WRAP_S,GL_TEXTURE_WRAP_T};
	int *texture = NULL;
	int i;
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_texture_wrap()\n");

	regl.GenTextures(1, &tid);

	dump_before();

	/*regl.ActiveTexture(GL_TEXTURE2_ARB);*/
	regl.Enable(GL_TEXTURE_2D);

	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	for (i=0; i<2; i++) {
		regl.TexParameteri(GL_TEXTURE_2D, wrap_coord[i], GL_CLAMP);
		tri();
		regl.TexParameteri(GL_TEXTURE_2D, wrap_coord[i], GL_CLAMP_TO_EDGE);
		tri();
		regl.TexParameteri(GL_TEXTURE_2D, wrap_coord[i], GL_CLAMP_TO_BORDER);
		tri();

#ifdef GL_ARB_texture_mirrored_repeat
		if (strstr(extensions, "GL_ARB_texture_mirrored_repeat")) {
			regl.TexParameteri(GL_TEXTURE_2D, wrap_coord[i], GL_MIRRORED_REPEAT_ARB);
			tri();
		} else
#endif
		{
			printf("# No ARB_texture_mirrored_repeat\n");
		}

#ifdef GL_EXT_texture_mirror_clamp
		if (strstr(extensions, "GL_EXT_texture_mirror_clamp")) {
			regl.TexParameteri(GL_TEXTURE_2D, wrap_coord[i], GL_MIRROR_CLAMP_EXT);
			tex1_tri();
			regl.TexParameteri(GL_TEXTURE_2D, wrap_coord[i], GL_MIRROR_CLAMP_TO_EDGE_EXT);
			tex1_tri();
			regl.TexParameteri(GL_TEXTURE_2D, wrap_coord[i], GL_MIRROR_CLAMP_TO_BORDER_EXT);
			tex1_tri();
		} else
#endif
		{
			printf("# No EXT_texture_mirror_clamp\n");
		}

		/* Restore default */
		regl.TexParameteri(GL_TEXTURE_2D, wrap_coord[i], GL_REPEAT);
		tri();
	}

	regl.Disable(GL_TEXTURE_2D);
	regl.DeleteTextures(1, &tid);
	dump_after(0);

	free(texture);
	TEST_EPILOGUE;
}

void test_draw_buffer(void)
{
	const GLenum bufname[]={
		GL_FRONT, GL_BACK, GL_LEFT, GL_RIGHT,
		GL_FRONT_LEFT, GL_FRONT_RIGHT,
		GL_BACK_LEFT, GL_BACK_RIGHT,
		GL_FRONT_AND_BACK,
		GL_AUX0, GL_AUX1, GL_AUX2, GL_AUX3,
		GL_NONE
	};
	// 0 means: available in the default GL context
	const SDL_GLattr attr[] = {0, SDL_GL_DOUBLEBUFFER, 0, SDL_GL_STEREO,
				0, SDL_GL_STEREO ,
				SDL_GL_DOUBLEBUFFER, SDL_GL_DOUBLEBUFFER | SDL_GL_STEREO,
				SDL_GL_DOUBLEBUFFER,
				0, 0, 0, 0, 0};
	int dobuff, stereo;
	SDL_GL_GetAttribute(SDL_GL_DOUBLEBUFFER, &dobuff);
	SDL_GL_GetAttribute(SDL_GL_STEREO, &stereo);
	fprintf(stderr, "\n");
	if (!stereo)
	   fprintf(stderr, "No Stereo context. Skipping stereo tests\n");
	if (!dobuff)
	fprintf(stderr, "No double buffering. Skipping double buffer tests\n");
	TEST_PROLOGUE;
	printf("# test_draw_buffer()\n");
	dump_before();

	int i;
	for (i=0;i< sizeof(bufname)/ sizeof(bufname[0]);i++) {
		if ((stereo && (attr[i] & SDL_GL_STEREO)) ||
		    (dobuff && (attr[i] & SDL_GL_DOUBLEBUFFER)) ||
		    (attr[i] == 0))
		{
			printf("# Testing buff %d\n",i);
			regl.DrawBuffer(bufname[i]);
			tri();
		}
	}

	dump_after(0);
	TEST_EPILOGUE;

	//regl.DrawBuffer(GL_BACK);	/* default for double buffer */
	regl.DrawBuffer(GL_FRONT);	/* default for single buffer */
}

void test_ext_texture_filter_anisotropic(void)
{
#ifndef GL_EXT_texture_filter_anisotropic
	printf("# Sorry, no GL_EXT_texture_filter_anisotropic\n");
#else
	GLuint tid;
	float i;
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_texture_filter_anisotropic")) {
		fprintf(stderr,"No GL_EXT_texture_filter_anisotropic extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_ext_texture_filter_anisotropic()\n");

	float max_aniso;
	regl.GetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT,&max_aniso);
	printf("# max anisotropy: %.3f\n", max_aniso);

	regl.GenTextures(1, &tid);

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	regl.TexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS,GL_TRUE);

	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	for (i=1;i<= max_aniso;i+=1.0) {
		printf("# coef = %f\n", i);

		dump_before();
		regl.TexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, i);
		tex1_tri();
		dump_after(0);
	}

	regl.TexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 1.0);
	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_single_vertex(void)
{
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	int maxtex = 1, i;
	if (strstr(extensions, "GL_ARB_multitexture")) {
		regl.GetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &maxtex);
	}

	TEST_PROLOGUE;
	printf("# test_single_vertex()\n");
	dump_before();

	/* color 3f */
	regl.Color3f(0.01,0.02,0.03);
	regl.Begin(GL_POINTS);
		regl.Vertex2s(1,2);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	/* color 4f */
	regl.Color4f(0.01,0.02,0.03,0.04);

	/* secondary color 3f */
	if (strstr(extensions, "GL_EXT_secondary_color")) {
		regl.Enable(GL_COLOR_SUM_EXT);
		regl.SecondaryColor3f(0.05,0.06,0.07);
	}

	/* vertex weight 1f */
	if (strstr(extensions, "GL_EXT_vertex_weighting")) {
		regl.Enable(GL_VERTEX_WEIGHTING_EXT);
		regl.VertexWeightfEXT(0.07);
	}

	/* fog 1f */
	if (strstr(extensions, "GL_EXT_fog_coord")) {
		regl.FogCoordf(0.08);
	}

	/* normal 3f */
	regl.Normal3f(0.09,0.10,0.11);

	/* texcoord 2f */
	regl.TexCoord2f(0.12,0.13);
	if (maxtex>1) {
		for (i=1;i<maxtex;i++) {
			regl.MultiTexCoord2f(GL_TEXTURE0_ARB+i, 0.14+i*2, 0.15+i*2);
		}
	}

	regl.Begin(GL_POINTS);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	/* texcoord 4f */
	regl.TexCoord4f(0.12,0.13,0.14,0.15);
	if (maxtex>1) {
		for (i=1;i<maxtex;i++) {
			regl.MultiTexCoord4f(GL_TEXTURE0_ARB+i, 0.16+i*4, 0.17+i*4,0.18+i*4,0.19+i*4);
		}
	}

	regl.Begin(GL_POINTS);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	/* color 4i */
	regl.Color4ub(10,20,30,40);

	/* secondary color 3i */
	if (strstr(extensions, "GL_EXT_secondary_color")) {
		regl.SecondaryColor3ub(50,60,70);
	}

	/* texcoord 2i */
	regl.TexCoord2s(80,90);
	if (maxtex>1) {
		for (i=1;i<maxtex;i++) {
			regl.MultiTexCoord2s(GL_TEXTURE0_ARB+i, 80+i*2, 90+i*2);
		}
	}

	/* normal 3i */
	regl.Normal3s(120,130,140);

	regl.Begin(GL_POINTS);
		regl.Vertex4i(40,50,60,70);
	regl.End();

	/* texcoord 4i */
	regl.TexCoord4s(80,90,100,110);
	if (maxtex>1) {
		for (i=1;i<maxtex;i++) {
			regl.MultiTexCoord4s(GL_TEXTURE0_ARB+i, 80+i*4, 90+i*4, 100+i*4, 110+i*4);
		}
	}

	regl.Begin(GL_POINTS);
		regl.Vertex3i(40,50,60);
	regl.End();

	dump_after(0);

	if (strstr(extensions, "GL_EXT_vertex_weighting")) {
		regl.Disable(GL_VERTEX_WEIGHTING_EXT);
	}
	if (strstr(extensions, "GL_EXT_secondary_color")) {
		regl.Disable(GL_COLOR_SUM_EXT);
	}
	regl.MatrixMode(GL_TEXTURE);
	regl.LoadIdentity();
	regl.MatrixMode(GL_MODELVIEW);
	regl.LoadIdentity();
	regl.Color4f(1.0,1.0,1.0,1.0);

	TEST_EPILOGUE;
}

void test_nv_texture_shader()
{
#ifndef GL_NV_texture_shader
	printf("# Sorry, no GL_NV_texture_shader\n");
#else
	const int cull_modes[4]={GL_LESS,GL_GEQUAL,GL_LESS,GL_GEQUAL};
	const int cull_modes_default[4]={GL_GEQUAL,GL_GEQUAL,GL_GEQUAL,GL_GEQUAL};
	const GLfloat offsets[4] = {0.25, 0.5, 0.75, 0};
	const GLfloat offsets_default[4] = {1.0, 0.0, 0.0, 1.0};
	GLuint tid[4];
	int i,j;
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_texture_shader")) {
		fprintf(stderr,"No GL_NV_texture_shader extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_nv_texture_shader\n");

	regl.GenTextures(4, tid);

	regl.ActiveTexture(GL_TEXTURE0);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[0]);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_DSDT_NV, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_DSDT_NV, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.ActiveTexture(GL_TEXTURE1);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[1]);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.ActiveTexture(GL_TEXTURE2);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[2]);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.ActiveTexture(GL_TEXTURE3);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[3]);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	/* Simple enabling */
	printf("# Enabling texture shader\n");
	dump_before();
	tex1_tri();
	regl.Enable(GL_TEXTURE_SHADER_NV);
	tex1_tri();
	regl.ActiveTexture(GL_TEXTURE0);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE1);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE2);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE3);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);
	tex1_tri();
	dump_after(0);

	printf("# Previous inputs\n");
	dump_before();

	/* Set DSDT texture on all units */
	for (i=0; i<4; i++) {
		regl.ActiveTexture(GL_TEXTURE0+i);
		regl.TexImage2D(GL_TEXTURE_2D, 0, GL_DSDT_NV, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
			GL_DSDT_NV, GL_UNSIGNED_BYTE, texture);
	}

	/* GL_OFFSET_TEXTURE_2D_NV requires a prior texture shader result. 
	* It won't work for texture 0.*/
	for (i=1; i<4; i++) {
		/* GL_PREVIOUS_TEXTURE_INPUT_NV requires a valid prior texture
		 * shader result. However, it must be a texture unit less than
		 * ours. It seems you cannot use texture units higher than the
		 * current.*/
		regl.ActiveTexture(GL_TEXTURE0+i);
		regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_OFFSET_TEXTURE_2D_NV);
		for (j=0;j<i;j++) {
			if (j == i)
				continue;
			regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB+j);
			CHECK_ERROR;
			tex1_tri();
		}
		regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);
		regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
	}
	dump_after(0);

	/* Restore RGBA texture, except for TX0 */
	for (i=1; i<4; i++) {
		regl.ActiveTexture(GL_TEXTURE0+i);
		regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
			GL_RGBA, GL_UNSIGNED_BYTE, texture);
	}
	regl.ActiveTexture(GL_TEXTURE2);

	printf("# Pass through\n");
	dump_before();
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_PASS_THROUGH_NV);
	tex1_tri();
	dump_after(0);

	printf("# Cull fragment\n");
	dump_before();
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_CULL_FRAGMENT_NV);
	regl.TexEnviv(GL_TEXTURE_SHADER_NV, GL_CULL_MODES_NV, cull_modes);
	tex1_tri();
	regl.TexEnviv(GL_TEXTURE_SHADER_NV, GL_CULL_MODES_NV, cull_modes_default);
	tex1_tri();
	dump_after(0);

	printf("# Offset texture\n");
	dump_before();
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_OFFSET_TEXTURE_2D_NV);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
	regl.TexEnvf(GL_TEXTURE_SHADER_NV, GL_OFFSET_TEXTURE_SCALE_NV, 2.0);
	regl.TexEnvf(GL_TEXTURE_SHADER_NV, GL_OFFSET_TEXTURE_BIAS_NV, 1.0);
	regl.TexEnvfv(GL_TEXTURE_SHADER_NV, GL_OFFSET_TEXTURE_MATRIX_NV, offsets);
	tex1_tri();
	regl.TexEnvf(GL_TEXTURE_SHADER_NV, GL_OFFSET_TEXTURE_SCALE_NV, 1.0);
	tex1_tri();
	regl.TexEnvf(GL_TEXTURE_SHADER_NV, GL_OFFSET_TEXTURE_BIAS_NV, 0.0);
	regl.TexEnvfv(GL_TEXTURE_SHADER_NV, GL_OFFSET_TEXTURE_MATRIX_NV, offsets_default);
	tex1_tri();
	dump_after(0);

	printf("# Dependent texture\n");
	dump_before();
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DEPENDENT_AR_TEXTURE_2D_NV);
	tex1_tri();
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DEPENDENT_GB_TEXTURE_2D_NV);
	tex1_tri();
	dump_after(0);

	/* Set a HILO texture for dot product */
	regl.ActiveTexture(GL_TEXTURE0);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_HILO_NV, TEXTURE_WIDTH/2,TEXTURE_HEIGHT/2, 0,
		GL_HILO_NV, GL_UNSIGNED_SHORT, texture);
	regl.ActiveTexture(GL_TEXTURE2);

	printf("# Dot product texture\n");
	dump_before();
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DOT_PRODUCT_NV);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
	tex1_tri();
	dump_after(0);

	printf("# Dot product texture 2D\n");
	dump_before();
	regl.ActiveTexture(GL_TEXTURE0);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE1);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DOT_PRODUCT_NV);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
	regl.ActiveTexture(GL_TEXTURE2);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DOT_PRODUCT_TEXTURE_2D_NV);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
	tex1_tri();
	dump_after(0);

	/*	TODO?
		dot product reflect cube map
		dot product const eye reflect cube map
		dot product diffuse cube map
	*/

	printf("# Dot product depth replace\n");
	dump_before();
	regl.ActiveTexture(GL_TEXTURE0);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE1);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DOT_PRODUCT_NV);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
	regl.ActiveTexture(GL_TEXTURE2);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DOT_PRODUCT_DEPTH_REPLACE_NV);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
	tex1_tri();
	dump_after(0);

	/* Restore defaults */
	regl.ActiveTexture(GL_TEXTURE3);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_NONE);
	regl.Disable(GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE2);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_NONE);
	regl.Disable(GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE1);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_NONE);
	regl.Disable(GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE0);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_NONE);
	regl.Disable(GL_TEXTURE_2D);
	regl.Disable(GL_TEXTURE_SHADER_NV);
	regl.DeleteTextures(4, tid);
	free(texture);


#if 0 /* --- Old code --- */
	dump_before();

	GLuint tid[10];
	GLuint tdata_uint[128*128*4];
#if 0
	GLint cullModes[4] = {GL_LESS, GL_LESS, GL_GEQUAL, GL_LESS};
#endif
	GLfloat offsets[4] = {0.25, 0.5, 0.75, 0};

	regl.Enable(GL_TEXTURE_SHADER_NV);

	regl.GenTextures(10, tid);

	regl.ActiveTexture(GL_TEXTURE0);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[0]);
	regl.TexImage2D(GL_TEXTURE_2D, 0,
			GL_DSDT_NV,
			64, 64,
			0,
			GL_DSDT_NV, GL_UNSIGNED_INT,
			tdata_uint
			);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);


	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);

	regl.ActiveTexture(GL_TEXTURE1);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[0]);
	regl.TexImage2D(GL_TEXTURE_2D, 0,
			GL_RGBA,
			64, 64,
			0,
			GL_RGBA, GL_UNSIGNED_INT,
			tdata_uint
			);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);


	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_OFFSET_TEXTURE_2D_NV);
	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0);
	regl.TexEnvfv(GL_TEXTURE_SHADER_NV, GL_OFFSET_TEXTURE_MATRIX_NV, offsets);

#if 0
	regl.ActiveTexture(GL_TEXTURE2);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[1]);
	regl.TexImage2D(GL_TEXTURE_2D, 0,
			GL_RGBA,
			64, 64,
			0,
			GL_RGBA, GL_UNSIGNED_INT,
			tdata_uint
			);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);

	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE1);
//	regl.TexEnviv(GL_TEXTURE_SHADER_NV, GL_CULL_MODES_NV, cullModes);

	regl.ActiveTexture(GL_TEXTURE3);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid[1]);
	regl.TexImage2D(GL_TEXTURE_2D, 0,
			GL_RGBA,
			64, 64,
			0,
			GL_RGBA, GL_UNSIGNED_INT,
			tdata_uint
			);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);

	regl.TexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE1);
//	regl.TexEnviv(GL_TEXTURE_SHADER_NV, GL_CULL_MODES_NV, cullModes);
#endif

	tex_tri();

	dump_after(0);
	regl.ActiveTexture(GL_TEXTURE0);
	regl.Disable(GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE1);
	regl.Disable(GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE2);
	regl.Disable(GL_TEXTURE_2D);
	regl.ActiveTexture(GL_TEXTURE3);
	regl.Disable(GL_TEXTURE_2D);
	regl.Disable(GL_TEXTURE_SHADER_NV);
#endif
#endif /* --- Old code ---*/
	TEST_EPILOGUE;
}

void test_nv_point_sprite(void)
{
#ifndef GL_NV_point_sprite
	printf("# Sorry, no GL_NV_point_sprite\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_point_sprite")) {
		fprintf(stderr,"No GL_NV_point_sprite.\n");
		return;
	}

	int maxtex = 1;
	if (strstr(extensions, "GL_ARB_multitexture")) {
		regl.GetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &maxtex);
	}
	if (maxtex<4) {
		printf("# Not enough texture units: %d<4\n", maxtex);
		return;
	}

	GLuint tid;
	int *texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_nv_point_sprite()\n");

	regl.GenTextures(1, &tid);

	dump_before();
	regl.Enable(GL_POINT_SPRITE_NV);

	regl.ActiveTexture(GL_TEXTURE3);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH, TEXTURE_HEIGHT,
			0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.Begin(GL_POINTS);
		regl.MultiTexCoord2f(GL_TEXTURE3_ARB, 0.1, 0.2);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	regl.TexEnvi(GL_POINT_SPRITE_NV, GL_COORD_REPLACE_NV, GL_TRUE);
	regl.PointParameteriNV(GL_POINT_SPRITE_R_MODE_NV, GL_S);

	regl.Begin(GL_POINTS);
		regl.MultiTexCoord2f(GL_TEXTURE3_ARB, 0.1, 0.2);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	regl.PointParameteriNV(GL_POINT_SPRITE_R_MODE_NV, GL_R);

	regl.Begin(GL_POINTS);
		regl.MultiTexCoord2f(GL_TEXTURE3_ARB, 0.1, 0.2);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	regl.PointParameteriNV(GL_POINT_SPRITE_R_MODE_NV, GL_ZERO);

	regl.Begin(GL_POINTS);
		regl.MultiTexCoord2f(GL_TEXTURE3_ARB, 0.1, 0.2);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	dump_after(0);
	regl.ActiveTexture(GL_TEXTURE3);
	regl.Disable(GL_TEXTURE_2D);
	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_POINT_SPRITE_NV);
	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_arb_point_sprite(void)
{
#ifndef GL_ARB_point_sprite
	printf("# Sorry, no GL_ARB_point_sprite\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_point_sprite")) {
		fprintf(stderr,"No GL_ARB_point_sprite.\n");
		return;
	}

	GLuint tid;
	int *texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_arb_point_sprite()\n");

	regl.GenTextures(1, &tid);

	dump_before();
	regl.Enable(GL_POINT_SPRITE_ARB);

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH, TEXTURE_HEIGHT,
			0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.Begin(GL_POINTS);
		regl.TexCoord2f(0.1, 0.2);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	regl.TexEnvi(GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE);

	regl.Begin(GL_POINTS);
		regl.TexCoord2f(0.1, 0.2);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	dump_after(0);
	regl.Disable(GL_TEXTURE_2D);
	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_POINT_SPRITE_ARB);
	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_arb_point_parameters(void)
{
#ifndef GL_ARB_point_parameters
	printf("# Sorry, no GL_ARB_point_parameters\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_point_parameters")) {
		fprintf(stderr,"No GL_ARB_point_parameters.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# test_arb_point_parameters()\n");

	float smin, smax, sthre, distatt[3];
	regl.GetFloatv(GL_POINT_SIZE_MIN_ARB, &smin);
	regl.GetFloatv(GL_POINT_SIZE_MAX_ARB, &smax);
	regl.GetFloatv(GL_POINT_FADE_THRESHOLD_SIZE_ARB, &sthre);
	printf("#  size min: %.3f, size max: %.3f, threshold: %.3f\n", smin, smax, sthre);
	regl.GetFloatv(GL_POINT_DISTANCE_ATTENUATION_ARB, distatt);
	printf("#  dist atten: %.3f,%.3f,%.3f\n", distatt[0],distatt[1],distatt[2]);

	dump_before();

	distatt[0] = 0.4;
	distatt[1] = 0.5;
	distatt[2] = 0.6;
	regl.PointParameterfvARB(GL_POINT_DISTANCE_ATTENUATION_ARB, distatt);

	regl.Begin(GL_POINTS);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	regl.PointParameterfARB(GL_POINT_SIZE_MIN_ARB, smin + 7.0);

	regl.Begin(GL_POINTS);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	regl.PointParameterfARB(GL_POINT_SIZE_MAX_ARB, smax - 8.0);

	regl.Begin(GL_POINTS);
		regl.Vertex3f(1.0,2.0,3.0);
	regl.End();

	dump_after(0);

#if TEST_POINTPARAMETERS_PARAMS
	float i;

	regl.PointParameterfARB(GL_POINT_SIZE_MIN_ARB, smin);
	regl.PointParameterfARB(GL_POINT_SIZE_MAX_ARB, smax);

	/* test point size min */
	for (i=0.0;i<smax/2.0;i+=10.0) {
		printf("# arb_point_parameters: point size min = %f\n", i);
		float amp = smax - (i<1.0 ? 1.0 : i);
		printf("#  point param a = %f\n", distatt[0]*amp*amp);
		printf("#  point param b = %f\n", distatt[1]*amp*amp);
		printf("#  point param c = %f\n", distatt[2]*amp*amp);
		printf("#  point param d = %f\n", amp);
		printf("#  point param e = %f\n", amp);
		printf("#  point param f = %f\n", amp);
		printf("#  point param g = %f\n", -1.0/amp);
		printf("#  point param h = %f\n", i<1.0 ? 1.0 : i);
		dump_before();
		regl.PointParameterfARB(GL_POINT_SIZE_MIN_ARB, i);
		regl.Begin(GL_POINTS);
			regl.Vertex3f(1.0,2.0,3.0);
		regl.End();
		dump_after(0);
	}
	regl.PointParameterfARB(GL_POINT_SIZE_MIN_ARB, smin);

	/* test point size max */
	for (i=smax/2.0;i<smax;i+=10.0) {
		printf("# arb_point_parameters: point size max = %f\n", i);
		float amp = i - (smin<1.0 ? 1.0 : smin);
		printf("#  point param a = %f\n", distatt[0]*amp*amp);
		printf("#  point param b = %f\n", distatt[1]*amp*amp);
		printf("#  point param c = %f\n", distatt[2]*amp*amp);
		printf("#  point param d = %f\n", amp);
		printf("#  point param e = %f\n", amp);
		printf("#  point param f = %f\n", amp);
		printf("#  point param g = %f\n", -1.0/amp);
		printf("#  point param h = %f\n", smin<1.0 ? 1.0 : smin);
		dump_before();
		regl.PointParameterfARB(GL_POINT_SIZE_MAX_ARB, i);
		regl.Begin(GL_POINTS);
			regl.Vertex3f(1.0,2.0,3.0);
		regl.End();
		dump_after(0);
	}

#endif
	/* Restore default */
	regl.PointParameterfARB(GL_POINT_SIZE_MIN_ARB, smin);
	regl.PointParameterfARB(GL_POINT_SIZE_MAX_ARB, smax);
	distatt[0] = 1.0;
	distatt[1] = 0.0;
	distatt[2] = 0.0;
	regl.PointParameterfvARB(GL_POINT_DISTANCE_ATTENUATION_ARB, distatt);

	TEST_EPILOGUE;
#endif
}

static void do_reg_test(uint32_t* regs, int size)
{
	int i,j;
	for(i=0;i<size;i+=2)
	{
		printf("# ===================================\n");
		printf("# Regs @ 0x%06x\n",regs[i]);
		for(j=0;j<regs[i+1];j++)
			printf("# [0x%06x/4] = 0x%08x\n",regs[i]+j*4,all_regs[regs[i]/4+j]);
	}
}

void test_interesting_regs()
{
	TEST_PROLOGUE;
	uint32_t regs[]={
	/* NV_PFB regs */
	0x100200,8,
	/* nv10 & original nv40 surfaces */
	0x100240,31,
	/* PGRAPH_DEBUGs and friends */
	0x400080,96,
	/* PFIFO */
	0x2100,1,
	0x2140,1,
	0x2210,5,
	0x2500,4,
	0x3200,16,
	0x3100,4,
	0x3800,4,
	/* PGRAPH_CHANNEL_CTX */
	0x400780,3,
	};
	do_reg_test(regs,sizeof(regs)/sizeof(uint32_t));

	uint32_t nv40_regs[]={
	/* clock regs */
	0x1000,384,
	0x4000,24,
	/* nv40+ surfaces */
	0x100600,64,
	/* PGRAPH_DEBUGs and friends */
	0x406900,64,
	0x2230,1,
	};
	if (card_family>=NV40)
		do_reg_test(nv40_regs,sizeof(nv40_regs)/sizeof(uint32_t));

	uint32_t no_c51_regs[]={
	0x400600,640,
	};

	if ((card_family>=NV40) && ((pci_device_id & 0x0ff0)!=0x240)) {
		do_reg_test(no_c51_regs,sizeof(no_c51_regs)/sizeof(uint32_t));
	}

	uint32_t no_g80_regs[]={
	/* PTIMER */
	/* it seems it has moved on G80 and reading it causes crashes */
	0x9000,144,
	0x720000,128*32
	};
	if (card_family<G80)
		do_reg_test(no_g80_regs,sizeof(no_g80_regs)/sizeof(uint32_t));

	uint32_t g80_regs[]={
	/* clock regs, 0xC000 makes nv40 crash */
	0xC000,24,
	};
	if (card_family>=G80)
		do_reg_test(g80_regs,sizeof(g80_regs)/sizeof(uint32_t));

	TEST_EPILOGUE;
}

void test_nv_fence(void)
{
#ifndef GL_NV_fence
	printf("# Sorry, no GL_NV_fence\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_fence")) {
		fprintf(stderr,"No GL_NV_fence.\n");
		return;
	}

	GLuint tid = 0;
	int *texture = NULL;

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_nv_fence()\n");
	dump_before();

	unsigned int fences[3];
	regl.GenFencesNV(3, fences);

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH, TEXTURE_HEIGHT,
			0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.SetFenceNV(0, GL_ALL_COMPLETED_NV);
	tri();
	regl.SetFenceNV(1, GL_ALL_COMPLETED_NV);
	tex_tri();
	regl.SetFenceNV(2, GL_ALL_COMPLETED_NV);
	tri();

        regl.FinishFenceNV(0);
        regl.FinishFenceNV(1);
        regl.FinishFenceNV(2);

	/* Now without fences */
	tri();
	tex_tri();
	tri();

	regl.DeleteFencesNV(3, fences);
	regl.Disable(GL_TEXTURE_2D);
	dump_after(0);
	TEST_EPILOGUE;

	free(texture);
#endif
}

void test_point_size()
{
    TEST_PROLOGUE;
    dump_before();
    regl.PointSize(0.5);
    tri();
    regl.PointSize(2.0);
    tri();
    regl.PointSize(5.0);
    tri();
    regl.PointSize(1.0);
    tri();
    dump_after(0);
    TEST_EPILOGUE;
}

void test_render_mode()
{
	GLuint sbuf[128] = {0};
	GLfloat fbuf[128] = {0};
	TEST_PROLOGUE;

	dump_before();
	regl.SelectBuffer(128, sbuf);
	regl.RenderMode(GL_SELECT);
	tri();
	dump_after(0);
	
	dump_before(0);
	regl.FeedbackBuffer(128, GL_2D, fbuf);
	regl.RenderMode(GL_FEEDBACK);
	tri();
	dump_after(0);

	dump_before();
	regl.RenderMode(GL_RENDER);
	tri();
	dump_after(0);
	
	TEST_EPILOGUE;
}

void test_arb_imaging(void)
{
#if !defined(GL_ARB_imaging) && !defined(GL_VERSION_1_4)
	fprintf(stderr, "Sorry, no GL_ARB_imaging\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_imaging")) {
		fprintf(stderr,"No GL_ARB_imaging.\n");
		return;
	}

	int *texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_arb_imaging\n");

	/* Color table */
	int i;
	uint32_t palette[256];
	for (i=0;i<256;i++) {
		palette[i] = (i<<24)|(i<<16)|(i<<8)|i;
	}
	GLfloat values[16];
	for (i=0;i<16;i++) {
		values[i] = i / 16.0;
	}

	dump_before();
	printf("# testing GL_COLOR_TABLE\n");
	regl.ColorTable(GL_COLOR_TABLE, GL_RGBA, 256, GL_RGBA, GL_UNSIGNED_BYTE, palette);
	regl.Enable(GL_COLOR_TABLE);
	regl.RasterPos2i(10,200);
	regl.DrawPixels(TEXTURE_WIDTH,TEXTURE_HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.Flush();
	regl.ColorTableParameterfv(GL_COLOR_TABLE, GL_COLOR_TABLE_SCALE, &values[0]);
	regl.ColorTableParameterfv(GL_COLOR_TABLE, GL_COLOR_TABLE_BIAS, &values[4]);
	regl.RasterPos2i(100,200);
	regl.DrawPixels(TEXTURE_WIDTH,TEXTURE_HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.Flush();
	regl.Disable(GL_COLOR_TABLE);
	dump_after(0);

	/* Color matrix */
	dump_before();
	printf("# testing Color matrix\n");
	regl.MatrixMode(GL_COLOR);
	regl.LoadMatrixf(values);
	regl.RasterPos2i(10,200);
	regl.DrawPixels(TEXTURE_WIDTH,TEXTURE_HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.Flush();
	regl.MatrixMode(GL_COLOR);
	regl.LoadIdentity();
	regl.MatrixMode(GL_MODELVIEW);
	dump_after(0);

	/* Convolution */
	GLfloat conv[3][3]={{0.0,-1.0,0.0}, {0.0,1.0,0.0}, {0.0,0.0,0.0}};	

	dump_before();
	printf("# testing convolution\n");
	regl.Enable(GL_CONVOLUTION_2D);
	regl.ConvolutionFilter2D(GL_CONVOLUTION_2D, GL_LUMINANCE, 3, 3, GL_LUMINANCE, GL_FLOAT, conv);
	regl.RasterPos2i(10,200);
	regl.DrawPixels(TEXTURE_WIDTH,TEXTURE_HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.Flush();
	regl.Disable(GL_CONVOLUTION_2D);
	dump_after(0);

	/* Histogram */
/*	dump_before();
	printf("# testing histogram\n");
	regl.Enable(GL_HISTOGRAM);
	regl.Histogram(GL_HISTOGRAM, 256, GL_RGBA, GL_FALSE);
	regl.RasterPos2i(10,200);
	regl.DrawPixels(TEXTURE_WIDTH,TEXTURE_HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.Flush();
	regl.GetHistogram(GL_HISTOGRAM, GL_TRUE, GL_RGBA, GL_UNSIGNED_SHORT, texture);
	regl.ResetHistogram(GL_HISTOGRAM);
	regl.Disable(GL_HISTOGRAM);
	dump_after(0);
*/	
	/* Minmax */
	dump_before();
	printf("# testing GL_MINMAX\n");
	regl.Enable(GL_MINMAX);
	regl.Minmax(GL_MINMAX, GL_RGBA, GL_FALSE);
	regl.RasterPos2i(10,200);
	regl.DrawPixels(TEXTURE_WIDTH,TEXTURE_HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.Flush();
	regl.GetMinmax(GL_MINMAX, GL_TRUE, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.ResetMinmax(GL_MINMAX);
	regl.Disable(GL_MINMAX);
	dump_after(0);

	TEST_EPILOGUE;
	free(texture);
#endif
}

void test_evaluators(void)
{
	GLfloat ctrlpoints[4][3]={
		{-4.0,-4.0,0.0},{-2.0,4.0,0.0},
		{2.0,-4.0,0.0},{4.0,4.0,0.0}
	};
	int i;

	TEST_PROLOGUE;

	/* Evaluator 1D */
	regl.Map1f(GL_MAP1_VERTEX_3, 0.0, 1.0, 3, 4, &ctrlpoints[0][0]);
	regl.Enable(GL_MAP1_VERTEX_3);
	regl.Begin(GL_LINE_STRIP);
	for (i=0; i<20; i++) {
		regl.EvalCoord1f((GLfloat) i / 20.0);
	}
	regl.End();
	regl.Disable(GL_MAP1_VERTEX_3);

	/* TODO: Evaluator 2D */

	TEST_EPILOGUE;
}

void test_ext_framebuffer_object(void)
{
#ifndef GL_EXT_framebuffer_object
	printf("# Sorry, no GL_EXT_framebuffer_object\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_framebuffer_object")) {
		fprintf(stderr,"No GL_EXT_framebuffer_object extension.\n");
		return;
	}

#define RENOUVEAU_MAX_DRAW_BUFFERS 4
	GLuint fb, rb[RENOUVEAU_MAX_DRAW_BUFFERS];
	GLuint fp_obj[RENOUVEAU_MAX_DRAW_BUFFERS];
	GLint nr_bufs;
	GLenum bufs[] = { GL_COLOR_ATTACHMENT0_EXT,
			  GL_COLOR_ATTACHMENT1_EXT,
			  GL_COLOR_ATTACHMENT2_EXT,
			  GL_COLOR_ATTACHMENT3_EXT,
			  GL_COLOR_ATTACHMENT3_EXT,
			  GL_COLOR_ATTACHMENT4_EXT,
			  GL_COLOR_ATTACHMENT5_EXT,
			  GL_COLOR_ATTACHMENT6_EXT,
			  GL_COLOR_ATTACHMENT7_EXT };
	char fp[1024];
	int i;
	int *texture = NULL;

	TEST_PROLOGUE;

/* defined as per http://oss.sgi.com/projects/ogl-sample/registry/ARB/draw_buffers.txt */
#define MAX_DRAW_BUFFERS_ARB 0x8824 
	regl.GetIntegerv(MAX_DRAW_BUFFERS_ARB, &nr_bufs);
	printf("# %d draw buffers supported\n", nr_bufs);
	if (nr_bufs == 1 && !strstr(extensions, "GL_ARB_draw_buffers"))
		printf("# No GL_ARB_draw_buffers extension\n");
	if (nr_bufs > RENOUVEAU_MAX_DRAW_BUFFERS)
		nr_bufs = RENOUVEAU_MAX_DRAW_BUFFERS;

	/* setup, really uninteresting..  */
	regl.Enable(GL_FRAGMENT_PROGRAM_ARB);
	regl.GenFramebuffersEXT(1, &fb);
	regl.GenRenderbuffersEXT(nr_bufs, rb);
	regl.BindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	for (i=0; i<nr_bufs; i++) {
	    regl.BindRenderbufferEXT(GL_RENDERBUFFER_EXT, rb[i]);
	    regl.RenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_RGBA, 32, 32);
	}
	regl.GenProgramsARB(RENOUVEAU_MAX_DRAW_BUFFERS, fp_obj);
	if (nr_bufs == 1 && !strstr(extensions, "GL_ARB_draw_buffers")) {
		/* cards with GL_EXT_framebuffer_object and no GL_ARB_draw_buffers 0322 */
		fp[0] = '\0';
		strcat(fp, "!!ARBfp1.0\nMOV result.color, fragment.color;\nEND\n");
		printf("# Fragprog is:\n");
		print_commented(fp);
		regl.BindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fp_obj[0]);
		regl.ProgramStringARB(GL_FRAGMENT_PROGRAM_ARB,
				      GL_PROGRAM_FORMAT_ASCII_ARB,
				      strlen(fp), fp);
	} else {
		for (i=0; i<nr_bufs; i++) {
			int buf;

			fp[0] = '\0';
			strcat(fp, "!!ARBfp1.0\nOPTION ARB_draw_buffers;\n");
			for (buf=0; buf<=i; buf++) {
				char l[100];
				sprintf(l, "MOV result.color[%d], fragment.color;\n", buf);
				strcat(fp, l);
		
			}
			strcat(fp, "END\n");
			printf("# Fragprog is:\n");
			print_commented(fp);
			regl.BindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fp_obj[i]);
			regl.ProgramStringARB(GL_FRAGMENT_PROGRAM_ARB,
					      GL_PROGRAM_FORMAT_ASCII_ARB,
					      strlen(fp), fp);
		}
	}
	regl.Finish();

	for (i=0; i<nr_bufs; i++) {
	    int buf;

	    printf("# Testing %d draw buffers\n", i+1);
	    dump_before();
	    for (buf=0; buf<=i; buf++) {
		regl.FramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
						GL_COLOR_ATTACHMENT0_EXT + buf,
						GL_RENDERBUFFER_EXT,
						rb[buf]);
	    }
	    printf("# FB status: 0x%08x\n",
		    regl.CheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT));
	    regl.DrawBuffers(i+1, bufs);
	    regl.BindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fp_obj[i]);
	    tri();
	    regl.Finish();
	    dump_after(0);
	}
	
	regl.DeleteProgramsARB(RENOUVEAU_MAX_DRAW_BUFFERS, fp_obj);
	regl.Disable(GL_FRAGMENT_PROGRAM_ARB);
	
	if (nr_bufs)
		regl.DeleteRenderbuffersEXT(nr_bufs, rb);

	/* Render to depth texture, no color buffer */
	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (texture) {
		GLenum fbo_status;
		GLuint tid = 0;

		printf("# Testing render to depth texture only\n");

		dump_before();

		regl.TexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT,
			TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
			GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, NULL);

		regl.FramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
			GL_DEPTH_ATTACHMENT_EXT,
			GL_TEXTURE_2D, tid, 0);
		regl.DrawBuffer(GL_NONE);
		regl.ReadBuffer(GL_NONE);

		tri();

		fbo_status = regl.CheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);

		regl.BindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

		dump_after(0);
		free(texture);
	}

	if (fb)
		regl.DeleteFramebuffersEXT(1, &fb);

	TEST_EPILOGUE;
#endif
}

void test_texture_border(void)
{
	GLuint tid = 0;
	int *texture = NULL;
	GLfloat tx_border_color[]={0.2,0.3,0.4,0.5};

	texture = create_random_texture(TEXTURE_WIDTH+2,TEXTURE_HEIGHT+2);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_texture_border()\n");
	dump_before();

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH+2, TEXTURE_HEIGHT+2,
			1, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	tex1_tri();

	regl.TexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, tx_border_color);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	tex1_tri();

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	tex1_tri();

	regl.Disable(GL_TEXTURE_2D);
	dump_after(0);
	TEST_EPILOGUE;

	free(texture);
}

void test_nehe02(void)
{
#if 0
	// This code was created by Jeff Molofee '99
	// (ported to SDL by Sam Lantinga '2000)
	int Width = 640, Height=480;

	TEST_PROLOGUE;
	printf("# test_nehe02()\n");
	dump_before();

	regl.Viewport(0, 0, Width, Height);
	regl.ClearColor(0.0f, 0.0f, 0.0f, 0.0f);		// This Will Clear The Background Color To Black
	regl.ClearDepth(1.0);				// Enables Clearing Of The Depth Buffer
	regl.DepthFunc(GL_LESS);				// The Type Of Depth Test To Do
	regl.Enable(GL_DEPTH_TEST);			// Enables Depth Testing
	regl.ShadeModel(GL_SMOOTH);			// Enables Smooth Color Shading

	regl.MatrixMode(GL_PROJECTION);
	regl.LoadIdentity();				// Reset The Projection Matrix

	gluPerspective(45.0f,(GLfloat)Width/(GLfloat)Height,0.1f,100.0f);	// Calculate The Aspect Ratio Of The Window
	//regl.Frustum(-50.0,50.0,-50.0,50.0,0.1,100.0);

	regl.MatrixMode(GL_MODELVIEW);
	regl.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Clear The Screen And The Depth Buffer
	regl.LoadIdentity();				// Reset The View

	regl.Translatef(-1.5f,0.0f,-6.0f);		// Move Left 1.5 Units And Into The Screen 6.0
	
	// draw a triangle
	regl.Begin(GL_POLYGON); 			   // start drawing a polygon
	regl.Vertex3f( 0.0f, 1.0f, 0.0f);		// Top
	regl.Vertex3f( 1.0f,-1.0f, 0.0f);		// Bottom Right
	regl.Vertex3f(-1.0f,-1.0f, 0.0f);		// Bottom Left	
	regl.End();					// we're done with the polygon

	regl.Translatef(3.0f,0.0f,0.0f);		        // Move Right 3 Units
	
	// draw a square (quadrilateral)
	regl.Begin(GL_QUADS);				// start drawing a polygon (4 sided)
	regl.Vertex3f(-1.0f, 1.0f, 0.0f);		// Top Left
	regl.Vertex3f( 1.0f, 1.0f, 0.0f);		// Top Right
	regl.Vertex3f( 1.0f,-1.0f, 0.0f);		// Bottom Right
	regl.Vertex3f(-1.0f,-1.0f, 0.0f);		// Bottom Left	
	regl.End();					// done with the polygon

	dump_after(0);
	TEST_EPILOGUE;
#endif
}

/*
 * This test is a bit different from other tests, it replicates EXA composite's A8 Mask behaviour using opengl
 * Look at the renouveau dump, replicate in the DDX, you get composite for dummies !
 * Blend function setup left as an exercise for the reader
 */
void test_exa(void)
{

#ifndef GL_ARB_texture_rectangle
        printf("# Sorry, no GL_ARB_texture_rectangle\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_texture_rectangle")) {
		fprintf(stderr,"No GL_ARB_texture_rectangle extension.");
		return;
	}

	TEST_PROLOGUE;
	GLuint screen_texnum,modulate_texnum;

	#define texw 20
	#define texh 10

	#define texmw 16
	#define texmh 4

	#define w 512
	#define h 512
	int i;
	dump_before();
	regl.Viewport(0,0,w,h);
	regl.MatrixMode(GL_PROJECTION);
	regl.LoadIdentity();
	regl.Ortho(0,w,h,0,-1.0, 1.0);
	regl.MatrixMode(GL_MODELVIEW);
	regl.LoadIdentity();

	regl.Disable(GL_FOG);
	regl.Disable(GL_LIGHTING);
	regl.Disable(GL_CULL_FACE);
	regl.Disable(GL_DEPTH_TEST);
	regl.Disable(GL_BLEND);
	regl.Disable(GL_NORMALIZE);
	regl.Disable(GL_ALPHA_TEST);
	regl.Enable(GL_TEXTURE_RECTANGLE_ARB);
	regl.BlendFunc (GL_SRC_ALPHA, GL_ONE);

	regl.ActiveTextureARB(GL_TEXTURE1_ARB);
	regl.Enable(GL_TEXTURE_RECTANGLE_ARB);
	regl.GenTextures(1,&modulate_texnum);
	regl.BindTexture(GL_TEXTURE_RECTANGLE_ARB,modulate_texnum);
	regl.TexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.Color4f(1.0,1.0,1.0,1.0);

	regl.ActiveTextureARB(GL_TEXTURE0_ARB);
	Uint8* pixels=(Uint8*)malloc(texw*texh*4);
	for(i=0;i<texw*texh*4;i++)
		pixels[i]=rand();
	regl.GenTextures(1,&screen_texnum);
	regl.BindTexture(GL_TEXTURE_RECTANGLE_ARB,screen_texnum);
	regl.TexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0,GL_RGBA8,texw,texh, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);

	regl.TexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
	regl.Color4f(1.0,1.0,1.0,1.0);
	dump_after(0);

	Uint8 *modulate_texture=malloc(texmw*texmh);
	uint8_t step=0;

	printf("# init ok, now uploading\n");
	dump_before();
	do
	{
		step++;
		regl.ActiveTextureARB(GL_TEXTURE1_ARB);
		regl.Enable(GL_TEXTURE_RECTANGLE_ARB);
		regl.BindTexture(GL_TEXTURE_RECTANGLE_ARB,modulate_texnum);
		for(i=0;i<texmw*texmh;i++)
				modulate_texture[i]=i;
		regl.TexImage2D(GL_TEXTURE_RECTANGLE_ARB,0,GL_INTENSITY8,texmw,texmh,0,GL_LUMINANCE,GL_UNSIGNED_BYTE, modulate_texture);
		regl.TexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
		dump_after(0);
		printf("# tex upload ok, now blitting\n");
		dump_before();

		regl.ActiveTextureARB(GL_TEXTURE0_ARB);
		regl.Enable(GL_TEXTURE_RECTANGLE_ARB);
		regl.BindTexture(GL_TEXTURE_RECTANGLE_ARB,screen_texnum);

		regl.Begin(GL_QUADS);
		regl.Color4f(1.0,1.0,1.0,1.0);

		regl.MultiTexCoord2f(GL_TEXTURE0_ARB,0.f, 0.f);
		regl.MultiTexCoord2f(GL_TEXTURE1_ARB,0.f, 0.f);
		regl.Vertex2i(0, 0);

		regl.MultiTexCoord2f(GL_TEXTURE0_ARB,0.f, texh);
		regl.MultiTexCoord2f(GL_TEXTURE1_ARB,0.f, texmh);
		regl.Vertex2i(0, h);

		regl.MultiTexCoord2f(GL_TEXTURE0_ARB,texw, texh);
		regl.MultiTexCoord2f(GL_TEXTURE1_ARB,texmw, texmh);
		regl.Vertex2i(w, h);

		regl.MultiTexCoord2f(GL_TEXTURE0_ARB,texw, 0.f);
		regl.MultiTexCoord2f(GL_TEXTURE1_ARB,texmw, 0.f);
		regl.Vertex2i(w, 0);
		regl.End();

		//SDL_GL_SwapBuffers();
	}	
	while(0);
	dump_after(0);
	printf("# blitting ok\n");
	TEST_EPILOGUE;
#endif
}

void test_ext_abgr(void)
{
#ifndef GL_EXT_abgr
	printf("# Sorry, no GL_EXT_abgr\n");
#else
	GLuint tid;
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_abgr")) {
		fprintf(stderr,"No GL_EXT_abgr extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing GL_EXT_abgr\n");

	regl.GenTextures(1, &tid);
	dump_before();

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_ABGR_EXT, GL_UNSIGNED_BYTE, texture);

	tex1_tri();

	dump_after(0);
	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_ext_bgra(void)
{
#ifndef GL_EXT_bgra
	printf("# Sorry, no GL_EXT_bgra\n");
#else
	GLuint tid;
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_bgra")) {
		fprintf(stderr,"No GL_EXT_bgra extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing GL_EXT_bgra\n");

	regl.GenTextures(1, &tid);
	dump_before();

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_BGR_EXT, GL_UNSIGNED_BYTE, texture);

	tex1_tri();

	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_BGRA_EXT, GL_UNSIGNED_BYTE, texture);

	tex1_tri();

	dump_after(0);
	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_texture_blend_color(void)
{
	GLuint tid;
	int *texture = NULL;
	const GLfloat blendColor[4]={0.2,0.3,0.4,0.5};
	const GLfloat blackColor[4]={0.0,0.0,0.0,0.0};

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing texture blend color\n");

	regl.GenTextures(1, &tid);
	dump_before();

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
	regl.TexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, blendColor);

	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	tex1_tri();

	dump_after(0);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	regl.TexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, blackColor);
	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	free(texture);
	TEST_EPILOGUE;
}

void test_bias()
{
	char *texture = NULL;
	GLuint tid;
	int i;
	texture=malloc(256*256);
	regl.GenTextures(1,&tid);
	for(i=0;i<256*256;i++)
		texture[i]=rand()%256;

	TEST_PROLOGUE;
	dump_before();
	printf("# Testing texture blend color\n");
//	regl.Enable(GL_TEXTURE_SHADER_NV);
	regl.ActiveTextureARB(GL_TEXTURE0_ARB);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	regl.TexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP);
	regl.TexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//	regl.TexEnvi(GL_TEXTURE_SHADER_NV,GL_SHADER_OPERATION_NV,GL_TEXTURE_3D);
	regl.Disable(GL_TEXTURE_3D);
	regl.Enable(GL_TEXTURE_2D);
	CHECK_ERROR;

	regl.Enable(GL_REGISTER_COMBINERS_NV);
	regl.CombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV, 1);
	CHECK_ERROR;
	regl.TexImage2D(GL_TEXTURE_2D, 0,GL_ALPHA,256,256,0,GL_ALPHA, GL_UNSIGNED_BYTE,texture);
	CHECK_ERROR;

	int mode;
	if (0)
		mode=GL_UNSIGNED_IDENTITY_NV;
	else
		mode=GL_HALF_BIAS_NORMAL_NV;
	regl.CombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_A_NV,GL_TEXTURE0, mode, GL_RGB);
	regl.CombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_B_NV,GL_TEXTURE0, mode, GL_ALPHA);
	regl.CombinerOutputNV(GL_COMBINER0_NV, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV,GL_SPARE0_NV, GL_SCALE_BY_FOUR_NV, GL_NONE, GL_FALSE,GL_FALSE, GL_FALSE);
	regl.CombinerOutputNV(GL_COMBINER0_NV, GL_ALPHA, GL_DISCARD_NV, GL_DISCARD_NV,GL_SPARE0_NV, GL_SCALE_BY_FOUR_NV, GL_NONE, GL_FALSE,GL_FALSE, GL_FALSE);
	regl.FinalCombinerInputNV(GL_VARIABLE_A_NV,GL_ZERO,GL_UNSIGNED_IDENTITY_NV,GL_RGB);
	regl.FinalCombinerInputNV(GL_VARIABLE_A_NV,GL_ZERO,GL_UNSIGNED_IDENTITY_NV,GL_ALPHA);
	CHECK_ERROR;

	tex1_tri();
	dump_after(0);

	TEST_EPILOGUE;
}

void test_yuv()
{
#define W 44
#define H 22
	char *texture_y = NULL;
	char *texture_uv = NULL;
	GLuint tex_y;
	GLuint tex_uv;
	GLuint tex_coef;
	int i;

	texture_y=malloc(W*H);
	texture_uv=malloc((W/2)*(H/2)*2);
	regl.GenTextures(1,&tex_y);
	regl.GenTextures(1,&tex_uv);
	regl.GenTextures(1,&tex_coef);
	for(i=0;i<W*H;i++)
		texture_y[i]=rand()%256;
	for(i=0;i<(W/2)*(H/2);i++)
	{
		texture_uv[2*i]=rand()%256;
		texture_uv[2*i+1]=rand()%256;
	}

	TEST_PROLOGUE;
	dump_before();
	printf("# Testing yuv\n");

	regl.ActiveTextureARB(GL_TEXTURE0_ARB);
	regl.Enable(GL_TEXTURE_1D);
	regl.BindTexture(GL_TEXTURE_1D, tex_coef);
	regl.TexImage1D(GL_TEXTURE_1D, 0,GL_SIGNED_RGBA8_NV,256,0,GL_RGBA, GL_UNSIGNED_BYTE,texture_uv);
	regl.TexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexParameteri (GL_TEXTURE_1D, GL_TEXTURE_WRAP_R, GL_REPEAT);
	CHECK_ERROR;

#define TEX_TYPE GL_TEXTURE_RECTANGLE_ARB
	regl.ActiveTextureARB(GL_TEXTURE1_ARB);
	regl.Disable(GL_TEXTURE_2D);
	regl.Enable(TEX_TYPE);
	regl.BindTexture(TEX_TYPE, tex_y);
	regl.TexImage2D(TEX_TYPE, 0,GL_ALPHA8,W,H,0,GL_LUMINANCE, GL_UNSIGNED_BYTE,texture_y);
	regl.TexParameteri(TEX_TYPE, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(TEX_TYPE, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	regl.TexParameteri (TEX_TYPE, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	regl.TexParameteri (TEX_TYPE, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	CHECK_ERROR;

	regl.ActiveTextureARB(GL_TEXTURE2_ARB);
	regl.Disable(GL_TEXTURE_2D);
	regl.Enable(TEX_TYPE);
	regl.BindTexture(TEX_TYPE, tex_uv);
	regl.TexImage2D(TEX_TYPE, 0,GL_LUMINANCE8_ALPHA8_EXT,W/2,H/2,0,GL_LUMINANCE, GL_UNSIGNED_BYTE,texture_uv);
	regl.TexParameteri(TEX_TYPE, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(TEX_TYPE, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	regl.TexParameteri (TEX_TYPE, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	regl.TexParameteri (TEX_TYPE, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	CHECK_ERROR;

#if 1
	// cubic attempt
	const char *fp =
	"!!ARBfp1.0"
	"OPTION ARB_precision_hint_fastest;"
	"TEMP coord, coord2, cdelta, parmx, parmy, a, b, yuv;"
	"ADD coord, fragment.texcoord[0], {0.5, 0.5};"
	"TEX parmx, coord.x, texture[2], 1D;"
	"MUL cdelta.xz, parmx.rrgg, {-1, 0, 1, 0};"
	"TEX parmy, coord.y, texture[2], 1D;"
	"MUL cdelta.yw, parmy.rrgg, {0, -1, 0, 1};"
	"ADD coord, fragment.texcoord[0].xyxy, cdelta.xyxw;"
	"ADD coord2, fragment.texcoord[0].xyxy, cdelta.zyzw;"
	"TEX a.r, coord.xyxy, texture[0], 2D;"
	"TEX a.g, coord.zwzw, texture[0], 2D;"
	"TEX b.r, coord2.xyxy, texture[0], 2D;"
	"TEX b.g, coord2.zwzw, texture[0], 2D;"
	"LRP a.b, parmy.b, a.rrrr, a.gggg;"
	"LRP a.a, parmy.b, b.rrrr, b.gggg;"
	"LRP yuv.r, parmx.b, a.bbbb, a.aaaa;"
	"TEX yuv.gb, fragment.texcoord[1], texture[1], 2D;"
	"PARAM ycoef =   { 1.1643  ,  1.1643  , 1.1643  };"
	"PARAM ucoef =   { 0.0000  , -0.39173 , 2.017   };"
	"PARAM vcoef =   { 1.5958  , -0.81290 , 0.0000  };"
	"PARAM offsets = { -0.87066,  0.52954 ,-1.08128 };"
	"TEMP res;"
	"MAD res.rgb, yuv.rrrr, ycoef, offsets;"
	"MAD res.rgb, yuv.gggg, ucoef, res;"
	"MAD result.color.rgb, yuv.bbbb, vcoef, res;"
	"END";
#else

	const char* fp=
	"!!ARBfp1.0"
	"OPTION ARB_precision_hint_fastest;"
	"TEMP yuv;"
	"TEX yuv.r, fragment.texcoord[0], texture[1], 2D;"
	"TEX yuv.gb, fragment.texcoord[1], texture[2], 2D;"
	"PARAM ycoef =   { 1.1643  ,  1.1643  , 1.1643  };"
	"PARAM ucoef =   { 0.0000  , -0.39173 , 2.017   };"
	"PARAM vcoef =   { 1.5958  , -0.81290 , 0.0000  };"
	"PARAM offsets = { -0.87066,  0.52954 ,-1.08128 };"
	"TEMP res;"
	"MAD res.rgb, yuv.rrrr, ycoef, offsets;"
	"MAD res.rgb, yuv.gggg, ucoef, res;"
	"MAD result.color.rgb, yuv.bbbb, vcoef, res;"
	"END";
#endif
	GLuint s;
//	regl.Enable(GL_FRAGMENT_PROGRAM_ARB);
	regl.GenProgramsARB(1, &s);
	regl.BindProgramARB(GL_FRAGMENT_PROGRAM_ARB, s);
	regl.GetError();
	regl.ProgramStringARB(GL_FRAGMENT_PROGRAM_ARB,
			GL_PROGRAM_FORMAT_ASCII_ARB, strlen(fp), fp);
	if (regl.GetError()) { printf("# err - fp_obj\n"); return; }

	regl.Begin(GL_TRIANGLES);
	regl.MultiTexCoord2f(GL_TEXTURE0_ARB, 0.0, 0.0);
	regl.MultiTexCoord2f(GL_TEXTURE1_ARB, 0.0, 0.0);
	regl.Vertex3f(1.0,0.0,0.0);
	regl.MultiTexCoord2f(GL_TEXTURE0_ARB, 0.0, 10.0);
	regl.MultiTexCoord2f(GL_TEXTURE1_ARB, 0.0, 5.0);
	regl.Vertex3f(0.0,1.0,0.0);
	regl.MultiTexCoord2f(GL_TEXTURE0_ARB, 10.0, 10.0);
	regl.MultiTexCoord2f(GL_TEXTURE1_ARB, 5.0, 5.0);
	regl.Vertex3f(1.0,1.0,0.0);
	regl.End();
	dump_after(0);

	TEST_EPILOGUE;
}


void test_float_tex()
{
	TEST_PROLOGUE;
	dump_before();
	printf("# Testing float texture\n");

	int TS=4;
	float* texture=(float*)malloc(sizeof(float)*TS*TS*4);
	int i;
	for(i=0;i<TS*4;i++)
	{	
		switch(i%8)
		{
			case 0:texture[i]=0.0;break;// R
			case 1:texture[i]=0.1;break;// G
			case 2:texture[i]=0.2;break;// B
			case 3:texture[i]=0.3;break;// A
			case 4:texture[i]=0.4;break;// R
			case 5:texture[i]=0.5;break;// G
			case 6:texture[i]=0.6;break;// B
			case 7:texture[i]=0.7;break;// A
		}
	}
	regl.ActiveTexture(GL_TEXTURE0);
	regl.Enable(GL_TEXTURE_1D);
	regl.BindTexture(GL_TEXTURE_1D, 1);
	regl.TexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	regl.TexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	regl.TexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexImage1D(GL_TEXTURE_1D, 0, GL_RGBA16F_ARB, TS, 0, GL_RGBA, GL_FLOAT, texture);

	tex1_tri();

	dump_after(0);

	TEST_EPILOGUE;
}

void test_arb_shadow(void)
{
#if !defined(GL_ARB_shadow) || !defined(GL_ARB_depth_texture)
	printf("# Sorry, no GL_ARB_shadow\n");
#else
	GLuint tid;
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_shadow")
	  || !strstr(extensions, "GL_ARB_depth_texture")
	) {
		fprintf(stderr,"No GL_ARB_shadow extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_arb_shadow()\n");

	regl.GenTextures(1, &tid);
	dump_before();

	regl.Enable(GL_DEPTH_TEST);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, TEXTURE_WIDTH, TEXTURE_HEIGHT,
			0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	tex1_tri();

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
	tex1_tri();

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_GEQUAL);
	tex1_tri();

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_NONE);

	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	regl.Disable(GL_DEPTH_TEST);
	dump_after(0);
	TEST_EPILOGUE;

	free(texture);
#endif
}

void test_ext_shadow_funcs(void)
{
#if !defined(GL_EXT_shadow_funcs) || !defined(GL_ARB_shadow) || !defined(GL_ARB_depth_texture)
	printf("# Sorry, no GL_EXT_shadow_funcs\n");
#else
	GLuint tid;
	int *texture = NULL;
	int i;
	const GLenum txcomp[8]={
		GL_LEQUAL,GL_GEQUAL, GL_LESS, GL_GREATER,
		GL_EQUAL, GL_NOTEQUAL, GL_ALWAYS, GL_NEVER
	};

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_shadow_funcs")
	  || !strstr(extensions, "GL_ARB_shadow")
	  || !strstr(extensions, "GL_ARB_depth_texture")
	) {
		fprintf(stderr,"No GL_EXT_shadow_funcs extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_ext_shadow_funcs()\n");

	regl.GenTextures(1, &tid);
	dump_before();

	regl.Enable(GL_DEPTH_TEST);
	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, TEXTURE_WIDTH, TEXTURE_HEIGHT,
			0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, texture);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	tex1_tri();

	for (i=0; i<8; i++) {
		regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE);
		regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, txcomp[i]);
		tex1_tri();
	}

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_NONE);

	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	regl.Disable(GL_DEPTH_TEST);
	dump_after(0);
	TEST_EPILOGUE;

	free(texture);
#endif
}

void test_ext_texture_lod_bias(void)
{
#ifndef GL_EXT_texture_lod_bias
	printf("# Sorry, no GL_EXT_texture_lod_bias\n");
#else
	GLuint tid;
	GLint max_bias;
	int *texture = NULL;
	int i;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_texture_lod_bias")) {
		fprintf(stderr,"No GL_EXT_texture_lod_bias extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing texture lod bias\n");

	regl.GetIntegerv(GL_MAX_TEXTURE_LOD_BIAS_EXT, &max_bias);
	printf("# Max lod bias: %d\n", max_bias);

	regl.GenTextures(1, &tid);

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	for (i=0; i<=max_bias; i++) {
		printf("# Set bias to %d\n", i);
		dump_before();

		regl.TexEnvi(GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT, i);
		tex1_tri();

		dump_after(0);
	}

	regl.TexEnvi(GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT, 0);

	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_nv_texture_expand_normal(void)
{
#ifndef GL_NV_texture_expand_normal
	printf("# Sorry, no GL_NV_texture_expand_normal\n");
#else
	GLuint tid;
	int *texture = NULL;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_texture_expand_normal")) {
		fprintf(stderr,"No GL_NV_texture_expand_normal extension.\n");
		return;
	}

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing texture expand normal\n");

	regl.GenTextures(1, &tid);

	dump_before();

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_UNSIGNED_REMAP_MODE_NV, GL_EXPAND_NORMAL_NV);
	tex1_tri();
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_UNSIGNED_REMAP_MODE_NV, GL_NONE);
	tex1_tri();

	dump_after(0);

	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	free(texture);
	TEST_EPILOGUE;
#endif
}

void test_texture_mipmap_levels(void)
{
	GLuint tid;
	int *texture = NULL;
	int i;

	texture = create_random_texture(256,256);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing texture mipmap levels\n");

	regl.GenTextures(1, &tid);

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	for (i=0; i<8; i++) {
		printf("# Add texture for mipmap level %d\n", i);

		dump_before();
		regl.TexImage2D(GL_TEXTURE_2D, i, GL_RGBA, 256>>i, 256>>i, 0,
			GL_RGBA, GL_UNSIGNED_BYTE, texture);
		regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, i);
		tex1_tri();
		dump_after(0);
	}

	for (i=7; i>=0; i--) {
		printf("# Set texture mipmap min lod %d\n", i);

		dump_before();

		regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, i);
		tex1_tri();

		dump_after(0);
	}

	for (i=0; i<8; i++) {
		printf("# Set texture mipmap max lod %d\n", i);

		dump_before();

		regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LOD, i);
		tex1_tri();

		dump_after(0);
	}

	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	free(texture);
	TEST_EPILOGUE;
}

void test_texture_size(void)
{
	GLuint tid;
	int *texture = NULL;
	GLint max_size;
	int i;

	TEST_PROLOGUE;
	printf("# Testing texture sizes\n");

	regl.GetIntegerv(GL_MAX_TEXTURE_SIZE, &max_size);
	printf("# Max texture dimension: %d\n", max_size);

	regl.GenTextures(1, &tid);

	regl.Enable(GL_TEXTURE_2D);
	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	texture = create_random_texture(max_size,16);
	if (texture) {
		for (i=4; i<max_size; i<<=1) {
			printf("# Trying %dx16 texture\n", i);

			dump_before();
			regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, i, 16, 0,
				GL_RGBA, GL_UNSIGNED_BYTE, texture);
			tex1_tri();
			dump_after(0);

			printf("# Trying 16x%d texture\n", i);

			dump_before();
			regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 16, i, 0,
				GL_RGBA, GL_UNSIGNED_BYTE, texture);
			tex1_tri();
			dump_after(0);
		}
		free(texture);
	}

	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	TEST_EPILOGUE;
}

void test_arb_multisample(void)
{
#ifndef GL_ARB_multisample
	printf("# Sorry, no GL_ARB_multisample\n");
#else
	int numBuf=0, numSam=0, i;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_ARB_multisample")) {
		fprintf(stderr,"No GL_ARB_multisample extension.\n");
		return;
	}

	SDL_GL_GetAttribute(SDL_GL_MULTISAMPLEBUFFERS, &numBuf);
	if (numBuf==0) {
		fprintf(stderr, "No multisample buffers. Skipping ARB_multisample test\n");
		return;
	}
	SDL_GL_GetAttribute(SDL_GL_MULTISAMPLESAMPLES, &numSam);
	if (numSam==0) {
		fprintf(stderr, "No multisample samples. Skipping ARB_multisample test\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing arb_multisample\n");

	dump_before();

	regl.Enable(GL_MULTISAMPLE_ARB);
	tri();

	regl.Enable(GL_SAMPLE_ALPHA_TO_COVERAGE_ARB);
	tri();
	regl.Disable(GL_SAMPLE_ALPHA_TO_COVERAGE_ARB);

	regl.Enable(GL_SAMPLE_ALPHA_TO_ONE_ARB);
	tri();
	regl.Disable(GL_SAMPLE_ALPHA_TO_ONE_ARB);

	regl.Enable(GL_SAMPLE_COVERAGE_ARB);
	tri();
	for (i=0; i<=16; i++) {
		regl.SampleCoverageARB(((float) i) / 16.0, GL_TRUE);
		tri();
		regl.SampleCoverageARB(((float) i) / 16.0, GL_FALSE);
		tri();
	}
	regl.Disable(GL_SAMPLE_COVERAGE_ARB);

	regl.Disable(GL_MULTISAMPLE_ARB);

	dump_after(0);

	TEST_EPILOGUE;
#endif
}

void test_nv_half_float(void)
{
#ifndef GL_NV_half_float
	printf("# Sorry, no GL_NV_half_float\n");
#else
	int maxtex = 1, i;

	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_half_float")) {
		fprintf(stderr,"No GL_NV_half_float extension.\n");
		return;
	}
	if (strstr(extensions, "GL_ARB_multitexture")) {
		regl.GetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &maxtex);
	}

	TEST_PROLOGUE;
	printf("# Testing nv_half_float\n");

	dump_before();

	/* color 3h */
	regl.Color3hNV(100,200,300);
	regl.Begin(GL_POINTS);
		regl.Vertex2hNV(400,500);
		regl.Vertex3hNV(600,700,800);
	regl.End();

	/* color 4h */
	regl.Color4hNV(900,1000,1100,1200);

	/* secondary color 3h */
	if (strstr(extensions, "GL_EXT_secondary_color")) {
		regl.Enable(GL_COLOR_SUM_EXT);
		regl.SecondaryColor3hNV(1300,1400,1500);
	}

	/* vertex weight */
	if (strstr(extensions, "GL_EXT_vertex_weighting")) {
		regl.Enable(GL_VERTEX_WEIGHTING_EXT);
		regl.VertexWeighthNV(1600);
	}

	/* fog */
	if (strstr(extensions, "GL_EXT_fog_coord")) {
		regl.FogCoordhNV(1700);
	}

	/* normal 3h */
	regl.Normal3hNV(1800,1900,2000);

	/* texcoord 1 */
	regl.TexCoord1hNV(2100);
	if (maxtex>1) {
		for (i=1;i<maxtex;i++) {
			regl.MultiTexCoord1hNV(GL_TEXTURE0_ARB+i, 2100+i);
		}
	}

	regl.Begin(GL_POINTS);
		regl.Vertex4hNV(2200,2300,2400,2500);
	regl.End();

	/* texcoord 2 */
	regl.TexCoord2hNV(2600,2700);
	if (maxtex>1) {
		for (i=1;i<maxtex;i++) {
			regl.MultiTexCoord2hNV(GL_TEXTURE0_ARB+i, 2600+i*2,2700+i*2);
		}
	}

	regl.Begin(GL_POINTS);
		regl.Vertex4hNV(2800,2900,3000,3100);
	regl.End();

	/* texcoord 3 */
	regl.TexCoord3hNV(3200,3300,3400);
	if (maxtex>1) {
		for (i=1;i<maxtex;i++) {
			regl.MultiTexCoord3hNV(GL_TEXTURE0_ARB+i, 3200+i*3,3300+i*3,3400+i*3);
		}
	}

	regl.Begin(GL_POINTS);
		regl.Vertex4hNV(3500,3600,3700,3800);
	regl.End();

	/* texcoord 4f */
	regl.TexCoord4hNV(3900,4000,4100,4200);
	if (maxtex>1) {
		for (i=1;i<maxtex;i++) {
			regl.MultiTexCoord4hNV(GL_TEXTURE0_ARB+i, 3900+i*4,4000+i*4,4100+i*4,4200+i*4);
		}
	}

	regl.Begin(GL_POINTS);
		regl.Vertex2hNV(4300,4400);
	regl.End();

	/* TODO: check VertexAttrib*hNV() functions */

	dump_after(0);

	if (strstr(extensions, "GL_EXT_vertex_weighting")) {
		regl.Disable(GL_VERTEX_WEIGHTING_EXT);
	}
	if (strstr(extensions, "GL_EXT_secondary_color")) {
		regl.Disable(GL_COLOR_SUM_EXT);
	}
	regl.MatrixMode(GL_TEXTURE);
	regl.LoadIdentity();
	regl.MatrixMode(GL_MODELVIEW);
	regl.LoadIdentity();
	regl.Color4f(1.0,1.0,1.0,1.0);

	tri();
	dump_after(0);

	TEST_EPILOGUE;
#endif
}

void test_texture_spherical_map(void)
{
	GLuint tid;
	int *texture;
	
	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing texture, texgen spherical mapping\n");

	regl.GenTextures(1, &tid);

	/*regl.ActiveTexture(GL_TEXTURE0_ARB+1);*/
	regl.Enable(GL_TEXTURE_2D);
	regl.Enable(GL_TEXTURE_GEN_S);
	regl.Enable(GL_TEXTURE_GEN_T);
	regl.Enable(GL_TEXTURE_GEN_R);
	regl.Enable(GL_TEXTURE_GEN_Q);

	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	dump_before();
	regl.TexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	tex1_tri();

	regl.TexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	tex1_tri();

	regl.TexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	tex1_tri();

	regl.TexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	tex1_tri();

	/* Restore default */
	regl.TexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	regl.TexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	regl.TexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	regl.TexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	tex1_tri();
	dump_after(0);

	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_GEN_S);
	regl.Disable(GL_TEXTURE_GEN_T);
	regl.Disable(GL_TEXTURE_GEN_R);
	regl.Disable(GL_TEXTURE_GEN_Q);
	regl.Disable(GL_TEXTURE_2D);
	free(texture);
	TEST_EPILOGUE;
}

void test_nv_depth_clamp(void)
{
#ifndef GL_NV_depth_clamp
	printf("# Sorry, no GL_NV_depth_clamp\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_NV_depth_clamp")) {
		fprintf(stderr,"No GL_NV_depth_clamp extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing nv_half_float\n");

	dump_before();

	regl.Enable(GL_DEPTH_CLAMP_NV);
	tri();
	regl.Disable(GL_DEPTH_CLAMP_NV);
	tri();

	dump_after(0);
	TEST_EPILOGUE;
#endif
}

void test_ext_framebuffer_blit(void)
{
#ifndef GL_EXT_framebuffer_blit
	printf("# Sorry, no GL_EXT_framebuffer_blit\n");
#else
	const char *extensions = (const char *)regl.GetString(GL_EXTENSIONS);
	if (!strstr(extensions, "GL_EXT_framebuffer_blit")) {
		fprintf(stderr,"No GL_EXT_framebuffer_blit extension.\n");
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing ext_framebuffer_blit\n");

	dump_before();

	regl.BlitFramebufferEXT(
		0,0,100,100,
		100,0,100,100,
		GL_COLOR_BUFFER_BIT, GL_NEAREST
	);
	regl.BlitFramebufferEXT(
		0,100,100,100,
		100,100,100,100,
		GL_COLOR_BUFFER_BIT, GL_LINEAR
	);

	dump_after(0);
	TEST_EPILOGUE;
#endif
}

void test_texture_env(void)
{
	GLuint tid;
	GLfloat envColor[4]={0.1, 0.2, 0.3, 0.4};	
	GLfloat blackColor[4]={0.0, 0.0, 0.0, 0.0};	
	int *texture;

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# Testing texture environment modes\n");

	regl.GenTextures(1, &tid);

	regl.Enable(GL_TEXTURE_2D);

	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	dump_before();
	regl.TexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
	tex1_tri();

	regl.TexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	tex1_tri();

	regl.TexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, envColor);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
	tex1_tri();

	/* Restore default */
	regl.TexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, blackColor);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	tex1_tri();

	dump_after(0);

	regl.DeleteTextures(1, &tid);
	regl.Disable(GL_TEXTURE_2D);
	free(texture);
	TEST_EPILOGUE;
}

void test_fb_to_tex(void)
{
	GLuint tid;
	int *texture = NULL;

	texture = create_random_texture(TEXTURE_WIDTH,TEXTURE_HEIGHT);
	if (!texture) {
		return;
	}

	TEST_PROLOGUE;
	printf("# test_fb_to_tex()\n");

	regl.GenTextures(1, &tid);

	dump_before();

	regl.Enable(GL_TEXTURE_2D);

	regl.BindTexture(GL_TEXTURE_2D, tid);
	regl.TexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	/* from 0,0, pot sized */
	tex1_tri();
	regl.CopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0,0, TEXTURE_WIDTH,TEXTURE_HEIGHT, 0);

	/* from 0,0, npot sized */
	tex1_tri();
	regl.CopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0,0, 7,5, 0);

	/* from non 0,0, npot sized */
	tex1_tri();
	regl.CopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 1,3, 7,5, 0);

	regl.Disable(GL_TEXTURE_2D);
	regl.DeleteTextures(1, &tid);
	dump_after(0);

	free(texture);
	TEST_EPILOGUE;
}

void test_shadow_tex(void)
{
#define DEG_TO_RAD (3.14159 / 180.0)
	GLenum depthFormat = GL_DEPTH_STENCIL_EXT;
	GLenum depthType = GL_UNSIGNED_INT_24_8_EXT;
	GLint ShadowTexWidth = 256, ShadowTexHeight = 256;
	GLfloat k = 6;
	GLfloat lightModelview[16];
	GLfloat lightProjection[16];
	GLuint ShadowTexture, ShadowFBO;
	const GLfloat borderColor[4] = {1.0, 0.0, 0.0, 0.0};
	GLfloat LightDist = 10;
	GLfloat LightLatitude = 45.0;
	GLfloat LightLongitude = 45.0;
	GLfloat lightPos[4];
	GLfloat spotDir[3];
	GLfloat spotAngle = 40.0 * DEG_TO_RAD;
	GLfloat shadowNear = 4.0, shadowFar = 24.0;
	GLfloat d = shadowNear * tan(spotAngle);
	GLfloat Blue[4] = {0, 0, 1, 1};

	lightPos[0] = LightDist * sin(LightLongitude * DEG_TO_RAD);
	lightPos[1] = LightDist * sin(LightLatitude * DEG_TO_RAD);
	lightPos[2] = LightDist * cos(LightLatitude * DEG_TO_RAD) * cos(LightLongitude * DEG_TO_RAD);
	lightPos[3] = 1;
	spotDir[0] = -lightPos[0];
	spotDir[1] = -lightPos[1];
	spotDir[2] = -lightPos[2];

	TEST_PROLOGUE;
	printf("# test_shadow_tex()\n");
	dump_before();

	/*--- Calc shadow matrix ---*/
	regl.MatrixMode(GL_PROJECTION);
	regl.PushMatrix();
	regl.LoadIdentity();
	regl.Frustum(-d, d, -d, d, shadowNear, shadowFar);
	regl.GetFloatv(GL_PROJECTION_MATRIX, lightProjection);
	regl.PopMatrix();

	regl.MatrixMode(GL_MODELVIEW);
	regl.PushMatrix();
	regl.LoadIdentity();
	gluLookAt(lightPos[0], lightPos[1], lightPos[2],
		lightPos[0] + spotDir[0],
		lightPos[1] + spotDir[1],
		lightPos[2] + spotDir[2],
		0.0, 1.0, 0.0);
	regl.GetFloatv(GL_MODELVIEW_MATRIX, lightModelview);
	regl.PopMatrix();

	/*--- Create shadow texture ---*/
	regl.GenTextures(1, &ShadowTexture);
	regl.BindTexture(GL_TEXTURE_2D, ShadowTexture);
	regl.TexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, borderColor);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB,
		GL_COMPARE_R_TO_TEXTURE_ARB);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB,
		GL_LEQUAL);

	/*--- Create FBO ---*/
	regl.GenFramebuffersEXT(1, &ShadowFBO);
	regl.BindFramebufferEXT(GL_FRAMEBUFFER_EXT, ShadowFBO);
	regl.FramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
		GL_RENDERBUFFER_EXT, 0);
	regl.FramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
		GL_TEXTURE_2D, ShadowTexture, 0);

	regl.BindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	regl.Enable(GL_DEPTH_TEST);
	regl.Enable(GL_LIGHTING);
	regl.Enable(GL_LIGHT0);

	/*--- Render shadow map ---*/
	regl.MatrixMode(GL_PROJECTION);
	regl.LoadMatrixf(lightProjection);

	regl.MatrixMode(GL_MODELVIEW);
	regl.LoadMatrixf(lightModelview);

	regl.TexImage2D(GL_TEXTURE_2D, 0, depthFormat,
			ShadowTexWidth, ShadowTexHeight, 0,
			depthFormat, depthType, NULL);
	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	regl.BindFramebufferEXT(GL_FRAMEBUFFER_EXT, ShadowFBO);
	regl.DrawBuffer(GL_NONE);
	regl.ReadBuffer(GL_NONE);

	regl.Viewport(0, 0, ShadowTexWidth, ShadowTexHeight);
	regl.Clear(GL_DEPTH_BUFFER_BIT);
	regl.Enable(GL_DEPTH_TEST);

	/* Draw scene */
	regl.PushMatrix();
	regl.Translatef(0, -1.1, 0);
	regl.Materialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, Blue);
	regl.Color4fv(Blue);
	regl.Normal3f(0, 1, 0);
	regl.Begin(GL_POLYGON);
		regl.Vertex3f(-k, 0, -k);
		regl.Vertex3f( k, 0, -k);
		regl.Vertex3f( k, 0,  k);
		regl.Vertex3f(-k, 0,  k);
	regl.End();
	regl.PopMatrix();

	regl.BindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	/*--- Show shadow map ---*/
	regl.Viewport(0, 0, 512, 512);
	regl.Clear(GL_COLOR_BUFFER_BIT);

	regl.MatrixMode(GL_TEXTURE);
	regl.LoadIdentity();

	regl.MatrixMode(GL_PROJECTION);
	regl.LoadIdentity();
	regl.Ortho(0, 512, 0, 512, -1, 1);

	regl.MatrixMode(GL_MODELVIEW);
	regl.LoadIdentity();

	regl.Disable(GL_DEPTH_TEST);
	regl.Disable(GL_LIGHTING);

	regl.Enable(GL_TEXTURE_2D);

	regl.Disable(GL_TEXTURE_GEN_S);
	regl.Disable(GL_TEXTURE_GEN_T);
	regl.Disable(GL_TEXTURE_GEN_R);
	regl.Disable(GL_TEXTURE_GEN_Q);

	regl.TexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_NONE);
	regl.TexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE_ARB, GL_LUMINANCE);
	regl.TexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	regl.Begin(GL_POLYGON);
		regl.TexCoord2f(0, 0);  regl.Vertex2f(0, 0);
		regl.TexCoord2f(1, 0);  regl.Vertex2f(ShadowTexWidth, 0);
		regl.TexCoord2f(1, 1);  regl.Vertex2f(ShadowTexWidth, ShadowTexHeight);
		regl.TexCoord2f(0, 1);  regl.Vertex2f(0, ShadowTexHeight);
	regl.End();

	regl.Disable(GL_TEXTURE_2D);
	regl.Enable(GL_DEPTH_TEST);
	regl.Enable(GL_LIGHTING);

	dump_after(0);
	TEST_EPILOGUE;
}
